function __vite__mapDeps(indexes) {
  if (!__vite__mapDeps.viteFileDeps) {
    __vite__mapDeps.viteFileDeps = ["./Selection-AWIjaBb7.js","./three.module-BW0MBeo0.js","./ModelViewerWrapper-CThr6USm.js"]
  }
  return indexes.map((i) => __vite__mapDeps.viteFileDeps[i])
}
var fm=Object.defineProperty;var hm=(t,e,n)=>e in t?fm(t,e,{enumerable:!0,configurable:!0,writable:!0,value:n}):t[e]=n;var Qt=(t,e,n)=>(hm(t,typeof e!="symbol"?e+"":e,n),n);(function(){const e=document.createElement("link").relList;if(e&&e.supports&&e.supports("modulepreload"))return;for(const r of document.querySelectorAll('link[rel="modulepreload"]'))s(r);new MutationObserver(r=>{for(const i of r)if(i.type==="childList")for(const o of i.addedNodes)o.tagName==="LINK"&&o.rel==="modulepreload"&&s(o)}).observe(document,{childList:!0,subtree:!0});function n(r){const i={};return r.integrity&&(i.integrity=r.integrity),r.referrerPolicy&&(i.referrerPolicy=r.referrerPolicy),r.crossOrigin==="use-credentials"?i.credentials="include":r.crossOrigin==="anonymous"?i.credentials="omit":i.credentials="same-origin",i}function s(r){if(r.ep)return;r.ep=!0;const i=n(r);fetch(r.href,i)}})();/**
* @vue/shared v3.4.21
* (c) 2018-present Yuxi (Evan) You and Vue contributors
* @license MIT
**/function Pa(t,e){const n=new Set(t.split(","));return e?s=>n.has(s.toLowerCase()):s=>n.has(s)}const Ae={},Ms=[],Ct=()=>{},mm=()=>!1,Wi=t=>t.charCodeAt(0)===111&&t.charCodeAt(1)===110&&(t.charCodeAt(2)>122||t.charCodeAt(2)<97),Oa=t=>t.startsWith("onUpdate:"),De=Object.assign,Na=(t,e)=>{const n=t.indexOf(e);n>-1&&t.splice(n,1)},gm=Object.prototype.hasOwnProperty,_e=(t,e)=>gm.call(t,e),oe=Array.isArray,Ts=t=>Gi(t)==="[object Map]",Kc=t=>Gi(t)==="[object Set]",fe=t=>typeof t=="function",Ne=t=>typeof t=="string",Os=t=>typeof t=="symbol",Ie=t=>t!==null&&typeof t=="object",Jc=t=>(Ie(t)||fe(t))&&fe(t.then)&&fe(t.catch),Qc=Object.prototype.toString,Gi=t=>Qc.call(t),vm=t=>Gi(t).slice(8,-1),ed=t=>Gi(t)==="[object Object]",Ba=t=>Ne(t)&&t!=="NaN"&&t[0]!=="-"&&""+parseInt(t,10)===t,Qs=Pa(",key,ref,ref_for,ref_key,onVnodeBeforeMount,onVnodeMounted,onVnodeBeforeUpdate,onVnodeUpdated,onVnodeBeforeUnmount,onVnodeUnmounted"),qi=t=>{const e=Object.create(null);return n=>e[n]||(e[n]=t(n))},pm=/-(\w)/g,bt=qi(t=>t.replace(pm,(e,n)=>n?n.toUpperCase():"")),ym=/\B([A-Z])/g,rs=qi(t=>t.replace(ym,"-$1").toLowerCase()),cn=qi(t=>t.charAt(0).toUpperCase()+t.slice(1)),bo=qi(t=>t?`on${cn(t)}`:""),Wt=(t,e)=>!Object.is(t,e),xo=(t,e)=>{for(let n=0;n<t.length;n++)t[n](e)},vi=(t,e,n)=>{Object.defineProperty(t,e,{configurable:!0,enumerable:!1,value:n})},bm=t=>{const e=parseFloat(t);return isNaN(e)?t:e},xm=t=>{const e=Ne(t)?Number(t):NaN;return isNaN(e)?t:e};let Ll;const td=()=>Ll||(Ll=typeof globalThis<"u"?globalThis:typeof self<"u"?self:typeof window<"u"?window:typeof global<"u"?global:{});function Yi(t){if(oe(t)){const e={};for(let n=0;n<t.length;n++){const s=t[n],r=Ne(s)?Cm(s):Yi(s);if(r)for(const i in r)e[i]=r[i]}return e}else if(Ne(t)||Ie(t))return t}const wm=/;(?![^(]*\))/g,_m=/:([^]+)/,Sm=/\/\*[^]*?\*\//g;function Cm(t){const e={};return t.replace(Sm,"").split(wm).forEach(n=>{if(n){const s=n.split(_m);s.length>1&&(e[s[0].trim()]=s[1].trim())}}),e}function Xi(t){let e="";if(Ne(t))e=t;else if(oe(t))for(let n=0;n<t.length;n++){const s=Xi(t[n]);s&&(e+=s+" ")}else if(Ie(t))for(const n in t)t[n]&&(e+=n+" ");return e.trim()}const Em="itemscope,allowfullscreen,formnovalidate,ismap,nomodule,novalidate,readonly",Mm=Pa(Em);function nd(t){return!!t||t===""}const sn=t=>Ne(t)?t:t==null?"":oe(t)||Ie(t)&&(t.toString===Qc||!fe(t.toString))?JSON.stringify(t,sd,2):String(t),sd=(t,e)=>e&&e.__v_isRef?sd(t,e.value):Ts(e)?{[`Map(${e.size})`]:[...e.entries()].reduce((n,[s,r],i)=>(n[wo(s,i)+" =>"]=r,n),{})}:Kc(e)?{[`Set(${e.size})`]:[...e.values()].map(n=>wo(n))}:Os(e)?wo(e):Ie(e)&&!oe(e)&&!ed(e)?String(e):e,wo=(t,e="")=>{var n;return Os(t)?`Symbol(${(n=t.description)!=null?n:e})`:t};/**
* @vue/reactivity v3.4.21
* (c) 2018-present Yuxi (Evan) You and Vue contributors
* @license MIT
**/let mt;class rd{constructor(e=!1){this.detached=e,this._active=!0,this.effects=[],this.cleanups=[],this.parent=mt,!e&&mt&&(this.index=(mt.scopes||(mt.scopes=[])).push(this)-1)}get active(){return this._active}run(e){if(this._active){const n=mt;try{return mt=this,e()}finally{mt=n}}}on(){mt=this}off(){mt=this.parent}stop(e){if(this._active){let n,s;for(n=0,s=this.effects.length;n<s;n++)this.effects[n].stop();for(n=0,s=this.cleanups.length;n<s;n++)this.cleanups[n]();if(this.scopes)for(n=0,s=this.scopes.length;n<s;n++)this.scopes[n].stop(!0);if(!this.detached&&this.parent&&!e){const r=this.parent.scopes.pop();r&&r!==this&&(this.parent.scopes[this.index]=r,r.index=this.index)}this.parent=void 0,this._active=!1}}}function za(t){return new rd(t)}function Tm(t,e=mt){e&&e.active&&e.effects.push(t)}function Am(){return mt}function At(t){mt&&mt.cleanups.push(t)}let Xn;class Da{constructor(e,n,s,r){this.fn=e,this.trigger=n,this.scheduler=s,this.active=!0,this.deps=[],this._dirtyLevel=4,this._trackId=0,this._runnings=0,this._shouldSchedule=!1,this._depsLength=0,Tm(this,r)}get dirty(){if(this._dirtyLevel===2||this._dirtyLevel===3){this._dirtyLevel=1,is();for(let e=0;e<this._depsLength;e++){const n=this.deps[e];if(n.computed&&(Rm(n.computed),this._dirtyLevel>=4))break}this._dirtyLevel===1&&(this._dirtyLevel=0),os()}return this._dirtyLevel>=4}set dirty(e){this._dirtyLevel=e?4:0}run(){if(this._dirtyLevel=0,!this.active)return this.fn();let e=An,n=Xn;try{return An=!0,Xn=this,this._runnings++,Pl(this),this.fn()}finally{Ol(this),this._runnings--,Xn=n,An=e}}stop(){var e;this.active&&(Pl(this),Ol(this),(e=this.onStop)==null||e.call(this),this.active=!1)}}function Rm(t){return t.value}function Pl(t){t._trackId++,t._depsLength=0}function Ol(t){if(t.deps.length>t._depsLength){for(let e=t._depsLength;e<t.deps.length;e++)id(t.deps[e],t);t.deps.length=t._depsLength}}function id(t,e){const n=t.get(e);n!==void 0&&e._trackId!==n&&(t.delete(e),t.size===0&&t.cleanup())}let An=!0,ea=0;const od=[];function is(){od.push(An),An=!1}function os(){const t=od.pop();An=t===void 0?!0:t}function $a(){ea++}function Ha(){for(ea--;!ea&&ta.length;)ta.shift()()}function ad(t,e,n){if(e.get(t)!==t._trackId){e.set(t,t._trackId);const s=t.deps[t._depsLength];s!==e?(s&&id(s,t),t.deps[t._depsLength++]=e):t._depsLength++}}const ta=[];function ld(t,e,n){$a();for(const s of t.keys()){let r;s._dirtyLevel<e&&(r??(r=t.get(s)===s._trackId))&&(s._shouldSchedule||(s._shouldSchedule=s._dirtyLevel===0),s._dirtyLevel=e),s._shouldSchedule&&(r??(r=t.get(s)===s._trackId))&&(s.trigger(),(!s._runnings||s.allowRecurse)&&s._dirtyLevel!==2&&(s._shouldSchedule=!1,s.scheduler&&ta.push(s.scheduler)))}Ha()}const ud=(t,e)=>{const n=new Map;return n.cleanup=t,n.computed=e,n},pi=new WeakMap,Zn=Symbol(""),na=Symbol("");function lt(t,e,n){if(An&&Xn){let s=pi.get(t);s||pi.set(t,s=new Map);let r=s.get(n);r||s.set(n,r=ud(()=>s.delete(n))),ad(Xn,r)}}function rn(t,e,n,s,r,i){const o=pi.get(t);if(!o)return;let a=[];if(e==="clear")a=[...o.values()];else if(n==="length"&&oe(t)){const l=Number(s);o.forEach((u,c)=>{(c==="length"||!Os(c)&&c>=l)&&a.push(u)})}else switch(n!==void 0&&a.push(o.get(n)),e){case"add":oe(t)?Ba(n)&&a.push(o.get("length")):(a.push(o.get(Zn)),Ts(t)&&a.push(o.get(na)));break;case"delete":oe(t)||(a.push(o.get(Zn)),Ts(t)&&a.push(o.get(na)));break;case"set":Ts(t)&&a.push(o.get(Zn));break}$a();for(const l of a)l&&ld(l,4);Ha()}function Im(t,e){var n;return(n=pi.get(t))==null?void 0:n.get(e)}const km=Pa("__proto__,__v_isRef,__isVue"),cd=new Set(Object.getOwnPropertyNames(Symbol).filter(t=>t!=="arguments"&&t!=="caller").map(t=>Symbol[t]).filter(Os)),Nl=Fm();function Fm(){const t={};return["includes","indexOf","lastIndexOf"].forEach(e=>{t[e]=function(...n){const s=pe(this);for(let i=0,o=this.length;i<o;i++)lt(s,"get",i+"");const r=s[e](...n);return r===-1||r===!1?s[e](...n.map(pe)):r}}),["push","pop","shift","unshift","splice"].forEach(e=>{t[e]=function(...n){is(),$a();const s=pe(this)[e].apply(this,n);return Ha(),os(),s}}),t}function Vm(t){const e=pe(this);return lt(e,"has",t),e.hasOwnProperty(t)}class dd{constructor(e=!1,n=!1){this._isReadonly=e,this._isShallow=n}get(e,n,s){const r=this._isReadonly,i=this._isShallow;if(n==="__v_isReactive")return!r;if(n==="__v_isReadonly")return r;if(n==="__v_isShallow")return i;if(n==="__v_raw")return s===(r?i?Gm:gd:i?md:hd).get(e)||Object.getPrototypeOf(e)===Object.getPrototypeOf(s)?e:void 0;const o=oe(e);if(!r){if(o&&_e(Nl,n))return Reflect.get(Nl,n,s);if(n==="hasOwnProperty")return Vm}const a=Reflect.get(e,n,s);return(Os(n)?cd.has(n):km(n))||(r||lt(e,"get",n),i)?a:Te(a)?o&&Ba(n)?a:a.value:Ie(a)?r?_r(a):tt(a):a}}class fd extends dd{constructor(e=!1){super(!1,e)}set(e,n,s,r){let i=e[n];if(!this._isShallow){const l=ks(i);if(!yi(s)&&!ks(s)&&(i=pe(i),s=pe(s)),!oe(e)&&Te(i)&&!Te(s))return l?!1:(i.value=s,!0)}const o=oe(e)&&Ba(n)?Number(n)<e.length:_e(e,n),a=Reflect.set(e,n,s,r);return e===pe(r)&&(o?Wt(s,i)&&rn(e,"set",n,s):rn(e,"add",n,s)),a}deleteProperty(e,n){const s=_e(e,n);e[n];const r=Reflect.deleteProperty(e,n);return r&&s&&rn(e,"delete",n,void 0),r}has(e,n){const s=Reflect.has(e,n);return(!Os(n)||!cd.has(n))&&lt(e,"has",n),s}ownKeys(e){return lt(e,"iterate",oe(e)?"length":Zn),Reflect.ownKeys(e)}}class Lm extends dd{constructor(e=!1){super(!0,e)}set(e,n){return!0}deleteProperty(e,n){return!0}}const Pm=new fd,Om=new Lm,Nm=new fd(!0),Ua=t=>t,Zi=t=>Reflect.getPrototypeOf(t);function Pr(t,e,n=!1,s=!1){t=t.__v_raw;const r=pe(t),i=pe(e);n||(Wt(e,i)&&lt(r,"get",e),lt(r,"get",i));const{has:o}=Zi(r),a=s?Ua:n?Ga:ur;if(o.call(r,e))return a(t.get(e));if(o.call(r,i))return a(t.get(i));t!==r&&t.get(e)}function Or(t,e=!1){const n=this.__v_raw,s=pe(n),r=pe(t);return e||(Wt(t,r)&&lt(s,"has",t),lt(s,"has",r)),t===r?n.has(t):n.has(t)||n.has(r)}function Nr(t,e=!1){return t=t.__v_raw,!e&&lt(pe(t),"iterate",Zn),Reflect.get(t,"size",t)}function Bl(t){t=pe(t);const e=pe(this);return Zi(e).has.call(e,t)||(e.add(t),rn(e,"add",t,t)),this}function zl(t,e){e=pe(e);const n=pe(this),{has:s,get:r}=Zi(n);let i=s.call(n,t);i||(t=pe(t),i=s.call(n,t));const o=r.call(n,t);return n.set(t,e),i?Wt(e,o)&&rn(n,"set",t,e):rn(n,"add",t,e),this}function Dl(t){const e=pe(this),{has:n,get:s}=Zi(e);let r=n.call(e,t);r||(t=pe(t),r=n.call(e,t)),s&&s.call(e,t);const i=e.delete(t);return r&&rn(e,"delete",t,void 0),i}function $l(){const t=pe(this),e=t.size!==0,n=t.clear();return e&&rn(t,"clear",void 0,void 0),n}function Br(t,e){return function(s,r){const i=this,o=i.__v_raw,a=pe(o),l=e?Ua:t?Ga:ur;return!t&&lt(a,"iterate",Zn),o.forEach((u,c)=>s.call(r,l(u),l(c),i))}}function zr(t,e,n){return function(...s){const r=this.__v_raw,i=pe(r),o=Ts(i),a=t==="entries"||t===Symbol.iterator&&o,l=t==="keys"&&o,u=r[t](...s),c=n?Ua:e?Ga:ur;return!e&&lt(i,"iterate",l?na:Zn),{next(){const{value:d,done:f}=u.next();return f?{value:d,done:f}:{value:a?[c(d[0]),c(d[1])]:c(d),done:f}},[Symbol.iterator](){return this}}}}function vn(t){return function(...e){return t==="delete"?!1:t==="clear"?void 0:this}}function Bm(){const t={get(i){return Pr(this,i)},get size(){return Nr(this)},has:Or,add:Bl,set:zl,delete:Dl,clear:$l,forEach:Br(!1,!1)},e={get(i){return Pr(this,i,!1,!0)},get size(){return Nr(this)},has:Or,add:Bl,set:zl,delete:Dl,clear:$l,forEach:Br(!1,!0)},n={get(i){return Pr(this,i,!0)},get size(){return Nr(this,!0)},has(i){return Or.call(this,i,!0)},add:vn("add"),set:vn("set"),delete:vn("delete"),clear:vn("clear"),forEach:Br(!0,!1)},s={get(i){return Pr(this,i,!0,!0)},get size(){return Nr(this,!0)},has(i){return Or.call(this,i,!0)},add:vn("add"),set:vn("set"),delete:vn("delete"),clear:vn("clear"),forEach:Br(!0,!0)};return["keys","values","entries",Symbol.iterator].forEach(i=>{t[i]=zr(i,!1,!1),n[i]=zr(i,!0,!1),e[i]=zr(i,!1,!0),s[i]=zr(i,!0,!0)}),[t,n,e,s]}const[zm,Dm,$m,Hm]=Bm();function ja(t,e){const n=e?t?Hm:$m:t?Dm:zm;return(s,r,i)=>r==="__v_isReactive"?!t:r==="__v_isReadonly"?t:r==="__v_raw"?s:Reflect.get(_e(n,r)&&r in s?n:s,r,i)}const Um={get:ja(!1,!1)},jm={get:ja(!1,!0)},Wm={get:ja(!0,!1)},hd=new WeakMap,md=new WeakMap,gd=new WeakMap,Gm=new WeakMap;function qm(t){switch(t){case"Object":case"Array":return 1;case"Map":case"Set":case"WeakMap":case"WeakSet":return 2;default:return 0}}function Ym(t){return t.__v_skip||!Object.isExtensible(t)?0:qm(vm(t))}function tt(t){return ks(t)?t:Wa(t,!1,Pm,Um,hd)}function Xm(t){return Wa(t,!1,Nm,jm,md)}function _r(t){return Wa(t,!0,Om,Wm,gd)}function Wa(t,e,n,s,r){if(!Ie(t)||t.__v_raw&&!(e&&t.__v_isReactive))return t;const i=r.get(t);if(i)return i;const o=Ym(t);if(o===0)return t;const a=new Proxy(t,o===2?s:n);return r.set(t,a),a}function As(t){return ks(t)?As(t.__v_raw):!!(t&&t.__v_isReactive)}function ks(t){return!!(t&&t.__v_isReadonly)}function yi(t){return!!(t&&t.__v_isShallow)}function vd(t){return As(t)||ks(t)}function pe(t){const e=t&&t.__v_raw;return e?pe(e):t}function pd(t){return Object.isExtensible(t)&&vi(t,"__v_skip",!0),t}const ur=t=>Ie(t)?tt(t):t,Ga=t=>Ie(t)?_r(t):t;class yd{constructor(e,n,s,r){this.getter=e,this._setter=n,this.dep=void 0,this.__v_isRef=!0,this.__v_isReadonly=!1,this.effect=new Da(()=>e(this._value),()=>Rs(this,this.effect._dirtyLevel===2?2:3)),this.effect.computed=this,this.effect.active=this._cacheable=!r,this.__v_isReadonly=s}get value(){const e=pe(this);return(!e._cacheable||e.effect.dirty)&&Wt(e._value,e._value=e.effect.run())&&Rs(e,4),qa(e),e.effect._dirtyLevel>=2&&Rs(e,2),e._value}set value(e){this._setter(e)}get _dirty(){return this.effect.dirty}set _dirty(e){this.effect.dirty=e}}function Zm(t,e,n=!1){let s,r;const i=fe(t);return i?(s=t,r=Ct):(s=t.get,r=t.set),new yd(s,r,i||!r,n)}function qa(t){var e;An&&Xn&&(t=pe(t),ad(Xn,(e=t.dep)!=null?e:t.dep=ud(()=>t.dep=void 0,t instanceof yd?t:void 0)))}function Rs(t,e=4,n){t=pe(t);const s=t.dep;s&&ld(s,e)}function Te(t){return!!(t&&t.__v_isRef===!0)}function se(t){return bd(t,!1)}function be(t){return bd(t,!0)}function bd(t,e){return Te(t)?t:new Km(t,e)}class Km{constructor(e,n){this.__v_isShallow=n,this.dep=void 0,this.__v_isRef=!0,this._rawValue=n?e:pe(e),this._value=n?e:ur(e)}get value(){return qa(this),this._value}set value(e){const n=this.__v_isShallow||yi(e)||ks(e);e=n?e:pe(e),Wt(e,this._rawValue)&&(this._rawValue=e,this._value=n?e:ur(e),Rs(this,4))}}function Hl(t){Rs(t,4)}function F(t){return Te(t)?t.value:t}const Jm={get:(t,e,n)=>F(Reflect.get(t,e,n)),set:(t,e,n,s)=>{const r=t[e];return Te(r)&&!Te(n)?(r.value=n,!0):Reflect.set(t,e,n,s)}};function xd(t){return As(t)?t:new Proxy(t,Jm)}class Qm{constructor(e){this.dep=void 0,this.__v_isRef=!0;const{get:n,set:s}=e(()=>qa(this),()=>Rs(this));this._get=n,this._set=s}get value(){return this._get()}set value(e){this._set(e)}}function eg(t){return new Qm(t)}function Ya(t){const e=oe(t)?new Array(t.length):{};for(const n in t)e[n]=wd(t,n);return e}class tg{constructor(e,n,s){this._object=e,this._key=n,this._defaultValue=s,this.__v_isRef=!0}get value(){const e=this._object[this._key];return e===void 0?this._defaultValue:e}set value(e){this._object[this._key]=e}get dep(){return Im(pe(this._object),this._key)}}class ng{constructor(e){this._getter=e,this.__v_isRef=!0,this.__v_isReadonly=!0}get value(){return this._getter()}}function de(t,e,n){return Te(t)?t:fe(t)?new ng(t):Ie(t)&&arguments.length>1?wd(t,e,n):se(t)}function wd(t,e,n){const s=t[e];return Te(s)?s:new tg(t,e,n)}/**
* @vue/runtime-core v3.4.21
* (c) 2018-present Yuxi (Evan) You and Vue contributors
* @license MIT
**/function Rn(t,e,n,s){try{return s?t(...s):t()}catch(r){Sr(r,e,n)}}function Mt(t,e,n,s){if(fe(t)){const i=Rn(t,e,n,s);return i&&Jc(i)&&i.catch(o=>{Sr(o,e,n)}),i}const r=[];for(let i=0;i<t.length;i++)r.push(Mt(t[i],e,n,s));return r}function Sr(t,e,n,s=!0){const r=e?e.vnode:null;if(e){let i=e.parent;const o=e.proxy,a=`https://vuejs.org/error-reference/#runtime-${n}`;for(;i;){const u=i.ec;if(u){for(let c=0;c<u.length;c++)if(u[c](t,o,a)===!1)return}i=i.parent}const l=e.appContext.config.errorHandler;if(l){Rn(l,null,10,[t,o,a]);return}}sg(t,n,r,s)}function sg(t,e,n,s=!0){console.error(t)}let cr=!1,sa=!1;const Ze=[];let $t=0;const Is=[];let Cn=null,Wn=0;const _d=Promise.resolve();let Xa=null;function xt(t){const e=Xa||_d;return t?e.then(this?t.bind(this):t):e}function rg(t){let e=$t+1,n=Ze.length;for(;e<n;){const s=e+n>>>1,r=Ze[s],i=dr(r);i<t||i===t&&r.pre?e=s+1:n=s}return e}function Ki(t){(!Ze.length||!Ze.includes(t,cr&&t.allowRecurse?$t+1:$t))&&(t.id==null?Ze.push(t):Ze.splice(rg(t.id),0,t),Sd())}function Sd(){!cr&&!sa&&(sa=!0,Xa=_d.then(Ed))}function ig(t){const e=Ze.indexOf(t);e>$t&&Ze.splice(e,1)}function og(t){oe(t)?Is.push(...t):(!Cn||!Cn.includes(t,t.allowRecurse?Wn+1:Wn))&&Is.push(t),Sd()}function Ul(t,e,n=cr?$t+1:0){for(;n<Ze.length;n++){const s=Ze[n];if(s&&s.pre){if(t&&s.id!==t.uid)continue;Ze.splice(n,1),n--,s()}}}function Cd(t){if(Is.length){const e=[...new Set(Is)].sort((n,s)=>dr(n)-dr(s));if(Is.length=0,Cn){Cn.push(...e);return}for(Cn=e,Wn=0;Wn<Cn.length;Wn++)Cn[Wn]();Cn=null,Wn=0}}const dr=t=>t.id==null?1/0:t.id,ag=(t,e)=>{const n=dr(t)-dr(e);if(n===0){if(t.pre&&!e.pre)return-1;if(e.pre&&!t.pre)return 1}return n};function Ed(t){sa=!1,cr=!0,Ze.sort(ag);try{for($t=0;$t<Ze.length;$t++){const e=Ze[$t];e&&e.active!==!1&&Rn(e,null,14)}}finally{$t=0,Ze.length=0,Cd(),cr=!1,Xa=null,(Ze.length||Is.length)&&Ed()}}function lg(t,e,...n){if(t.isUnmounted)return;const s=t.vnode.props||Ae;let r=n;const i=e.startsWith("update:"),o=i&&e.slice(7);if(o&&o in s){const c=`${o==="modelValue"?"model":o}Modifiers`,{number:d,trim:f}=s[c]||Ae;f&&(r=n.map(h=>Ne(h)?h.trim():h)),d&&(r=n.map(bm))}let a,l=s[a=bo(e)]||s[a=bo(bt(e))];!l&&i&&(l=s[a=bo(rs(e))]),l&&Mt(l,t,6,r);const u=s[a+"Once"];if(u){if(!t.emitted)t.emitted={};else if(t.emitted[a])return;t.emitted[a]=!0,Mt(u,t,6,r)}}function Md(t,e,n=!1){const s=e.emitsCache,r=s.get(t);if(r!==void 0)return r;const i=t.emits;let o={},a=!1;if(!fe(t)){const l=u=>{const c=Md(u,e,!0);c&&(a=!0,De(o,c))};!n&&e.mixins.length&&e.mixins.forEach(l),t.extends&&l(t.extends),t.mixins&&t.mixins.forEach(l)}return!i&&!a?(Ie(t)&&s.set(t,null),null):(oe(i)?i.forEach(l=>o[l]=null):De(o,i),Ie(t)&&s.set(t,o),o)}function Ji(t,e){return!t||!Wi(e)?!1:(e=e.slice(2).replace(/Once$/,""),_e(t,e[0].toLowerCase()+e.slice(1))||_e(t,rs(e))||_e(t,e))}let $e=null,Td=null;function bi(t){const e=$e;return $e=t,Td=t&&t.type.__scopeId||null,e}function X(t,e=$e,n){if(!e||t._n)return t;const s=(...r)=>{s._d&&su(-1);const i=bi(e);let o;try{o=t(...r)}finally{bi(i),s._d&&su(1)}return o};return s._n=!0,s._c=!0,s._d=!0,s}function _o(t){const{type:e,vnode:n,proxy:s,withProxy:r,props:i,propsOptions:[o],slots:a,attrs:l,emit:u,render:c,renderCache:d,data:f,setupState:h,ctx:g,inheritAttrs:y}=t;let v,S;const m=bi(t);try{if(n.shapeFlag&4){const w=r||s,b=w;v=Dt(c.call(b,w,d,i,h,f,g)),S=l}else{const w=e;v=Dt(w.length>1?w(i,{attrs:l,slots:a,emit:u}):w(i,null)),S=e.props?l:ug(l)}}catch(w){rr.length=0,Sr(w,t,1),v=_(pt)}let p=v;if(S&&y!==!1){const w=Object.keys(S),{shapeFlag:b}=p;w.length&&b&7&&(o&&w.some(Oa)&&(S=cg(S,o)),p=Vn(p,S))}return n.dirs&&(p=Vn(p),p.dirs=p.dirs?p.dirs.concat(n.dirs):n.dirs),n.transition&&(p.transition=n.transition),v=p,bi(m),v}const ug=t=>{let e;for(const n in t)(n==="class"||n==="style"||Wi(n))&&((e||(e={}))[n]=t[n]);return e},cg=(t,e)=>{const n={};for(const s in t)(!Oa(s)||!(s.slice(9)in e))&&(n[s]=t[s]);return n};function dg(t,e,n){const{props:s,children:r,component:i}=t,{props:o,children:a,patchFlag:l}=e,u=i.emitsOptions;if(e.dirs||e.transition)return!0;if(n&&l>=0){if(l&1024)return!0;if(l&16)return s?jl(s,o,u):!!o;if(l&8){const c=e.dynamicProps;for(let d=0;d<c.length;d++){const f=c[d];if(o[f]!==s[f]&&!Ji(u,f))return!0}}}else return(r||a)&&(!a||!a.$stable)?!0:s===o?!1:s?o?jl(s,o,u):!0:!!o;return!1}function jl(t,e,n){const s=Object.keys(e);if(s.length!==Object.keys(t).length)return!0;for(let r=0;r<s.length;r++){const i=s[r];if(e[i]!==t[i]&&!Ji(n,i))return!0}return!1}function fg({vnode:t,parent:e},n){for(;e;){const s=e.subTree;if(s.suspense&&s.suspense.activeBranch===t&&(s.el=t.el),s===t)(t=e.vnode).el=n,e=e.parent;else break}}const Ad="components",hg="directives",Rd=Symbol.for("v-ndc");function mg(t){return Ne(t)?Id(Ad,t,!1)||t:t||Rd}function as(t){return Id(hg,t)}function Id(t,e,n=!0,s=!1){const r=$e||je;if(r){const i=r.type;if(t===Ad){const a=av(i,!1);if(a&&(a===e||a===bt(e)||a===cn(bt(e))))return i}const o=Wl(r[t]||i[t],e)||Wl(r.appContext[t],e);return!o&&s?i:o}}function Wl(t,e){return t&&(t[e]||t[bt(e)]||t[cn(bt(e))])}const gg=t=>t.__isSuspense;function vg(t,e){e&&e.pendingBranch?oe(t)?e.effects.push(...t):e.effects.push(t):og(t)}const pg=Symbol.for("v-scx"),yg=()=>Pe(pg);function Pn(t,e){return Qi(t,null,e)}function bg(t,e){return Qi(t,null,{flush:"sync"})}const Dr={};function le(t,e,n){return Qi(t,e,n)}function Qi(t,e,{immediate:n,deep:s,flush:r,once:i,onTrack:o,onTrigger:a}=Ae){if(e&&i){const x=e;e=(...C)=>{x(...C),b()}}const l=je,u=x=>s===!0?x:qn(x,s===!1?1:void 0);let c,d=!1,f=!1;if(Te(t)?(c=()=>t.value,d=yi(t)):As(t)?(c=()=>u(t),d=!0):oe(t)?(f=!0,d=t.some(x=>As(x)||yi(x)),c=()=>t.map(x=>{if(Te(x))return x.value;if(As(x))return u(x);if(fe(x))return Rn(x,l,2)})):fe(t)?e?c=()=>Rn(t,l,2):c=()=>(h&&h(),Mt(t,l,3,[g])):c=Ct,e&&s){const x=c;c=()=>qn(x())}let h,g=x=>{h=p.onStop=()=>{Rn(x,l,4),h=p.onStop=void 0}},y;if(Ar)if(g=Ct,e?n&&Mt(e,l,3,[c(),f?[]:void 0,g]):c(),r==="sync"){const x=yg();y=x.__watcherHandles||(x.__watcherHandles=[])}else return Ct;let v=f?new Array(t.length).fill(Dr):Dr;const S=()=>{if(!(!p.active||!p.dirty))if(e){const x=p.run();(s||d||(f?x.some((C,M)=>Wt(C,v[M])):Wt(x,v)))&&(h&&h(),Mt(e,l,3,[x,v===Dr?void 0:f&&v[0]===Dr?[]:v,g]),v=x)}else p.run()};S.allowRecurse=!!e;let m;r==="sync"?m=S:r==="post"?m=()=>rt(S,l&&l.suspense):(S.pre=!0,l&&(S.id=l.uid),m=()=>Ki(S));const p=new Da(c,Ct,m),w=Am(),b=()=>{p.stop(),w&&Na(w.effects,p)};return e?n?S():v=p.run():r==="post"?rt(p.run.bind(p),l&&l.suspense):p.run(),y&&y.push(b),b}function xg(t,e,n){const s=this.proxy,r=Ne(t)?t.includes(".")?kd(s,t):()=>s[t]:t.bind(s,s);let i;fe(e)?i=e:(i=e.handler,n=e);const o=Tr(this),a=Qi(r,i.bind(s),n);return o(),a}function kd(t,e){const n=e.split(".");return()=>{let s=t;for(let r=0;r<n.length&&s;r++)s=s[n[r]];return s}}function qn(t,e,n=0,s){if(!Ie(t)||t.__v_skip)return t;if(e&&e>0){if(n>=e)return t;n++}if(s=s||new Set,s.has(t))return t;if(s.add(t),Te(t))qn(t.value,e,n,s);else if(oe(t))for(let r=0;r<t.length;r++)qn(t[r],e,n,s);else if(Kc(t)||Ts(t))t.forEach(r=>{qn(r,e,n,s)});else if(ed(t))for(const r in t)qn(t[r],e,n,s);return t}function Gt(t,e){if($e===null)return t;const n=ro($e)||$e.proxy,s=t.dirs||(t.dirs=[]);for(let r=0;r<e.length;r++){let[i,o,a,l=Ae]=e[r];i&&(fe(i)&&(i={mounted:i,updated:i}),i.deep&&qn(o),s.push({dir:i,instance:n,value:o,oldValue:void 0,arg:a,modifiers:l}))}return t}function Bn(t,e,n,s){const r=t.dirs,i=e&&e.dirs;for(let o=0;o<r.length;o++){const a=r[o];i&&(a.oldValue=i[o].value);let l=a.dir[s];l&&(is(),Mt(l,n,8,[t.el,a,t,e]),os())}}const En=Symbol("_leaveCb"),$r=Symbol("_enterCb");function Fd(){const t={isMounted:!1,isLeaving:!1,isUnmounting:!1,leavingVNodes:new Map};return Yt(()=>{t.isMounted=!0}),Rt(()=>{t.isUnmounting=!0}),t}const St=[Function,Array],Vd={mode:String,appear:Boolean,persisted:Boolean,onBeforeEnter:St,onEnter:St,onAfterEnter:St,onEnterCancelled:St,onBeforeLeave:St,onLeave:St,onAfterLeave:St,onLeaveCancelled:St,onBeforeAppear:St,onAppear:St,onAfterAppear:St,onAppearCancelled:St},wg={name:"BaseTransition",props:Vd,setup(t,{slots:e}){const n=so(),s=Fd();return()=>{const r=e.default&&Za(e.default(),!0);if(!r||!r.length)return;let i=r[0];if(r.length>1){for(const f of r)if(f.type!==pt){i=f;break}}const o=pe(t),{mode:a}=o;if(s.isLeaving)return So(i);const l=Gl(i);if(!l)return So(i);const u=fr(l,o,s,n);hr(l,u);const c=n.subTree,d=c&&Gl(c);if(d&&d.type!==pt&&!Gn(l,d)){const f=fr(d,o,s,n);if(hr(d,f),a==="out-in")return s.isLeaving=!0,f.afterLeave=()=>{s.isLeaving=!1,n.update.active!==!1&&(n.effect.dirty=!0,n.update())},So(i);a==="in-out"&&l.type!==pt&&(f.delayLeave=(h,g,y)=>{const v=Ld(s,d);v[String(d.key)]=d,h[En]=()=>{g(),h[En]=void 0,delete u.delayedLeave},u.delayedLeave=y})}return i}}},_g=wg;function Ld(t,e){const{leavingVNodes:n}=t;let s=n.get(e.type);return s||(s=Object.create(null),n.set(e.type,s)),s}function fr(t,e,n,s){const{appear:r,mode:i,persisted:o=!1,onBeforeEnter:a,onEnter:l,onAfterEnter:u,onEnterCancelled:c,onBeforeLeave:d,onLeave:f,onAfterLeave:h,onLeaveCancelled:g,onBeforeAppear:y,onAppear:v,onAfterAppear:S,onAppearCancelled:m}=e,p=String(t.key),w=Ld(n,t),b=(M,T)=>{M&&Mt(M,s,9,T)},x=(M,T)=>{const V=T[1];b(M,T),oe(M)?M.every(L=>L.length<=1)&&V():M.length<=1&&V()},C={mode:i,persisted:o,beforeEnter(M){let T=a;if(!n.isMounted)if(r)T=y||a;else return;M[En]&&M[En](!0);const V=w[p];V&&Gn(t,V)&&V.el[En]&&V.el[En](),b(T,[M])},enter(M){let T=l,V=u,L=c;if(!n.isMounted)if(r)T=v||l,V=S||u,L=m||c;else return;let A=!1;const k=M[$r]=O=>{A||(A=!0,O?b(L,[M]):b(V,[M]),C.delayedLeave&&C.delayedLeave(),M[$r]=void 0)};T?x(T,[M,k]):k()},leave(M,T){const V=String(t.key);if(M[$r]&&M[$r](!0),n.isUnmounting)return T();b(d,[M]);let L=!1;const A=M[En]=k=>{L||(L=!0,T(),k?b(g,[M]):b(h,[M]),M[En]=void 0,w[V]===t&&delete w[V])};w[V]=t,f?x(f,[M,A]):A()},clone(M){return fr(M,e,n,s)}};return C}function So(t){if(Cr(t))return t=Vn(t),t.children=null,t}function Gl(t){return Cr(t)?t.children?t.children[0]:void 0:t}function hr(t,e){t.shapeFlag&6&&t.component?hr(t.component.subTree,e):t.shapeFlag&128?(t.ssContent.transition=e.clone(t.ssContent),t.ssFallback.transition=e.clone(t.ssFallback)):t.transition=e}function Za(t,e=!1,n){let s=[],r=0;for(let i=0;i<t.length;i++){let o=t[i];const a=n==null?o.key:String(n)+String(o.key!=null?o.key:i);o.type===Re?(o.patchFlag&128&&r++,s=s.concat(Za(o.children,e,a))):(e||o.type!==pt)&&s.push(a!=null?Vn(o,{key:a}):o)}if(r>1)for(let i=0;i<s.length;i++)s[i].patchFlag=-2;return s}/*! #__NO_SIDE_EFFECTS__ */function dn(t,e){return fe(t)?De({name:t.name},e,{setup:t}):t}const er=t=>!!t.type.__asyncLoader;/*! #__NO_SIDE_EFFECTS__ */function ra(t){fe(t)&&(t={loader:t});const{loader:e,loadingComponent:n,errorComponent:s,delay:r=200,timeout:i,suspensible:o=!0,onError:a}=t;let l=null,u,c=0;const d=()=>(c++,l=null,f()),f=()=>{let h;return l||(h=l=e().catch(g=>{if(g=g instanceof Error?g:new Error(String(g)),a)return new Promise((y,v)=>{a(g,()=>y(d()),()=>v(g),c+1)});throw g}).then(g=>h!==l&&l?l:(g&&(g.__esModule||g[Symbol.toStringTag]==="Module")&&(g=g.default),u=g,g)))};return dn({name:"AsyncComponentWrapper",__asyncLoader:f,get __asyncResolved(){return u},setup(){const h=je;if(u)return()=>Co(u,h);const g=m=>{l=null,Sr(m,h,13,!s)};if(o&&h.suspense||Ar)return f().then(m=>()=>Co(m,h)).catch(m=>(g(m),()=>s?_(s,{error:m}):null));const y=se(!1),v=se(),S=se(!!r);return r&&setTimeout(()=>{S.value=!1},r),i!=null&&setTimeout(()=>{if(!y.value&&!v.value){const m=new Error(`Async component timed out after ${i}ms.`);g(m),v.value=m}},i),f().then(()=>{y.value=!0,h.parent&&Cr(h.parent.vnode)&&(h.parent.effect.dirty=!0,Ki(h.parent.update))}).catch(m=>{g(m),v.value=m}),()=>{if(y.value&&u)return Co(u,h);if(v.value&&s)return _(s,{error:v.value});if(n&&!S.value)return _(n)}}})}function Co(t,e){const{ref:n,props:s,children:r,ce:i}=e.vnode,o=_(t,s,r);return o.ref=n,o.ce=i,delete e.vnode.ce,o}const Cr=t=>t.type.__isKeepAlive;function Pd(t,e){Nd(t,"a",e)}function Od(t,e){Nd(t,"da",e)}function Nd(t,e,n=je){const s=t.__wdc||(t.__wdc=()=>{let r=n;for(;r;){if(r.isDeactivated)return;r=r.parent}return t()});if(eo(e,s,n),n){let r=n.parent;for(;r&&r.parent;)Cr(r.parent.vnode)&&Sg(s,e,n,r),r=r.parent}}function Sg(t,e,n,s){const r=eo(e,t,s,!0);Bd(()=>{Na(s[e],r)},n)}function eo(t,e,n=je,s=!1){if(n){const r=n[t]||(n[t]=[]),i=e.__weh||(e.__weh=(...o)=>{if(n.isUnmounted)return;is();const a=Tr(n),l=Mt(e,n,t,o);return a(),os(),l});return s?r.unshift(i):r.push(i),i}}const fn=t=>(e,n=je)=>(!Ar||t==="sp")&&eo(t,(...s)=>e(...s),n),to=fn("bm"),Yt=fn("m"),Cg=fn("bu"),Ka=fn("u"),Rt=fn("bum"),Bd=fn("um"),Eg=fn("sp"),Mg=fn("rtg"),Tg=fn("rtc");function Ag(t,e=je){eo("ec",t,e)}function Rg(t,e,n,s){let r;const i=n&&n[s];if(oe(t)||Ne(t)){r=new Array(t.length);for(let o=0,a=t.length;o<a;o++)r[o]=e(t[o],o,void 0,i&&i[o])}else if(typeof t=="number"){r=new Array(t);for(let o=0;o<t;o++)r[o]=e(o+1,o,void 0,i&&i[o])}else if(Ie(t))if(t[Symbol.iterator])r=Array.from(t,(o,a)=>e(o,a,void 0,i&&i[a]));else{const o=Object.keys(t);r=new Array(o.length);for(let a=0,l=o.length;a<l;a++){const u=o[a];r[a]=e(t[u],u,a,i&&i[a])}}else r=[];return n&&(n[s]=r),r}function Gs(t,e,n={},s,r){if($e.isCE||$e.parent&&er($e.parent)&&$e.parent.isCE)return e!=="default"&&(n.name=e),_("slot",n,s&&s());let i=t[e];i&&i._c&&(i._d=!1),at();const o=i&&zd(i(n)),a=ln(Re,{key:n.key||o&&o.key||`_${e}`},o||(s?s():[]),o&&t._===1?64:-2);return!r&&a.scopeId&&(a.slotScopeIds=[a.scopeId+"-s"]),i&&i._c&&(i._d=!0),a}function zd(t){return t.some(e=>gr(e)?!(e.type===pt||e.type===Re&&!zd(e.children)):!0)?t:null}const ia=t=>t?Kd(t)?ro(t)||t.proxy:ia(t.parent):null,tr=De(Object.create(null),{$:t=>t,$el:t=>t.vnode.el,$data:t=>t.data,$props:t=>t.props,$attrs:t=>t.attrs,$slots:t=>t.slots,$refs:t=>t.refs,$parent:t=>ia(t.parent),$root:t=>ia(t.root),$emit:t=>t.emit,$options:t=>Ja(t),$forceUpdate:t=>t.f||(t.f=()=>{t.effect.dirty=!0,Ki(t.update)}),$nextTick:t=>t.n||(t.n=xt.bind(t.proxy)),$watch:t=>xg.bind(t)}),Eo=(t,e)=>t!==Ae&&!t.__isScriptSetup&&_e(t,e),Ig={get({_:t},e){const{ctx:n,setupState:s,data:r,props:i,accessCache:o,type:a,appContext:l}=t;let u;if(e[0]!=="$"){const h=o[e];if(h!==void 0)switch(h){case 1:return s[e];case 2:return r[e];case 4:return n[e];case 3:return i[e]}else{if(Eo(s,e))return o[e]=1,s[e];if(r!==Ae&&_e(r,e))return o[e]=2,r[e];if((u=t.propsOptions[0])&&_e(u,e))return o[e]=3,i[e];if(n!==Ae&&_e(n,e))return o[e]=4,n[e];oa&&(o[e]=0)}}const c=tr[e];let d,f;if(c)return e==="$attrs"&&lt(t,"get",e),c(t);if((d=a.__cssModules)&&(d=d[e]))return d;if(n!==Ae&&_e(n,e))return o[e]=4,n[e];if(f=l.config.globalProperties,_e(f,e))return f[e]},set({_:t},e,n){const{data:s,setupState:r,ctx:i}=t;return Eo(r,e)?(r[e]=n,!0):s!==Ae&&_e(s,e)?(s[e]=n,!0):_e(t.props,e)||e[0]==="$"&&e.slice(1)in t?!1:(i[e]=n,!0)},has({_:{data:t,setupState:e,accessCache:n,ctx:s,appContext:r,propsOptions:i}},o){let a;return!!n[o]||t!==Ae&&_e(t,o)||Eo(e,o)||(a=i[0])&&_e(a,o)||_e(s,o)||_e(tr,o)||_e(r.config.globalProperties,o)},defineProperty(t,e,n){return n.get!=null?t._.accessCache[e]=0:_e(n,"value")&&this.set(t,e,n.value,null),Reflect.defineProperty(t,e,n)}};function xi(t){return oe(t)?t.reduce((e,n)=>(e[n]=null,e),{}):t}function ql(t,e){return!t||!e?t||e:oe(t)&&oe(e)?t.concat(e):De({},xi(t),xi(e))}let oa=!0;function kg(t){const e=Ja(t),n=t.proxy,s=t.ctx;oa=!1,e.beforeCreate&&Yl(e.beforeCreate,t,"bc");const{data:r,computed:i,methods:o,watch:a,provide:l,inject:u,created:c,beforeMount:d,mounted:f,beforeUpdate:h,updated:g,activated:y,deactivated:v,beforeDestroy:S,beforeUnmount:m,destroyed:p,unmounted:w,render:b,renderTracked:x,renderTriggered:C,errorCaptured:M,serverPrefetch:T,expose:V,inheritAttrs:L,components:A,directives:k,filters:O}=e;if(u&&Fg(u,s,null),o)for(const j in o){const Y=o[j];fe(Y)&&(s[j]=Y.bind(n))}if(r){const j=r.call(n,n);Ie(j)&&(t.data=tt(j))}if(oa=!0,i)for(const j in i){const Y=i[j],ae=fe(Y)?Y.bind(n,n):fe(Y.get)?Y.get.bind(n,n):Ct,Q=!fe(Y)&&fe(Y.set)?Y.set.bind(n):Ct,te=I({get:ae,set:Q});Object.defineProperty(s,j,{enumerable:!0,configurable:!0,get:()=>te.value,set:ge=>te.value=ge})}if(a)for(const j in a)Dd(a[j],s,n,j);if(l){const j=fe(l)?l.call(n):l;Reflect.ownKeys(j).forEach(Y=>{ut(Y,j[Y])})}c&&Yl(c,t,"c");function q(j,Y){oe(Y)?Y.forEach(ae=>j(ae.bind(n))):Y&&j(Y.bind(n))}if(q(to,d),q(Yt,f),q(Cg,h),q(Ka,g),q(Pd,y),q(Od,v),q(Ag,M),q(Tg,x),q(Mg,C),q(Rt,m),q(Bd,w),q(Eg,T),oe(V))if(V.length){const j=t.exposed||(t.exposed={});V.forEach(Y=>{Object.defineProperty(j,Y,{get:()=>n[Y],set:ae=>n[Y]=ae})})}else t.exposed||(t.exposed={});b&&t.render===Ct&&(t.render=b),L!=null&&(t.inheritAttrs=L),A&&(t.components=A),k&&(t.directives=k)}function Fg(t,e,n=Ct){oe(t)&&(t=aa(t));for(const s in t){const r=t[s];let i;Ie(r)?"default"in r?i=Pe(r.from||s,r.default,!0):i=Pe(r.from||s):i=Pe(r),Te(i)?Object.defineProperty(e,s,{enumerable:!0,configurable:!0,get:()=>i.value,set:o=>i.value=o}):e[s]=i}}function Yl(t,e,n){Mt(oe(t)?t.map(s=>s.bind(e.proxy)):t.bind(e.proxy),e,n)}function Dd(t,e,n,s){const r=s.includes(".")?kd(n,s):()=>n[s];if(Ne(t)){const i=e[t];fe(i)&&le(r,i)}else if(fe(t))le(r,t.bind(n));else if(Ie(t))if(oe(t))t.forEach(i=>Dd(i,e,n,s));else{const i=fe(t.handler)?t.handler.bind(n):e[t.handler];fe(i)&&le(r,i,t)}}function Ja(t){const e=t.type,{mixins:n,extends:s}=e,{mixins:r,optionsCache:i,config:{optionMergeStrategies:o}}=t.appContext,a=i.get(e);let l;return a?l=a:!r.length&&!n&&!s?l=e:(l={},r.length&&r.forEach(u=>wi(l,u,o,!0)),wi(l,e,o)),Ie(e)&&i.set(e,l),l}function wi(t,e,n,s=!1){const{mixins:r,extends:i}=e;i&&wi(t,i,n,!0),r&&r.forEach(o=>wi(t,o,n,!0));for(const o in e)if(!(s&&o==="expose")){const a=Vg[o]||n&&n[o];t[o]=a?a(t[o],e[o]):e[o]}return t}const Vg={data:Xl,props:Zl,emits:Zl,methods:qs,computed:qs,beforeCreate:et,created:et,beforeMount:et,mounted:et,beforeUpdate:et,updated:et,beforeDestroy:et,beforeUnmount:et,destroyed:et,unmounted:et,activated:et,deactivated:et,errorCaptured:et,serverPrefetch:et,components:qs,directives:qs,watch:Pg,provide:Xl,inject:Lg};function Xl(t,e){return e?t?function(){return De(fe(t)?t.call(this,this):t,fe(e)?e.call(this,this):e)}:e:t}function Lg(t,e){return qs(aa(t),aa(e))}function aa(t){if(oe(t)){const e={};for(let n=0;n<t.length;n++)e[t[n]]=t[n];return e}return t}function et(t,e){return t?[...new Set([].concat(t,e))]:e}function qs(t,e){return t?De(Object.create(null),t,e):e}function Zl(t,e){return t?oe(t)&&oe(e)?[...new Set([...t,...e])]:De(Object.create(null),xi(t),xi(e??{})):e}function Pg(t,e){if(!t)return e;if(!e)return t;const n=De(Object.create(null),t);for(const s in e)n[s]=et(t[s],e[s]);return n}function $d(){return{app:null,config:{isNativeTag:mm,performance:!1,globalProperties:{},optionMergeStrategies:{},errorHandler:void 0,warnHandler:void 0,compilerOptions:{}},mixins:[],components:{},directives:{},provides:Object.create(null),optionsCache:new WeakMap,propsCache:new WeakMap,emitsCache:new WeakMap}}let Og=0;function Ng(t,e){return function(s,r=null){fe(s)||(s=De({},s)),r!=null&&!Ie(r)&&(r=null);const i=$d(),o=new WeakSet;let a=!1;const l=i.app={_uid:Og++,_component:s,_props:r,_container:null,_context:i,_instance:null,version:uv,get config(){return i.config},set config(u){},use(u,...c){return o.has(u)||(u&&fe(u.install)?(o.add(u),u.install(l,...c)):fe(u)&&(o.add(u),u(l,...c))),l},mixin(u){return i.mixins.includes(u)||i.mixins.push(u),l},component(u,c){return c?(i.components[u]=c,l):i.components[u]},directive(u,c){return c?(i.directives[u]=c,l):i.directives[u]},mount(u,c,d){if(!a){const f=_(s,r);return f.appContext=i,d===!0?d="svg":d===!1&&(d=void 0),c&&e?e(f,u):t(f,u,d),a=!0,l._container=u,u.__vue_app__=l,ro(f.component)||f.component.proxy}},unmount(){a&&(t(null,l._container),delete l._container.__vue_app__)},provide(u,c){return i.provides[u]=c,l},runWithContext(u){const c=nr;nr=l;try{return u()}finally{nr=c}}};return l}}let nr=null;function ut(t,e){if(je){let n=je.provides;const s=je.parent&&je.parent.provides;s===n&&(n=je.provides=Object.create(s)),n[t]=e}}function Pe(t,e,n=!1){const s=je||$e;if(s||nr){const r=s?s.parent==null?s.vnode.appContext&&s.vnode.appContext.provides:s.parent.provides:nr._context.provides;if(r&&t in r)return r[t];if(arguments.length>1)return n&&fe(e)?e.call(s&&s.proxy):e}}function Bg(t,e,n,s=!1){const r={},i={};vi(i,no,1),t.propsDefaults=Object.create(null),Hd(t,e,r,i);for(const o in t.propsOptions[0])o in r||(r[o]=void 0);n?t.props=s?r:Xm(r):t.type.props?t.props=r:t.props=i,t.attrs=i}function zg(t,e,n,s){const{props:r,attrs:i,vnode:{patchFlag:o}}=t,a=pe(r),[l]=t.propsOptions;let u=!1;if((s||o>0)&&!(o&16)){if(o&8){const c=t.vnode.dynamicProps;for(let d=0;d<c.length;d++){let f=c[d];if(Ji(t.emitsOptions,f))continue;const h=e[f];if(l)if(_e(i,f))h!==i[f]&&(i[f]=h,u=!0);else{const g=bt(f);r[g]=la(l,a,g,h,t,!1)}else h!==i[f]&&(i[f]=h,u=!0)}}}else{Hd(t,e,r,i)&&(u=!0);let c;for(const d in a)(!e||!_e(e,d)&&((c=rs(d))===d||!_e(e,c)))&&(l?n&&(n[d]!==void 0||n[c]!==void 0)&&(r[d]=la(l,a,d,void 0,t,!0)):delete r[d]);if(i!==a)for(const d in i)(!e||!_e(e,d))&&(delete i[d],u=!0)}u&&rn(t,"set","$attrs")}function Hd(t,e,n,s){const[r,i]=t.propsOptions;let o=!1,a;if(e)for(let l in e){if(Qs(l))continue;const u=e[l];let c;r&&_e(r,c=bt(l))?!i||!i.includes(c)?n[c]=u:(a||(a={}))[c]=u:Ji(t.emitsOptions,l)||(!(l in s)||u!==s[l])&&(s[l]=u,o=!0)}if(i){const l=pe(n),u=a||Ae;for(let c=0;c<i.length;c++){const d=i[c];n[d]=la(r,l,d,u[d],t,!_e(u,d))}}return o}function la(t,e,n,s,r,i){const o=t[n];if(o!=null){const a=_e(o,"default");if(a&&s===void 0){const l=o.default;if(o.type!==Function&&!o.skipFactory&&fe(l)){const{propsDefaults:u}=r;if(n in u)s=u[n];else{const c=Tr(r);s=u[n]=l.call(null,e),c()}}else s=l}o[0]&&(i&&!a?s=!1:o[1]&&(s===""||s===rs(n))&&(s=!0))}return s}function Ud(t,e,n=!1){const s=e.propsCache,r=s.get(t);if(r)return r;const i=t.props,o={},a=[];let l=!1;if(!fe(t)){const c=d=>{l=!0;const[f,h]=Ud(d,e,!0);De(o,f),h&&a.push(...h)};!n&&e.mixins.length&&e.mixins.forEach(c),t.extends&&c(t.extends),t.mixins&&t.mixins.forEach(c)}if(!i&&!l)return Ie(t)&&s.set(t,Ms),Ms;if(oe(i))for(let c=0;c<i.length;c++){const d=bt(i[c]);Kl(d)&&(o[d]=Ae)}else if(i)for(const c in i){const d=bt(c);if(Kl(d)){const f=i[c],h=o[d]=oe(f)||fe(f)?{type:f}:De({},f);if(h){const g=eu(Boolean,h.type),y=eu(String,h.type);h[0]=g>-1,h[1]=y<0||g<y,(g>-1||_e(h,"default"))&&a.push(d)}}}const u=[o,a];return Ie(t)&&s.set(t,u),u}function Kl(t){return t[0]!=="$"&&!Qs(t)}function Jl(t){return t===null?"null":typeof t=="function"?t.name||"":typeof t=="object"&&t.constructor&&t.constructor.name||""}function Ql(t,e){return Jl(t)===Jl(e)}function eu(t,e){return oe(e)?e.findIndex(n=>Ql(n,t)):fe(e)&&Ql(e,t)?0:-1}const jd=t=>t[0]==="_"||t==="$stable",Qa=t=>oe(t)?t.map(Dt):[Dt(t)],Dg=(t,e,n)=>{if(e._n)return e;const s=X((...r)=>Qa(e(...r)),n);return s._c=!1,s},Wd=(t,e,n)=>{const s=t._ctx;for(const r in t){if(jd(r))continue;const i=t[r];if(fe(i))e[r]=Dg(r,i,s);else if(i!=null){const o=Qa(i);e[r]=()=>o}}},Gd=(t,e)=>{const n=Qa(e);t.slots.default=()=>n},$g=(t,e)=>{if(t.vnode.shapeFlag&32){const n=e._;n?(t.slots=pe(e),vi(e,"_",n)):Wd(e,t.slots={})}else t.slots={},e&&Gd(t,e);vi(t.slots,no,1)},Hg=(t,e,n)=>{const{vnode:s,slots:r}=t;let i=!0,o=Ae;if(s.shapeFlag&32){const a=e._;a?n&&a===1?i=!1:(De(r,e),!n&&a===1&&delete r._):(i=!e.$stable,Wd(e,r)),o=e}else e&&(Gd(t,e),o={default:1});if(i)for(const a in r)!jd(a)&&o[a]==null&&delete r[a]};function ua(t,e,n,s,r=!1){if(oe(t)){t.forEach((f,h)=>ua(f,e&&(oe(e)?e[h]:e),n,s,r));return}if(er(s)&&!r)return;const i=s.shapeFlag&4?ro(s.component)||s.component.proxy:s.el,o=r?null:i,{i:a,r:l}=t,u=e&&e.r,c=a.refs===Ae?a.refs={}:a.refs,d=a.setupState;if(u!=null&&u!==l&&(Ne(u)?(c[u]=null,_e(d,u)&&(d[u]=null)):Te(u)&&(u.value=null)),fe(l))Rn(l,a,12,[o,c]);else{const f=Ne(l),h=Te(l);if(f||h){const g=()=>{if(t.f){const y=f?_e(d,l)?d[l]:c[l]:l.value;r?oe(y)&&Na(y,i):oe(y)?y.includes(i)||y.push(i):f?(c[l]=[i],_e(d,l)&&(d[l]=c[l])):(l.value=[i],t.k&&(c[t.k]=l.value))}else f?(c[l]=o,_e(d,l)&&(d[l]=o)):h&&(l.value=o,t.k&&(c[t.k]=o))};o?(g.id=-1,rt(g,n)):g()}}}const rt=vg;function Ug(t){return jg(t)}function jg(t,e){const n=td();n.__VUE__=!0;const{insert:s,remove:r,patchProp:i,createElement:o,createText:a,createComment:l,setText:u,setElementText:c,parentNode:d,nextSibling:f,setScopeId:h=Ct,insertStaticContent:g}=t,y=(E,R,P,N=null,B=null,$=null,W=void 0,D=null,U=!!R.dynamicChildren)=>{if(E===R)return;E&&!Gn(E,R)&&(N=xe(E),ge(E,B,$,!0),E=null),R.patchFlag===-2&&(U=!1,R.dynamicChildren=null);const{type:z,ref:K,shapeFlag:re}=R;switch(z){case Er:v(E,R,P,N);break;case pt:S(E,R,P,N);break;case To:E==null&&m(R,P,N,W);break;case Re:A(E,R,P,N,B,$,W,D,U);break;default:re&1?b(E,R,P,N,B,$,W,D,U):re&6?k(E,R,P,N,B,$,W,D,U):(re&64||re&128)&&z.process(E,R,P,N,B,$,W,D,U,_t)}K!=null&&B&&ua(K,E&&E.ref,$,R||E,!R)},v=(E,R,P,N)=>{if(E==null)s(R.el=a(R.children),P,N);else{const B=R.el=E.el;R.children!==E.children&&u(B,R.children)}},S=(E,R,P,N)=>{E==null?s(R.el=l(R.children||""),P,N):R.el=E.el},m=(E,R,P,N)=>{[E.el,E.anchor]=g(E.children,R,P,N,E.el,E.anchor)},p=({el:E,anchor:R},P,N)=>{let B;for(;E&&E!==R;)B=f(E),s(E,P,N),E=B;s(R,P,N)},w=({el:E,anchor:R})=>{let P;for(;E&&E!==R;)P=f(E),r(E),E=P;r(R)},b=(E,R,P,N,B,$,W,D,U)=>{R.type==="svg"?W="svg":R.type==="math"&&(W="mathml"),E==null?x(R,P,N,B,$,W,D,U):T(E,R,B,$,W,D,U)},x=(E,R,P,N,B,$,W,D)=>{let U,z;const{props:K,shapeFlag:re,transition:ee,dirs:ce}=E;if(U=E.el=o(E.type,$,K&&K.is,K),re&8?c(U,E.children):re&16&&M(E.children,U,null,N,B,Mo(E,$),W,D),ce&&Bn(E,null,N,"created"),C(U,E,E.scopeId,W,N),K){for(const Me in K)Me!=="value"&&!Qs(Me)&&i(U,Me,null,K[Me],$,E.children,N,B,Z);"value"in K&&i(U,"value",null,K.value,$),(z=K.onVnodeBeforeMount)&&zt(z,N,E)}ce&&Bn(E,null,N,"beforeMount");const ve=Wg(B,ee);ve&&ee.beforeEnter(U),s(U,R,P),((z=K&&K.onVnodeMounted)||ve||ce)&&rt(()=>{z&&zt(z,N,E),ve&&ee.enter(U),ce&&Bn(E,null,N,"mounted")},B)},C=(E,R,P,N,B)=>{if(P&&h(E,P),N)for(let $=0;$<N.length;$++)h(E,N[$]);if(B){let $=B.subTree;if(R===$){const W=B.vnode;C(E,W,W.scopeId,W.slotScopeIds,B.parent)}}},M=(E,R,P,N,B,$,W,D,U=0)=>{for(let z=U;z<E.length;z++){const K=E[z]=D?Mn(E[z]):Dt(E[z]);y(null,K,R,P,N,B,$,W,D)}},T=(E,R,P,N,B,$,W)=>{const D=R.el=E.el;let{patchFlag:U,dynamicChildren:z,dirs:K}=R;U|=E.patchFlag&16;const re=E.props||Ae,ee=R.props||Ae;let ce;if(P&&zn(P,!1),(ce=ee.onVnodeBeforeUpdate)&&zt(ce,P,R,E),K&&Bn(R,E,P,"beforeUpdate"),P&&zn(P,!0),z?V(E.dynamicChildren,z,D,P,N,Mo(R,B),$):W||Y(E,R,D,null,P,N,Mo(R,B),$,!1),U>0){if(U&16)L(D,R,re,ee,P,N,B);else if(U&2&&re.class!==ee.class&&i(D,"class",null,ee.class,B),U&4&&i(D,"style",re.style,ee.style,B),U&8){const ve=R.dynamicProps;for(let Me=0;Me<ve.length;Me++){const Fe=ve[Me],Ue=re[Fe],kt=ee[Fe];(kt!==Ue||Fe==="value")&&i(D,Fe,Ue,kt,B,E.children,P,N,Z)}}U&1&&E.children!==R.children&&c(D,R.children)}else!W&&z==null&&L(D,R,re,ee,P,N,B);((ce=ee.onVnodeUpdated)||K)&&rt(()=>{ce&&zt(ce,P,R,E),K&&Bn(R,E,P,"updated")},N)},V=(E,R,P,N,B,$,W)=>{for(let D=0;D<R.length;D++){const U=E[D],z=R[D],K=U.el&&(U.type===Re||!Gn(U,z)||U.shapeFlag&70)?d(U.el):P;y(U,z,K,null,N,B,$,W,!0)}},L=(E,R,P,N,B,$,W)=>{if(P!==N){if(P!==Ae)for(const D in P)!Qs(D)&&!(D in N)&&i(E,D,P[D],null,W,R.children,B,$,Z);for(const D in N){if(Qs(D))continue;const U=N[D],z=P[D];U!==z&&D!=="value"&&i(E,D,z,U,W,R.children,B,$,Z)}"value"in N&&i(E,"value",P.value,N.value,W)}},A=(E,R,P,N,B,$,W,D,U)=>{const z=R.el=E?E.el:a(""),K=R.anchor=E?E.anchor:a("");let{patchFlag:re,dynamicChildren:ee,slotScopeIds:ce}=R;ce&&(D=D?D.concat(ce):ce),E==null?(s(z,P,N),s(K,P,N),M(R.children||[],P,K,B,$,W,D,U)):re>0&&re&64&&ee&&E.dynamicChildren?(V(E.dynamicChildren,ee,P,B,$,W,D),(R.key!=null||B&&R===B.subTree)&&el(E,R,!0)):Y(E,R,P,K,B,$,W,D,U)},k=(E,R,P,N,B,$,W,D,U)=>{R.slotScopeIds=D,E==null?R.shapeFlag&512?B.ctx.activate(R,P,N,W,U):O(R,P,N,B,$,W,U):H(E,R,U)},O=(E,R,P,N,B,$,W)=>{const D=E.component=nv(E,N,B);if(Cr(E)&&(D.ctx.renderer=_t),sv(D),D.asyncDep){if(B&&B.registerDep(D,q),!E.el){const U=D.subTree=_(pt);S(null,U,R,P)}}else q(D,E,R,P,B,$,W)},H=(E,R,P)=>{const N=R.component=E.component;if(dg(E,R,P))if(N.asyncDep&&!N.asyncResolved){j(N,R,P);return}else N.next=R,ig(N.update),N.effect.dirty=!0,N.update();else R.el=E.el,N.vnode=R},q=(E,R,P,N,B,$,W)=>{const D=()=>{if(E.isMounted){let{next:K,bu:re,u:ee,parent:ce,vnode:ve}=E;{const hs=qd(E);if(hs){K&&(K.el=ve.el,j(E,K,W)),hs.asyncDep.then(()=>{E.isUnmounted||D()});return}}let Me=K,Fe;zn(E,!1),K?(K.el=ve.el,j(E,K,W)):K=ve,re&&xo(re),(Fe=K.props&&K.props.onVnodeBeforeUpdate)&&zt(Fe,ce,K,ve),zn(E,!0);const Ue=_o(E),kt=E.subTree;E.subTree=Ue,y(kt,Ue,d(kt.el),xe(kt),E,B,$),K.el=Ue.el,Me===null&&fg(E,Ue.el),ee&&rt(ee,B),(Fe=K.props&&K.props.onVnodeUpdated)&&rt(()=>zt(Fe,ce,K,ve),B)}else{let K;const{el:re,props:ee}=R,{bm:ce,m:ve,parent:Me}=E,Fe=er(R);if(zn(E,!1),ce&&xo(ce),!Fe&&(K=ee&&ee.onVnodeBeforeMount)&&zt(K,Me,R),zn(E,!0),re&&Nn){const Ue=()=>{E.subTree=_o(E),Nn(re,E.subTree,E,B,null)};Fe?R.type.__asyncLoader().then(()=>!E.isUnmounted&&Ue()):Ue()}else{const Ue=E.subTree=_o(E);y(null,Ue,P,N,E,B,$),R.el=Ue.el}if(ve&&rt(ve,B),!Fe&&(K=ee&&ee.onVnodeMounted)){const Ue=R;rt(()=>zt(K,Me,Ue),B)}(R.shapeFlag&256||Me&&er(Me.vnode)&&Me.vnode.shapeFlag&256)&&E.a&&rt(E.a,B),E.isMounted=!0,R=P=N=null}},U=E.effect=new Da(D,Ct,()=>Ki(z),E.scope),z=E.update=()=>{U.dirty&&U.run()};z.id=E.uid,zn(E,!0),z()},j=(E,R,P)=>{R.component=E;const N=E.vnode.props;E.vnode=R,E.next=null,zg(E,R.props,N,P),Hg(E,R.children,P),is(),Ul(E),os()},Y=(E,R,P,N,B,$,W,D,U=!1)=>{const z=E&&E.children,K=E?E.shapeFlag:0,re=R.children,{patchFlag:ee,shapeFlag:ce}=R;if(ee>0){if(ee&128){Q(z,re,P,N,B,$,W,D,U);return}else if(ee&256){ae(z,re,P,N,B,$,W,D,U);return}}ce&8?(K&16&&Z(z,B,$),re!==z&&c(P,re)):K&16?ce&16?Q(z,re,P,N,B,$,W,D,U):Z(z,B,$,!0):(K&8&&c(P,""),ce&16&&M(re,P,N,B,$,W,D,U))},ae=(E,R,P,N,B,$,W,D,U)=>{E=E||Ms,R=R||Ms;const z=E.length,K=R.length,re=Math.min(z,K);let ee;for(ee=0;ee<re;ee++){const ce=R[ee]=U?Mn(R[ee]):Dt(R[ee]);y(E[ee],ce,P,null,B,$,W,D,U)}z>K?Z(E,B,$,!0,!1,re):M(R,P,N,B,$,W,D,U,re)},Q=(E,R,P,N,B,$,W,D,U)=>{let z=0;const K=R.length;let re=E.length-1,ee=K-1;for(;z<=re&&z<=ee;){const ce=E[z],ve=R[z]=U?Mn(R[z]):Dt(R[z]);if(Gn(ce,ve))y(ce,ve,P,null,B,$,W,D,U);else break;z++}for(;z<=re&&z<=ee;){const ce=E[re],ve=R[ee]=U?Mn(R[ee]):Dt(R[ee]);if(Gn(ce,ve))y(ce,ve,P,null,B,$,W,D,U);else break;re--,ee--}if(z>re){if(z<=ee){const ce=ee+1,ve=ce<K?R[ce].el:N;for(;z<=ee;)y(null,R[z]=U?Mn(R[z]):Dt(R[z]),P,ve,B,$,W,D,U),z++}}else if(z>ee)for(;z<=re;)ge(E[z],B,$,!0),z++;else{const ce=z,ve=z,Me=new Map;for(z=ve;z<=ee;z++){const ft=R[z]=U?Mn(R[z]):Dt(R[z]);ft.key!=null&&Me.set(ft.key,z)}let Fe,Ue=0;const kt=ee-ve+1;let hs=!1,kl=0;const zs=new Array(kt);for(z=0;z<kt;z++)zs[z]=0;for(z=ce;z<=re;z++){const ft=E[z];if(Ue>=kt){ge(ft,B,$,!0);continue}let Bt;if(ft.key!=null)Bt=Me.get(ft.key);else for(Fe=ve;Fe<=ee;Fe++)if(zs[Fe-ve]===0&&Gn(ft,R[Fe])){Bt=Fe;break}Bt===void 0?ge(ft,B,$,!0):(zs[Bt-ve]=z+1,Bt>=kl?kl=Bt:hs=!0,y(ft,R[Bt],P,null,B,$,W,D,U),Ue++)}const Fl=hs?Gg(zs):Ms;for(Fe=Fl.length-1,z=kt-1;z>=0;z--){const ft=ve+z,Bt=R[ft],Vl=ft+1<K?R[ft+1].el:N;zs[z]===0?y(null,Bt,P,Vl,B,$,W,D,U):hs&&(Fe<0||z!==Fl[Fe]?te(Bt,P,Vl,2):Fe--)}}},te=(E,R,P,N,B=null)=>{const{el:$,type:W,transition:D,children:U,shapeFlag:z}=E;if(z&6){te(E.component.subTree,R,P,N);return}if(z&128){E.suspense.move(R,P,N);return}if(z&64){W.move(E,R,P,_t);return}if(W===Re){s($,R,P);for(let re=0;re<U.length;re++)te(U[re],R,P,N);s(E.anchor,R,P);return}if(W===To){p(E,R,P);return}if(N!==2&&z&1&&D)if(N===0)D.beforeEnter($),s($,R,P),rt(()=>D.enter($),B);else{const{leave:re,delayLeave:ee,afterLeave:ce}=D,ve=()=>s($,R,P),Me=()=>{re($,()=>{ve(),ce&&ce()})};ee?ee($,ve,Me):Me()}else s($,R,P)},ge=(E,R,P,N=!1,B=!1)=>{const{type:$,props:W,ref:D,children:U,dynamicChildren:z,shapeFlag:K,patchFlag:re,dirs:ee}=E;if(D!=null&&ua(D,null,P,E,!0),K&256){R.ctx.deactivate(E);return}const ce=K&1&&ee,ve=!er(E);let Me;if(ve&&(Me=W&&W.onVnodeBeforeUnmount)&&zt(Me,R,E),K&6)We(E.component,P,N);else{if(K&128){E.suspense.unmount(P,N);return}ce&&Bn(E,null,R,"beforeUnmount"),K&64?E.type.remove(E,R,P,B,_t,N):z&&($!==Re||re>0&&re&64)?Z(z,R,P,!1,!0):($===Re&&re&384||!B&&K&16)&&Z(U,R,P),N&&me(E)}(ve&&(Me=W&&W.onVnodeUnmounted)||ce)&&rt(()=>{Me&&zt(Me,R,E),ce&&Bn(E,null,R,"unmounted")},P)},me=E=>{const{type:R,el:P,anchor:N,transition:B}=E;if(R===Re){Ce(P,N);return}if(R===To){w(E);return}const $=()=>{r(P),B&&!B.persisted&&B.afterLeave&&B.afterLeave()};if(E.shapeFlag&1&&B&&!B.persisted){const{leave:W,delayLeave:D}=B,U=()=>W(P,$);D?D(E.el,$,U):U()}else $()},Ce=(E,R)=>{let P;for(;E!==R;)P=f(E),r(E),E=P;r(R)},We=(E,R,P)=>{const{bum:N,scope:B,update:$,subTree:W,um:D}=E;N&&xo(N),B.stop(),$&&($.active=!1,ge(W,E,R,P)),D&&rt(D,R),rt(()=>{E.isUnmounted=!0},R),R&&R.pendingBranch&&!R.isUnmounted&&E.asyncDep&&!E.asyncResolved&&E.suspenseId===R.pendingId&&(R.deps--,R.deps===0&&R.resolve())},Z=(E,R,P,N=!1,B=!1,$=0)=>{for(let W=$;W<E.length;W++)ge(E[W],R,P,N,B)},xe=E=>E.shapeFlag&6?xe(E.component.subTree):E.shapeFlag&128?E.suspense.next():f(E.anchor||E.el);let gn=!1;const ke=(E,R,P)=>{E==null?R._vnode&&ge(R._vnode,null,null,!0):y(R._vnode||null,E,R,null,null,null,P),gn||(gn=!0,Ul(),Cd(),gn=!1),R._vnode=E},_t={p:y,um:ge,m:te,r:me,mt:O,mc:M,pc:Y,pbc:V,n:xe,o:t};let Qe,Nn;return e&&([Qe,Nn]=e(_t)),{render:ke,hydrate:Qe,createApp:Ng(ke,Qe)}}function Mo({type:t,props:e},n){return n==="svg"&&t==="foreignObject"||n==="mathml"&&t==="annotation-xml"&&e&&e.encoding&&e.encoding.includes("html")?void 0:n}function zn({effect:t,update:e},n){t.allowRecurse=e.allowRecurse=n}function Wg(t,e){return(!t||t&&!t.pendingBranch)&&e&&!e.persisted}function el(t,e,n=!1){const s=t.children,r=e.children;if(oe(s)&&oe(r))for(let i=0;i<s.length;i++){const o=s[i];let a=r[i];a.shapeFlag&1&&!a.dynamicChildren&&((a.patchFlag<=0||a.patchFlag===32)&&(a=r[i]=Mn(r[i]),a.el=o.el),n||el(o,a)),a.type===Er&&(a.el=o.el)}}function Gg(t){const e=t.slice(),n=[0];let s,r,i,o,a;const l=t.length;for(s=0;s<l;s++){const u=t[s];if(u!==0){if(r=n[n.length-1],t[r]<u){e[s]=r,n.push(s);continue}for(i=0,o=n.length-1;i<o;)a=i+o>>1,t[n[a]]<u?i=a+1:o=a;u<t[n[i]]&&(i>0&&(e[s]=n[i-1]),n[i]=s)}}for(i=n.length,o=n[i-1];i-- >0;)n[i]=o,o=e[o];return n}function qd(t){const e=t.subTree.component;if(e)return e.asyncDep&&!e.asyncResolved?e:qd(e)}const qg=t=>t.__isTeleport,sr=t=>t&&(t.disabled||t.disabled===""),tu=t=>typeof SVGElement<"u"&&t instanceof SVGElement,nu=t=>typeof MathMLElement=="function"&&t instanceof MathMLElement,ca=(t,e)=>{const n=t&&t.to;return Ne(n)?e?e(n):null:n},Yg={name:"Teleport",__isTeleport:!0,process(t,e,n,s,r,i,o,a,l,u){const{mc:c,pc:d,pbc:f,o:{insert:h,querySelector:g,createText:y,createComment:v}}=u,S=sr(e.props);let{shapeFlag:m,children:p,dynamicChildren:w}=e;if(t==null){const b=e.el=y(""),x=e.anchor=y("");h(b,n,s),h(x,n,s);const C=e.target=ca(e.props,g),M=e.targetAnchor=y("");C&&(h(M,C),o==="svg"||tu(C)?o="svg":(o==="mathml"||nu(C))&&(o="mathml"));const T=(V,L)=>{m&16&&c(p,V,L,r,i,o,a,l)};S?T(n,x):C&&T(C,M)}else{e.el=t.el;const b=e.anchor=t.anchor,x=e.target=t.target,C=e.targetAnchor=t.targetAnchor,M=sr(t.props),T=M?n:x,V=M?b:C;if(o==="svg"||tu(x)?o="svg":(o==="mathml"||nu(x))&&(o="mathml"),w?(f(t.dynamicChildren,w,T,r,i,o,a),el(t,e,!0)):l||d(t,e,T,V,r,i,o,a,!1),S)M?e.props&&t.props&&e.props.to!==t.props.to&&(e.props.to=t.props.to):Hr(e,n,b,u,1);else if((e.props&&e.props.to)!==(t.props&&t.props.to)){const L=e.target=ca(e.props,g);L&&Hr(e,L,null,u,0)}else M&&Hr(e,x,C,u,1)}Yd(e)},remove(t,e,n,s,{um:r,o:{remove:i}},o){const{shapeFlag:a,children:l,anchor:u,targetAnchor:c,target:d,props:f}=t;if(d&&i(c),o&&i(u),a&16){const h=o||!sr(f);for(let g=0;g<l.length;g++){const y=l[g];r(y,e,n,h,!!y.dynamicChildren)}}},move:Hr,hydrate:Xg};function Hr(t,e,n,{o:{insert:s},m:r},i=2){i===0&&s(t.targetAnchor,e,n);const{el:o,anchor:a,shapeFlag:l,children:u,props:c}=t,d=i===2;if(d&&s(o,e,n),(!d||sr(c))&&l&16)for(let f=0;f<u.length;f++)r(u[f],e,n,2);d&&s(a,e,n)}function Xg(t,e,n,s,r,i,{o:{nextSibling:o,parentNode:a,querySelector:l}},u){const c=e.target=ca(e.props,l);if(c){const d=c._lpa||c.firstChild;if(e.shapeFlag&16)if(sr(e.props))e.anchor=u(o(t),e,a(t),n,s,r,i),e.targetAnchor=d;else{e.anchor=o(t);let f=d;for(;f;)if(f=o(f),f&&f.nodeType===8&&f.data==="teleport anchor"){e.targetAnchor=f,c._lpa=e.targetAnchor&&o(e.targetAnchor);break}u(d,e,c,n,s,r,i)}Yd(e)}return e.anchor&&o(e.anchor)}const Zg=Yg;function Yd(t){const e=t.ctx;if(e&&e.ut){let n=t.children[0].el;for(;n&&n!==t.targetAnchor;)n.nodeType===1&&n.setAttribute("data-v-owner",e.uid),n=n.nextSibling;e.ut()}}const Re=Symbol.for("v-fgt"),Er=Symbol.for("v-txt"),pt=Symbol.for("v-cmt"),To=Symbol.for("v-stc"),rr=[];let Ot=null;function at(t=!1){rr.push(Ot=t?null:[])}function Kg(){rr.pop(),Ot=rr[rr.length-1]||null}let mr=1;function su(t){mr+=t}function Xd(t){return t.dynamicChildren=mr>0?Ot||Ms:null,Kg(),mr>0&&Ot&&Ot.push(t),t}function Mr(t,e,n,s,r,i){return Xd(jt(t,e,n,s,r,i,!0))}function ln(t,e,n,s,r){return Xd(_(t,e,n,s,r,!0))}function gr(t){return t?t.__v_isVNode===!0:!1}function Gn(t,e){return t.type===e.type&&t.key===e.key}const no="__vInternal",Zd=({key:t})=>t??null,ui=({ref:t,ref_key:e,ref_for:n})=>(typeof t=="number"&&(t=""+t),t!=null?Ne(t)||Te(t)||fe(t)?{i:$e,r:t,k:e,f:!!n}:t:null);function jt(t,e=null,n=null,s=0,r=null,i=t===Re?0:1,o=!1,a=!1){const l={__v_isVNode:!0,__v_skip:!0,type:t,props:e,key:e&&Zd(e),ref:e&&ui(e),scopeId:Td,slotScopeIds:null,children:n,component:null,suspense:null,ssContent:null,ssFallback:null,dirs:null,transition:null,el:null,anchor:null,target:null,targetAnchor:null,staticCount:0,shapeFlag:i,patchFlag:s,dynamicProps:r,dynamicChildren:null,appContext:null,ctx:$e};return a?(tl(l,n),i&128&&t.normalize(l)):n&&(l.shapeFlag|=Ne(n)?8:16),mr>0&&!o&&Ot&&(l.patchFlag>0||i&6)&&l.patchFlag!==32&&Ot.push(l),l}const _=Jg;function Jg(t,e=null,n=null,s=0,r=null,i=!1){if((!t||t===Rd)&&(t=pt),gr(t)){const a=Vn(t,e,!0);return n&&tl(a,n),mr>0&&!i&&Ot&&(a.shapeFlag&6?Ot[Ot.indexOf(t)]=a:Ot.push(a)),a.patchFlag|=-2,a}if(lv(t)&&(t=t.__vccOpts),e){e=Qg(e);let{class:a,style:l}=e;a&&!Ne(a)&&(e.class=Xi(a)),Ie(l)&&(vd(l)&&!oe(l)&&(l=De({},l)),e.style=Yi(l))}const o=Ne(t)?1:gg(t)?128:qg(t)?64:Ie(t)?4:fe(t)?2:0;return jt(t,e,n,s,r,o,i,!0)}function Qg(t){return t?vd(t)||no in t?De({},t):t:null}function Vn(t,e,n=!1){const{props:s,ref:r,patchFlag:i,children:o}=t,a=e?Le(s||{},e):s;return{__v_isVNode:!0,__v_skip:!0,type:t.type,props:a,key:a&&Zd(a),ref:e&&e.ref?n&&r?oe(r)?r.concat(ui(e)):[r,ui(e)]:ui(e):r,scopeId:t.scopeId,slotScopeIds:t.slotScopeIds,children:o,target:t.target,targetAnchor:t.targetAnchor,staticCount:t.staticCount,shapeFlag:t.shapeFlag,patchFlag:e&&t.type!==Re?i===-1?16:i|16:i,dynamicProps:t.dynamicProps,dynamicChildren:t.dynamicChildren,appContext:t.appContext,dirs:t.dirs,transition:t.transition,component:t.component,suspense:t.suspense,ssContent:t.ssContent&&Vn(t.ssContent),ssFallback:t.ssFallback&&Vn(t.ssFallback),el:t.el,anchor:t.anchor,ctx:t.ctx,ce:t.ce}}function Ve(t=" ",e=0){return _(Er,null,t,e)}function da(t="",e=!1){return e?(at(),ln(pt,null,t)):_(pt,null,t)}function Dt(t){return t==null||typeof t=="boolean"?_(pt):oe(t)?_(Re,null,t.slice()):typeof t=="object"?Mn(t):_(Er,null,String(t))}function Mn(t){return t.el===null&&t.patchFlag!==-1||t.memo?t:Vn(t)}function tl(t,e){let n=0;const{shapeFlag:s}=t;if(e==null)e=null;else if(oe(e))n=16;else if(typeof e=="object")if(s&65){const r=e.default;r&&(r._c&&(r._d=!1),tl(t,r()),r._c&&(r._d=!0));return}else{n=32;const r=e._;!r&&!(no in e)?e._ctx=$e:r===3&&$e&&($e.slots._===1?e._=1:(e._=2,t.patchFlag|=1024))}else fe(e)?(e={default:e,_ctx:$e},n=32):(e=String(e),s&64?(n=16,e=[Ve(e)]):n=8);t.children=e,t.shapeFlag|=n}function Le(...t){const e={};for(let n=0;n<t.length;n++){const s=t[n];for(const r in s)if(r==="class")e.class!==s.class&&(e.class=Xi([e.class,s.class]));else if(r==="style")e.style=Yi([e.style,s.style]);else if(Wi(r)){const i=e[r],o=s[r];o&&i!==o&&!(oe(i)&&i.includes(o))&&(e[r]=i?[].concat(i,o):o)}else r!==""&&(e[r]=s[r])}return e}function zt(t,e,n,s=null){Mt(t,e,7,[n,s])}const ev=$d();let tv=0;function nv(t,e,n){const s=t.type,r=(e?e.appContext:t.appContext)||ev,i={uid:tv++,vnode:t,type:s,parent:e,appContext:r,root:null,next:null,subTree:null,effect:null,update:null,scope:new rd(!0),render:null,proxy:null,exposed:null,exposeProxy:null,withProxy:null,provides:e?e.provides:Object.create(r.provides),accessCache:null,renderCache:[],components:null,directives:null,propsOptions:Ud(s,r),emitsOptions:Md(s,r),emit:null,emitted:null,propsDefaults:Ae,inheritAttrs:s.inheritAttrs,ctx:Ae,data:Ae,props:Ae,attrs:Ae,slots:Ae,refs:Ae,setupState:Ae,setupContext:null,attrsProxy:null,slotsProxy:null,suspense:n,suspenseId:n?n.pendingId:0,asyncDep:null,asyncResolved:!1,isMounted:!1,isUnmounted:!1,isDeactivated:!1,bc:null,c:null,bm:null,m:null,bu:null,u:null,um:null,bum:null,da:null,a:null,rtg:null,rtc:null,ec:null,sp:null};return i.ctx={_:i},i.root=e?e.root:i,i.emit=lg.bind(null,i),t.ce&&t.ce(i),i}let je=null;const so=()=>je||$e;let _i,fa;{const t=td(),e=(n,s)=>{let r;return(r=t[n])||(r=t[n]=[]),r.push(s),i=>{r.length>1?r.forEach(o=>o(i)):r[0](i)}};_i=e("__VUE_INSTANCE_SETTERS__",n=>je=n),fa=e("__VUE_SSR_SETTERS__",n=>Ar=n)}const Tr=t=>{const e=je;return _i(t),t.scope.on(),()=>{t.scope.off(),_i(e)}},ru=()=>{je&&je.scope.off(),_i(null)};function Kd(t){return t.vnode.shapeFlag&4}let Ar=!1;function sv(t,e=!1){e&&fa(e);const{props:n,children:s}=t.vnode,r=Kd(t);Bg(t,n,r,e),$g(t,s);const i=r?rv(t,e):void 0;return e&&fa(!1),i}function rv(t,e){const n=t.type;t.accessCache=Object.create(null),t.proxy=pd(new Proxy(t.ctx,Ig));const{setup:s}=n;if(s){const r=t.setupContext=s.length>1?ov(t):null,i=Tr(t);is();const o=Rn(s,t,0,[t.props,r]);if(os(),i(),Jc(o)){if(o.then(ru,ru),e)return o.then(a=>{iu(t,a,e)}).catch(a=>{Sr(a,t,0)});t.asyncDep=o}else iu(t,o,e)}else Jd(t,e)}function iu(t,e,n){fe(e)?t.type.__ssrInlineRender?t.ssrRender=e:t.render=e:Ie(e)&&(t.setupState=xd(e)),Jd(t,n)}let ou;function Jd(t,e,n){const s=t.type;if(!t.render){if(!e&&ou&&!s.render){const r=s.template||Ja(t).template;if(r){const{isCustomElement:i,compilerOptions:o}=t.appContext.config,{delimiters:a,compilerOptions:l}=s,u=De(De({isCustomElement:i,delimiters:a},o),l);s.render=ou(r,u)}}t.render=s.render||Ct}{const r=Tr(t);is();try{kg(t)}finally{os(),r()}}}function iv(t){return t.attrsProxy||(t.attrsProxy=new Proxy(t.attrs,{get(e,n){return lt(t,"get","$attrs"),e[n]}}))}function ov(t){const e=n=>{t.exposed=n||{}};return{get attrs(){return iv(t)},slots:t.slots,emit:t.emit,expose:e}}function ro(t){if(t.exposed)return t.exposeProxy||(t.exposeProxy=new Proxy(xd(pd(t.exposed)),{get(e,n){if(n in e)return e[n];if(n in tr)return tr[n](t)},has(e,n){return n in e||n in tr}}))}function av(t,e=!0){return fe(t)?t.displayName||t.name:t.name||e&&t.__name}function lv(t){return fe(t)&&"__vccOpts"in t}const I=(t,e)=>Zm(t,e,Ar);function au(t,e,n=Ae){const s=so(),r=bt(e),i=rs(e),o=eg((l,u)=>{let c;return bg(()=>{const d=t[e];Wt(c,d)&&(c=d,u())}),{get(){return l(),n.get?n.get(c):c},set(d){const f=s.vnode.props;!(f&&(e in f||r in f||i in f)&&(`onUpdate:${e}`in f||`onUpdate:${r}`in f||`onUpdate:${i}`in f))&&Wt(d,c)&&(c=d,u()),s.emit(`update:${e}`,n.set?n.set(d):d)}}}),a=e==="modelValue"?"modelModifiers":`${e}Modifiers`;return o[Symbol.iterator]=()=>{let l=0;return{next(){return l<2?{value:l++?t[a]||{}:o,done:!1}:{done:!0}}}},o}function On(t,e,n){const s=arguments.length;return s===2?Ie(e)&&!oe(e)?gr(e)?_(t,null,[e]):_(t,e):_(t,null,e):(s>3?n=Array.prototype.slice.call(arguments,2):s===3&&gr(n)&&(n=[n]),_(t,e,n))}const uv="3.4.21";/**
* @vue/runtime-dom v3.4.21
* (c) 2018-present Yuxi (Evan) You and Vue contributors
* @license MIT
**/const cv="http://www.w3.org/2000/svg",dv="http://www.w3.org/1998/Math/MathML",Tn=typeof document<"u"?document:null,lu=Tn&&Tn.createElement("template"),fv={insert:(t,e,n)=>{e.insertBefore(t,n||null)},remove:t=>{const e=t.parentNode;e&&e.removeChild(t)},createElement:(t,e,n,s)=>{const r=e==="svg"?Tn.createElementNS(cv,t):e==="mathml"?Tn.createElementNS(dv,t):Tn.createElement(t,n?{is:n}:void 0);return t==="select"&&s&&s.multiple!=null&&r.setAttribute("multiple",s.multiple),r},createText:t=>Tn.createTextNode(t),createComment:t=>Tn.createComment(t),setText:(t,e)=>{t.nodeValue=e},setElementText:(t,e)=>{t.textContent=e},parentNode:t=>t.parentNode,nextSibling:t=>t.nextSibling,querySelector:t=>Tn.querySelector(t),setScopeId(t,e){t.setAttribute(e,"")},insertStaticContent(t,e,n,s,r,i){const o=n?n.previousSibling:e.lastChild;if(r&&(r===i||r.nextSibling))for(;e.insertBefore(r.cloneNode(!0),n),!(r===i||!(r=r.nextSibling)););else{lu.innerHTML=s==="svg"?`<svg>${t}</svg>`:s==="mathml"?`<math>${t}</math>`:t;const a=lu.content;if(s==="svg"||s==="mathml"){const l=a.firstChild;for(;l.firstChild;)a.appendChild(l.firstChild);a.removeChild(l)}e.insertBefore(a,n)}return[o?o.nextSibling:e.firstChild,n?n.previousSibling:e.lastChild]}},pn="transition",Ds="animation",Fs=Symbol("_vtc"),qt=(t,{slots:e})=>On(_g,ef(t),e);qt.displayName="Transition";const Qd={name:String,type:String,css:{type:Boolean,default:!0},duration:[String,Number,Object],enterFromClass:String,enterActiveClass:String,enterToClass:String,appearFromClass:String,appearActiveClass:String,appearToClass:String,leaveFromClass:String,leaveActiveClass:String,leaveToClass:String},hv=qt.props=De({},Vd,Qd),Dn=(t,e=[])=>{oe(t)?t.forEach(n=>n(...e)):t&&t(...e)},uu=t=>t?oe(t)?t.some(e=>e.length>1):t.length>1:!1;function ef(t){const e={};for(const A in t)A in Qd||(e[A]=t[A]);if(t.css===!1)return e;const{name:n="v",type:s,duration:r,enterFromClass:i=`${n}-enter-from`,enterActiveClass:o=`${n}-enter-active`,enterToClass:a=`${n}-enter-to`,appearFromClass:l=i,appearActiveClass:u=o,appearToClass:c=a,leaveFromClass:d=`${n}-leave-from`,leaveActiveClass:f=`${n}-leave-active`,leaveToClass:h=`${n}-leave-to`}=t,g=mv(r),y=g&&g[0],v=g&&g[1],{onBeforeEnter:S,onEnter:m,onEnterCancelled:p,onLeave:w,onLeaveCancelled:b,onBeforeAppear:x=S,onAppear:C=m,onAppearCancelled:M=p}=e,T=(A,k,O)=>{_n(A,k?c:a),_n(A,k?u:o),O&&O()},V=(A,k)=>{A._isLeaving=!1,_n(A,d),_n(A,h),_n(A,f),k&&k()},L=A=>(k,O)=>{const H=A?C:m,q=()=>T(k,A,O);Dn(H,[k,q]),cu(()=>{_n(k,A?l:i),nn(k,A?c:a),uu(H)||du(k,s,y,q)})};return De(e,{onBeforeEnter(A){Dn(S,[A]),nn(A,i),nn(A,o)},onBeforeAppear(A){Dn(x,[A]),nn(A,l),nn(A,u)},onEnter:L(!1),onAppear:L(!0),onLeave(A,k){A._isLeaving=!0;const O=()=>V(A,k);nn(A,d),nf(),nn(A,f),cu(()=>{A._isLeaving&&(_n(A,d),nn(A,h),uu(w)||du(A,s,v,O))}),Dn(w,[A,O])},onEnterCancelled(A){T(A,!1),Dn(p,[A])},onAppearCancelled(A){T(A,!0),Dn(M,[A])},onLeaveCancelled(A){V(A),Dn(b,[A])}})}function mv(t){if(t==null)return null;if(Ie(t))return[Ao(t.enter),Ao(t.leave)];{const e=Ao(t);return[e,e]}}function Ao(t){return xm(t)}function nn(t,e){e.split(/\s+/).forEach(n=>n&&t.classList.add(n)),(t[Fs]||(t[Fs]=new Set)).add(e)}function _n(t,e){e.split(/\s+/).forEach(s=>s&&t.classList.remove(s));const n=t[Fs];n&&(n.delete(e),n.size||(t[Fs]=void 0))}function cu(t){requestAnimationFrame(()=>{requestAnimationFrame(t)})}let gv=0;function du(t,e,n,s){const r=t._endId=++gv,i=()=>{r===t._endId&&s()};if(n)return setTimeout(i,n);const{type:o,timeout:a,propCount:l}=tf(t,e);if(!o)return s();const u=o+"end";let c=0;const d=()=>{t.removeEventListener(u,f),i()},f=h=>{h.target===t&&++c>=l&&d()};setTimeout(()=>{c<l&&d()},a+1),t.addEventListener(u,f)}function tf(t,e){const n=window.getComputedStyle(t),s=g=>(n[g]||"").split(", "),r=s(`${pn}Delay`),i=s(`${pn}Duration`),o=fu(r,i),a=s(`${Ds}Delay`),l=s(`${Ds}Duration`),u=fu(a,l);let c=null,d=0,f=0;e===pn?o>0&&(c=pn,d=o,f=i.length):e===Ds?u>0&&(c=Ds,d=u,f=l.length):(d=Math.max(o,u),c=d>0?o>u?pn:Ds:null,f=c?c===pn?i.length:l.length:0);const h=c===pn&&/\b(transform|all)(,|$)/.test(s(`${pn}Property`).toString());return{type:c,timeout:d,propCount:f,hasTransform:h}}function fu(t,e){for(;t.length<e.length;)t=t.concat(t);return Math.max(...e.map((n,s)=>hu(n)+hu(t[s])))}function hu(t){return t==="auto"?0:Number(t.slice(0,-1).replace(",","."))*1e3}function nf(){return document.body.offsetHeight}function vv(t,e,n){const s=t[Fs];s&&(e=(e?[e,...s]:[...s]).join(" ")),e==null?t.removeAttribute("class"):n?t.setAttribute("class",e):t.className=e}const Si=Symbol("_vod"),sf=Symbol("_vsh"),io={beforeMount(t,{value:e},{transition:n}){t[Si]=t.style.display==="none"?"":t.style.display,n&&e?n.beforeEnter(t):$s(t,e)},mounted(t,{value:e},{transition:n}){n&&e&&n.enter(t)},updated(t,{value:e,oldValue:n},{transition:s}){!e!=!n&&(s?e?(s.beforeEnter(t),$s(t,!0),s.enter(t)):s.leave(t,()=>{$s(t,!1)}):$s(t,e))},beforeUnmount(t,{value:e}){$s(t,e)}};function $s(t,e){t.style.display=e?t[Si]:"none",t[sf]=!e}const pv=Symbol(""),yv=/(^|;)\s*display\s*:/;function bv(t,e,n){const s=t.style,r=Ne(n);let i=!1;if(n&&!r){if(e)if(Ne(e))for(const o of e.split(";")){const a=o.slice(0,o.indexOf(":")).trim();n[a]==null&&ci(s,a,"")}else for(const o in e)n[o]==null&&ci(s,o,"");for(const o in n)o==="display"&&(i=!0),ci(s,o,n[o])}else if(r){if(e!==n){const o=s[pv];o&&(n+=";"+o),s.cssText=n,i=yv.test(n)}}else e&&t.removeAttribute("style");Si in t&&(t[Si]=i?s.display:"",t[sf]&&(s.display="none"))}const mu=/\s*!important$/;function ci(t,e,n){if(oe(n))n.forEach(s=>ci(t,e,s));else if(n==null&&(n=""),e.startsWith("--"))t.setProperty(e,n);else{const s=xv(t,e);mu.test(n)?t.setProperty(rs(s),n.replace(mu,""),"important"):t[s]=n}}const gu=["Webkit","Moz","ms"],Ro={};function xv(t,e){const n=Ro[e];if(n)return n;let s=bt(e);if(s!=="filter"&&s in t)return Ro[e]=s;s=cn(s);for(let r=0;r<gu.length;r++){const i=gu[r]+s;if(i in t)return Ro[e]=i}return e}const vu="http://www.w3.org/1999/xlink";function wv(t,e,n,s,r){if(s&&e.startsWith("xlink:"))n==null?t.removeAttributeNS(vu,e.slice(6,e.length)):t.setAttributeNS(vu,e,n);else{const i=Mm(e);n==null||i&&!nd(n)?t.removeAttribute(e):t.setAttribute(e,i?"":n)}}function _v(t,e,n,s,r,i,o){if(e==="innerHTML"||e==="textContent"){s&&o(s,r,i),t[e]=n??"";return}const a=t.tagName;if(e==="value"&&a!=="PROGRESS"&&!a.includes("-")){const u=a==="OPTION"?t.getAttribute("value")||"":t.value,c=n??"";(u!==c||!("_value"in t))&&(t.value=c),n==null&&t.removeAttribute(e),t._value=n;return}let l=!1;if(n===""||n==null){const u=typeof t[e];u==="boolean"?n=nd(n):n==null&&u==="string"?(n="",l=!0):u==="number"&&(n=0,l=!0)}try{t[e]=n}catch{}l&&t.removeAttribute(e)}function Sv(t,e,n,s){t.addEventListener(e,n,s)}function Cv(t,e,n,s){t.removeEventListener(e,n,s)}const pu=Symbol("_vei");function Ev(t,e,n,s,r=null){const i=t[pu]||(t[pu]={}),o=i[e];if(s&&o)o.value=s;else{const[a,l]=Mv(e);if(s){const u=i[e]=Rv(s,r);Sv(t,a,u,l)}else o&&(Cv(t,a,o,l),i[e]=void 0)}}const yu=/(?:Once|Passive|Capture)$/;function Mv(t){let e;if(yu.test(t)){e={};let s;for(;s=t.match(yu);)t=t.slice(0,t.length-s[0].length),e[s[0].toLowerCase()]=!0}return[t[2]===":"?t.slice(3):rs(t.slice(2)),e]}let Io=0;const Tv=Promise.resolve(),Av=()=>Io||(Tv.then(()=>Io=0),Io=Date.now());function Rv(t,e){const n=s=>{if(!s._vts)s._vts=Date.now();else if(s._vts<=n.attached)return;Mt(Iv(s,n.value),e,5,[s])};return n.value=t,n.attached=Av(),n}function Iv(t,e){if(oe(e)){const n=t.stopImmediatePropagation;return t.stopImmediatePropagation=()=>{n.call(t),t._stopped=!0},e.map(s=>r=>!r._stopped&&s&&s(r))}else return e}const bu=t=>t.charCodeAt(0)===111&&t.charCodeAt(1)===110&&t.charCodeAt(2)>96&&t.charCodeAt(2)<123,kv=(t,e,n,s,r,i,o,a,l)=>{const u=r==="svg";e==="class"?vv(t,s,u):e==="style"?bv(t,n,s):Wi(e)?Oa(e)||Ev(t,e,n,s,o):(e[0]==="."?(e=e.slice(1),!0):e[0]==="^"?(e=e.slice(1),!1):Fv(t,e,s,u))?_v(t,e,s,i,o,a,l):(e==="true-value"?t._trueValue=s:e==="false-value"&&(t._falseValue=s),wv(t,e,s,u))};function Fv(t,e,n,s){if(s)return!!(e==="innerHTML"||e==="textContent"||e in t&&bu(e)&&fe(n));if(e==="spellcheck"||e==="draggable"||e==="translate"||e==="form"||e==="list"&&t.tagName==="INPUT"||e==="type"&&t.tagName==="TEXTAREA")return!1;if(e==="width"||e==="height"){const r=t.tagName;if(r==="IMG"||r==="VIDEO"||r==="CANVAS"||r==="SOURCE")return!1}return bu(e)&&Ne(n)?!1:e in t}const rf=new WeakMap,of=new WeakMap,Ci=Symbol("_moveCb"),xu=Symbol("_enterCb"),af={name:"TransitionGroup",props:De({},hv,{tag:String,moveClass:String}),setup(t,{slots:e}){const n=so(),s=Fd();let r,i;return Ka(()=>{if(!r.length)return;const o=t.moveClass||`${t.name||"v"}-move`;if(!Nv(r[0].el,n.vnode.el,o))return;r.forEach(Lv),r.forEach(Pv);const a=r.filter(Ov);nf(),a.forEach(l=>{const u=l.el,c=u.style;nn(u,o),c.transform=c.webkitTransform=c.transitionDuration="";const d=u[Ci]=f=>{f&&f.target!==u||(!f||/transform$/.test(f.propertyName))&&(u.removeEventListener("transitionend",d),u[Ci]=null,_n(u,o))};u.addEventListener("transitionend",d)})}),()=>{const o=pe(t),a=ef(o);let l=o.tag||Re;r=i,i=e.default?Za(e.default()):[];for(let u=0;u<i.length;u++){const c=i[u];c.key!=null&&hr(c,fr(c,a,s,n))}if(r)for(let u=0;u<r.length;u++){const c=r[u];hr(c,fr(c,a,s,n)),rf.set(c,c.el.getBoundingClientRect())}return _(l,null,i)}}},Vv=t=>delete t.mode;af.props;const lf=af;function Lv(t){const e=t.el;e[Ci]&&e[Ci](),e[xu]&&e[xu]()}function Pv(t){of.set(t,t.el.getBoundingClientRect())}function Ov(t){const e=rf.get(t),n=of.get(t),s=e.left-n.left,r=e.top-n.top;if(s||r){const i=t.el.style;return i.transform=i.webkitTransform=`translate(${s}px,${r}px)`,i.transitionDuration="0s",t}}function Nv(t,e,n){const s=t.cloneNode(),r=t[Fs];r&&r.forEach(a=>{a.split(/\s+/).forEach(l=>l&&s.classList.remove(l))}),n.split(/\s+/).forEach(a=>a&&s.classList.add(a)),s.style.display="none";const i=e.nodeType===1?e:e.parentNode;i.appendChild(s);const{hasTransform:o}=tf(s);return i.removeChild(s),o}const Bv=["ctrl","shift","alt","meta"],zv={stop:t=>t.stopPropagation(),prevent:t=>t.preventDefault(),self:t=>t.target!==t.currentTarget,ctrl:t=>!t.ctrlKey,shift:t=>!t.shiftKey,alt:t=>!t.altKey,meta:t=>!t.metaKey,left:t=>"button"in t&&t.button!==0,middle:t=>"button"in t&&t.button!==1,right:t=>"button"in t&&t.button!==2,exact:(t,e)=>Bv.some(n=>t[`${n}Key`]&&!e.includes(n))},wu=(t,e)=>{const n=t._withMods||(t._withMods={}),s=e.join(".");return n[s]||(n[s]=(r,...i)=>{for(let o=0;o<e.length;o++){const a=zv[e[o]];if(a&&a(r,e))return}return t(r,...i)})},Dv=De({patchProp:kv},fv);let _u;function $v(){return _u||(_u=Ug(Dv))}const Hv=(...t)=>{const e=$v().createApp(...t),{mount:n}=e;return e.mount=s=>{const r=jv(s);if(!r)return;const i=e._component;!fe(i)&&!i.render&&!i.template&&(i.template=r.innerHTML),r.innerHTML="";const o=n(r,!1,Uv(r));return r instanceof Element&&(r.removeAttribute("v-cloak"),r.setAttribute("data-v-app","")),o},e};function Uv(t){if(t instanceof SVGElement)return"svg";if(typeof MathMLElement=="function"&&t instanceof MathMLElement)return"mathml"}function jv(t){return Ne(t)?document.querySelector(t):t}const Wv="modulepreload",Gv=function(t,e){return new URL(t,e).href},Su={},ha=function(e,n,s){let r=Promise.resolve();if(n&&n.length>0){const i=document.getElementsByTagName("link");r=Promise.all(n.map(o=>{if(o=Gv(o,s),o in Su)return;Su[o]=!0;const a=o.endsWith(".css"),l=a?'[rel="stylesheet"]':"";if(!!s)for(let d=i.length-1;d>=0;d--){const f=i[d];if(f.href===o&&(!a||f.rel==="stylesheet"))return}else if(document.querySelector(`link[href="${o}"]${l}`))return;const c=document.createElement("link");if(c.rel=a?"stylesheet":Wv,a||(c.as="script",c.crossOrigin=""),c.href=o,document.head.appendChild(c),a)return new Promise((d,f)=>{c.addEventListener("load",d),c.addEventListener("error",()=>f(new Error(`Unable to preload CSS for ${o}`)))})}))}return r.then(()=>e()).catch(i=>{const o=new Event("vite:preloadError",{cancelable:!0});if(o.payload=i,window.dispatchEvent(o),!o.defaultPrevented)throw i})};function G(t,e){return n=>Object.keys(t).reduce((s,r)=>{const o=typeof t[r]=="object"&&t[r]!=null&&!Array.isArray(t[r])?t[r]:{type:t[r]};return n&&r in n?s[r]={...o,default:n[r]}:s[r]=o,e&&!s[r].source&&(s[r].source=e),s},{})}const we=G({class:[String,Array],style:{type:[String,Array,Object],default:null}},"component"),Oe=typeof window<"u",nl=Oe&&"IntersectionObserver"in window,qv=Oe&&("ontouchstart"in window||window.navigator.maxTouchPoints>0);function Cu(t,e,n){Yv(t,e),e.set(t,n)}function Yv(t,e){if(e.has(t))throw new TypeError("Cannot initialize the same private elements twice on an object")}function Xv(t,e,n){var s=uf(t,e,"set");return Zv(t,s,n),n}function Zv(t,e,n){if(e.set)e.set.call(t,n);else{if(!e.writable)throw new TypeError("attempted to set read only private field");e.value=n}}function $n(t,e){var n=uf(t,e,"get");return Kv(t,n)}function uf(t,e,n){if(!e.has(t))throw new TypeError("attempted to "+n+" private field on non-instance");return e.get(t)}function Kv(t,e){return e.get?e.get.call(t):e.value}function cf(t,e,n){const s=e.length-1;if(s<0)return t===void 0?n:t;for(let r=0;r<s;r++){if(t==null)return n;t=t[e[r]]}return t==null||t[e[s]]===void 0?n:t[e[s]]}function oo(t,e){if(t===e)return!0;if(t instanceof Date&&e instanceof Date&&t.getTime()!==e.getTime()||t!==Object(t)||e!==Object(e))return!1;const n=Object.keys(t);return n.length!==Object.keys(e).length?!1:n.every(s=>oo(t[s],e[s]))}function ma(t,e,n){return t==null||!e||typeof e!="string"?n:t[e]!==void 0?t[e]:(e=e.replace(/\[(\w+)\]/g,".$1"),e=e.replace(/^\./,""),cf(t,e.split("."),n))}function _w(t,e,n){if(e===!0)return t===void 0?n:t;if(e==null||typeof e=="boolean")return n;if(t!==Object(t)){if(typeof e!="function")return n;const r=e(t,n);return typeof r>"u"?n:r}if(typeof e=="string")return ma(t,e,n);if(Array.isArray(e))return cf(t,e,n);if(typeof e!="function")return n;const s=e(t,n);return typeof s>"u"?n:s}function sl(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0;return Array.from({length:t},(n,s)=>e+s)}function ie(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:"px";if(!(t==null||t===""))return isNaN(+t)?String(t):isFinite(+t)?`${Number(t)}${e}`:void 0}function ga(t){return t!==null&&typeof t=="object"&&!Array.isArray(t)}function Ei(t){if(t&&"$el"in t){const e=t.$el;return(e==null?void 0:e.nodeType)===Node.TEXT_NODE?e.nextElementSibling:e}return t}const Eu=Object.freeze({enter:13,tab:9,delete:46,esc:27,space:32,up:38,down:40,left:37,right:39,end:35,home:36,del:46,backspace:8,insert:45,pageup:33,pagedown:34,shift:16}),Jv=Object.freeze({enter:"Enter",tab:"Tab",delete:"Delete",esc:"Escape",space:"Space",up:"ArrowUp",down:"ArrowDown",left:"ArrowLeft",right:"ArrowRight",end:"End",home:"Home",del:"Delete",backspace:"Backspace",insert:"Insert",pageup:"PageUp",pagedown:"PageDown",shift:"Shift"});function df(t){return Object.keys(t)}function ko(t,e){return e.every(n=>t.hasOwnProperty(n))}function Qv(t,e){const n={},s=new Set(Object.keys(t));for(const r of e)s.has(r)&&(n[r]=t[r]);return n}function Mu(t,e,n){const s=Object.create(null),r=Object.create(null);for(const i in t)e.some(o=>o instanceof RegExp?o.test(i):o===i)&&!(n!=null&&n.some(o=>o===i))?s[i]=t[i]:r[i]=t[i];return[s,r]}function rl(t,e){const n={...t};return e.forEach(s=>delete n[s]),n}const ff=/^on[^a-z]/,hf=t=>ff.test(t),ep=["onAfterscriptexecute","onAnimationcancel","onAnimationend","onAnimationiteration","onAnimationstart","onAuxclick","onBeforeinput","onBeforescriptexecute","onChange","onClick","onCompositionend","onCompositionstart","onCompositionupdate","onContextmenu","onCopy","onCut","onDblclick","onFocusin","onFocusout","onFullscreenchange","onFullscreenerror","onGesturechange","onGestureend","onGesturestart","onGotpointercapture","onInput","onKeydown","onKeypress","onKeyup","onLostpointercapture","onMousedown","onMousemove","onMouseout","onMouseover","onMouseup","onMousewheel","onPaste","onPointercancel","onPointerdown","onPointerenter","onPointerleave","onPointermove","onPointerout","onPointerover","onPointerup","onReset","onSelect","onSubmit","onTouchcancel","onTouchend","onTouchmove","onTouchstart","onTransitioncancel","onTransitionend","onTransitionrun","onTransitionstart","onWheel"];function tp(t){const[e,n]=Mu(t,[ff]),s=rl(e,ep),[r,i]=Mu(n,["class","style","id",/^data-/]);return Object.assign(r,e),Object.assign(i,s),[r,i]}function In(t){return t==null?[]:Array.isArray(t)?t:[t]}function Sw(t,e){let n=0;const s=function(){for(var r=arguments.length,i=new Array(r),o=0;o<r;o++)i[o]=arguments[o];clearTimeout(n),n=setTimeout(()=>t(...i),F(e))};return s.clear=()=>{clearTimeout(n)},s.immediate=t,s}function vr(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0,n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:1;return Math.max(e,Math.min(n,t))}function Tu(t){const e=t.toString().trim();return e.includes(".")?e.length-e.indexOf(".")-1:0}function Au(t,e){let n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:"0";return t+n.repeat(Math.max(0,e-t.length))}function Ru(t,e){return(arguments.length>2&&arguments[2]!==void 0?arguments[2]:"0").repeat(Math.max(0,e-t.length))+t}function np(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:1;const n=[];let s=0;for(;s<t.length;)n.push(t.substr(s,e)),s+=e;return n}function vt(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},n=arguments.length>2?arguments[2]:void 0;const s={};for(const r in t)s[r]=t[r];for(const r in e){const i=t[r],o=e[r];if(ga(i)&&ga(o)){s[r]=vt(i,o,n);continue}if(Array.isArray(i)&&Array.isArray(o)&&n){s[r]=n(i,o);continue}s[r]=o}return s}function mf(t){return t.map(e=>e.type===Re?mf(e.children):e).flat()}function Kn(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:"";if(Kn.cache.has(t))return Kn.cache.get(t);const e=t.replace(/[^a-z]/gi,"-").replace(/\B([A-Z])/g,"-$1").toLowerCase();return Kn.cache.set(t,e),e}Kn.cache=new Map;function ir(t,e){if(!e||typeof e!="object")return[];if(Array.isArray(e))return e.map(n=>ir(t,n)).flat(1);if(Array.isArray(e.children))return e.children.map(n=>ir(t,n)).flat(1);if(e.component){if(Object.getOwnPropertySymbols(e.component.provides).includes(t))return[e.component];if(e.component.subTree)return ir(t,e.component.subTree).flat(1)}return[]}var Ur=new WeakMap,ms=new WeakMap;class sp{constructor(e){Cu(this,Ur,{writable:!0,value:[]}),Cu(this,ms,{writable:!0,value:0}),this.size=e}push(e){$n(this,Ur)[$n(this,ms)]=e,Xv(this,ms,($n(this,ms)+1)%this.size)}values(){return $n(this,Ur).slice($n(this,ms)).concat($n(this,Ur).slice(0,$n(this,ms)))}}function il(t){const e=tt({}),n=I(t);return Pn(()=>{for(const s in n.value)e[s]=n.value[s]},{flush:"sync"}),Ya(e)}function Mi(t,e){return t.includes(e)}function gf(t){return t[2].toLowerCase()+t.slice(3)}const Ti=()=>[Function,Array];function Iu(t,e){return e="on"+cn(e),!!(t[e]||t[`${e}Once`]||t[`${e}Capture`]||t[`${e}OnceCapture`]||t[`${e}CaptureOnce`])}function Cw(t){for(var e=arguments.length,n=new Array(e>1?e-1:0),s=1;s<e;s++)n[s-1]=arguments[s];if(Array.isArray(t))for(const r of t)r(...n);else typeof t=="function"&&t(...n)}function vf(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:!0;const n=["button","[href]",'input:not([type="hidden"])',"select","textarea","[tabindex]"].map(s=>`${s}${e?':not([tabindex="-1"])':""}:not([disabled])`).join(", ");return[...t.querySelectorAll(n)]}function rp(t,e,n){let s,r=t.indexOf(document.activeElement);const i=e==="next"?1:-1;do r+=i,s=t[r];while((!s||s.offsetParent==null||!((n==null?void 0:n(s))??!0))&&r<t.length&&r>=0);return s}function ip(t,e){var s,r,i,o;const n=vf(t);if(!e)(t===document.activeElement||!t.contains(document.activeElement))&&((s=n[0])==null||s.focus());else if(e==="first")(r=n[0])==null||r.focus();else if(e==="last")(i=n.at(-1))==null||i.focus();else if(typeof e=="number")(o=n[e])==null||o.focus();else{const a=rp(n,e);a?a.focus():ip(t,e==="next"?"first":"last")}}function pf(t,e){if(!(Oe&&typeof CSS<"u"&&typeof CSS.supports<"u"&&CSS.supports(`selector(${e})`)))return null;try{return!!t&&t.matches(e)}catch{return null}}function op(t){return t.some(e=>gr(e)?e.type===pt?!1:e.type!==Re||op(e.children):!0)?t:null}function ap(t,e){if(!Oe||t===0)return e(),()=>{};const n=window.setTimeout(e,t);return()=>window.clearTimeout(n)}function Ew(t,e){const n=t.clientX,s=t.clientY,r=e.getBoundingClientRect(),i=r.left,o=r.top,a=r.right,l=r.bottom;return n>=i&&n<=a&&s>=o&&s<=l}const yf=["top","bottom"],lp=["start","end","left","right"];function va(t,e){let[n,s]=t.split(" ");return s||(s=Mi(yf,n)?"start":Mi(lp,n)?"top":"center"),{side:pa(n,e),align:pa(s,e)}}function pa(t,e){return t==="start"?e?"right":"left":t==="end"?e?"left":"right":t}function Fo(t){return{side:{center:"center",top:"bottom",bottom:"top",left:"right",right:"left"}[t.side],align:t.align}}function Vo(t){return{side:t.side,align:{center:"center",top:"bottom",bottom:"top",left:"right",right:"left"}[t.align]}}function ku(t){return{side:t.align,align:t.side}}function Fu(t){return Mi(yf,t.side)?"y":"x"}class Jn{constructor(e){let{x:n,y:s,width:r,height:i}=e;this.x=n,this.y=s,this.width=r,this.height=i}get top(){return this.y}get bottom(){return this.y+this.height}get left(){return this.x}get right(){return this.x+this.width}}function Vu(t,e){return{x:{before:Math.max(0,e.left-t.left),after:Math.max(0,t.right-e.right)},y:{before:Math.max(0,e.top-t.top),after:Math.max(0,t.bottom-e.bottom)}}}function bf(t){return Array.isArray(t)?new Jn({x:t[0],y:t[1],width:0,height:0}):t.getBoundingClientRect()}function xf(t){const e=t.getBoundingClientRect(),n=getComputedStyle(t),s=n.transform;if(s){let r,i,o,a,l;if(s.startsWith("matrix3d("))r=s.slice(9,-1).split(/, /),i=+r[0],o=+r[5],a=+r[12],l=+r[13];else if(s.startsWith("matrix("))r=s.slice(7,-1).split(/, /),i=+r[0],o=+r[3],a=+r[4],l=+r[5];else return new Jn(e);const u=n.transformOrigin,c=e.x-a-(1-i)*parseFloat(u),d=e.y-l-(1-o)*parseFloat(u.slice(u.indexOf(" ")+1)),f=i?e.width/i:t.offsetWidth+1,h=o?e.height/o:t.offsetHeight+1;return new Jn({x:c,y:d,width:f,height:h})}else return new Jn(e)}function Ys(t,e,n){if(typeof t.animate>"u")return{finished:Promise.resolve()};let s;try{s=t.animate(e,n)}catch{return{finished:Promise.resolve()}}return typeof s.finished>"u"&&(s.finished=new Promise(r=>{s.onfinish=()=>{r(s)}})),s}const di=new WeakMap;function up(t,e){Object.keys(e).forEach(n=>{if(hf(n)){const s=gf(n),r=di.get(t);if(e[n]==null)r==null||r.forEach(i=>{const[o,a]=i;o===s&&(t.removeEventListener(s,a),r.delete(i))});else if(!r||![...r].some(i=>i[0]===s&&i[1]===e[n])){t.addEventListener(s,e[n]);const i=r||new Set;i.add([s,e[n]]),di.has(t)||di.set(t,i)}}else e[n]==null?t.removeAttribute(n):t.setAttribute(n,e[n])})}function cp(t,e){Object.keys(e).forEach(n=>{if(hf(n)){const s=gf(n),r=di.get(t);r==null||r.forEach(i=>{const[o,a]=i;o===s&&(t.removeEventListener(s,a),r.delete(i))})}else t.removeAttribute(n)})}const gs=2.4,Lu=.2126729,Pu=.7151522,Ou=.072175,dp=.55,fp=.58,hp=.57,mp=.62,jr=.03,Nu=1.45,gp=5e-4,vp=1.25,pp=1.25,Bu=.078,zu=12.82051282051282,Wr=.06,Du=.001;function $u(t,e){const n=(t.r/255)**gs,s=(t.g/255)**gs,r=(t.b/255)**gs,i=(e.r/255)**gs,o=(e.g/255)**gs,a=(e.b/255)**gs;let l=n*Lu+s*Pu+r*Ou,u=i*Lu+o*Pu+a*Ou;if(l<=jr&&(l+=(jr-l)**Nu),u<=jr&&(u+=(jr-u)**Nu),Math.abs(u-l)<gp)return 0;let c;if(u>l){const d=(u**dp-l**fp)*vp;c=d<Du?0:d<Bu?d-d*zu*Wr:d-Wr}else{const d=(u**mp-l**hp)*pp;c=d>-Du?0:d>-Bu?d-d*zu*Wr:d+Wr}return c*100}function Mw(t,e){e=Array.isArray(e)?e.slice(0,-1).map(n=>`'${n}'`).join(", ")+` or '${e.at(-1)}'`:`'${e}'`}const Ai=.20689655172413793,yp=t=>t>Ai**3?Math.cbrt(t):t/(3*Ai**2)+4/29,bp=t=>t>Ai?t**3:3*Ai**2*(t-4/29);function wf(t){const e=yp,n=e(t[1]);return[116*n-16,500*(e(t[0]/.95047)-n),200*(n-e(t[2]/1.08883))]}function _f(t){const e=bp,n=(t[0]+16)/116;return[e(n+t[1]/500)*.95047,e(n),e(n-t[2]/200)*1.08883]}const xp=[[3.2406,-1.5372,-.4986],[-.9689,1.8758,.0415],[.0557,-.204,1.057]],wp=t=>t<=.0031308?t*12.92:1.055*t**(1/2.4)-.055,_p=[[.4124,.3576,.1805],[.2126,.7152,.0722],[.0193,.1192,.9505]],Sp=t=>t<=.04045?t/12.92:((t+.055)/1.055)**2.4;function Sf(t){const e=Array(3),n=wp,s=xp;for(let r=0;r<3;++r)e[r]=Math.round(vr(n(s[r][0]*t[0]+s[r][1]*t[1]+s[r][2]*t[2]))*255);return{r:e[0],g:e[1],b:e[2]}}function ol(t){let{r:e,g:n,b:s}=t;const r=[0,0,0],i=Sp,o=_p;e=i(e/255),n=i(n/255),s=i(s/255);for(let a=0;a<3;++a)r[a]=o[a][0]*e+o[a][1]*n+o[a][2]*s;return r}function ya(t){return!!t&&/^(#|var\(--|(rgb|hsl)a?\()/.test(t)}function Cp(t){return ya(t)&&!/^((rgb|hsl)a?\()?var\(--/.test(t)}const Hu=/^(?<fn>(?:rgb|hsl)a?)\((?<values>.+)\)/,Ep={rgb:(t,e,n,s)=>({r:t,g:e,b:n,a:s}),rgba:(t,e,n,s)=>({r:t,g:e,b:n,a:s}),hsl:(t,e,n,s)=>Uu({h:t,s:e,l:n,a:s}),hsla:(t,e,n,s)=>Uu({h:t,s:e,l:n,a:s}),hsv:(t,e,n,s)=>pr({h:t,s:e,v:n,a:s}),hsva:(t,e,n,s)=>pr({h:t,s:e,v:n,a:s})};function Ut(t){if(typeof t=="number")return{r:(t&16711680)>>16,g:(t&65280)>>8,b:t&255};if(typeof t=="string"&&Hu.test(t)){const{groups:e}=t.match(Hu),{fn:n,values:s}=e,r=s.split(/,\s*/).map(i=>i.endsWith("%")&&["hsl","hsla","hsv","hsva"].includes(n)?parseFloat(i)/100:parseFloat(i));return Ep[n](...r)}else if(typeof t=="string"){let e=t.startsWith("#")?t.slice(1):t;return[3,4].includes(e.length)?e=e.split("").map(n=>n+n).join(""):[6,8].includes(e.length),Tp(e)}else if(typeof t=="object"){if(ko(t,["r","g","b"]))return t;if(ko(t,["h","s","l"]))return pr(Cf(t));if(ko(t,["h","s","v"]))return pr(t)}throw new TypeError(`Invalid color: ${t==null?t:String(t)||t.constructor.name}
Expected #hex, #hexa, rgb(), rgba(), hsl(), hsla(), object or number`)}function pr(t){const{h:e,s:n,v:s,a:r}=t,i=a=>{const l=(a+e/60)%6;return s-s*n*Math.max(Math.min(l,4-l,1),0)},o=[i(5),i(3),i(1)].map(a=>Math.round(a*255));return{r:o[0],g:o[1],b:o[2],a:r}}function Uu(t){return pr(Cf(t))}function Cf(t){const{h:e,s:n,l:s,a:r}=t,i=s+n*Math.min(s,1-s),o=i===0?0:2-2*s/i;return{h:e,s:o,v:i,a:r}}function Gr(t){const e=Math.round(t).toString(16);return("00".substr(0,2-e.length)+e).toUpperCase()}function Mp(t){let{r:e,g:n,b:s,a:r}=t;return`#${[Gr(e),Gr(n),Gr(s),r!==void 0?Gr(Math.round(r*255)):""].join("")}`}function Tp(t){t=Ap(t);let[e,n,s,r]=np(t,2).map(i=>parseInt(i,16));return r=r===void 0?r:r/255,{r:e,g:n,b:s,a:r}}function Ap(t){return t.startsWith("#")&&(t=t.slice(1)),t=t.replace(/([^0-9a-f])/gi,"F"),(t.length===3||t.length===4)&&(t=t.split("").map(e=>e+e).join("")),t.length!==6&&(t=Au(Au(t,6),8,"F")),t}function Rp(t,e){const n=wf(ol(t));return n[0]=n[0]+e*10,Sf(_f(n))}function Ip(t,e){const n=wf(ol(t));return n[0]=n[0]-e*10,Sf(_f(n))}function kp(t){const e=Ut(t);return ol(e)[1]}function Ef(t){const e=Math.abs($u(Ut(0),Ut(t)));return Math.abs($u(Ut(16777215),Ut(t)))>Math.min(e,50)?"#fff":"#000"}const Vs=Symbol.for("vuetify:defaults");function Fp(t){return se(t)}function al(){const t=Pe(Vs);if(!t)throw new Error("[Vuetify] Could not find defaults instance");return t}function hn(t,e){const n=al(),s=se(t),r=I(()=>{if(F(e==null?void 0:e.disabled))return n.value;const o=F(e==null?void 0:e.scoped),a=F(e==null?void 0:e.reset),l=F(e==null?void 0:e.root);if(s.value==null&&!(o||a||l))return n.value;let u=vt(s.value,{prev:n.value});if(o)return u;if(a||l){const c=Number(a||1/0);for(let d=0;d<=c&&!(!u||!("prev"in u));d++)u=u.prev;return u&&typeof l=="string"&&l in u&&(u=vt(vt(u,{prev:u}),u[l])),u}return u.prev?vt(u.prev,u):u});return ut(Vs,r),r}function Vp(t,e){var n,s;return typeof((n=t.props)==null?void 0:n[e])<"u"||typeof((s=t.props)==null?void 0:s[Kn(e)])<"u"}function Lp(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},e=arguments.length>1?arguments[1]:void 0,n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:al();const s=Ge("useDefaults");if(e=e??s.type.name??s.type.__name,!e)throw new Error("[Vuetify] Could not determine component name");const r=I(()=>{var l;return(l=n.value)==null?void 0:l[t._as??e]}),i=new Proxy(t,{get(l,u){var d,f,h,g;const c=Reflect.get(l,u);return u==="class"||u==="style"?[(d=r.value)==null?void 0:d[u],c].filter(y=>y!=null):typeof u=="string"&&!Vp(s.vnode,u)?((f=r.value)==null?void 0:f[u])??((g=(h=n.value)==null?void 0:h.global)==null?void 0:g[u])??c:c}}),o=be();Pn(()=>{if(r.value){const l=Object.entries(r.value).filter(u=>{let[c]=u;return c.startsWith(c[0].toUpperCase())});o.value=l.length?Object.fromEntries(l):void 0}else o.value=void 0});function a(){const l=zp(Vs,s);ut(Vs,I(()=>o.value?vt((l==null?void 0:l.value)??{},o.value):l==null?void 0:l.value))}return{props:i,provideSubDefaults:a}}function Rr(t){if(t._setup=t._setup??t.setup,!t.name)return t;if(t._setup){t.props=G(t.props??{},t.name)();const e=Object.keys(t.props).filter(n=>n!=="class"&&n!=="style");t.filterProps=function(s){return Qv(s,e)},t.props._as=String,t.setup=function(s,r){const i=al();if(!i.value)return t._setup(s,r);const{props:o,provideSubDefaults:a}=Lp(s,s._as??t.name,i),l=t._setup(o,r);return a(),l}}return t}function he(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:!0;return e=>(t?Rr:dn)(e)}function ao(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:"div",n=arguments.length>2?arguments[2]:void 0;return he()({name:n??cn(bt(t.replace(/__/g,"-"))),props:{tag:{type:String,default:e},...we()},setup(s,r){let{slots:i}=r;return()=>{var o;return On(s.tag,{class:[t,s.class],style:s.style},(o=i.default)==null?void 0:o.call(i))}}})}function Mf(t){if(typeof t.getRootNode!="function"){for(;t.parentNode;)t=t.parentNode;return t!==document?null:document}const e=t.getRootNode();return e!==document&&e.getRootNode({composed:!0})!==document?null:e}const ba="cubic-bezier(0.4, 0, 0.2, 1)",Pp="cubic-bezier(0.0, 0, 0.2, 1)",Op="cubic-bezier(0.4, 0, 1, 1)";function Ge(t,e){const n=so();if(!n)throw new Error(`[Vuetify] ${t} ${e||"must be called from inside a setup function"}`);return n}function Xt(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:"composables";const e=Ge(t).type;return Kn((e==null?void 0:e.aliasName)||(e==null?void 0:e.name))}let Tf=0,fi=new WeakMap;function mn(){const t=Ge("getUid");if(fi.has(t))return fi.get(t);{const e=Tf++;return fi.set(t,e),e}}mn.reset=()=>{Tf=0,fi=new WeakMap};function Np(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:!1;for(;t;){if(e?Bp(t):ll(t))return t;t=t.parentElement}return document.scrollingElement}function Ri(t,e){const n=[];if(e&&t&&!e.contains(t))return n;for(;t&&(ll(t)&&n.push(t),t!==e);)t=t.parentElement;return n}function ll(t){if(!t||t.nodeType!==Node.ELEMENT_NODE)return!1;const e=window.getComputedStyle(t);return e.overflowY==="scroll"||e.overflowY==="auto"&&t.scrollHeight>t.clientHeight}function Bp(t){if(!t||t.nodeType!==Node.ELEMENT_NODE)return!1;const e=window.getComputedStyle(t);return["scroll","auto"].includes(e.overflowY)}function zp(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Ge("injectSelf");const{provides:n}=e;if(n&&t in n)return n[t]}function Dp(t){for(;t;){if(window.getComputedStyle(t).position==="fixed")return!0;t=t.offsetParent}return!1}function ye(t){const e=Ge("useRender");e.render=t}function Af(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:"content";const n=se(),s=se();if(Oe){const r=new ResizeObserver(i=>{t==null||t(i,r),i.length&&(e==="content"?s.value=i[0].contentRect:s.value=i[0].target.getBoundingClientRect())});Rt(()=>{r.disconnect()}),le(n,(i,o)=>{o&&(r.unobserve(Ei(o)),s.value=void 0),i&&r.observe(Ei(i))},{flush:"post"})}return{resizeRef:n,contentRect:_r(s)}}const Ii=Symbol.for("vuetify:layout"),Rf=Symbol.for("vuetify:layout-item"),ju=1e3,$p=G({overlaps:{type:Array,default:()=>[]},fullHeight:Boolean},"layout"),Hp=G({name:{type:String},order:{type:[Number,String],default:0},absolute:Boolean},"layout-item");function Up(){const t=Pe(Ii);if(!t)throw new Error("[Vuetify] Could not find injected layout");return{getLayoutItem:t.getLayoutItem,mainRect:t.mainRect,mainStyles:t.mainStyles}}function jp(t){const e=Pe(Ii);if(!e)throw new Error("[Vuetify] Could not find injected layout");const n=t.id??`layout-item-${mn()}`,s=Ge("useLayoutItem");ut(Rf,{id:n});const r=be(!1);Od(()=>r.value=!0),Pd(()=>r.value=!1);const{layoutItemStyles:i,layoutItemScrimStyles:o}=e.register(s,{...t,active:I(()=>r.value?!1:t.active.value),id:n});return Rt(()=>e.unregister(n)),{layoutItemStyles:i,layoutRect:e.layoutRect,layoutItemScrimStyles:o}}const Wp=(t,e,n,s)=>{let r={top:0,left:0,right:0,bottom:0};const i=[{id:"",layer:{...r}}];for(const o of t){const a=e.get(o),l=n.get(o),u=s.get(o);if(!a||!l||!u)continue;const c={...r,[a.value]:parseInt(r[a.value],10)+(u.value?parseInt(l.value,10):0)};i.push({id:o,layer:c}),r=c}return i};function Gp(t){const e=Pe(Ii,null),n=I(()=>e?e.rootZIndex.value-100:ju),s=se([]),r=tt(new Map),i=tt(new Map),o=tt(new Map),a=tt(new Map),l=tt(new Map),{resizeRef:u,contentRect:c}=Af(),d=I(()=>{const x=new Map,C=t.overlaps??[];for(const M of C.filter(T=>T.includes(":"))){const[T,V]=M.split(":");if(!s.value.includes(T)||!s.value.includes(V))continue;const L=r.get(T),A=r.get(V),k=i.get(T),O=i.get(V);!L||!A||!k||!O||(x.set(V,{position:L.value,amount:parseInt(k.value,10)}),x.set(T,{position:A.value,amount:-parseInt(O.value,10)}))}return x}),f=I(()=>{const x=[...new Set([...o.values()].map(M=>M.value))].sort((M,T)=>M-T),C=[];for(const M of x){const T=s.value.filter(V=>{var L;return((L=o.get(V))==null?void 0:L.value)===M});C.push(...T)}return Wp(C,r,i,a)}),h=I(()=>!Array.from(l.values()).some(x=>x.value)),g=I(()=>f.value[f.value.length-1].layer),y=I(()=>({"--v-layout-left":ie(g.value.left),"--v-layout-right":ie(g.value.right),"--v-layout-top":ie(g.value.top),"--v-layout-bottom":ie(g.value.bottom),...h.value?void 0:{transition:"none"}})),v=I(()=>f.value.slice(1).map((x,C)=>{let{id:M}=x;const{layer:T}=f.value[C],V=i.get(M),L=r.get(M);return{id:M,...T,size:Number(V.value),position:L.value}})),S=x=>v.value.find(C=>C.id===x),m=Ge("createLayout"),p=be(!1);Yt(()=>{p.value=!0}),ut(Ii,{register:(x,C)=>{let{id:M,order:T,position:V,layoutSize:L,elementSize:A,active:k,disableTransitions:O,absolute:H}=C;o.set(M,T),r.set(M,V),i.set(M,L),a.set(M,k),O&&l.set(M,O);const j=ir(Rf,m==null?void 0:m.vnode).indexOf(x);j>-1?s.value.splice(j,0,M):s.value.push(M);const Y=I(()=>v.value.findIndex(ge=>ge.id===M)),ae=I(()=>n.value+f.value.length*2-Y.value*2),Q=I(()=>{const ge=V.value==="left"||V.value==="right",me=V.value==="right",Ce=V.value==="bottom",We={[V.value]:0,zIndex:ae.value,transform:`translate${ge?"X":"Y"}(${(k.value?0:-110)*(me||Ce?-1:1)}%)`,position:H.value||n.value!==ju?"absolute":"fixed",...h.value?void 0:{transition:"none"}};if(!p.value)return We;const Z=v.value[Y.value];if(!Z)throw new Error(`[Vuetify] Could not find layout item "${M}"`);const xe=d.value.get(M);return xe&&(Z[xe.position]+=xe.amount),{...We,height:ge?`calc(100% - ${Z.top}px - ${Z.bottom}px)`:A.value?`${A.value}px`:void 0,left:me?void 0:`${Z.left}px`,right:me?`${Z.right}px`:void 0,top:V.value!=="bottom"?`${Z.top}px`:void 0,bottom:V.value!=="top"?`${Z.bottom}px`:void 0,width:ge?A.value?`${A.value}px`:void 0:`calc(100% - ${Z.left}px - ${Z.right}px)`}}),te=I(()=>({zIndex:ae.value-1}));return{layoutItemStyles:Q,layoutItemScrimStyles:te,zIndex:ae}},unregister:x=>{o.delete(x),r.delete(x),i.delete(x),a.delete(x),l.delete(x),s.value=s.value.filter(C=>C!==x)},mainRect:g,mainStyles:y,getLayoutItem:S,items:v,layoutRect:c,rootZIndex:n});const w=I(()=>["v-layout",{"v-layout--full-height":t.fullHeight}]),b=I(()=>({zIndex:e?n.value:void 0,position:e?"relative":void 0,overflow:e?"hidden":void 0}));return{layoutClasses:w,layoutStyles:b,getLayoutItem:S,items:v,layoutRect:c,layoutRef:u}}function on(t,e){let n;function s(){n=za(),n.run(()=>e.length?e(()=>{n==null||n.stop(),s()}):e())}le(t,r=>{r&&!n?s():r||(n==null||n.stop(),n=void 0)},{immediate:!0}),At(()=>{n==null||n.stop()})}function ct(t,e,n){let s=arguments.length>3&&arguments[3]!==void 0?arguments[3]:d=>d,r=arguments.length>4&&arguments[4]!==void 0?arguments[4]:d=>d;const i=Ge("useProxiedModel"),o=se(t[e]!==void 0?t[e]:n),a=Kn(e),u=I(a!==e?()=>{var d,f,h,g;return t[e],!!(((d=i.vnode.props)!=null&&d.hasOwnProperty(e)||(f=i.vnode.props)!=null&&f.hasOwnProperty(a))&&((h=i.vnode.props)!=null&&h.hasOwnProperty(`onUpdate:${e}`)||(g=i.vnode.props)!=null&&g.hasOwnProperty(`onUpdate:${a}`)))}:()=>{var d,f;return t[e],!!((d=i.vnode.props)!=null&&d.hasOwnProperty(e)&&((f=i.vnode.props)!=null&&f.hasOwnProperty(`onUpdate:${e}`)))});on(()=>!u.value,()=>{le(()=>t[e],d=>{o.value=d})});const c=I({get(){const d=t[e];return s(u.value?d:o.value)},set(d){const f=r(d),h=pe(u.value?t[e]:o.value);h===f||s(h)===d||(o.value=f,i==null||i.emit(`update:${e}`,f))}});return Object.defineProperty(c,"externalValue",{get:()=>u.value?t[e]:o.value}),c}const qp={badge:"Badge",open:"Open",close:"Close",confirmEdit:{ok:"OK",cancel:"Cancel"},dataIterator:{noResultsText:"No matching records found",loadingText:"Loading items..."},dataTable:{itemsPerPageText:"Rows per page:",ariaLabel:{sortDescending:"Sorted descending.",sortAscending:"Sorted ascending.",sortNone:"Not sorted.",activateNone:"Activate to remove sorting.",activateDescending:"Activate to sort descending.",activateAscending:"Activate to sort ascending."},sortBy:"Sort by"},dataFooter:{itemsPerPageText:"Items per page:",itemsPerPageAll:"All",nextPage:"Next page",prevPage:"Previous page",firstPage:"First page",lastPage:"Last page",pageText:"{0}-{1} of {2}"},dateRangeInput:{divider:"to"},datePicker:{itemsSelected:"{0} selected",range:{title:"Select dates",header:"Enter dates"},title:"Select date",header:"Enter date",input:{placeholder:"Enter date"}},noDataText:"No data available",carousel:{prev:"Previous visual",next:"Next visual",ariaLabel:{delimiter:"Carousel slide {0} of {1}"}},calendar:{moreEvents:"{0} more",today:"Today"},input:{clear:"Clear {0}",prependAction:"{0} prepended action",appendAction:"{0} appended action",otp:"Please enter OTP character {0}"},fileInput:{counter:"{0} files",counterSize:"{0} files ({1} in total)"},timePicker:{am:"AM",pm:"PM"},pagination:{ariaLabel:{root:"Pagination Navigation",next:"Next page",previous:"Previous page",page:"Go to page {0}",currentPage:"Page {0}, Current page",first:"First page",last:"Last page"}},stepper:{next:"Next",prev:"Previous"},rating:{ariaLabel:{item:"Rating {0} of {1}"}},loading:"Loading...",infiniteScroll:{loadMore:"Load more",empty:"No more"}},Wu="$vuetify.",Gu=(t,e)=>t.replace(/\{(\d+)\}/g,(n,s)=>String(e[+s])),If=(t,e,n)=>function(s){for(var r=arguments.length,i=new Array(r>1?r-1:0),o=1;o<r;o++)i[o-1]=arguments[o];if(!s.startsWith(Wu))return Gu(s,i);const a=s.replace(Wu,""),l=t.value&&n.value[t.value],u=e.value&&n.value[e.value];let c=ma(l,a,null);return c||(`${s}${t.value}`,c=ma(u,a,null)),c||(c=s),typeof c!="string"&&(c=s),Gu(c,i)};function kf(t,e){return(n,s)=>new Intl.NumberFormat([t.value,e.value],s).format(n)}function Lo(t,e,n){const s=ct(t,e,t[e]??n.value);return s.value=t[e]??n.value,le(n,r=>{t[e]==null&&(s.value=n.value)}),s}function Ff(t){return e=>{const n=Lo(e,"locale",t.current),s=Lo(e,"fallback",t.fallback),r=Lo(e,"messages",t.messages);return{name:"vuetify",current:n,fallback:s,messages:r,t:If(n,s,r),n:kf(n,s),provide:Ff({current:n,fallback:s,messages:r})}}}function Yp(t){const e=be((t==null?void 0:t.locale)??"en"),n=be((t==null?void 0:t.fallback)??"en"),s=se({en:qp,...t==null?void 0:t.messages});return{name:"vuetify",current:e,fallback:n,messages:s,t:If(e,n,s),n:kf(e,n),provide:Ff({current:e,fallback:n,messages:s})}}const ki=Symbol.for("vuetify:locale");function Xp(t){return t.name!=null}function Zp(t){const e=t!=null&&t.adapter&&Xp(t==null?void 0:t.adapter)?t==null?void 0:t.adapter:Yp(t),n=Qp(e,t);return{...e,...n}}function Kp(){const t=Pe(ki);if(!t)throw new Error("[Vuetify] Could not find injected locale instance");return t}function Jp(){return{af:!1,ar:!0,bg:!1,ca:!1,ckb:!1,cs:!1,de:!1,el:!1,en:!1,es:!1,et:!1,fa:!0,fi:!1,fr:!1,hr:!1,hu:!1,he:!0,id:!1,it:!1,ja:!1,km:!1,ko:!1,lv:!1,lt:!1,nl:!1,no:!1,pl:!1,pt:!1,ro:!1,ru:!1,sk:!1,sl:!1,srCyrl:!1,srLatn:!1,sv:!1,th:!1,tr:!1,az:!1,uk:!1,vi:!1,zhHans:!1,zhHant:!1}}function Qp(t,e){const n=se((e==null?void 0:e.rtl)??Jp()),s=I(()=>n.value[t.current.value]??!1);return{isRtl:s,rtl:n,rtlClasses:I(()=>`v-locale--is-${s.value?"rtl":"ltr"}`)}}function Zt(){const t=Pe(ki);if(!t)throw new Error("[Vuetify] Could not find injected rtl instance");return{isRtl:t.isRtl,rtlClasses:t.rtlClasses}}const Fi=Symbol.for("vuetify:theme"),dt=G({theme:String},"theme");function qu(){return{defaultTheme:"light",variations:{colors:[],lighten:0,darken:0},themes:{light:{dark:!1,colors:{background:"#FFFFFF",surface:"#FFFFFF","surface-bright":"#FFFFFF","surface-light":"#EEEEEE","surface-variant":"#424242","on-surface-variant":"#EEEEEE",primary:"#1867C0","primary-darken-1":"#1F5592",secondary:"#48A9A6","secondary-darken-1":"#018786",error:"#B00020",info:"#2196F3",success:"#4CAF50",warning:"#FB8C00"},variables:{"border-color":"#000000","border-opacity":.12,"high-emphasis-opacity":.87,"medium-emphasis-opacity":.6,"disabled-opacity":.38,"idle-opacity":.04,"hover-opacity":.04,"focus-opacity":.12,"selected-opacity":.08,"activated-opacity":.12,"pressed-opacity":.12,"dragged-opacity":.08,"theme-kbd":"#212529","theme-on-kbd":"#FFFFFF","theme-code":"#F5F5F5","theme-on-code":"#000000"}},dark:{dark:!0,colors:{background:"#121212",surface:"#212121","surface-bright":"#ccbfd6","surface-light":"#424242","surface-variant":"#a3a3a3","on-surface-variant":"#424242",primary:"#2196F3","primary-darken-1":"#277CC1",secondary:"#54B6B2","secondary-darken-1":"#48A9A6",error:"#CF6679",info:"#2196F3",success:"#4CAF50",warning:"#FB8C00"},variables:{"border-color":"#FFFFFF","border-opacity":.12,"high-emphasis-opacity":1,"medium-emphasis-opacity":.7,"disabled-opacity":.5,"idle-opacity":.1,"hover-opacity":.04,"focus-opacity":.12,"selected-opacity":.08,"activated-opacity":.12,"pressed-opacity":.16,"dragged-opacity":.08,"theme-kbd":"#212529","theme-on-kbd":"#FFFFFF","theme-code":"#343434","theme-on-code":"#CCCCCC"}}}}}function ey(){var s,r;let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:qu();const e=qu();if(!t)return{...e,isDisabled:!0};const n={};for(const[i,o]of Object.entries(t.themes??{})){const a=o.dark||i==="dark"?(s=e.themes)==null?void 0:s.dark:(r=e.themes)==null?void 0:r.light;n[i]=vt(a,o)}return vt(e,{...t,themes:n})}function ty(t){const e=ey(t),n=se(e.defaultTheme),s=se(e.themes),r=I(()=>{const c={};for(const[d,f]of Object.entries(s.value)){const h=c[d]={...f,colors:{...f.colors}};if(e.variations)for(const g of e.variations.colors){const y=h.colors[g];if(y)for(const v of["lighten","darken"]){const S=v==="lighten"?Rp:Ip;for(const m of sl(e.variations[v],1))h.colors[`${g}-${v}-${m}`]=Mp(S(Ut(y),m))}}for(const g of Object.keys(h.colors)){if(/^on-[a-z]/.test(g)||h.colors[`on-${g}`])continue;const y=`on-${g}`,v=Ut(h.colors[g]);h.colors[y]=Ef(v)}}return c}),i=I(()=>r.value[n.value]),o=I(()=>{const c=[];i.value.dark&&Hn(c,":root",["color-scheme: dark"]),Hn(c,":root",Yu(i.value));for(const[g,y]of Object.entries(r.value))Hn(c,`.v-theme--${g}`,[`color-scheme: ${y.dark?"dark":"normal"}`,...Yu(y)]);const d=[],f=[],h=new Set(Object.values(r.value).flatMap(g=>Object.keys(g.colors)));for(const g of h)/^on-[a-z]/.test(g)?Hn(f,`.${g}`,[`color: rgb(var(--v-theme-${g})) !important`]):(Hn(d,`.bg-${g}`,[`--v-theme-overlay-multiplier: var(--v-theme-${g}-overlay-multiplier)`,`background-color: rgb(var(--v-theme-${g})) !important`,`color: rgb(var(--v-theme-on-${g})) !important`]),Hn(f,`.text-${g}`,[`color: rgb(var(--v-theme-${g})) !important`]),Hn(f,`.border-${g}`,[`--v-border-color: var(--v-theme-${g})`]));return c.push(...d,...f),c.map((g,y)=>y===0?g:`    ${g}`).join("")});function a(){return{style:[{children:o.value,id:"vuetify-theme-stylesheet",nonce:e.cspNonce||!1}]}}function l(c){if(e.isDisabled)return;const d=c._context.provides.usehead;if(d)if(d.push){const f=d.push(a);Oe&&le(o,()=>{f.patch(a)})}else Oe?(d.addHeadObjs(I(a)),Pn(()=>d.updateDOM())):d.addHeadObjs(a());else{let h=function(){if(typeof document<"u"&&!f){const g=document.createElement("style");g.type="text/css",g.id="vuetify-theme-stylesheet",e.cspNonce&&g.setAttribute("nonce",e.cspNonce),f=g,document.head.appendChild(f)}f&&(f.innerHTML=o.value)},f=Oe?document.getElementById("vuetify-theme-stylesheet"):null;Oe?le(o,h,{immediate:!0}):h()}}const u=I(()=>e.isDisabled?void 0:`v-theme--${n.value}`);return{install:l,isDisabled:e.isDisabled,name:n,themes:s,current:i,computedThemes:r,themeClasses:u,styles:o,global:{name:n,current:i}}}function It(t){Ge("provideTheme");const e=Pe(Fi,null);if(!e)throw new Error("Could not find Vuetify theme injection");const n=I(()=>t.theme??e.name.value),s=I(()=>e.themes.value[n.value]),r=I(()=>e.isDisabled?void 0:`v-theme--${n.value}`),i={...e,name:n,current:s,themeClasses:r};return ut(Fi,i),i}function Hn(t,e,n){t.push(`${e} {
`,...n.map(s=>`  ${s};
`),`}
`)}function Yu(t){const e=t.dark?2:1,n=t.dark?1:2,s=[];for(const[r,i]of Object.entries(t.colors)){const o=Ut(i);s.push(`--v-theme-${r}: ${o.r},${o.g},${o.b}`),r.startsWith("on-")||s.push(`--v-theme-${r}-overlay-multiplier: ${kp(i)>.18?e:n}`)}for(const[r,i]of Object.entries(t.variables)){const o=typeof i=="string"&&i.startsWith("#")?Ut(i):void 0,a=o?`${o.r}, ${o.g}, ${o.b}`:void 0;s.push(`--v-${r}: ${a??i}`)}return s}const Je=G({tag:{type:String,default:"div"}},"tag"),ny=G({text:String,...we(),...Je()},"VToolbarTitle"),Vi=he()({name:"VToolbarTitle",props:ny(),setup(t,e){let{slots:n}=e;return ye(()=>{const s=!!(n.default||n.text||t.text);return _(t.tag,{class:["v-toolbar-title",t.class],style:t.style},{default:()=>{var r;return[s&&_("div",{class:"v-toolbar-title__placeholder"},[n.text?n.text():t.text,(r=n.default)==null?void 0:r.call(n)])]}})}),{}}}),sy=G({disabled:Boolean,group:Boolean,hideOnLeave:Boolean,leaveAbsolute:Boolean,mode:String,origin:String},"transition");function wt(t,e,n){return he()({name:t,props:sy({mode:n,origin:e}),setup(s,r){let{slots:i}=r;const o={onBeforeEnter(a){s.origin&&(a.style.transformOrigin=s.origin)},onLeave(a){if(s.leaveAbsolute){const{offsetTop:l,offsetLeft:u,offsetWidth:c,offsetHeight:d}=a;a._transitionInitialStyles={position:a.style.position,top:a.style.top,left:a.style.left,width:a.style.width,height:a.style.height},a.style.position="absolute",a.style.top=`${l}px`,a.style.left=`${u}px`,a.style.width=`${c}px`,a.style.height=`${d}px`}s.hideOnLeave&&a.style.setProperty("display","none","important")},onAfterLeave(a){if(s.leaveAbsolute&&(a!=null&&a._transitionInitialStyles)){const{position:l,top:u,left:c,width:d,height:f}=a._transitionInitialStyles;delete a._transitionInitialStyles,a.style.position=l||"",a.style.top=u||"",a.style.left=c||"",a.style.width=d||"",a.style.height=f||""}}};return()=>{const a=s.group?lf:qt;return On(a,{name:s.disabled?"":t,css:!s.disabled,...s.group?void 0:{mode:s.mode},...s.disabled?{}:o},i.default)}}})}function Vf(t,e){let n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:"in-out";return he()({name:t,props:{mode:{type:String,default:n},disabled:Boolean},setup(s,r){let{slots:i}=r;return()=>On(qt,{name:s.disabled?"":t,css:!s.disabled,...s.disabled?{}:e},i.default)}})}function Lf(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:"";const n=(arguments.length>1&&arguments[1]!==void 0?arguments[1]:!1)?"width":"height",s=bt(`offset-${n}`);return{onBeforeEnter(o){o._parent=o.parentNode,o._initialStyle={transition:o.style.transition,overflow:o.style.overflow,[n]:o.style[n]}},onEnter(o){const a=o._initialStyle;o.style.setProperty("transition","none","important"),o.style.overflow="hidden";const l=`${o[s]}px`;o.style[n]="0",o.offsetHeight,o.style.transition=a.transition,t&&o._parent&&o._parent.classList.add(t),requestAnimationFrame(()=>{o.style[n]=l})},onAfterEnter:i,onEnterCancelled:i,onLeave(o){o._initialStyle={transition:"",overflow:o.style.overflow,[n]:o.style[n]},o.style.overflow="hidden",o.style[n]=`${o[s]}px`,o.offsetHeight,requestAnimationFrame(()=>o.style[n]="0")},onAfterLeave:r,onLeaveCancelled:r};function r(o){t&&o._parent&&o._parent.classList.remove(t),i(o)}function i(o){const a=o._initialStyle[n];o.style.overflow=o._initialStyle.overflow,a!=null&&(o.style[n]=a),delete o._initialStyle}}const ry=G({target:[Object,Array]},"v-dialog-transition"),iy=he()({name:"VDialogTransition",props:ry(),setup(t,e){let{slots:n}=e;const s={onBeforeEnter(r){r.style.pointerEvents="none",r.style.visibility="hidden"},async onEnter(r,i){var f;await new Promise(h=>requestAnimationFrame(h)),await new Promise(h=>requestAnimationFrame(h)),r.style.visibility="";const{x:o,y:a,sx:l,sy:u,speed:c}=Zu(t.target,r),d=Ys(r,[{transform:`translate(${o}px, ${a}px) scale(${l}, ${u})`,opacity:0},{}],{duration:225*c,easing:Pp});(f=Xu(r))==null||f.forEach(h=>{Ys(h,[{opacity:0},{opacity:0,offset:.33},{}],{duration:225*2*c,easing:ba})}),d.finished.then(()=>i())},onAfterEnter(r){r.style.removeProperty("pointer-events")},onBeforeLeave(r){r.style.pointerEvents="none"},async onLeave(r,i){var f;await new Promise(h=>requestAnimationFrame(h));const{x:o,y:a,sx:l,sy:u,speed:c}=Zu(t.target,r);Ys(r,[{},{transform:`translate(${o}px, ${a}px) scale(${l}, ${u})`,opacity:0}],{duration:125*c,easing:Op}).finished.then(()=>i()),(f=Xu(r))==null||f.forEach(h=>{Ys(h,[{},{opacity:0,offset:.2},{opacity:0}],{duration:125*2*c,easing:ba})})},onAfterLeave(r){r.style.removeProperty("pointer-events")}};return()=>t.target?_(qt,Le({name:"dialog-transition"},s,{css:!1}),n):_(qt,{name:"dialog-transition"},n)}});function Xu(t){var n;const e=(n=t.querySelector(":scope > .v-card, :scope > .v-sheet, :scope > .v-list"))==null?void 0:n.children;return e&&[...e]}function Zu(t,e){const n=bf(t),s=xf(e),[r,i]=getComputedStyle(e).transformOrigin.split(" ").map(S=>parseFloat(S)),[o,a]=getComputedStyle(e).getPropertyValue("--v-overlay-anchor-origin").split(" ");let l=n.left+n.width/2;o==="left"||a==="left"?l-=n.width/2:(o==="right"||a==="right")&&(l+=n.width/2);let u=n.top+n.height/2;o==="top"||a==="top"?u-=n.height/2:(o==="bottom"||a==="bottom")&&(u+=n.height/2);const c=n.width/s.width,d=n.height/s.height,f=Math.max(1,c,d),h=c/f||0,g=d/f||0,y=s.width*s.height/(window.innerWidth*window.innerHeight),v=y>.12?Math.min(1.5,(y-.12)*10+1):1;return{x:l-(r+s.left),y:u-(i+s.top),sx:h,sy:g,speed:v}}wt("fab-transition","center center","out-in");wt("dialog-bottom-transition");wt("dialog-top-transition");const Tw=wt("fade-transition"),oy=wt("scale-transition");wt("scroll-x-transition");wt("scroll-x-reverse-transition");wt("scroll-y-transition");wt("scroll-y-reverse-transition");wt("slide-x-transition");wt("slide-x-reverse-transition");const ay=wt("slide-y-transition");wt("slide-y-reverse-transition");const Pf=Vf("expand-transition",Lf()),Aw=Vf("expand-x-transition",Lf("",!0)),ly=G({defaults:Object,disabled:Boolean,reset:[Number,String],root:[Boolean,String],scoped:Boolean},"VDefaultsProvider"),Nt=he(!1)({name:"VDefaultsProvider",props:ly(),setup(t,e){let{slots:n}=e;const{defaults:s,disabled:r,reset:i,root:o,scoped:a}=Ya(t);return hn(s,{reset:i,root:o,scoped:a,disabled:r}),()=>{var l;return(l=n.default)==null?void 0:l.call(n)}}}),lo=G({height:[Number,String],maxHeight:[Number,String],maxWidth:[Number,String],minHeight:[Number,String],minWidth:[Number,String],width:[Number,String]},"dimension");function uo(t){return{dimensionStyles:I(()=>({height:ie(t.height),maxHeight:ie(t.maxHeight),maxWidth:ie(t.maxWidth),minHeight:ie(t.minHeight),minWidth:ie(t.minWidth),width:ie(t.width)}))}}function uy(t){return{aspectStyles:I(()=>{const e=Number(t.aspectRatio);return e?{paddingBottom:String(1/e*100)+"%"}:void 0})}}const Of=G({aspectRatio:[String,Number],contentClass:String,inline:Boolean,...we(),...lo()},"VResponsive"),Ku=he()({name:"VResponsive",props:Of(),setup(t,e){let{slots:n}=e;const{aspectStyles:s}=uy(t),{dimensionStyles:r}=uo(t);return ye(()=>{var i;return _("div",{class:["v-responsive",{"v-responsive--inline":t.inline},t.class],style:[r.value,t.style]},[_("div",{class:"v-responsive__sizer",style:s.value},null),(i=n.additional)==null?void 0:i.call(n),n.default&&_("div",{class:["v-responsive__content",t.contentClass]},[n.default()])])}),{}}});function ul(t){return il(()=>{const e=[],n={};if(t.value.background)if(ya(t.value.background)){if(n.backgroundColor=t.value.background,!t.value.text&&Cp(t.value.background)){const s=Ut(t.value.background);if(s.a==null||s.a===1){const r=Ef(s);n.color=r,n.caretColor=r}}}else e.push(`bg-${t.value.background}`);return t.value.text&&(ya(t.value.text)?(n.color=t.value.text,n.caretColor=t.value.text):e.push(`text-${t.value.text}`)),{colorClasses:e,colorStyles:n}})}function Ln(t,e){const n=I(()=>({text:Te(t)?t.value:e?t[e]:null})),{colorClasses:s,colorStyles:r}=ul(n);return{textColorClasses:s,textColorStyles:r}}function Tt(t,e){const n=I(()=>({background:Te(t)?t.value:e?t[e]:null})),{colorClasses:s,colorStyles:r}=ul(n);return{backgroundColorClasses:s,backgroundColorStyles:r}}const Kt=G({rounded:{type:[Boolean,Number,String],default:void 0},tile:Boolean},"rounded");function Jt(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();return{roundedClasses:I(()=>{const s=Te(t)?t.value:t.rounded,r=Te(t)?t.value:t.tile,i=[];if(s===!0||s==="")i.push(`${e}--rounded`);else if(typeof s=="string"||s===0)for(const o of String(s).split(" "))i.push(`rounded-${o}`);else r&&i.push("rounded-0");return i})}}const cl=G({transition:{type:[Boolean,String,Object],default:"fade-transition",validator:t=>t!==!0}},"transition"),Cs=(t,e)=>{let{slots:n}=e;const{transition:s,disabled:r,group:i,...o}=t,{component:a=i?lf:qt,...l}=typeof s=="object"?s:{};return On(a,Le(typeof s=="string"?{name:r?"":s}:l,typeof s=="string"?{}:{disabled:r,group:i},o),n)};function cy(t,e){if(!nl)return;const n=e.modifiers||{},s=e.value,{handler:r,options:i}=typeof s=="object"?s:{handler:s,options:{}},o=new IntersectionObserver(function(){var d;let a=arguments.length>0&&arguments[0]!==void 0?arguments[0]:[],l=arguments.length>1?arguments[1]:void 0;const u=(d=t._observe)==null?void 0:d[e.instance.$.uid];if(!u)return;const c=a.some(f=>f.isIntersecting);r&&(!n.quiet||u.init)&&(!n.once||c||u.init)&&r(c,a,l),c&&n.once?Nf(t,e):u.init=!0},i);t._observe=Object(t._observe),t._observe[e.instance.$.uid]={init:!1,observer:o},o.observe(t)}function Nf(t,e){var s;const n=(s=t._observe)==null?void 0:s[e.instance.$.uid];n&&(n.observer.unobserve(t),delete t._observe[e.instance.$.uid])}const Bf={mounted:cy,unmounted:Nf},dy=G({alt:String,cover:Boolean,color:String,draggable:{type:[Boolean,String],default:void 0},eager:Boolean,gradient:String,lazySrc:String,options:{type:Object,default:()=>({root:void 0,rootMargin:void 0,threshold:void 0})},sizes:String,src:{type:[String,Object],default:""},crossorigin:String,referrerpolicy:String,srcset:String,position:String,...Of(),...we(),...Kt(),...cl()},"VImg"),dl=he()({name:"VImg",directives:{intersect:Bf},props:dy(),emits:{loadstart:t=>!0,load:t=>!0,error:t=>!0},setup(t,e){let{emit:n,slots:s}=e;const{backgroundColorClasses:r,backgroundColorStyles:i}=Tt(de(t,"color")),{roundedClasses:o}=Jt(t),a=Ge("VImg"),l=be(""),u=se(),c=be(t.eager?"loading":"idle"),d=be(),f=be(),h=I(()=>t.src&&typeof t.src=="object"?{src:t.src.src,srcset:t.srcset||t.src.srcset,lazySrc:t.lazySrc||t.src.lazySrc,aspect:Number(t.aspectRatio||t.src.aspect||0)}:{src:t.src,srcset:t.srcset,lazySrc:t.lazySrc,aspect:Number(t.aspectRatio||0)}),g=I(()=>h.value.aspect||d.value/f.value||0);le(()=>t.src,()=>{y(c.value!=="idle")}),le(g,(A,k)=>{!A&&k&&u.value&&w(u.value)}),to(()=>y());function y(A){if(!(t.eager&&A)&&!(nl&&!A&&!t.eager)){if(c.value="loading",h.value.lazySrc){const k=new Image;k.src=h.value.lazySrc,w(k,null)}h.value.src&&xt(()=>{var k;n("loadstart",((k=u.value)==null?void 0:k.currentSrc)||h.value.src),setTimeout(()=>{var O;if(!a.isUnmounted)if((O=u.value)!=null&&O.complete){if(u.value.naturalWidth||S(),c.value==="error")return;g.value||w(u.value,null),c.value==="loading"&&v()}else g.value||w(u.value),m()})})}}function v(){var A;a.isUnmounted||(m(),w(u.value),c.value="loaded",n("load",((A=u.value)==null?void 0:A.currentSrc)||h.value.src))}function S(){var A;a.isUnmounted||(c.value="error",n("error",((A=u.value)==null?void 0:A.currentSrc)||h.value.src))}function m(){const A=u.value;A&&(l.value=A.currentSrc||A.src)}let p=-1;Rt(()=>{clearTimeout(p)});function w(A){let k=arguments.length>1&&arguments[1]!==void 0?arguments[1]:100;const O=()=>{if(clearTimeout(p),a.isUnmounted)return;const{naturalHeight:H,naturalWidth:q}=A;H||q?(d.value=q,f.value=H):!A.complete&&c.value==="loading"&&k!=null?p=window.setTimeout(O,k):(A.currentSrc.endsWith(".svg")||A.currentSrc.startsWith("data:image/svg+xml"))&&(d.value=1,f.value=1)};O()}const b=I(()=>({"v-img__img--cover":t.cover,"v-img__img--contain":!t.cover})),x=()=>{var O;if(!h.value.src||c.value==="idle")return null;const A=_("img",{class:["v-img__img",b.value],style:{objectPosition:t.position},src:h.value.src,srcset:h.value.srcset,alt:t.alt,crossorigin:t.crossorigin,referrerpolicy:t.referrerpolicy,draggable:t.draggable,sizes:t.sizes,ref:u,onLoad:v,onError:S},null),k=(O=s.sources)==null?void 0:O.call(s);return _(Cs,{transition:t.transition,appear:!0},{default:()=>[Gt(k?_("picture",{class:"v-img__picture"},[k,A]):A,[[io,c.value==="loaded"]])]})},C=()=>_(Cs,{transition:t.transition},{default:()=>[h.value.lazySrc&&c.value!=="loaded"&&_("img",{class:["v-img__img","v-img__img--preload",b.value],style:{objectPosition:t.position},src:h.value.lazySrc,alt:t.alt,crossorigin:t.crossorigin,referrerpolicy:t.referrerpolicy,draggable:t.draggable},null)]}),M=()=>s.placeholder?_(Cs,{transition:t.transition,appear:!0},{default:()=>[(c.value==="loading"||c.value==="error"&&!s.error)&&_("div",{class:"v-img__placeholder"},[s.placeholder()])]}):null,T=()=>s.error?_(Cs,{transition:t.transition,appear:!0},{default:()=>[c.value==="error"&&_("div",{class:"v-img__error"},[s.error()])]}):null,V=()=>t.gradient?_("div",{class:"v-img__gradient",style:{backgroundImage:`linear-gradient(${t.gradient})`}},null):null,L=be(!1);{const A=le(g,k=>{k&&(requestAnimationFrame(()=>{requestAnimationFrame(()=>{L.value=!0})}),A())})}return ye(()=>{const A=Ku.filterProps(t);return Gt(_(Ku,Le({class:["v-img",{"v-img--booting":!L.value},r.value,o.value,t.class],style:[{width:ie(t.width==="auto"?d.value:t.width)},i.value,t.style]},A,{aspectRatio:g.value,"aria-label":t.alt,role:t.alt?"img":void 0}),{additional:()=>_(Re,null,[_(x,null,null),_(C,null,null),_(V,null,null),_(M,null,null),_(T,null,null)]),default:s.default}),[[as("intersect"),{handler:y,options:t.options},null,{once:!0}]])}),{currentSrc:l,image:u,state:c,naturalWidth:d,naturalHeight:f}}}),Ir=G({border:[Boolean,Number,String]},"border");function kr(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();return{borderClasses:I(()=>{const s=Te(t)?t.value:t.border,r=[];if(s===!0||s==="")r.push(`${e}--border`);else if(typeof s=="string"||s===0)for(const i of String(s).split(" "))r.push(`border-${i}`);return r})}}const ls=G({elevation:{type:[Number,String],validator(t){const e=parseInt(t);return!isNaN(e)&&e>=0&&e<=24}}},"elevation");function us(t){return{elevationClasses:I(()=>{const n=Te(t)?t.value:t.elevation,s=[];return n==null||s.push(`elevation-${n}`),s})}}const fy=[null,"prominent","default","comfortable","compact"],hy=G({absolute:Boolean,collapse:Boolean,color:String,density:{type:String,default:"default",validator:t=>fy.includes(t)},extended:Boolean,extensionHeight:{type:[Number,String],default:48},flat:Boolean,floating:Boolean,height:{type:[Number,String],default:64},image:String,title:String,...Ir(),...we(),...ls(),...Kt(),...Je({tag:"header"}),...dt()},"VToolbar"),zf=he()({name:"VToolbar",props:hy(),setup(t,e){var h;let{slots:n}=e;const{backgroundColorClasses:s,backgroundColorStyles:r}=Tt(de(t,"color")),{borderClasses:i}=kr(t),{elevationClasses:o}=us(t),{roundedClasses:a}=Jt(t),{themeClasses:l}=It(t),{rtlClasses:u}=Zt(),c=be(!!(t.extended||(h=n.extension)!=null&&h.call(n))),d=I(()=>parseInt(Number(t.height)+(t.density==="prominent"?Number(t.height):0)-(t.density==="comfortable"?8:0)-(t.density==="compact"?16:0),10)),f=I(()=>c.value?parseInt(Number(t.extensionHeight)+(t.density==="prominent"?Number(t.extensionHeight):0)-(t.density==="comfortable"?4:0)-(t.density==="compact"?8:0),10):0);return hn({VBtn:{variant:"text"}}),ye(()=>{var S;const g=!!(t.title||n.title),y=!!(n.image||t.image),v=(S=n.extension)==null?void 0:S.call(n);return c.value=!!(t.extended||v),_(t.tag,{class:["v-toolbar",{"v-toolbar--absolute":t.absolute,"v-toolbar--collapse":t.collapse,"v-toolbar--flat":t.flat,"v-toolbar--floating":t.floating,[`v-toolbar--density-${t.density}`]:!0},s.value,i.value,o.value,a.value,l.value,u.value,t.class],style:[r.value,t.style]},{default:()=>[y&&_("div",{key:"image",class:"v-toolbar__image"},[n.image?_(Nt,{key:"image-defaults",disabled:!t.image,defaults:{VImg:{cover:!0,src:t.image}}},n.image):_(dl,{key:"image-img",cover:!0,src:t.image},null)]),_(Nt,{defaults:{VTabs:{height:ie(d.value)}}},{default:()=>{var m,p,w;return[_("div",{class:"v-toolbar__content",style:{height:ie(d.value)}},[n.prepend&&_("div",{class:"v-toolbar__prepend"},[(m=n.prepend)==null?void 0:m.call(n)]),g&&_(Vi,{key:"title",text:t.title},{text:n.title}),(p=n.default)==null?void 0:p.call(n),n.append&&_("div",{class:"v-toolbar__append"},[(w=n.append)==null?void 0:w.call(n)])])]}}),_(Nt,{defaults:{VTabs:{height:ie(f.value)}}},{default:()=>[_(Pf,null,{default:()=>[c.value&&_("div",{class:"v-toolbar__extension",style:{height:ie(f.value)}},[v])]})]})]})}),{contentHeight:d,extensionHeight:f}}});function Df(){const t=be(!1);return Yt(()=>{window.requestAnimationFrame(()=>{t.value=!0})}),{ssrBootStyles:I(()=>t.value?void 0:{transition:"none !important"}),isBooted:_r(t)}}const my=[null,"default","comfortable","compact"],cs=G({density:{type:String,default:"default",validator:t=>my.includes(t)}},"density");function Ns(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();return{densityClasses:I(()=>`${e}--density-${t.density}`)}}const gy=["elevated","flat","tonal","outlined","text","plain"];function fl(t,e){return _(Re,null,[t&&_("span",{key:"overlay",class:`${e}__overlay`},null),_("span",{key:"underlay",class:`${e}__underlay`},null)])}const Fr=G({color:String,variant:{type:String,default:"elevated",validator:t=>gy.includes(t)}},"variant");function hl(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();const n=I(()=>{const{variant:i}=F(t);return`${e}--variant-${i}`}),{colorClasses:s,colorStyles:r}=ul(I(()=>{const{variant:i,color:o}=F(t);return{[["elevated","flat"].includes(i)?"background":"text"]:o}}));return{colorClasses:s,colorStyles:r,variantClasses:n}}const $f=G({divided:Boolean,...Ir(),...we(),...cs(),...ls(),...Kt(),...Je(),...dt(),...Fr()},"VBtnGroup"),Ju=he()({name:"VBtnGroup",props:$f(),setup(t,e){let{slots:n}=e;const{themeClasses:s}=It(t),{densityClasses:r}=Ns(t),{borderClasses:i}=kr(t),{elevationClasses:o}=us(t),{roundedClasses:a}=Jt(t);hn({VBtn:{height:"auto",color:de(t,"color"),density:de(t,"density"),flat:!0,variant:de(t,"variant")}}),ye(()=>_(t.tag,{class:["v-btn-group",{"v-btn-group--divided":t.divided},s.value,i.value,r.value,o.value,a.value,t.class],style:t.style},n))}}),Hf=G({modelValue:{type:null,default:void 0},multiple:Boolean,mandatory:[Boolean,String],max:Number,selectedClass:String,disabled:Boolean},"group"),Uf=G({value:null,disabled:Boolean,selectedClass:String},"group-item");function jf(t,e){let n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0;const s=Ge("useGroupItem");if(!s)throw new Error("[Vuetify] useGroupItem composable must be used inside a component setup function");const r=mn();ut(Symbol.for(`${e.description}:id`),r);const i=Pe(e,null);if(!i){if(!n)return i;throw new Error(`[Vuetify] Could not find useGroup injection with symbol ${e.description}`)}const o=de(t,"value"),a=I(()=>!!(i.disabled.value||t.disabled));i.register({id:r,value:o,disabled:a},s),Rt(()=>{i.unregister(r)});const l=I(()=>i.isSelected(r)),u=I(()=>l.value&&[i.selectedClass.value,t.selectedClass]);return le(l,c=>{s.emit("group:selected",{value:c})},{flush:"sync"}),{id:r,isSelected:l,toggle:()=>i.select(r,!l.value),select:c=>i.select(r,c),selectedClass:u,value:o,disabled:a,group:i}}function Wf(t,e){let n=!1;const s=tt([]),r=ct(t,"modelValue",[],f=>f==null?[]:Gf(s,In(f)),f=>{const h=py(s,f);return t.multiple?h:h[0]}),i=Ge("useGroup");function o(f,h){const g=f,y=Symbol.for(`${e.description}:id`),S=ir(y,i==null?void 0:i.vnode).indexOf(h);F(g.value)==null&&(g.value=S),S>-1?s.splice(S,0,g):s.push(g)}function a(f){if(n)return;l();const h=s.findIndex(g=>g.id===f);s.splice(h,1)}function l(){const f=s.find(h=>!h.disabled);f&&t.mandatory==="force"&&!r.value.length&&(r.value=[f.id])}Yt(()=>{l()}),Rt(()=>{n=!0});function u(f,h){const g=s.find(y=>y.id===f);if(!(h&&(g!=null&&g.disabled)))if(t.multiple){const y=r.value.slice(),v=y.findIndex(m=>m===f),S=~v;if(h=h??!S,S&&t.mandatory&&y.length<=1||!S&&t.max!=null&&y.length+1>t.max)return;v<0&&h?y.push(f):v>=0&&!h&&y.splice(v,1),r.value=y}else{const y=r.value.includes(f);if(t.mandatory&&y)return;r.value=h??!y?[f]:[]}}function c(f){if(t.multiple,r.value.length){const h=r.value[0],g=s.findIndex(S=>S.id===h);let y=(g+f)%s.length,v=s[y];for(;v.disabled&&y!==g;)y=(y+f)%s.length,v=s[y];if(v.disabled)return;r.value=[s[y].id]}else{const h=s.find(g=>!g.disabled);h&&(r.value=[h.id])}}const d={register:o,unregister:a,selected:r,select:u,disabled:de(t,"disabled"),prev:()=>c(s.length-1),next:()=>c(1),isSelected:f=>r.value.includes(f),selectedClass:I(()=>t.selectedClass),items:I(()=>s),getItemIndex:f=>vy(s,f)};return ut(e,d),d}function vy(t,e){const n=Gf(t,[e]);return n.length?t.findIndex(s=>s.id===n[0]):-1}function Gf(t,e){const n=[];return e.forEach(s=>{const r=t.find(o=>oo(s,o.value)),i=t[s];(r==null?void 0:r.value)!=null?n.push(r.id):i!=null&&n.push(i.id)}),n}function py(t,e){const n=[];return e.forEach(s=>{const r=t.findIndex(i=>i.id===s);if(~r){const i=t[r];n.push(i.value!=null?i.value:r)}}),n}const qf=Symbol.for("vuetify:v-btn-toggle"),yy=G({...$f(),...Hf()},"VBtnToggle"),by=he()({name:"VBtnToggle",props:yy(),emits:{"update:modelValue":t=>!0},setup(t,e){let{slots:n}=e;const{isSelected:s,next:r,prev:i,select:o,selected:a}=Wf(t,qf);return ye(()=>{const l=Ju.filterProps(t);return _(Ju,Le({class:["v-btn-toggle",t.class]},l,{style:t.style}),{default:()=>{var u;return[(u=n.default)==null?void 0:u.call(n,{isSelected:s,next:r,prev:i,select:o,selected:a})]}})}),{next:r,prev:i,select:o}}}),xy={collapse:"mdi-chevron-up",complete:"mdi-check",cancel:"mdi-close-circle",close:"mdi-close",delete:"mdi-close-circle",clear:"mdi-close-circle",success:"mdi-check-circle",info:"mdi-information",warning:"mdi-alert-circle",error:"mdi-close-circle",prev:"mdi-chevron-left",next:"mdi-chevron-right",checkboxOn:"mdi-checkbox-marked",checkboxOff:"mdi-checkbox-blank-outline",checkboxIndeterminate:"mdi-minus-box",delimiter:"mdi-circle",sortAsc:"mdi-arrow-up",sortDesc:"mdi-arrow-down",expand:"mdi-chevron-down",menu:"mdi-menu",subgroup:"mdi-menu-down",dropdown:"mdi-menu-down",radioOn:"mdi-radiobox-marked",radioOff:"mdi-radiobox-blank",edit:"mdi-pencil",ratingEmpty:"mdi-star-outline",ratingFull:"mdi-star",ratingHalf:"mdi-star-half-full",loading:"mdi-cached",first:"mdi-page-first",last:"mdi-page-last",unfold:"mdi-unfold-more-horizontal",file:"mdi-paperclip",plus:"mdi-plus",minus:"mdi-minus",calendar:"mdi-calendar",eyeDropper:"mdi-eyedropper"},wy={component:t=>On(Xf,{...t,class:"mdi"})},Ke=[String,Function,Object,Array],xa=Symbol.for("vuetify:icons"),co=G({icon:{type:Ke},tag:{type:String,required:!0}},"icon"),Qu=he()({name:"VComponentIcon",props:co(),setup(t,e){let{slots:n}=e;return()=>{const s=t.icon;return _(t.tag,null,{default:()=>{var r;return[t.icon?_(s,null,null):(r=n.default)==null?void 0:r.call(n)]}})}}}),Yf=Rr({name:"VSvgIcon",inheritAttrs:!1,props:co(),setup(t,e){let{attrs:n}=e;return()=>_(t.tag,Le(n,{style:null}),{default:()=>[_("svg",{class:"v-icon__svg",xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24",role:"img","aria-hidden":"true"},[Array.isArray(t.icon)?t.icon.map(s=>Array.isArray(s)?_("path",{d:s[0],"fill-opacity":s[1]},null):_("path",{d:s},null)):_("path",{d:t.icon},null)])]})}});Rr({name:"VLigatureIcon",props:co(),setup(t){return()=>_(t.tag,null,{default:()=>[t.icon]})}});const Xf=Rr({name:"VClassIcon",props:co(),setup(t){return()=>_(t.tag,{class:t.icon},null)}});function _y(){return{svg:{component:Yf},class:{component:Xf}}}function Sy(t){const e=_y(),n=(t==null?void 0:t.defaultSet)??"mdi";return n==="mdi"&&!e.mdi&&(e.mdi=wy),vt({defaultSet:n,sets:e,aliases:{...xy,vuetify:["M8.2241 14.2009L12 21L22 3H14.4459L8.2241 14.2009Z",["M7.26303 12.4733L7.00113 12L2 3H12.5261C12.5261 3 12.5261 3 12.5261 3L7.26303 12.4733Z",.6]],"vuetify-outline":"svg:M7.26 12.47 12.53 3H2L7.26 12.47ZM14.45 3 8.22 14.2 12 21 22 3H14.45ZM18.6 5 12 16.88 10.51 14.2 15.62 5ZM7.26 8.35 5.4 5H9.13L7.26 8.35Z"}},t)}const Cy=t=>{const e=Pe(xa);if(!e)throw new Error("Missing Vuetify Icons provide!");return{iconData:I(()=>{var l;const s=F(t);if(!s)return{component:Qu};let r=s;if(typeof r=="string"&&(r=r.trim(),r.startsWith("$")&&(r=(l=e.aliases)==null?void 0:l[r.slice(1)])),!r)throw new Error(`Could not find aliased icon "${s}"`);if(Array.isArray(r))return{component:Yf,icon:r};if(typeof r!="string")return{component:Qu,icon:r};const i=Object.keys(e.sets).find(u=>typeof r=="string"&&r.startsWith(`${u}:`)),o=i?r.slice(i.length+1):r;return{component:e.sets[i??e.defaultSet].component,icon:o}})}},Ey=["x-small","small","default","large","x-large"],fo=G({size:{type:[String,Number],default:"default"}},"size");function ho(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();return il(()=>{let n,s;return Mi(Ey,t.size)?n=`${e}--size-${t.size}`:t.size&&(s={width:ie(t.size),height:ie(t.size)}),{sizeClasses:n,sizeStyles:s}})}const My=G({color:String,start:Boolean,end:Boolean,icon:Ke,...we(),...fo(),...Je({tag:"i"}),...dt()},"VIcon"),an=he()({name:"VIcon",props:My(),setup(t,e){let{attrs:n,slots:s}=e;const r=se(),{themeClasses:i}=It(t),{iconData:o}=Cy(I(()=>r.value||t.icon)),{sizeClasses:a}=ho(t),{textColorClasses:l,textColorStyles:u}=Ln(de(t,"color"));return ye(()=>{var d,f;const c=(d=s.default)==null?void 0:d.call(s);return c&&(r.value=(f=mf(c).filter(h=>h.type===Er&&h.children&&typeof h.children=="string")[0])==null?void 0:f.children),_(o.value.component,{tag:t.tag,icon:o.value.icon,class:["v-icon","notranslate",i.value,a.value,l.value,{"v-icon--clickable":!!n.onClick,"v-icon--start":t.start,"v-icon--end":t.end},t.class],style:[a.value?void 0:{fontSize:ie(t.size),height:ie(t.size),width:ie(t.size)},u.value,t.style],role:n.onClick?"button":void 0,"aria-hidden":!n.onClick},{default:()=>[c]})}),{}}});function Zf(t,e){const n=se(),s=be(!1);if(nl){const r=new IntersectionObserver(i=>{t==null||t(i,r),s.value=!!i.find(o=>o.isIntersecting)},e);Rt(()=>{r.disconnect()}),le(n,(i,o)=>{o&&(r.unobserve(o),s.value=!1),i&&r.observe(i)},{flush:"post"})}return{intersectionRef:n,isIntersecting:s}}const Ty=G({bgColor:String,color:String,indeterminate:[Boolean,String],modelValue:{type:[Number,String],default:0},rotate:{type:[Number,String],default:0},width:{type:[Number,String],default:4},...we(),...fo(),...Je({tag:"div"}),...dt()},"VProgressCircular"),Kf=he()({name:"VProgressCircular",props:Ty(),setup(t,e){let{slots:n}=e;const s=20,r=2*Math.PI*s,i=se(),{themeClasses:o}=It(t),{sizeClasses:a,sizeStyles:l}=ho(t),{textColorClasses:u,textColorStyles:c}=Ln(de(t,"color")),{textColorClasses:d,textColorStyles:f}=Ln(de(t,"bgColor")),{intersectionRef:h,isIntersecting:g}=Zf(),{resizeRef:y,contentRect:v}=Af(),S=I(()=>Math.max(0,Math.min(100,parseFloat(t.modelValue)))),m=I(()=>Number(t.width)),p=I(()=>l.value?Number(t.size):v.value?v.value.width:Math.max(m.value,32)),w=I(()=>s/(1-m.value/p.value)*2),b=I(()=>m.value/p.value*w.value),x=I(()=>ie((100-S.value)/100*r));return Pn(()=>{h.value=i.value,y.value=i.value}),ye(()=>_(t.tag,{ref:i,class:["v-progress-circular",{"v-progress-circular--indeterminate":!!t.indeterminate,"v-progress-circular--visible":g.value,"v-progress-circular--disable-shrink":t.indeterminate==="disable-shrink"},o.value,a.value,u.value,t.class],style:[l.value,c.value,t.style],role:"progressbar","aria-valuemin":"0","aria-valuemax":"100","aria-valuenow":t.indeterminate?void 0:S.value},{default:()=>[_("svg",{style:{transform:`rotate(calc(-90deg + ${Number(t.rotate)}deg))`},xmlns:"http://www.w3.org/2000/svg",viewBox:`0 0 ${w.value} ${w.value}`},[_("circle",{class:["v-progress-circular__underlay",d.value],style:f.value,fill:"transparent",cx:"50%",cy:"50%",r:s,"stroke-width":b.value,"stroke-dasharray":r,"stroke-dashoffset":0},null),_("circle",{class:"v-progress-circular__overlay",fill:"transparent",cx:"50%",cy:"50%",r:s,"stroke-width":b.value,"stroke-dasharray":r,"stroke-dashoffset":x.value},null)]),n.default&&_("div",{class:"v-progress-circular__content"},[n.default({value:S.value})])]})),{}}}),ec={center:"center",top:"bottom",bottom:"top",left:"right",right:"left"},ml=G({location:String},"location");function gl(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:!1,n=arguments.length>2?arguments[2]:void 0;const{isRtl:s}=Zt();return{locationStyles:I(()=>{if(!t.location)return{};const{side:i,align:o}=va(t.location.split(" ").length>1?t.location:`${t.location} center`,s.value);function a(u){return n?n(u):0}const l={};return i!=="center"&&(e?l[ec[i]]=`calc(100% - ${a(i)}px)`:l[i]=0),o!=="center"?e?l[ec[o]]=`calc(100% - ${a(o)}px)`:l[o]=0:(i==="center"?l.top=l.left="50%":l[{top:"left",bottom:"left",left:"top",right:"top"}[i]]="50%",l.transform={top:"translateX(-50%)",bottom:"translateX(-50%)",left:"translateY(-50%)",right:"translateY(-50%)",center:"translate(-50%, -50%)"}[i]),l})}}const Ay=G({absolute:Boolean,active:{type:Boolean,default:!0},bgColor:String,bgOpacity:[Number,String],bufferValue:{type:[Number,String],default:0},clickable:Boolean,color:String,height:{type:[Number,String],default:4},indeterminate:Boolean,max:{type:[Number,String],default:100},modelValue:{type:[Number,String],default:0},reverse:Boolean,stream:Boolean,striped:Boolean,roundedBar:Boolean,...we(),...ml({location:"top"}),...Kt(),...Je(),...dt()},"VProgressLinear"),Ry=he()({name:"VProgressLinear",props:Ay(),emits:{"update:modelValue":t=>!0},setup(t,e){let{slots:n}=e;const s=ct(t,"modelValue"),{isRtl:r,rtlClasses:i}=Zt(),{themeClasses:o}=It(t),{locationStyles:a}=gl(t),{textColorClasses:l,textColorStyles:u}=Ln(t,"color"),{backgroundColorClasses:c,backgroundColorStyles:d}=Tt(I(()=>t.bgColor||t.color)),{backgroundColorClasses:f,backgroundColorStyles:h}=Tt(t,"color"),{roundedClasses:g}=Jt(t),{intersectionRef:y,isIntersecting:v}=Zf(),S=I(()=>parseInt(t.max,10)),m=I(()=>parseInt(t.height,10)),p=I(()=>parseFloat(t.bufferValue)/S.value*100),w=I(()=>parseFloat(s.value)/S.value*100),b=I(()=>r.value!==t.reverse),x=I(()=>t.indeterminate?"fade-transition":"slide-x-transition"),C=I(()=>t.bgOpacity==null?t.bgOpacity:parseFloat(t.bgOpacity));function M(T){if(!y.value)return;const{left:V,right:L,width:A}=y.value.getBoundingClientRect(),k=b.value?A-T.clientX+(L-A):T.clientX-V;s.value=Math.round(k/A*S.value)}return ye(()=>_(t.tag,{ref:y,class:["v-progress-linear",{"v-progress-linear--absolute":t.absolute,"v-progress-linear--active":t.active&&v.value,"v-progress-linear--reverse":b.value,"v-progress-linear--rounded":t.rounded,"v-progress-linear--rounded-bar":t.roundedBar,"v-progress-linear--striped":t.striped},g.value,o.value,i.value,t.class],style:[{bottom:t.location==="bottom"?0:void 0,top:t.location==="top"?0:void 0,height:t.active?ie(m.value):0,"--v-progress-linear-height":ie(m.value),...a.value},t.style],role:"progressbar","aria-hidden":t.active?"false":"true","aria-valuemin":"0","aria-valuemax":t.max,"aria-valuenow":t.indeterminate?void 0:w.value,onClick:t.clickable&&M},{default:()=>[t.stream&&_("div",{key:"stream",class:["v-progress-linear__stream",l.value],style:{...u.value,[b.value?"left":"right"]:ie(-m.value),borderTop:`${ie(m.value/2)} dotted`,opacity:C.value,top:`calc(50% - ${ie(m.value/4)})`,width:ie(100-p.value,"%"),"--v-progress-linear-stream-to":ie(m.value*(b.value?1:-1))}},null),_("div",{class:["v-progress-linear__background",c.value],style:[d.value,{opacity:C.value,width:ie(t.stream?p.value:100,"%")}]},null),_(qt,{name:x.value},{default:()=>[t.indeterminate?_("div",{class:"v-progress-linear__indeterminate"},[["long","short"].map(T=>_("div",{key:T,class:["v-progress-linear__indeterminate",T,f.value],style:h.value},null))]):_("div",{class:["v-progress-linear__determinate",f.value],style:[h.value,{width:ie(w.value,"%")}]},null)]}),n.default&&_("div",{class:"v-progress-linear__content"},[n.default({value:w.value,buffer:p.value})])]})),{}}}),Jf=G({loading:[Boolean,String]},"loader");function Qf(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();return{loaderClasses:I(()=>({[`${e}--loading`]:t.loading}))}}function Iy(t,e){var s;let{slots:n}=e;return _("div",{class:`${t.name}__loader`},[((s=n.default)==null?void 0:s.call(n,{color:t.color,isActive:t.active}))||_(Ry,{absolute:t.absolute,active:t.active,color:t.color,height:"2",indeterminate:!0},null)])}const ky=["static","relative","fixed","absolute","sticky"],eh=G({position:{type:String,validator:t=>ky.includes(t)}},"position");function th(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();return{positionClasses:I(()=>t.position?`${e}--${t.position}`:void 0)}}function Fy(){const t=Ge("useRoute");return I(()=>{var e;return(e=t==null?void 0:t.proxy)==null?void 0:e.$route})}function nh(){var t,e;return(e=(t=Ge("useRouter"))==null?void 0:t.proxy)==null?void 0:e.$router}function sh(t,e){const n=mg("RouterLink"),s=I(()=>!!(t.href||t.to)),r=I(()=>(s==null?void 0:s.value)||Iu(e,"click")||Iu(t,"click"));if(typeof n=="string")return{isLink:s,isClickable:r,href:de(t,"href")};const i=t.to?n.useLink(t):void 0,o=Fy();return{isLink:s,isClickable:r,route:i==null?void 0:i.route,navigate:i==null?void 0:i.navigate,isActive:i&&I(()=>{var a,l,u;return t.exact?o.value?((u=i.isExactActive)==null?void 0:u.value)&&oo(i.route.value.query,o.value.query):(l=i.isExactActive)==null?void 0:l.value:(a=i.isActive)==null?void 0:a.value}),href:I(()=>t.to?i==null?void 0:i.route.value.href:t.href)}}const rh=G({href:String,replace:Boolean,to:[String,Object],exact:Boolean},"router");let Po=!1;function Vy(t,e){let n=!1,s,r;Oe&&(xt(()=>{window.addEventListener("popstate",i),s=t==null?void 0:t.beforeEach((o,a,l)=>{Po?n?e(l):l():setTimeout(()=>n?e(l):l()),Po=!0}),r=t==null?void 0:t.afterEach(()=>{Po=!1})}),At(()=>{window.removeEventListener("popstate",i),s==null||s(),r==null||r()}));function i(o){var a;(a=o.state)!=null&&a.replaced||(n=!0,setTimeout(()=>n=!1))}}function Ly(t,e){le(()=>{var n;return(n=t.isActive)==null?void 0:n.value},n=>{t.isLink.value&&n&&e&&xt(()=>{e(!0)})},{immediate:!0})}const wa=Symbol("rippleStop"),Py=80;function tc(t,e){t.style.transform=e,t.style.webkitTransform=e}function _a(t){return t.constructor.name==="TouchEvent"}function ih(t){return t.constructor.name==="KeyboardEvent"}const Oy=function(t,e){var d;let n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{},s=0,r=0;if(!ih(t)){const f=e.getBoundingClientRect(),h=_a(t)?t.touches[t.touches.length-1]:t;s=h.clientX-f.left,r=h.clientY-f.top}let i=0,o=.3;(d=e._ripple)!=null&&d.circle?(o=.15,i=e.clientWidth/2,i=n.center?i:i+Math.sqrt((s-i)**2+(r-i)**2)/4):i=Math.sqrt(e.clientWidth**2+e.clientHeight**2)/2;const a=`${(e.clientWidth-i*2)/2}px`,l=`${(e.clientHeight-i*2)/2}px`,u=n.center?a:`${s-i}px`,c=n.center?l:`${r-i}px`;return{radius:i,scale:o,x:u,y:c,centerX:a,centerY:l}},Li={show(t,e){var h;let n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{};if(!((h=e==null?void 0:e._ripple)!=null&&h.enabled))return;const s=document.createElement("span"),r=document.createElement("span");s.appendChild(r),s.className="v-ripple__container",n.class&&(s.className+=` ${n.class}`);const{radius:i,scale:o,x:a,y:l,centerX:u,centerY:c}=Oy(t,e,n),d=`${i*2}px`;r.className="v-ripple__animation",r.style.width=d,r.style.height=d,e.appendChild(s);const f=window.getComputedStyle(e);f&&f.position==="static"&&(e.style.position="relative",e.dataset.previousPosition="static"),r.classList.add("v-ripple__animation--enter"),r.classList.add("v-ripple__animation--visible"),tc(r,`translate(${a}, ${l}) scale3d(${o},${o},${o})`),r.dataset.activated=String(performance.now()),setTimeout(()=>{r.classList.remove("v-ripple__animation--enter"),r.classList.add("v-ripple__animation--in"),tc(r,`translate(${u}, ${c}) scale3d(1,1,1)`)},0)},hide(t){var i;if(!((i=t==null?void 0:t._ripple)!=null&&i.enabled))return;const e=t.getElementsByClassName("v-ripple__animation");if(e.length===0)return;const n=e[e.length-1];if(n.dataset.isHiding)return;n.dataset.isHiding="true";const s=performance.now()-Number(n.dataset.activated),r=Math.max(250-s,0);setTimeout(()=>{n.classList.remove("v-ripple__animation--in"),n.classList.add("v-ripple__animation--out"),setTimeout(()=>{var a;t.getElementsByClassName("v-ripple__animation").length===1&&t.dataset.previousPosition&&(t.style.position=t.dataset.previousPosition,delete t.dataset.previousPosition),((a=n.parentNode)==null?void 0:a.parentNode)===t&&t.removeChild(n.parentNode)},300)},r)}};function oh(t){return typeof t>"u"||!!t}function yr(t){const e={},n=t.currentTarget;if(!(!(n!=null&&n._ripple)||n._ripple.touched||t[wa])){if(t[wa]=!0,_a(t))n._ripple.touched=!0,n._ripple.isTouch=!0;else if(n._ripple.isTouch)return;if(e.center=n._ripple.centered||ih(t),n._ripple.class&&(e.class=n._ripple.class),_a(t)){if(n._ripple.showTimerCommit)return;n._ripple.showTimerCommit=()=>{Li.show(t,n,e)},n._ripple.showTimer=window.setTimeout(()=>{var s;(s=n==null?void 0:n._ripple)!=null&&s.showTimerCommit&&(n._ripple.showTimerCommit(),n._ripple.showTimerCommit=null)},Py)}else Li.show(t,n,e)}}function nc(t){t[wa]=!0}function gt(t){const e=t.currentTarget;if(e!=null&&e._ripple){if(window.clearTimeout(e._ripple.showTimer),t.type==="touchend"&&e._ripple.showTimerCommit){e._ripple.showTimerCommit(),e._ripple.showTimerCommit=null,e._ripple.showTimer=window.setTimeout(()=>{gt(t)});return}window.setTimeout(()=>{e._ripple&&(e._ripple.touched=!1)}),Li.hide(e)}}function ah(t){const e=t.currentTarget;e!=null&&e._ripple&&(e._ripple.showTimerCommit&&(e._ripple.showTimerCommit=null),window.clearTimeout(e._ripple.showTimer))}let br=!1;function lh(t){!br&&(t.keyCode===Eu.enter||t.keyCode===Eu.space)&&(br=!0,yr(t))}function uh(t){br=!1,gt(t)}function ch(t){br&&(br=!1,gt(t))}function dh(t,e,n){const{value:s,modifiers:r}=e,i=oh(s);if(i||Li.hide(t),t._ripple=t._ripple??{},t._ripple.enabled=i,t._ripple.centered=r.center,t._ripple.circle=r.circle,ga(s)&&s.class&&(t._ripple.class=s.class),i&&!n){if(r.stop){t.addEventListener("touchstart",nc,{passive:!0}),t.addEventListener("mousedown",nc);return}t.addEventListener("touchstart",yr,{passive:!0}),t.addEventListener("touchend",gt,{passive:!0}),t.addEventListener("touchmove",ah,{passive:!0}),t.addEventListener("touchcancel",gt),t.addEventListener("mousedown",yr),t.addEventListener("mouseup",gt),t.addEventListener("mouseleave",gt),t.addEventListener("keydown",lh),t.addEventListener("keyup",uh),t.addEventListener("blur",ch),t.addEventListener("dragstart",gt,{passive:!0})}else!i&&n&&fh(t)}function fh(t){t.removeEventListener("mousedown",yr),t.removeEventListener("touchstart",yr),t.removeEventListener("touchend",gt),t.removeEventListener("touchmove",ah),t.removeEventListener("touchcancel",gt),t.removeEventListener("mouseup",gt),t.removeEventListener("mouseleave",gt),t.removeEventListener("keydown",lh),t.removeEventListener("keyup",uh),t.removeEventListener("dragstart",gt),t.removeEventListener("blur",ch)}function Ny(t,e){dh(t,e,!1)}function By(t){delete t._ripple,fh(t)}function zy(t,e){if(e.value===e.oldValue)return;const n=oh(e.oldValue);dh(t,e,n)}const Bs={mounted:Ny,unmounted:By,updated:zy},Dy=G({active:{type:Boolean,default:void 0},symbol:{type:null,default:qf},flat:Boolean,icon:[Boolean,String,Function,Object],prependIcon:Ke,appendIcon:Ke,block:Boolean,slim:Boolean,stacked:Boolean,ripple:{type:[Boolean,Object],default:!0},text:String,...Ir(),...we(),...cs(),...lo(),...ls(),...Uf(),...Jf(),...ml(),...eh(),...Kt(),...rh(),...fo(),...Je({tag:"button"}),...dt(),...Fr({variant:"elevated"})},"VBtn"),it=he()({name:"VBtn",directives:{Ripple:Bs},props:Dy(),emits:{"group:selected":t=>!0},setup(t,e){let{attrs:n,slots:s}=e;const{themeClasses:r}=It(t),{borderClasses:i}=kr(t),{colorClasses:o,colorStyles:a,variantClasses:l}=hl(t),{densityClasses:u}=Ns(t),{dimensionStyles:c}=uo(t),{elevationClasses:d}=us(t),{loaderClasses:f}=Qf(t),{locationStyles:h}=gl(t),{positionClasses:g}=th(t),{roundedClasses:y}=Jt(t),{sizeClasses:v,sizeStyles:S}=ho(t),m=jf(t,t.symbol,!1),p=sh(t,n),w=I(()=>{var T;return t.active!==void 0?t.active:p.isLink.value?(T=p.isActive)==null?void 0:T.value:m==null?void 0:m.isSelected.value}),b=I(()=>(m==null?void 0:m.disabled.value)||t.disabled),x=I(()=>t.variant==="elevated"&&!(t.disabled||t.flat||t.border)),C=I(()=>{if(!(t.value===void 0||typeof t.value=="symbol"))return Object(t.value)===t.value?JSON.stringify(t.value,null,0):t.value});function M(T){var V;b.value||p.isLink.value&&(T.metaKey||T.ctrlKey||T.shiftKey||T.button!==0||n.target==="_blank")||((V=p.navigate)==null||V.call(p,T),m==null||m.toggle())}return Ly(p,m==null?void 0:m.select),ye(()=>{var O,H;const T=p.isLink.value?"a":t.tag,V=!!(t.prependIcon||s.prepend),L=!!(t.appendIcon||s.append),A=!!(t.icon&&t.icon!==!0),k=(m==null?void 0:m.isSelected.value)&&(!p.isLink.value||((O=p.isActive)==null?void 0:O.value))||!m||((H=p.isActive)==null?void 0:H.value);return Gt(_(T,{type:T==="a"?void 0:"button",class:["v-btn",m==null?void 0:m.selectedClass.value,{"v-btn--active":w.value,"v-btn--block":t.block,"v-btn--disabled":b.value,"v-btn--elevated":x.value,"v-btn--flat":t.flat,"v-btn--icon":!!t.icon,"v-btn--loading":t.loading,"v-btn--slim":t.slim,"v-btn--stacked":t.stacked},r.value,i.value,k?o.value:void 0,u.value,d.value,f.value,g.value,y.value,v.value,l.value,t.class],style:[k?a.value:void 0,c.value,h.value,S.value,t.style],disabled:b.value||void 0,href:p.href.value,onClick:M,value:C.value},{default:()=>{var q;return[fl(!0,"v-btn"),!t.icon&&V&&_("span",{key:"prepend",class:"v-btn__prepend"},[s.prepend?_(Nt,{key:"prepend-defaults",disabled:!t.prependIcon,defaults:{VIcon:{icon:t.prependIcon}}},s.prepend):_(an,{key:"prepend-icon",icon:t.prependIcon},null)]),_("span",{class:"v-btn__content","data-no-activator":""},[!s.default&&A?_(an,{key:"content-icon",icon:t.icon},null):_(Nt,{key:"content-defaults",disabled:!A,defaults:{VIcon:{icon:t.icon}}},{default:()=>{var j;return[((j=s.default)==null?void 0:j.call(s))??t.text]}})]),!t.icon&&L&&_("span",{key:"append",class:"v-btn__append"},[s.append?_(Nt,{key:"append-defaults",disabled:!t.appendIcon,defaults:{VIcon:{icon:t.appendIcon}}},s.append):_(an,{key:"append-icon",icon:t.appendIcon},null)]),!!t.loading&&_("span",{key:"loader",class:"v-btn__loader"},[((q=s.loader)==null?void 0:q.call(s))??_(Kf,{color:typeof t.loading=="boolean"?void 0:t.loading,indeterminate:!0,size:"23",width:"2"},null)])]}}),[[as("ripple"),!b.value&&t.ripple,null]])}),{group:m}}}),$y=G({start:Boolean,end:Boolean,icon:Ke,image:String,text:String,...we(),...cs(),...Kt(),...fo(),...Je(),...dt(),...Fr({variant:"flat"})},"VAvatar"),sc=he()({name:"VAvatar",props:$y(),setup(t,e){let{slots:n}=e;const{themeClasses:s}=It(t),{colorClasses:r,colorStyles:i,variantClasses:o}=hl(t),{densityClasses:a}=Ns(t),{roundedClasses:l}=Jt(t),{sizeClasses:u,sizeStyles:c}=ho(t);return ye(()=>_(t.tag,{class:["v-avatar",{"v-avatar--start":t.start,"v-avatar--end":t.end},s.value,r.value,a.value,l.value,u.value,o.value,t.class],style:[i.value,c.value,t.style]},{default:()=>[n.default?_(Nt,{key:"content-defaults",defaults:{VImg:{cover:!0,image:t.image},VIcon:{icon:t.icon}}},{default:()=>[n.default()]}):t.image?_(dl,{key:"image",src:t.image,alt:"",cover:!0},null):t.icon?_(an,{key:"icon",icon:t.icon},null):t.text,fl(!1,"v-avatar")]})),{}}}),Hy=G({text:String,onClick:Ti(),...we(),...dt()},"VLabel"),hh=he()({name:"VLabel",props:Hy(),setup(t,e){let{slots:n}=e;return ye(()=>{var s;return _("label",{class:["v-label",{"v-label--clickable":!!t.onClick},t.class],style:t.style,onClick:t.onClick},[t.text,(s=n.default)==null?void 0:s.call(n)])}),{}}}),mh=Symbol.for("vuetify:selection-control-group"),gh=G({color:String,disabled:{type:Boolean,default:null},defaultsTarget:String,error:Boolean,id:String,inline:Boolean,falseIcon:Ke,trueIcon:Ke,ripple:{type:Boolean,default:!0},multiple:{type:Boolean,default:null},name:String,readonly:{type:Boolean,default:null},modelValue:null,type:String,valueComparator:{type:Function,default:oo},...we(),...cs(),...dt()},"SelectionControlGroup"),Uy=G({...gh({defaultsTarget:"VSelectionControl"})},"VSelectionControlGroup");he()({name:"VSelectionControlGroup",props:Uy(),emits:{"update:modelValue":t=>!0},setup(t,e){let{slots:n}=e;const s=ct(t,"modelValue"),r=mn(),i=I(()=>t.id||`v-selection-control-group-${r}`),o=I(()=>t.name||i.value),a=new Set;return ut(mh,{modelValue:s,forceUpdate:()=>{a.forEach(l=>l())},onForceUpdate:l=>{a.add(l),At(()=>{a.delete(l)})}}),hn({[t.defaultsTarget]:{color:de(t,"color"),disabled:de(t,"disabled"),density:de(t,"density"),error:de(t,"error"),inline:de(t,"inline"),modelValue:s,multiple:I(()=>!!t.multiple||t.multiple==null&&Array.isArray(s.value)),name:o,falseIcon:de(t,"falseIcon"),trueIcon:de(t,"trueIcon"),readonly:de(t,"readonly"),ripple:de(t,"ripple"),type:de(t,"type"),valueComparator:de(t,"valueComparator")}}),ye(()=>{var l;return _("div",{class:["v-selection-control-group",{"v-selection-control-group--inline":t.inline},t.class],style:t.style,role:t.type==="radio"?"radiogroup":void 0},[(l=n.default)==null?void 0:l.call(n)])}),{}}});const vh=G({label:String,baseColor:String,trueValue:null,falseValue:null,value:null,...we(),...gh()},"VSelectionControl");function jy(t){const e=Pe(mh,void 0),{densityClasses:n}=Ns(t),s=ct(t,"modelValue"),r=I(()=>t.trueValue!==void 0?t.trueValue:t.value!==void 0?t.value:!0),i=I(()=>t.falseValue!==void 0?t.falseValue:!1),o=I(()=>!!t.multiple||t.multiple==null&&Array.isArray(s.value)),a=I({get(){const h=e?e.modelValue.value:s.value;return o.value?In(h).some(g=>t.valueComparator(g,r.value)):t.valueComparator(h,r.value)},set(h){if(t.readonly)return;const g=h?r.value:i.value;let y=g;o.value&&(y=h?[...In(s.value),g]:In(s.value).filter(v=>!t.valueComparator(v,r.value))),e?e.modelValue.value=y:s.value=y}}),{textColorClasses:l,textColorStyles:u}=Ln(I(()=>{if(!(t.error||t.disabled))return a.value?t.color:t.baseColor})),{backgroundColorClasses:c,backgroundColorStyles:d}=Tt(I(()=>a.value&&!t.error&&!t.disabled?t.color:void 0)),f=I(()=>a.value?t.trueIcon:t.falseIcon);return{group:e,densityClasses:n,trueValue:r,falseValue:i,model:a,textColorClasses:l,textColorStyles:u,backgroundColorClasses:c,backgroundColorStyles:d,icon:f}}const rc=he()({name:"VSelectionControl",directives:{Ripple:Bs},inheritAttrs:!1,props:vh(),emits:{"update:modelValue":t=>!0},setup(t,e){let{attrs:n,slots:s}=e;const{group:r,densityClasses:i,icon:o,model:a,textColorClasses:l,textColorStyles:u,backgroundColorClasses:c,backgroundColorStyles:d,trueValue:f}=jy(t),h=mn(),g=be(!1),y=be(!1),v=se(),S=I(()=>t.id||`input-${h}`),m=I(()=>!t.disabled&&!t.readonly);r==null||r.onForceUpdate(()=>{v.value&&(v.value.checked=a.value)});function p(C){m.value&&(g.value=!0,pf(C.target,":focus-visible")!==!1&&(y.value=!0))}function w(){g.value=!1,y.value=!1}function b(C){C.stopPropagation()}function x(C){m.value&&(t.readonly&&r&&xt(()=>r.forceUpdate()),a.value=C.target.checked)}return ye(()=>{var L,A;const C=s.label?s.label({label:t.label,props:{for:S.value}}):t.label,[M,T]=tp(n),V=_("input",Le({ref:v,checked:a.value,disabled:!!t.disabled,id:S.value,onBlur:w,onFocus:p,onInput:x,"aria-disabled":!!t.disabled,type:t.type,value:f.value,name:t.name,"aria-checked":t.type==="checkbox"?a.value:void 0},T),null);return _("div",Le({class:["v-selection-control",{"v-selection-control--dirty":a.value,"v-selection-control--disabled":t.disabled,"v-selection-control--error":t.error,"v-selection-control--focused":g.value,"v-selection-control--focus-visible":y.value,"v-selection-control--inline":t.inline},i.value,t.class]},M,{style:t.style}),[_("div",{class:["v-selection-control__wrapper",l.value],style:u.value},[(L=s.default)==null?void 0:L.call(s,{backgroundColorClasses:c,backgroundColorStyles:d}),Gt(_("div",{class:["v-selection-control__input"]},[((A=s.input)==null?void 0:A.call(s,{model:a,textColorClasses:l,textColorStyles:u,backgroundColorClasses:c,backgroundColorStyles:d,inputNode:V,icon:o.value,props:{onFocus:p,onBlur:w,id:S.value}}))??_(Re,null,[o.value&&_(an,{key:"icon",icon:o.value},null),V])]),[[as("ripple"),t.ripple&&[!t.disabled&&!t.readonly,null,["center","circle"]]]])]),C&&_(hh,{for:S.value,onClick:b},{default:()=>[C]})])}),{isFocused:g,input:v}}}),Wy=G({indeterminate:Boolean,indeterminateIcon:{type:Ke,default:"$checkboxIndeterminate"},...vh({falseIcon:"$checkboxOff",trueIcon:"$checkboxOn"})},"VCheckboxBtn"),Oo=he()({name:"VCheckboxBtn",props:Wy(),emits:{"update:modelValue":t=>!0,"update:indeterminate":t=>!0},setup(t,e){let{slots:n}=e;const s=ct(t,"indeterminate"),r=ct(t,"modelValue");function i(l){s.value&&(s.value=!1)}const o=I(()=>s.value?t.indeterminateIcon:t.falseIcon),a=I(()=>s.value?t.indeterminateIcon:t.trueIcon);return ye(()=>{const l=rl(rc.filterProps(t),["modelValue"]);return _(rc,Le(l,{modelValue:r.value,"onUpdate:modelValue":[u=>r.value=u,i],class:["v-checkbox-btn",t.class],style:t.style,type:"checkbox",falseIcon:o.value,trueIcon:a.value,"aria-checked":s.value?"mixed":void 0}),n)}),{}}});function Gy(t){const{t:e}=Kp();function n(s){let{name:r}=s;const i={prepend:"prependAction",prependInner:"prependAction",append:"appendAction",appendInner:"appendAction",clear:"clear"}[r],o=t[`onClick:${r}`],a=o&&i?e(`$vuetify.input.${i}`,t.label??""):void 0;return _(an,{icon:t[`${r}Icon`],"aria-label":a,onClick:o},null)}return{InputIcon:n}}const qy=G({active:Boolean,color:String,messages:{type:[Array,String],default:()=>[]},...we(),...cl({transition:{component:ay,leaveAbsolute:!0,group:!0}})},"VMessages"),Yy=he()({name:"VMessages",props:qy(),setup(t,e){let{slots:n}=e;const s=I(()=>In(t.messages)),{textColorClasses:r,textColorStyles:i}=Ln(I(()=>t.color));return ye(()=>_(Cs,{transition:t.transition,tag:"div",class:["v-messages",r.value,t.class],style:[i.value,t.style],role:"alert","aria-live":"polite"},{default:()=>[t.active&&s.value.map((o,a)=>_("div",{class:"v-messages__message",key:`${a}-${s.value}`},[n.message?n.message({message:o}):o]))]})),{}}}),ph=G({focused:Boolean,"onUpdate:focused":Ti()},"focus");function Xy(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();const n=ct(t,"focused"),s=I(()=>({[`${e}--focused`]:n.value}));function r(){n.value=!0}function i(){n.value=!1}return{focusClasses:s,isFocused:n,focus:r,blur:i}}const Zy=Symbol.for("vuetify:form");function Ky(){return Pe(Zy,null)}const Jy=G({disabled:{type:Boolean,default:null},error:Boolean,errorMessages:{type:[Array,String],default:()=>[]},maxErrors:{type:[Number,String],default:1},name:String,label:String,readonly:{type:Boolean,default:null},rules:{type:Array,default:()=>[]},modelValue:null,validateOn:String,validationValue:null,...ph()},"validation");function Qy(t){let e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt(),n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:mn();const s=ct(t,"modelValue"),r=I(()=>t.validationValue===void 0?s.value:t.validationValue),i=Ky(),o=se([]),a=be(!0),l=I(()=>!!(In(s.value===""?null:s.value).length||In(r.value===""?null:r.value).length)),u=I(()=>!!(t.disabled??(i==null?void 0:i.isDisabled.value))),c=I(()=>!!(t.readonly??(i==null?void 0:i.isReadonly.value))),d=I(()=>{var w;return(w=t.errorMessages)!=null&&w.length?In(t.errorMessages).concat(o.value).slice(0,Math.max(0,+t.maxErrors)):o.value}),f=I(()=>{let w=(t.validateOn??(i==null?void 0:i.validateOn.value))||"input";w==="lazy"&&(w="input lazy");const b=new Set((w==null?void 0:w.split(" "))??[]);return{blur:b.has("blur")||b.has("input"),input:b.has("input"),submit:b.has("submit"),lazy:b.has("lazy")}}),h=I(()=>{var w;return t.error||(w=t.errorMessages)!=null&&w.length?!1:t.rules.length?a.value?o.value.length||f.value.lazy?null:!0:!o.value.length:!0}),g=be(!1),y=I(()=>({[`${e}--error`]:h.value===!1,[`${e}--dirty`]:l.value,[`${e}--disabled`]:u.value,[`${e}--readonly`]:c.value})),v=I(()=>t.name??F(n));to(()=>{i==null||i.register({id:v.value,validate:p,reset:S,resetValidation:m})}),Rt(()=>{i==null||i.unregister(v.value)}),Yt(async()=>{f.value.lazy||await p(!0),i==null||i.update(v.value,h.value,d.value)}),on(()=>f.value.input,()=>{le(r,()=>{if(r.value!=null)p();else if(t.focused){const w=le(()=>t.focused,b=>{b||p(),w()})}})}),on(()=>f.value.blur,()=>{le(()=>t.focused,w=>{w||p()})}),le([h,d],()=>{i==null||i.update(v.value,h.value,d.value)});function S(){s.value=null,xt(m)}function m(){a.value=!0,f.value.lazy?o.value=[]:p(!0)}async function p(){let w=arguments.length>0&&arguments[0]!==void 0?arguments[0]:!1;const b=[];g.value=!0;for(const x of t.rules){if(b.length>=+(t.maxErrors??1))break;const M=await(typeof x=="function"?x:()=>x)(r.value);if(M!==!0){if(M!==!1&&typeof M!="string"){console.warn(`${M} is not a valid value. Rule functions must return boolean true or a string.`);continue}b.push(M||"")}}return o.value=b,g.value=!1,a.value=w,o.value}return{errorMessages:d,isDirty:l,isDisabled:u,isReadonly:c,isPristine:a,isValid:h,isValidating:g,reset:S,resetValidation:m,validate:p,validationClasses:y}}const yh=G({id:String,appendIcon:Ke,centerAffix:{type:Boolean,default:!0},prependIcon:Ke,hideDetails:[Boolean,String],hideSpinButtons:Boolean,hint:String,persistentHint:Boolean,messages:{type:[Array,String],default:()=>[]},direction:{type:String,default:"horizontal",validator:t=>["horizontal","vertical"].includes(t)},"onClick:prepend":Ti(),"onClick:append":Ti(),...we(),...cs(),...Jy()},"VInput"),ic=he()({name:"VInput",props:{...yh()},emits:{"update:modelValue":t=>!0},setup(t,e){let{attrs:n,slots:s,emit:r}=e;const{densityClasses:i}=Ns(t),{rtlClasses:o}=Zt(),{InputIcon:a}=Gy(t),l=mn(),u=I(()=>t.id||`input-${l}`),c=I(()=>`${u.value}-messages`),{errorMessages:d,isDirty:f,isDisabled:h,isReadonly:g,isPristine:y,isValid:v,isValidating:S,reset:m,resetValidation:p,validate:w,validationClasses:b}=Qy(t,"v-input",u),x=I(()=>({id:u,messagesId:c,isDirty:f,isDisabled:h,isReadonly:g,isPristine:y,isValid:v,isValidating:S,reset:m,resetValidation:p,validate:w})),C=I(()=>{var M;return(M=t.errorMessages)!=null&&M.length||!y.value&&d.value.length?d.value:t.hint&&(t.persistentHint||t.focused)?t.hint:t.messages});return ye(()=>{var A,k,O,H;const M=!!(s.prepend||t.prependIcon),T=!!(s.append||t.appendIcon),V=C.value.length>0,L=!t.hideDetails||t.hideDetails==="auto"&&(V||!!s.details);return _("div",{class:["v-input",`v-input--${t.direction}`,{"v-input--center-affix":t.centerAffix,"v-input--hide-spin-buttons":t.hideSpinButtons},i.value,o.value,b.value,t.class],style:t.style},[M&&_("div",{key:"prepend",class:"v-input__prepend"},[(A=s.prepend)==null?void 0:A.call(s,x.value),t.prependIcon&&_(a,{key:"prepend-icon",name:"prepend"},null)]),s.default&&_("div",{class:"v-input__control"},[(k=s.default)==null?void 0:k.call(s,x.value)]),T&&_("div",{key:"append",class:"v-input__append"},[t.appendIcon&&_(a,{key:"append-icon",name:"append"},null),(O=s.append)==null?void 0:O.call(s,x.value)]),L&&_("div",{class:"v-input__details"},[_(Yy,{id:c.value,active:V,messages:C.value},{message:s.message}),(H=s.details)==null?void 0:H.call(s,x.value)])])}),{reset:m,resetValidation:p,validate:w,isValid:v,errorMessages:d}}}),mo=["sm","md","lg","xl","xxl"],Sa=Symbol.for("vuetify:display"),oc={mobileBreakpoint:"lg",thresholds:{xs:0,sm:600,md:960,lg:1280,xl:1920,xxl:2560}},e0=function(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:oc;return vt(oc,t)};function ac(t){return Oe&&!t?window.innerWidth:typeof t=="object"&&t.clientWidth||0}function lc(t){return Oe&&!t?window.innerHeight:typeof t=="object"&&t.clientHeight||0}function uc(t){const e=Oe&&!t?window.navigator.userAgent:"ssr";function n(g){return!!e.match(g)}const s=n(/android/i),r=n(/iphone|ipad|ipod/i),i=n(/cordova/i),o=n(/electron/i),a=n(/chrome/i),l=n(/edge/i),u=n(/firefox/i),c=n(/opera/i),d=n(/win/i),f=n(/mac/i),h=n(/linux/i);return{android:s,ios:r,cordova:i,electron:o,chrome:a,edge:l,firefox:u,opera:c,win:d,mac:f,linux:h,touch:qv,ssr:e==="ssr"}}function t0(t,e){const{thresholds:n,mobileBreakpoint:s}=e0(t),r=be(lc(e)),i=be(uc(e)),o=tt({}),a=be(ac(e));function l(){r.value=lc(),a.value=ac()}function u(){l(),i.value=uc()}return Pn(()=>{const c=a.value<n.sm,d=a.value<n.md&&!c,f=a.value<n.lg&&!(d||c),h=a.value<n.xl&&!(f||d||c),g=a.value<n.xxl&&!(h||f||d||c),y=a.value>=n.xxl,v=c?"xs":d?"sm":f?"md":h?"lg":g?"xl":"xxl",S=typeof s=="number"?s:n[s],m=a.value<S;o.xs=c,o.sm=d,o.md=f,o.lg=h,o.xl=g,o.xxl=y,o.smAndUp=!c,o.mdAndUp=!(c||d),o.lgAndUp=!(c||d||f),o.xlAndUp=!(c||d||f||h),o.smAndDown=!(f||h||g||y),o.mdAndDown=!(h||g||y),o.lgAndDown=!(g||y),o.xlAndDown=!y,o.name=v,o.height=r.value,o.width=a.value,o.mobile=m,o.mobileBreakpoint=s,o.platform=i.value,o.thresholds=n}),Oe&&window.addEventListener("resize",l,{passive:!0}),{...Ya(o),update:u,ssr:!!e}}const n0=G({mobileBreakpoint:[Number,String]},"display");function bh(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:Xt();const n=Pe(Sa);if(!n)throw new Error("Could not find Vuetify display injection");const s=I(()=>{if(!t.mobileBreakpoint)return n.mobile.value;const i=typeof t.mobileBreakpoint=="number"?t.mobileBreakpoint:n.thresholds.value[t.mobileBreakpoint];return n.width.value<i}),r=I(()=>e?{[`${e}--mobile`]:s.value}:{});return{...n,displayClasses:r,mobile:s}}const s0=G({color:String,inset:Boolean,length:[Number,String],thickness:[Number,String],vertical:Boolean,...we(),...dt()},"VDivider"),hi=he()({name:"VDivider",props:s0(),setup(t,e){let{attrs:n}=e;const{themeClasses:s}=It(t),{textColorClasses:r,textColorStyles:i}=Ln(de(t,"color")),o=I(()=>{const a={};return t.length&&(a[t.vertical?"maxHeight":"maxWidth"]=ie(t.length)),t.thickness&&(a[t.vertical?"borderRightWidth":"borderTopWidth"]=ie(t.thickness)),a});return ye(()=>_("hr",{class:[{"v-divider":!0,"v-divider--inset":t.inset,"v-divider--vertical":t.vertical},s.value,r.value,t.class],style:[o.value,i.value,t.style],"aria-orientation":!n.role||n.role==="separator"?t.vertical?"vertical":"horizontal":void 0,role:`${n.role||"separator"}`},null)),{}}});function No(t,e){return{x:t.x+e.x,y:t.y+e.y}}function r0(t,e){return{x:t.x-e.x,y:t.y-e.y}}function cc(t,e){if(t.side==="top"||t.side==="bottom"){const{side:n,align:s}=t,r=s==="left"?0:s==="center"?e.width/2:s==="right"?e.width:s,i=n==="top"?0:n==="bottom"?e.height:n;return No({x:r,y:i},e)}else if(t.side==="left"||t.side==="right"){const{side:n,align:s}=t,r=n==="left"?0:n==="right"?e.width:n,i=s==="top"?0:s==="center"?e.height/2:s==="bottom"?e.height:s;return No({x:r,y:i},e)}return No({x:e.width/2,y:e.height/2},e)}const xh={static:a0,connected:u0},i0=G({locationStrategy:{type:[String,Function],default:"static",validator:t=>typeof t=="function"||t in xh},location:{type:String,default:"bottom"},origin:{type:String,default:"auto"},offset:[Number,String,Array]},"VOverlay-location-strategies");function o0(t,e){const n=se({}),s=se();Oe&&on(()=>!!(e.isActive.value&&t.locationStrategy),i=>{var o,a;le(()=>t.locationStrategy,i),At(()=>{window.removeEventListener("resize",r),s.value=void 0}),window.addEventListener("resize",r,{passive:!0}),typeof t.locationStrategy=="function"?s.value=(o=t.locationStrategy(e,t,n))==null?void 0:o.updateLocation:s.value=(a=xh[t.locationStrategy](e,t,n))==null?void 0:a.updateLocation});function r(i){var o;(o=s.value)==null||o.call(s,i)}return{contentStyles:n,updateLocation:s}}function a0(){}function l0(t,e){e?t.style.removeProperty("left"):t.style.removeProperty("right");const n=xf(t);return e?n.x+=parseFloat(t.style.right||0):n.x-=parseFloat(t.style.left||0),n.y-=parseFloat(t.style.top||0),n}function u0(t,e,n){(Array.isArray(t.target.value)||Dp(t.target.value))&&Object.assign(n.value,{position:"fixed",top:0,[t.isRtl.value?"right":"left"]:0});const{preferredAnchor:r,preferredOrigin:i}=il(()=>{const g=va(e.location,t.isRtl.value),y=e.origin==="overlap"?g:e.origin==="auto"?Fo(g):va(e.origin,t.isRtl.value);return g.side===y.side&&g.align===Vo(y).align?{preferredAnchor:ku(g),preferredOrigin:ku(y)}:{preferredAnchor:g,preferredOrigin:y}}),[o,a,l,u]=["minWidth","minHeight","maxWidth","maxHeight"].map(g=>I(()=>{const y=parseFloat(e[g]);return isNaN(y)?1/0:y})),c=I(()=>{if(Array.isArray(e.offset))return e.offset;if(typeof e.offset=="string"){const g=e.offset.split(" ").map(parseFloat);return g.length<2&&g.push(0),g}return typeof e.offset=="number"?[e.offset,0]:[0,0]});let d=!1;const f=new ResizeObserver(()=>{d&&h()});le([t.target,t.contentEl],(g,y)=>{let[v,S]=g,[m,p]=y;m&&!Array.isArray(m)&&f.unobserve(m),v&&!Array.isArray(v)&&f.observe(v),p&&f.unobserve(p),S&&f.observe(S)},{immediate:!0}),At(()=>{f.disconnect()});function h(){if(d=!1,requestAnimationFrame(()=>d=!0),!t.target.value||!t.contentEl.value)return;const g=bf(t.target.value),y=l0(t.contentEl.value,t.isRtl.value),v=Ri(t.contentEl.value),S=12;v.length||(v.push(document.documentElement),t.contentEl.value.style.top&&t.contentEl.value.style.left||(y.x-=parseFloat(document.documentElement.style.getPropertyValue("--v-body-scroll-x")||0),y.y-=parseFloat(document.documentElement.style.getPropertyValue("--v-body-scroll-y")||0)));const m=v.reduce((L,A)=>{const k=A.getBoundingClientRect(),O=new Jn({x:A===document.documentElement?0:k.x,y:A===document.documentElement?0:k.y,width:A.clientWidth,height:A.clientHeight});return L?new Jn({x:Math.max(L.left,O.left),y:Math.max(L.top,O.top),width:Math.min(L.right,O.right)-Math.max(L.left,O.left),height:Math.min(L.bottom,O.bottom)-Math.max(L.top,O.top)}):O},void 0);m.x+=S,m.y+=S,m.width-=S*2,m.height-=S*2;let p={anchor:r.value,origin:i.value};function w(L){const A=new Jn(y),k=cc(L.anchor,g),O=cc(L.origin,A);let{x:H,y:q}=r0(k,O);switch(L.anchor.side){case"top":q-=c.value[0];break;case"bottom":q+=c.value[0];break;case"left":H-=c.value[0];break;case"right":H+=c.value[0];break}switch(L.anchor.align){case"top":q-=c.value[1];break;case"bottom":q+=c.value[1];break;case"left":H-=c.value[1];break;case"right":H+=c.value[1];break}return A.x+=H,A.y+=q,A.width=Math.min(A.width,l.value),A.height=Math.min(A.height,u.value),{overflows:Vu(A,m),x:H,y:q}}let b=0,x=0;const C={x:0,y:0},M={x:!1,y:!1};let T=-1;for(;!(T++>10);){const{x:L,y:A,overflows:k}=w(p);b+=L,x+=A,y.x+=L,y.y+=A;{const O=Fu(p.anchor),H=k.x.before||k.x.after,q=k.y.before||k.y.after;let j=!1;if(["x","y"].forEach(Y=>{if(Y==="x"&&H&&!M.x||Y==="y"&&q&&!M.y){const ae={anchor:{...p.anchor},origin:{...p.origin}},Q=Y==="x"?O==="y"?Vo:Fo:O==="y"?Fo:Vo;ae.anchor=Q(ae.anchor),ae.origin=Q(ae.origin);const{overflows:te}=w(ae);(te[Y].before<=k[Y].before&&te[Y].after<=k[Y].after||te[Y].before+te[Y].after<(k[Y].before+k[Y].after)/2)&&(p=ae,j=M[Y]=!0)}}),j)continue}k.x.before&&(b+=k.x.before,y.x+=k.x.before),k.x.after&&(b-=k.x.after,y.x-=k.x.after),k.y.before&&(x+=k.y.before,y.y+=k.y.before),k.y.after&&(x-=k.y.after,y.y-=k.y.after);{const O=Vu(y,m);C.x=m.width-O.x.before-O.x.after,C.y=m.height-O.y.before-O.y.after,b+=O.x.before,y.x+=O.x.before,x+=O.y.before,y.y+=O.y.before}break}const V=Fu(p.anchor);return Object.assign(n.value,{"--v-overlay-anchor-origin":`${p.anchor.side} ${p.anchor.align}`,transformOrigin:`${p.origin.side} ${p.origin.align}`,top:ie(Bo(x)),left:t.isRtl.value?void 0:ie(Bo(b)),right:t.isRtl.value?ie(Bo(-b)):void 0,minWidth:ie(V==="y"?Math.min(o.value,g.width):o.value),maxWidth:ie(dc(vr(C.x,o.value===1/0?0:o.value,l.value))),maxHeight:ie(dc(vr(C.y,a.value===1/0?0:a.value,u.value)))}),{available:C,contentBox:y}}return le(()=>[r.value,i.value,e.offset,e.minWidth,e.minHeight,e.maxWidth,e.maxHeight],()=>h()),xt(()=>{const g=h();if(!g)return;const{available:y,contentBox:v}=g;v.height>y.y&&requestAnimationFrame(()=>{h(),requestAnimationFrame(()=>{h()})})}),{updateLocation:h}}function Bo(t){return Math.round(t*devicePixelRatio)/devicePixelRatio}function dc(t){return Math.ceil(t*devicePixelRatio)/devicePixelRatio}let Ca=!0;const Pi=[];function c0(t){!Ca||Pi.length?(Pi.push(t),Ea()):(Ca=!1,t(),Ea())}let fc=-1;function Ea(){cancelAnimationFrame(fc),fc=requestAnimationFrame(()=>{const t=Pi.shift();t&&t(),Pi.length?Ea():Ca=!0})}const mi={none:null,close:h0,block:m0,reposition:g0},d0=G({scrollStrategy:{type:[String,Function],default:"block",validator:t=>typeof t=="function"||t in mi}},"VOverlay-scroll-strategies");function f0(t,e){if(!Oe)return;let n;Pn(async()=>{n==null||n.stop(),e.isActive.value&&t.scrollStrategy&&(n=za(),await xt(),n.active&&n.run(()=>{var s;typeof t.scrollStrategy=="function"?t.scrollStrategy(e,t,n):(s=mi[t.scrollStrategy])==null||s.call(mi,e,t,n)}))}),At(()=>{n==null||n.stop()})}function h0(t){function e(n){t.isActive.value=!1}wh(t.targetEl.value??t.contentEl.value,e)}function m0(t,e){var o;const n=(o=t.root.value)==null?void 0:o.offsetParent,s=[...new Set([...Ri(t.targetEl.value,e.contained?n:void 0),...Ri(t.contentEl.value,e.contained?n:void 0)])].filter(a=>!a.classList.contains("v-overlay-scroll-blocked")),r=window.innerWidth-document.documentElement.offsetWidth,i=(a=>ll(a)&&a)(n||document.documentElement);i&&t.root.value.classList.add("v-overlay--scroll-blocked"),s.forEach((a,l)=>{a.style.setProperty("--v-body-scroll-x",ie(-a.scrollLeft)),a.style.setProperty("--v-body-scroll-y",ie(-a.scrollTop)),a!==document.documentElement&&a.style.setProperty("--v-scrollbar-offset",ie(r)),a.classList.add("v-overlay-scroll-blocked")}),At(()=>{s.forEach((a,l)=>{const u=parseFloat(a.style.getPropertyValue("--v-body-scroll-x")),c=parseFloat(a.style.getPropertyValue("--v-body-scroll-y")),d=a.style.scrollBehavior;a.style.scrollBehavior="auto",a.style.removeProperty("--v-body-scroll-x"),a.style.removeProperty("--v-body-scroll-y"),a.style.removeProperty("--v-scrollbar-offset"),a.classList.remove("v-overlay-scroll-blocked"),a.scrollLeft=-u,a.scrollTop=-c,a.style.scrollBehavior=d}),i&&t.root.value.classList.remove("v-overlay--scroll-blocked")})}function g0(t,e,n){let s=!1,r=-1,i=-1;function o(a){c0(()=>{var c,d;const l=performance.now();(d=(c=t.updateLocation).value)==null||d.call(c,a),s=(performance.now()-l)/(1e3/60)>2})}i=(typeof requestIdleCallback>"u"?a=>a():requestIdleCallback)(()=>{n.run(()=>{wh(t.targetEl.value??t.contentEl.value,a=>{s?(cancelAnimationFrame(r),r=requestAnimationFrame(()=>{r=requestAnimationFrame(()=>{o(a)})})):o(a)})})}),At(()=>{typeof cancelIdleCallback<"u"&&cancelIdleCallback(i),cancelAnimationFrame(r)})}function wh(t,e){const n=[document,...Ri(t)];n.forEach(s=>{s.addEventListener("scroll",e,{passive:!0})}),At(()=>{n.forEach(s=>{s.removeEventListener("scroll",e)})})}const v0=Symbol.for("vuetify:v-menu"),p0=G({closeDelay:[Number,String],openDelay:[Number,String]},"delay");function y0(t,e){let n=()=>{};function s(o){n==null||n();const a=Number(o?t.openDelay:t.closeDelay);return new Promise(l=>{n=ap(a,()=>{e==null||e(o),l(o)})})}function r(){return s(!0)}function i(){return s(!1)}return{clearDelay:n,runOpenDelay:r,runCloseDelay:i}}const b0=G({target:[String,Object],activator:[String,Object],activatorProps:{type:Object,default:()=>({})},openOnClick:{type:Boolean,default:void 0},openOnHover:Boolean,openOnFocus:{type:Boolean,default:void 0},closeOnContentClick:Boolean,...p0()},"VOverlay-activator");function x0(t,e){let{isActive:n,isTop:s}=e;const r=Ge("useActivator"),i=se();let o=!1,a=!1,l=!0;const u=I(()=>t.openOnFocus||t.openOnFocus==null&&t.openOnHover),c=I(()=>t.openOnClick||t.openOnClick==null&&!t.openOnHover&&!u.value),{runOpenDelay:d,runCloseDelay:f}=y0(t,C=>{C===(t.openOnHover&&o||u.value&&a)&&!(t.openOnHover&&n.value&&!s.value)&&(n.value!==C&&(l=!0),n.value=C)}),h=se(),g={onClick:C=>{C.stopPropagation(),i.value=C.currentTarget||C.target,n.value||(h.value=[C.clientX,C.clientY]),n.value=!n.value},onMouseenter:C=>{var M;(M=C.sourceCapabilities)!=null&&M.firesTouchEvents||(o=!0,i.value=C.currentTarget||C.target,d())},onMouseleave:C=>{o=!1,f()},onFocus:C=>{pf(C.target,":focus-visible")!==!1&&(a=!0,C.stopPropagation(),i.value=C.currentTarget||C.target,d())},onBlur:C=>{a=!1,C.stopPropagation(),f()}},y=I(()=>{const C={};return c.value&&(C.onClick=g.onClick),t.openOnHover&&(C.onMouseenter=g.onMouseenter,C.onMouseleave=g.onMouseleave),u.value&&(C.onFocus=g.onFocus,C.onBlur=g.onBlur),C}),v=I(()=>{const C={};if(t.openOnHover&&(C.onMouseenter=()=>{o=!0,d()},C.onMouseleave=()=>{o=!1,f()}),u.value&&(C.onFocusin=()=>{a=!0,d()},C.onFocusout=()=>{a=!1,f()}),t.closeOnContentClick){const M=Pe(v0,null);C.onClick=()=>{n.value=!1,M==null||M.closeParents()}}return C}),S=I(()=>{const C={};return t.openOnHover&&(C.onMouseenter=()=>{l&&(o=!0,l=!1,d())},C.onMouseleave=()=>{o=!1,f()}),C});le(s,C=>{C&&(t.openOnHover&&!o&&(!u.value||!a)||u.value&&!a&&(!t.openOnHover||!o))&&(n.value=!1)}),le(n,C=>{C||setTimeout(()=>{h.value=void 0})},{flush:"post"});const m=se();Pn(()=>{m.value&&xt(()=>{i.value=Ei(m.value)})});const p=se(),w=I(()=>t.target==="cursor"&&h.value?h.value:p.value?Ei(p.value):_h(t.target,r)||i.value),b=I(()=>Array.isArray(w.value)?void 0:w.value);let x;return le(()=>!!t.activator,C=>{C&&Oe?(x=za(),x.run(()=>{w0(t,r,{activatorEl:i,activatorEvents:y})})):x&&x.stop()},{flush:"post",immediate:!0}),At(()=>{x==null||x.stop()}),{activatorEl:i,activatorRef:m,target:w,targetEl:b,targetRef:p,activatorEvents:y,contentEvents:v,scrimEvents:S}}function w0(t,e,n){let{activatorEl:s,activatorEvents:r}=n;le(()=>t.activator,(l,u)=>{if(u&&l!==u){const c=a(u);c&&o(c)}l&&xt(()=>i())},{immediate:!0}),le(()=>t.activatorProps,()=>{i()}),At(()=>{o()});function i(){let l=arguments.length>0&&arguments[0]!==void 0?arguments[0]:a(),u=arguments.length>1&&arguments[1]!==void 0?arguments[1]:t.activatorProps;l&&up(l,Le(r.value,u))}function o(){let l=arguments.length>0&&arguments[0]!==void 0?arguments[0]:a(),u=arguments.length>1&&arguments[1]!==void 0?arguments[1]:t.activatorProps;l&&cp(l,Le(r.value,u))}function a(){let l=arguments.length>0&&arguments[0]!==void 0?arguments[0]:t.activator;const u=_h(l,e);return s.value=(u==null?void 0:u.nodeType)===Node.ELEMENT_NODE?u:void 0,s.value}}function _h(t,e){var s,r;if(!t)return;let n;if(t==="parent"){let i=(r=(s=e==null?void 0:e.proxy)==null?void 0:s.$el)==null?void 0:r.parentNode;for(;i!=null&&i.hasAttribute("data-no-activator");)i=i.parentNode;n=i}else typeof t=="string"?n=document.querySelector(t):"$el"in t?n=t.$el:n=t;return n}function _0(){if(!Oe)return be(!1);const{ssr:t}=bh();if(t){const e=be(!1);return Yt(()=>{e.value=!0}),e}else return be(!0)}const vl=G({eager:Boolean},"lazy");function Sh(t,e){const n=be(!1),s=I(()=>n.value||t.eager||e.value);le(e,()=>n.value=!0);function r(){t.eager||(n.value=!1)}return{isBooted:n,hasContent:s,onAfterLeave:r}}function go(){const e=Ge("useScopeId").vnode.scopeId;return{scopeId:e?{[e]:""}:void 0}}const hc=Symbol.for("vuetify:stack"),Hs=tt([]);function S0(t,e,n){const s=Ge("useStack"),r=!n,i=Pe(hc,void 0),o=tt({activeChildren:new Set});ut(hc,o);const a=be(+e.value);on(t,()=>{var d;const c=(d=Hs.at(-1))==null?void 0:d[1];a.value=c?c+10:+e.value,r&&Hs.push([s.uid,a.value]),i==null||i.activeChildren.add(s.uid),At(()=>{if(r){const f=pe(Hs).findIndex(h=>h[0]===s.uid);Hs.splice(f,1)}i==null||i.activeChildren.delete(s.uid)})});const l=be(!0);r&&Pn(()=>{var d;const c=((d=Hs.at(-1))==null?void 0:d[0])===s.uid;setTimeout(()=>l.value=c)});const u=I(()=>!o.activeChildren.size);return{globalTop:_r(l),localTop:u,stackStyles:I(()=>({zIndex:a.value}))}}function C0(t){return{teleportTarget:I(()=>{const n=t.value;if(n===!0||!Oe)return;const s=n===!1?document.body:typeof n=="string"?document.querySelector(n):n;if(s==null)return;let r=s.querySelector(":scope > .v-overlay-container");return r||(r=document.createElement("div"),r.className="v-overlay-container",s.appendChild(r)),r})}}function E0(){return!0}function Ch(t,e,n){if(!t||Eh(t,n)===!1)return!1;const s=Mf(e);if(typeof ShadowRoot<"u"&&s instanceof ShadowRoot&&s.host===t.target)return!1;const r=(typeof n.value=="object"&&n.value.include||(()=>[]))();return r.push(e),!r.some(i=>i==null?void 0:i.contains(t.target))}function Eh(t,e){return(typeof e.value=="object"&&e.value.closeConditional||E0)(t)}function M0(t,e,n){const s=typeof n.value=="function"?n.value:n.value.handler;e._clickOutside.lastMousedownWasOutside&&Ch(t,e,n)&&setTimeout(()=>{Eh(t,n)&&s&&s(t)},0)}function mc(t,e){const n=Mf(t);e(document),typeof ShadowRoot<"u"&&n instanceof ShadowRoot&&e(n)}const Mh={mounted(t,e){const n=r=>M0(r,t,e),s=r=>{t._clickOutside.lastMousedownWasOutside=Ch(r,t,e)};mc(t,r=>{r.addEventListener("click",n,!0),r.addEventListener("mousedown",s,!0)}),t._clickOutside||(t._clickOutside={lastMousedownWasOutside:!1}),t._clickOutside[e.instance.$.uid]={onClick:n,onMousedown:s}},unmounted(t,e){t._clickOutside&&(mc(t,n=>{var i;if(!n||!((i=t._clickOutside)!=null&&i[e.instance.$.uid]))return;const{onClick:s,onMousedown:r}=t._clickOutside[e.instance.$.uid];n.removeEventListener("click",s,!0),n.removeEventListener("mousedown",r,!0)}),delete t._clickOutside[e.instance.$.uid])}};function T0(t){const{modelValue:e,color:n,...s}=t;return _(qt,{name:"fade-transition",appear:!0},{default:()=>[t.modelValue&&_("div",Le({class:["v-overlay__scrim",t.color.backgroundColorClasses.value],style:t.color.backgroundColorStyles.value},s),null)]})}const pl=G({absolute:Boolean,attach:[Boolean,String,Object],closeOnBack:{type:Boolean,default:!0},contained:Boolean,contentClass:null,contentProps:null,disabled:Boolean,opacity:[Number,String],noClickAnimation:Boolean,modelValue:Boolean,persistent:Boolean,scrim:{type:[Boolean,String],default:!0},zIndex:{type:[Number,String],default:2e3},...b0(),...we(),...lo(),...vl(),...i0(),...d0(),...dt(),...cl()},"VOverlay"),Oi=he()({name:"VOverlay",directives:{ClickOutside:Mh},inheritAttrs:!1,props:{_disableGlobalStack:Boolean,...pl()},emits:{"click:outside":t=>!0,"update:modelValue":t=>!0,afterLeave:()=>!0},setup(t,e){let{slots:n,attrs:s,emit:r}=e;const i=ct(t,"modelValue"),o=I({get:()=>i.value,set:me=>{me&&t.disabled||(i.value=me)}}),{teleportTarget:a}=C0(I(()=>t.attach||t.contained)),{themeClasses:l}=It(t),{rtlClasses:u,isRtl:c}=Zt(),{hasContent:d,onAfterLeave:f}=Sh(t,o),h=Tt(I(()=>typeof t.scrim=="string"?t.scrim:null)),{globalTop:g,localTop:y,stackStyles:v}=S0(o,de(t,"zIndex"),t._disableGlobalStack),{activatorEl:S,activatorRef:m,target:p,targetEl:w,targetRef:b,activatorEvents:x,contentEvents:C,scrimEvents:M}=x0(t,{isActive:o,isTop:y}),{dimensionStyles:T}=uo(t),V=_0(),{scopeId:L}=go();le(()=>t.disabled,me=>{me&&(o.value=!1)});const A=se(),k=se(),{contentStyles:O,updateLocation:H}=o0(t,{isRtl:c,contentEl:k,target:p,isActive:o});f0(t,{root:A,contentEl:k,targetEl:w,isActive:o,updateLocation:H});function q(me){r("click:outside",me),t.persistent?te():o.value=!1}function j(){return o.value&&g.value}Oe&&le(o,me=>{me?window.addEventListener("keydown",Y):window.removeEventListener("keydown",Y)},{immediate:!0}),Rt(()=>{Oe&&window.removeEventListener("keydown",Y)});function Y(me){var Ce,We;me.key==="Escape"&&g.value&&(t.persistent?te():(o.value=!1,(Ce=k.value)!=null&&Ce.contains(document.activeElement)&&((We=S.value)==null||We.focus())))}const ae=nh();on(()=>t.closeOnBack,()=>{Vy(ae,me=>{g.value&&o.value?(me(!1),t.persistent?te():o.value=!1):me()})});const Q=se();le(()=>o.value&&(t.absolute||t.contained)&&a.value==null,me=>{if(me){const Ce=Np(A.value);Ce&&Ce!==document.scrollingElement&&(Q.value=Ce.scrollTop)}});function te(){t.noClickAnimation||k.value&&Ys(k.value,[{transformOrigin:"center"},{transform:"scale(1.03)"},{transformOrigin:"center"}],{duration:150,easing:ba})}function ge(){f(),r("afterLeave")}return ye(()=>{var me;return _(Re,null,[(me=n.activator)==null?void 0:me.call(n,{isActive:o.value,props:Le({ref:m,targetRef:b},x.value,t.activatorProps)}),V.value&&d.value&&_(Zg,{disabled:!a.value,to:a.value},{default:()=>[_("div",Le({class:["v-overlay",{"v-overlay--absolute":t.absolute||t.contained,"v-overlay--active":o.value,"v-overlay--contained":t.contained},l.value,u.value,t.class],style:[v.value,{"--v-overlay-opacity":t.opacity,top:ie(Q.value)},t.style],ref:A},L,s),[_(T0,Le({color:h,modelValue:o.value&&!!t.scrim},M.value),null),_(Cs,{appear:!0,persisted:!0,transition:t.transition,target:p.value,onAfterLeave:ge},{default:()=>{var Ce;return[Gt(_("div",Le({ref:k,class:["v-overlay__content",t.contentClass],style:[T.value,O.value]},C.value,t.contentProps),[(Ce=n.default)==null?void 0:Ce.call(n,{isActive:o})]),[[io,o.value],[as("click-outside"),{handler:q,closeConditional:j,include:()=>[S.value]}]])]}})])]})])}),{activatorEl:S,target:p,animateClick:te,contentEl:k,globalTop:g,localTop:y,updateLocation:H}}}),zo=Symbol("Forwarded refs");function Do(t,e){let n=t;for(;n;){const s=Reflect.getOwnPropertyDescriptor(n,e);if(s)return s;n=Object.getPrototypeOf(n)}}function Th(t){for(var e=arguments.length,n=new Array(e>1?e-1:0),s=1;s<e;s++)n[s-1]=arguments[s];return t[zo]=n,new Proxy(t,{get(r,i){if(Reflect.has(r,i))return Reflect.get(r,i);if(!(typeof i=="symbol"||i.startsWith("$")||i.startsWith("__"))){for(const o of n)if(o.value&&Reflect.has(o.value,i)){const a=Reflect.get(o.value,i);return typeof a=="function"?a.bind(o.value):a}}},has(r,i){if(Reflect.has(r,i))return!0;if(typeof i=="symbol"||i.startsWith("$")||i.startsWith("__"))return!1;for(const o of n)if(o.value&&Reflect.has(o.value,i))return!0;return!1},set(r,i,o){if(Reflect.has(r,i))return Reflect.set(r,i,o);if(typeof i=="symbol"||i.startsWith("$")||i.startsWith("__"))return!1;for(const a of n)if(a.value&&Reflect.has(a.value,i))return Reflect.set(a.value,i,o);return!1},getOwnPropertyDescriptor(r,i){var a;const o=Reflect.getOwnPropertyDescriptor(r,i);if(o)return o;if(!(typeof i=="symbol"||i.startsWith("$")||i.startsWith("__"))){for(const l of n){if(!l.value)continue;const u=Do(l.value,i)??("_"in l.value?Do((a=l.value._)==null?void 0:a.setupState,i):void 0);if(u)return u}for(const l of n){const u=l.value&&l.value[zo];if(!u)continue;const c=u.slice();for(;c.length;){const d=c.shift(),f=Do(d.value,i);if(f)return f;const h=d.value&&d.value[zo];h&&c.push(...h)}}}}})}const A0=G({fullscreen:Boolean,retainFocus:{type:Boolean,default:!0},scrollable:Boolean,...pl({origin:"center center",scrollStrategy:"block",transition:{component:iy},zIndex:2400})},"VDialog"),R0=he()({name:"VDialog",props:A0(),emits:{"update:modelValue":t=>!0},setup(t,e){let{slots:n}=e;const s=ct(t,"modelValue"),{scopeId:r}=go(),i=se();function o(l){var d,f;const u=l.relatedTarget,c=l.target;if(u!==c&&((d=i.value)!=null&&d.contentEl)&&((f=i.value)!=null&&f.globalTop)&&![document,i.value.contentEl].includes(c)&&!i.value.contentEl.contains(c)){const h=vf(i.value.contentEl);if(!h.length)return;const g=h[0],y=h[h.length-1];u===g?y.focus():g.focus()}}Oe&&le(()=>s.value&&t.retainFocus,l=>{l?document.addEventListener("focusin",o):document.removeEventListener("focusin",o)},{immediate:!0}),le(s,async l=>{var u,c;await xt(),l?(u=i.value.contentEl)==null||u.focus({preventScroll:!0}):(c=i.value.activatorEl)==null||c.focus({preventScroll:!0})});const a=I(()=>Le({"aria-haspopup":"dialog","aria-expanded":String(s.value)},t.activatorProps));return ye(()=>{const l=Oi.filterProps(t);return _(Oi,Le({ref:i,class:["v-dialog",{"v-dialog--fullscreen":t.fullscreen,"v-dialog--scrollable":t.scrollable},t.class],style:t.style},l,{modelValue:s.value,"onUpdate:modelValue":u=>s.value=u,"aria-modal":"true",activatorProps:a.value,role:"dialog"},r),{activator:n.activator,default:function(){for(var u=arguments.length,c=new Array(u),d=0;d<u;d++)c[d]=arguments[d];return _(Nt,{root:"VDialog"},{default:()=>{var f;return[(f=n.default)==null?void 0:f.call(n,...c)]}})}})}),Th({},i)}}),I0=he()({name:"VCardActions",props:we(),setup(t,e){let{slots:n}=e;return hn({VBtn:{slim:!0,variant:"text"}}),ye(()=>{var s;return _("div",{class:["v-card-actions",t.class],style:t.style},[(s=n.default)==null?void 0:s.call(n)])}),{}}}),k0=ao("v-card-subtitle"),F0=ao("v-card-title"),V0=G({appendAvatar:String,appendIcon:Ke,prependAvatar:String,prependIcon:Ke,subtitle:[String,Number],title:[String,Number],...we(),...cs()},"VCardItem"),L0=he()({name:"VCardItem",props:V0(),setup(t,e){let{slots:n}=e;return ye(()=>{var u;const s=!!(t.prependAvatar||t.prependIcon),r=!!(s||n.prepend),i=!!(t.appendAvatar||t.appendIcon),o=!!(i||n.append),a=!!(t.title!=null||n.title),l=!!(t.subtitle!=null||n.subtitle);return _("div",{class:["v-card-item",t.class],style:t.style},[r&&_("div",{key:"prepend",class:"v-card-item__prepend"},[n.prepend?_(Nt,{key:"prepend-defaults",disabled:!s,defaults:{VAvatar:{density:t.density,image:t.prependAvatar},VIcon:{density:t.density,icon:t.prependIcon}}},n.prepend):_(Re,null,[t.prependAvatar&&_(sc,{key:"prepend-avatar",density:t.density,image:t.prependAvatar},null),t.prependIcon&&_(an,{key:"prepend-icon",density:t.density,icon:t.prependIcon},null)])]),_("div",{class:"v-card-item__content"},[a&&_(F0,{key:"title"},{default:()=>{var c;return[((c=n.title)==null?void 0:c.call(n))??t.title]}}),l&&_(k0,{key:"subtitle"},{default:()=>{var c;return[((c=n.subtitle)==null?void 0:c.call(n))??t.subtitle]}}),(u=n.default)==null?void 0:u.call(n)]),o&&_("div",{key:"append",class:"v-card-item__append"},[n.append?_(Nt,{key:"append-defaults",disabled:!i,defaults:{VAvatar:{density:t.density,image:t.appendAvatar},VIcon:{density:t.density,icon:t.appendIcon}}},n.append):_(Re,null,[t.appendIcon&&_(an,{key:"append-icon",density:t.density,icon:t.appendIcon},null),t.appendAvatar&&_(sc,{key:"append-avatar",density:t.density,image:t.appendAvatar},null)])])])}),{}}}),Ah=ao("v-card-text"),P0=G({appendAvatar:String,appendIcon:Ke,disabled:Boolean,flat:Boolean,hover:Boolean,image:String,link:{type:Boolean,default:void 0},prependAvatar:String,prependIcon:Ke,ripple:{type:[Boolean,Object],default:!0},subtitle:[String,Number],text:[String,Number],title:[String,Number],...Ir(),...we(),...cs(),...lo(),...ls(),...Jf(),...ml(),...eh(),...Kt(),...rh(),...Je(),...dt(),...Fr({variant:"elevated"})},"VCard"),O0=he()({name:"VCard",directives:{Ripple:Bs},props:P0(),setup(t,e){let{attrs:n,slots:s}=e;const{themeClasses:r}=It(t),{borderClasses:i}=kr(t),{colorClasses:o,colorStyles:a,variantClasses:l}=hl(t),{densityClasses:u}=Ns(t),{dimensionStyles:c}=uo(t),{elevationClasses:d}=us(t),{loaderClasses:f}=Qf(t),{locationStyles:h}=gl(t),{positionClasses:g}=th(t),{roundedClasses:y}=Jt(t),v=sh(t,n),S=I(()=>t.link!==!1&&v.isLink.value),m=I(()=>!t.disabled&&t.link!==!1&&(t.link||v.isClickable.value));return ye(()=>{const p=S.value?"a":t.tag,w=!!(s.title||t.title!=null),b=!!(s.subtitle||t.subtitle!=null),x=w||b,C=!!(s.append||t.appendAvatar||t.appendIcon),M=!!(s.prepend||t.prependAvatar||t.prependIcon),T=!!(s.image||t.image),V=x||M||C,L=!!(s.text||t.text!=null);return Gt(_(p,{class:["v-card",{"v-card--disabled":t.disabled,"v-card--flat":t.flat,"v-card--hover":t.hover&&!(t.disabled||t.flat),"v-card--link":m.value},r.value,i.value,o.value,u.value,d.value,f.value,g.value,y.value,l.value,t.class],style:[a.value,c.value,h.value,t.style],href:v.href.value,onClick:m.value&&v.navigate,tabindex:t.disabled?-1:void 0},{default:()=>{var A;return[T&&_("div",{key:"image",class:"v-card__image"},[s.image?_(Nt,{key:"image-defaults",disabled:!t.image,defaults:{VImg:{cover:!0,src:t.image}}},s.image):_(dl,{key:"image-img",cover:!0,src:t.image},null)]),_(Iy,{name:"v-card",active:!!t.loading,color:typeof t.loading=="boolean"?void 0:t.loading},{default:s.loader}),V&&_(L0,{key:"item",prependAvatar:t.prependAvatar,prependIcon:t.prependIcon,title:t.title,subtitle:t.subtitle,appendAvatar:t.appendAvatar,appendIcon:t.appendIcon},{default:s.item,prepend:s.prepend,title:s.title,subtitle:s.subtitle,append:s.append}),L&&_(Ah,{key:"text"},{default:()=>{var k;return[((k=s.text)==null?void 0:k.call(s))??t.text]}}),(A=s.default)==null?void 0:A.call(s),s.actions&&_(I0,null,{default:s.actions}),fl(m.value,"v-card")]}}),[[as("ripple"),m.value&&t.ripple]])}),{}}}),N0=t=>{const{touchstartX:e,touchendX:n,touchstartY:s,touchendY:r}=t,i=.5,o=16;t.offsetX=n-e,t.offsetY=r-s,Math.abs(t.offsetY)<i*Math.abs(t.offsetX)&&(t.left&&n<e-o&&t.left(t),t.right&&n>e+o&&t.right(t)),Math.abs(t.offsetX)<i*Math.abs(t.offsetY)&&(t.up&&r<s-o&&t.up(t),t.down&&r>s+o&&t.down(t))};function B0(t,e){var s;const n=t.changedTouches[0];e.touchstartX=n.clientX,e.touchstartY=n.clientY,(s=e.start)==null||s.call(e,{originalEvent:t,...e})}function z0(t,e){var s;const n=t.changedTouches[0];e.touchendX=n.clientX,e.touchendY=n.clientY,(s=e.end)==null||s.call(e,{originalEvent:t,...e}),N0(e)}function D0(t,e){var s;const n=t.changedTouches[0];e.touchmoveX=n.clientX,e.touchmoveY=n.clientY,(s=e.move)==null||s.call(e,{originalEvent:t,...e})}function $0(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};const e={touchstartX:0,touchstartY:0,touchendX:0,touchendY:0,touchmoveX:0,touchmoveY:0,offsetX:0,offsetY:0,left:t.left,right:t.right,up:t.up,down:t.down,start:t.start,move:t.move,end:t.end};return{touchstart:n=>B0(n,e),touchend:n=>z0(n,e),touchmove:n=>D0(n,e)}}function H0(t,e){var a;const n=e.value,s=n!=null&&n.parent?t.parentElement:t,r=(n==null?void 0:n.options)??{passive:!0},i=(a=e.instance)==null?void 0:a.$.uid;if(!s||!i)return;const o=$0(e.value);s._touchHandlers=s._touchHandlers??Object.create(null),s._touchHandlers[i]=o,df(o).forEach(l=>{s.addEventListener(l,o[l],r)})}function U0(t,e){var i,o;const n=(i=e.value)!=null&&i.parent?t.parentElement:t,s=(o=e.instance)==null?void 0:o.$.uid;if(!(n!=null&&n._touchHandlers)||!s)return;const r=n._touchHandlers[s];df(r).forEach(a=>{n.removeEventListener(a,r[a])}),delete n._touchHandlers[s]}const j0={mounted:H0,unmounted:U0},yl=Symbol.for("vuetify:v-slider");function W0(t,e,n){const s=n==="vertical",r=e.getBoundingClientRect(),i="touches"in t?t.touches[0]:t;return s?i.clientY-(r.top+r.height/2):i.clientX-(r.left+r.width/2)}function G0(t,e){return"touches"in t&&t.touches.length?t.touches[0][e]:"changedTouches"in t&&t.changedTouches.length?t.changedTouches[0][e]:t[e]}const q0=G({disabled:{type:Boolean,default:null},error:Boolean,readonly:{type:Boolean,default:null},max:{type:[Number,String],default:100},min:{type:[Number,String],default:0},step:{type:[Number,String],default:0},thumbColor:String,thumbLabel:{type:[Boolean,String],default:void 0,validator:t=>typeof t=="boolean"||t==="always"},thumbSize:{type:[Number,String],default:20},showTicks:{type:[Boolean,String],default:!1,validator:t=>typeof t=="boolean"||t==="always"},ticks:{type:[Array,Object]},tickSize:{type:[Number,String],default:2},color:String,trackColor:String,trackFillColor:String,trackSize:{type:[Number,String],default:4},direction:{type:String,default:"horizontal",validator:t=>["vertical","horizontal"].includes(t)},reverse:Boolean,...Kt(),...ls({elevation:2}),ripple:{type:Boolean,default:!0}},"Slider"),Y0=t=>{const e=I(()=>parseFloat(t.min)),n=I(()=>parseFloat(t.max)),s=I(()=>+t.step>0?parseFloat(t.step):0),r=I(()=>Math.max(Tu(s.value),Tu(e.value)));function i(o){if(o=parseFloat(o),s.value<=0)return o;const a=vr(o,e.value,n.value),l=e.value%s.value,u=Math.round((a-l)/s.value)*s.value+l;return parseFloat(Math.min(u,n.value).toFixed(r.value))}return{min:e,max:n,step:s,decimals:r,roundValue:i}},X0=t=>{let{props:e,steps:n,onSliderStart:s,onSliderMove:r,onSliderEnd:i,getActiveThumb:o}=t;const{isRtl:a}=Zt(),l=de(e,"reverse"),u=I(()=>e.direction==="vertical"),c=I(()=>u.value!==l.value),{min:d,max:f,step:h,decimals:g,roundValue:y}=n,v=I(()=>parseInt(e.thumbSize,10)),S=I(()=>parseInt(e.tickSize,10)),m=I(()=>parseInt(e.trackSize,10)),p=I(()=>(f.value-d.value)/h.value),w=de(e,"disabled"),b=I(()=>e.error||e.disabled?void 0:e.thumbColor??e.color),x=I(()=>e.error||e.disabled?void 0:e.trackColor??e.color),C=I(()=>e.error||e.disabled?void 0:e.trackFillColor??e.color),M=be(!1),T=be(0),V=se(),L=se();function A(Z){var P;const xe=e.direction==="vertical",gn=xe?"top":"left",ke=xe?"height":"width",_t=xe?"clientY":"clientX",{[gn]:Qe,[ke]:Nn}=(P=V.value)==null?void 0:P.$el.getBoundingClientRect(),E=G0(Z,_t);let R=Math.min(Math.max((E-Qe-T.value)/Nn,0),1)||0;return(xe?c.value:c.value!==a.value)&&(R=1-R),y(d.value+R*(f.value-d.value))}const k=Z=>{i({value:A(Z)}),M.value=!1,T.value=0},O=Z=>{L.value=o(Z),L.value&&(L.value.focus(),M.value=!0,L.value.contains(Z.target)?T.value=W0(Z,L.value,e.direction):(T.value=0,r({value:A(Z)})),s({value:A(Z)}))},H={passive:!0,capture:!0};function q(Z){r({value:A(Z)})}function j(Z){Z.stopPropagation(),Z.preventDefault(),k(Z),window.removeEventListener("mousemove",q,H),window.removeEventListener("mouseup",j)}function Y(Z){var xe;k(Z),window.removeEventListener("touchmove",q,H),(xe=Z.target)==null||xe.removeEventListener("touchend",Y)}function ae(Z){var xe;O(Z),window.addEventListener("touchmove",q,H),(xe=Z.target)==null||xe.addEventListener("touchend",Y,{passive:!1})}function Q(Z){Z.preventDefault(),O(Z),window.addEventListener("mousemove",q,H),window.addEventListener("mouseup",j,{passive:!1})}const te=Z=>{const xe=(Z-d.value)/(f.value-d.value)*100;return vr(isNaN(xe)?0:xe,0,100)},ge=de(e,"showTicks"),me=I(()=>ge.value?e.ticks?Array.isArray(e.ticks)?e.ticks.map(Z=>({value:Z,position:te(Z),label:Z.toString()})):Object.keys(e.ticks).map(Z=>({value:parseFloat(Z),position:te(parseFloat(Z)),label:e.ticks[Z]})):p.value!==1/0?sl(p.value+1).map(Z=>{const xe=d.value+Z*h.value;return{value:xe,position:te(xe)}}):[]:[]),Ce=I(()=>me.value.some(Z=>{let{label:xe}=Z;return!!xe})),We={activeThumbRef:L,color:de(e,"color"),decimals:g,disabled:w,direction:de(e,"direction"),elevation:de(e,"elevation"),hasLabels:Ce,isReversed:l,indexFromEnd:c,min:d,max:f,mousePressed:M,numTicks:p,onSliderMousedown:Q,onSliderTouchstart:ae,parsedTicks:me,parseMouseMove:A,position:te,readonly:de(e,"readonly"),rounded:de(e,"rounded"),roundValue:y,showTicks:ge,startOffset:T,step:h,thumbSize:v,thumbColor:b,thumbLabel:de(e,"thumbLabel"),ticks:de(e,"ticks"),tickSize:S,trackColor:x,trackContainerRef:V,trackFillColor:C,trackSize:m,vertical:u};return ut(yl,We),We},Z0=G({focused:Boolean,max:{type:Number,required:!0},min:{type:Number,required:!0},modelValue:{type:Number,required:!0},position:{type:Number,required:!0},ripple:{type:[Boolean,Object],default:!0},...we()},"VSliderThumb"),K0=he()({name:"VSliderThumb",directives:{Ripple:Bs},props:Z0(),emits:{"update:modelValue":t=>!0},setup(t,e){let{slots:n,emit:s}=e;const r=Pe(yl),{isRtl:i,rtlClasses:o}=Zt();if(!r)throw new Error("[Vuetify] v-slider-thumb must be used inside v-slider or v-range-slider");const{thumbColor:a,step:l,disabled:u,thumbSize:c,thumbLabel:d,direction:f,isReversed:h,vertical:g,readonly:y,elevation:v,mousePressed:S,decimals:m,indexFromEnd:p}=r,w=I(()=>u.value?void 0:v.value),{elevationClasses:b}=us(w),{textColorClasses:x,textColorStyles:C}=Ln(a),{pageup:M,pagedown:T,end:V,home:L,left:A,right:k,down:O,up:H}=Jv,q=[M,T,V,L,A,k,O,H],j=I(()=>l.value?[1,2,3]:[1,5,10]);function Y(Q,te){if(!q.includes(Q.key))return;Q.preventDefault();const ge=l.value||.1,me=(t.max-t.min)/ge;if([A,k,O,H].includes(Q.key)){const We=(g.value?[i.value?A:k,h.value?O:H]:p.value!==i.value?[A,H]:[k,H]).includes(Q.key)?1:-1,Z=Q.shiftKey?2:Q.ctrlKey?1:0;te=te+We*ge*j.value[Z]}else if(Q.key===L)te=t.min;else if(Q.key===V)te=t.max;else{const Ce=Q.key===T?1:-1;te=te-Ce*ge*(me>100?me/10:10)}return Math.max(t.min,Math.min(t.max,te))}function ae(Q){const te=Y(Q,t.modelValue);te!=null&&s("update:modelValue",te)}return ye(()=>{const Q=ie(p.value?100-t.position:t.position,"%");return _("div",{class:["v-slider-thumb",{"v-slider-thumb--focused":t.focused,"v-slider-thumb--pressed":t.focused&&S.value},t.class,o.value],style:[{"--v-slider-thumb-position":Q,"--v-slider-thumb-size":ie(c.value)},t.style],role:"slider",tabindex:u.value?-1:0,"aria-valuemin":t.min,"aria-valuemax":t.max,"aria-valuenow":t.modelValue,"aria-readonly":!!y.value,"aria-orientation":f.value,onKeydown:y.value?void 0:ae},[_("div",{class:["v-slider-thumb__surface",x.value,b.value],style:{...C.value}},null),Gt(_("div",{class:["v-slider-thumb__ripple",x.value],style:C.value},null),[[as("ripple"),t.ripple,null,{circle:!0,center:!0}]]),_(oy,{origin:"bottom center"},{default:()=>{var te;return[Gt(_("div",{class:"v-slider-thumb__label-container"},[_("div",{class:["v-slider-thumb__label"]},[_("div",null,[((te=n["thumb-label"])==null?void 0:te.call(n,{modelValue:t.modelValue}))??t.modelValue.toFixed(l.value?m.value:1)])])]),[[io,d.value&&t.focused||d.value==="always"]])]}})])}),{}}}),J0=G({start:{type:Number,required:!0},stop:{type:Number,required:!0},...we()},"VSliderTrack"),Q0=he()({name:"VSliderTrack",props:J0(),emits:{},setup(t,e){let{slots:n}=e;const s=Pe(yl);if(!s)throw new Error("[Vuetify] v-slider-track must be inside v-slider or v-range-slider");const{color:r,parsedTicks:i,rounded:o,showTicks:a,tickSize:l,trackColor:u,trackFillColor:c,trackSize:d,vertical:f,min:h,max:g,indexFromEnd:y}=s,{roundedClasses:v}=Jt(o),{backgroundColorClasses:S,backgroundColorStyles:m}=Tt(c),{backgroundColorClasses:p,backgroundColorStyles:w}=Tt(u),b=I(()=>`inset-${f.value?"block":"inline"}-${y.value?"end":"start"}`),x=I(()=>f.value?"height":"width"),C=I(()=>({[b.value]:"0%",[x.value]:"100%"})),M=I(()=>t.stop-t.start),T=I(()=>({[b.value]:ie(t.start,"%"),[x.value]:ie(M.value,"%")})),V=I(()=>a.value?(f.value?i.value.slice().reverse():i.value).map((A,k)=>{var H;const O=A.value!==h.value&&A.value!==g.value?ie(A.position,"%"):void 0;return _("div",{key:A.value,class:["v-slider-track__tick",{"v-slider-track__tick--filled":A.position>=t.start&&A.position<=t.stop,"v-slider-track__tick--first":A.value===h.value,"v-slider-track__tick--last":A.value===g.value}],style:{[b.value]:O}},[(A.label||n["tick-label"])&&_("div",{class:"v-slider-track__tick-label"},[((H=n["tick-label"])==null?void 0:H.call(n,{tick:A,index:k}))??A.label])])}):[]);return ye(()=>_("div",{class:["v-slider-track",v.value,t.class],style:[{"--v-slider-track-size":ie(d.value),"--v-slider-tick-size":ie(l.value)},t.style]},[_("div",{class:["v-slider-track__background",p.value,{"v-slider-track__background--opacity":!!r.value||!c.value}],style:{...C.value,...w.value}},null),_("div",{class:["v-slider-track__fill",S.value],style:{...T.value,...m.value}},null),a.value&&_("div",{class:["v-slider-track__ticks",{"v-slider-track__ticks--always-show":a.value==="always"}]},[V.value])])),{}}}),e1=G({...ph(),...q0(),...yh(),modelValue:{type:[Number,String],default:0}},"VSlider"),qr=he()({name:"VSlider",props:e1(),emits:{"update:focused":t=>!0,"update:modelValue":t=>!0,start:t=>!0,end:t=>!0},setup(t,e){let{slots:n,emit:s}=e;const r=se(),{rtlClasses:i}=Zt(),o=Y0(t),a=ct(t,"modelValue",void 0,x=>o.roundValue(x??o.min.value)),{min:l,max:u,mousePressed:c,roundValue:d,onSliderMousedown:f,onSliderTouchstart:h,trackContainerRef:g,position:y,hasLabels:v,readonly:S}=X0({props:t,steps:o,onSliderStart:()=>{s("start",a.value)},onSliderEnd:x=>{let{value:C}=x;const M=d(C);a.value=M,s("end",M)},onSliderMove:x=>{let{value:C}=x;return a.value=d(C)},getActiveThumb:()=>{var x;return(x=r.value)==null?void 0:x.$el}}),{isFocused:m,focus:p,blur:w}=Xy(t),b=I(()=>y(a.value));return ye(()=>{const x=ic.filterProps(t),C=!!(t.label||n.label||n.prepend);return _(ic,Le({class:["v-slider",{"v-slider--has-labels":!!n["tick-label"]||v.value,"v-slider--focused":m.value,"v-slider--pressed":c.value,"v-slider--disabled":t.disabled},i.value,t.class],style:t.style},x,{focused:m.value}),{...n,prepend:C?M=>{var T,V;return _(Re,null,[((T=n.label)==null?void 0:T.call(n,M))??(t.label?_(hh,{id:M.id.value,class:"v-slider__label",text:t.label},null):void 0),(V=n.prepend)==null?void 0:V.call(n,M)])}:void 0,default:M=>{let{id:T,messagesId:V}=M;return _("div",{class:"v-slider__container",onMousedown:S.value?void 0:f,onTouchstartPassive:S.value?void 0:h},[_("input",{id:T.value,name:t.name||T.value,disabled:!!t.disabled,readonly:!!t.readonly,tabindex:"-1",value:a.value},null),_(Q0,{ref:g,start:0,stop:b.value},{"tick-label":n["tick-label"]}),_(K0,{ref:r,"aria-describedby":V.value,focused:m.value,min:l.value,max:u.value,modelValue:a.value,"onUpdate:modelValue":L=>a.value=L,position:b.value,elevation:t.elevation,onFocus:p,onBlur:w,ripple:t.ripple},{"thumb-label":n["thumb-label"]})])}})}),{}}}),Ma={"001":1,AD:1,AE:6,AF:6,AG:0,AI:1,AL:1,AM:1,AN:1,AR:1,AS:0,AT:1,AU:1,AX:1,AZ:1,BA:1,BD:0,BE:1,BG:1,BH:6,BM:1,BN:1,BR:0,BS:0,BT:0,BW:0,BY:1,BZ:0,CA:0,CH:1,CL:1,CM:1,CN:1,CO:0,CR:1,CY:1,CZ:1,DE:1,DJ:6,DK:1,DM:0,DO:0,DZ:6,EC:1,EE:1,EG:6,ES:1,ET:0,FI:1,FJ:1,FO:1,FR:1,GB:1,"GB-alt-variant":0,GE:1,GF:1,GP:1,GR:1,GT:0,GU:0,HK:0,HN:0,HR:1,HU:1,ID:0,IE:1,IL:0,IN:0,IQ:6,IR:6,IS:1,IT:1,JM:0,JO:6,JP:0,KE:0,KG:1,KH:0,KR:0,KW:6,KZ:1,LA:0,LB:1,LI:1,LK:1,LT:1,LU:1,LV:1,LY:6,MC:1,MD:1,ME:1,MH:0,MK:1,MM:0,MN:1,MO:0,MQ:1,MT:0,MV:5,MX:0,MY:1,MZ:0,NI:0,NL:1,NO:1,NP:0,NZ:1,OM:6,PA:0,PE:0,PH:0,PK:0,PL:1,PR:0,PT:0,PY:0,QA:6,RE:1,RO:1,RS:1,RU:1,SA:0,SD:6,SE:1,SG:0,SI:1,SK:1,SM:1,SV:0,SY:6,TH:0,TJ:1,TM:1,TR:1,TT:0,TW:0,UA:1,UM:0,US:0,UY:1,UZ:1,VA:1,VE:0,VI:0,VN:1,WS:0,XK:1,YE:0,ZA:0,ZW:0};function t1(t,e){const n=[];let s=[];const r=Rh(t),i=Ih(t),o=(r.getDay()-Ma[e.slice(-2).toUpperCase()]+7)%7,a=(i.getDay()-Ma[e.slice(-2).toUpperCase()]+7)%7;for(let l=0;l<o;l++){const u=new Date(r);u.setDate(u.getDate()-(o-l)),s.push(u)}for(let l=1;l<=i.getDate();l++){const u=new Date(t.getFullYear(),t.getMonth(),l);s.push(u),s.length===7&&(n.push(s),s=[])}for(let l=1;l<7-a;l++){const u=new Date(i);u.setDate(u.getDate()+l),s.push(u)}return s.length>0&&n.push(s),n}function n1(t){const e=new Date(t);for(;e.getDay()!==0;)e.setDate(e.getDate()-1);return e}function s1(t){const e=new Date(t);for(;e.getDay()!==6;)e.setDate(e.getDate()+1);return e}function Rh(t){return new Date(t.getFullYear(),t.getMonth(),1)}function Ih(t){return new Date(t.getFullYear(),t.getMonth()+1,0)}function r1(t){const e=t.split("-").map(Number);return new Date(e[0],e[1]-1,e[2])}const i1=/^([12]\d{3}-([1-9]|0[1-9]|1[0-2])-([1-9]|0[1-9]|[12]\d|3[01]))$/;function kh(t){if(t==null)return new Date;if(t instanceof Date)return t;if(typeof t=="string"){let e;if(i1.test(t))return r1(t);if(e=Date.parse(t),!isNaN(e))return new Date(e)}return null}const gc=new Date(2e3,0,2);function o1(t){const e=Ma[t.slice(-2).toUpperCase()];return sl(7).map(n=>{const s=new Date(gc);return s.setDate(gc.getDate()+e+n),new Intl.DateTimeFormat(t,{weekday:"narrow"}).format(s)})}function a1(t,e,n,s){const r=kh(t)??new Date,i=s==null?void 0:s[e];if(typeof i=="function")return i(r,e,n);let o={};switch(e){case"fullDateWithWeekday":o={weekday:"long",day:"numeric",month:"long",year:"numeric"};break;case"hours12h":o={hour:"numeric",hour12:!0};break;case"normalDateWithWeekday":o={weekday:"short",day:"numeric",month:"short"};break;case"keyboardDate":o={day:"2-digit",month:"2-digit",year:"numeric"};break;case"monthAndDate":o={month:"long",day:"numeric"};break;case"monthAndYear":o={month:"long",year:"numeric"};break;case"month":o={month:"long"};break;case"monthShort":o={month:"short"};break;case"dayOfMonth":return new Intl.NumberFormat(n).format(r.getDate());case"shortDate":o={year:"2-digit",month:"numeric",day:"numeric"};break;case"weekdayShort":o={weekday:"short"};break;case"year":o={year:"numeric"};break;default:o=i??{timeZone:"UTC",timeZoneName:"short"}}return new Intl.DateTimeFormat(n,o).format(r)}function l1(t,e){const n=t.toJsDate(e),s=n.getFullYear(),r=Ru(String(n.getMonth()+1),2,"0"),i=Ru(String(n.getDate()),2,"0");return`${s}-${r}-${i}`}function u1(t){const[e,n,s]=t.split("-").map(Number);return new Date(e,n-1,s)}function c1(t,e){const n=new Date(t);return n.setMinutes(n.getMinutes()+e),n}function d1(t,e){const n=new Date(t);return n.setHours(n.getHours()+e),n}function f1(t,e){const n=new Date(t);return n.setDate(n.getDate()+e),n}function h1(t,e){const n=new Date(t);return n.setDate(n.getDate()+e*7),n}function m1(t,e){const n=new Date(t);return n.setMonth(n.getMonth()+e),n}function g1(t){return t.getFullYear()}function v1(t){return t.getMonth()}function p1(t){return new Date(t.getFullYear(),t.getMonth()+1,1)}function y1(t){return t.getHours()}function b1(t){return t.getMinutes()}function x1(t){return new Date(t.getFullYear(),0,1)}function w1(t){return new Date(t.getFullYear(),11,31)}function _1(t,e){return Ta(t,e[0])&&C1(t,e[1])}function S1(t){const e=new Date(t);return e instanceof Date&&!isNaN(e.getTime())}function Ta(t,e){return t.getTime()>e.getTime()}function C1(t,e){return t.getTime()<e.getTime()}function vc(t,e){return t.getTime()===e.getTime()}function E1(t,e){return t.getDate()===e.getDate()&&t.getMonth()===e.getMonth()&&t.getFullYear()===e.getFullYear()}function M1(t,e){return t.getMonth()===e.getMonth()&&t.getFullYear()===e.getFullYear()}function T1(t,e,n){const s=new Date(t),r=new Date(e);switch(n){case"years":return s.getFullYear()-r.getFullYear();case"quarters":return Math.floor((s.getMonth()-r.getMonth()+(s.getFullYear()-r.getFullYear())*12)/4);case"months":return s.getMonth()-r.getMonth()+(s.getFullYear()-r.getFullYear())*12;case"weeks":return Math.floor((s.getTime()-r.getTime())/(1e3*60*60*24*7));case"days":return Math.floor((s.getTime()-r.getTime())/(1e3*60*60*24));case"hours":return Math.floor((s.getTime()-r.getTime())/(1e3*60*60));case"minutes":return Math.floor((s.getTime()-r.getTime())/(1e3*60));case"seconds":return Math.floor((s.getTime()-r.getTime())/1e3);default:return s.getTime()-r.getTime()}}function A1(t,e){const n=new Date(t);return n.setHours(e),n}function R1(t,e){const n=new Date(t);return n.setMinutes(e),n}function I1(t,e){const n=new Date(t);return n.setMonth(e),n}function k1(t,e){const n=new Date(t);return n.setFullYear(e),n}function F1(t){return new Date(t.getFullYear(),t.getMonth(),t.getDate())}function V1(t){return new Date(t.getFullYear(),t.getMonth(),t.getDate(),23,59,59,999)}class L1{constructor(e){this.locale=e.locale,this.formats=e.formats}date(e){return kh(e)}toJsDate(e){return e}toISO(e){return l1(this,e)}parseISO(e){return u1(e)}addMinutes(e,n){return c1(e,n)}addHours(e,n){return d1(e,n)}addDays(e,n){return f1(e,n)}addWeeks(e,n){return h1(e,n)}addMonths(e,n){return m1(e,n)}getWeekArray(e){return t1(e,this.locale)}startOfWeek(e){return n1(e)}endOfWeek(e){return s1(e)}startOfMonth(e){return Rh(e)}endOfMonth(e){return Ih(e)}format(e,n){return a1(e,n,this.locale,this.formats)}isEqual(e,n){return vc(e,n)}isValid(e){return S1(e)}isWithinRange(e,n){return _1(e,n)}isAfter(e,n){return Ta(e,n)}isBefore(e,n){return!Ta(e,n)&&!vc(e,n)}isSameDay(e,n){return E1(e,n)}isSameMonth(e,n){return M1(e,n)}setMinutes(e,n){return R1(e,n)}setHours(e,n){return A1(e,n)}setMonth(e,n){return I1(e,n)}setYear(e,n){return k1(e,n)}getDiff(e,n,s){return T1(e,n,s)}getWeekdays(){return o1(this.locale)}getYear(e){return g1(e)}getMonth(e){return v1(e)}getNextMonth(e){return p1(e)}getHours(e){return y1(e)}getMinutes(e){return b1(e)}startOfDay(e){return F1(e)}endOfDay(e){return V1(e)}startOfYear(e){return x1(e)}endOfYear(e){return w1(e)}}const P1=Symbol.for("vuetify:date-options"),pc=Symbol.for("vuetify:date-adapter");function O1(t,e){const n=vt({adapter:L1,locale:{af:"af-ZA",bg:"bg-BG",ca:"ca-ES",ckb:"",cs:"cs-CZ",de:"de-DE",el:"el-GR",en:"en-US",et:"et-EE",fa:"fa-IR",fi:"fi-FI",hr:"hr-HR",hu:"hu-HU",he:"he-IL",id:"id-ID",it:"it-IT",ja:"ja-JP",ko:"ko-KR",lv:"lv-LV",lt:"lt-LT",nl:"nl-NL",no:"no-NO",pl:"pl-PL",pt:"pt-PT",ro:"ro-RO",ru:"ru-RU",sk:"sk-SK",sl:"sl-SI",srCyrl:"sr-SP",srLatn:"sr-SP",sv:"sv-SE",th:"th-TH",tr:"tr-TR",az:"az-AZ",uk:"uk-UA",vi:"vi-VN",zhHans:"zh-CN",zhHant:"zh-TW"}},t);return{options:n,instance:N1(n,e)}}function N1(t,e){const n=tt(typeof t.adapter=="function"?new t.adapter({locale:t.locale[e.current.value]??e.current.value,formats:t.formats}):t.adapter);return le(e.current,s=>{n.locale=t.locale[s]??s??n.locale}),n}const B1=Symbol.for("vuetify:goto");function z1(){return{container:void 0,duration:300,layout:!1,offset:0,easing:"easeInOutCubic",patterns:{linear:t=>t,easeInQuad:t=>t**2,easeOutQuad:t=>t*(2-t),easeInOutQuad:t=>t<.5?2*t**2:-1+(4-2*t)*t,easeInCubic:t=>t**3,easeOutCubic:t=>--t**3+1,easeInOutCubic:t=>t<.5?4*t**3:(t-1)*(2*t-2)*(2*t-2)+1,easeInQuart:t=>t**4,easeOutQuart:t=>1- --t**4,easeInOutQuart:t=>t<.5?8*t**4:1-8*--t**4,easeInQuint:t=>t**5,easeOutQuint:t=>1+--t**5,easeInOutQuint:t=>t<.5?16*t**5:1+16*--t**5}}}function D1(t,e){return{rtl:e.isRtl,options:vt(z1(),t)}}const $1=G({fluid:{type:Boolean,default:!1},...we(),...Je()},"VContainer"),H1=he()({name:"VContainer",props:$1(),setup(t,e){let{slots:n}=e;const{rtlClasses:s}=Zt();return ye(()=>_(t.tag,{class:["v-container",{"v-container--fluid":t.fluid},s.value,t.class],style:t.style},n)),{}}}),Fh=mo.reduce((t,e)=>(t[e]={type:[Boolean,String,Number],default:!1},t),{}),Vh=mo.reduce((t,e)=>{const n="offset"+cn(e);return t[n]={type:[String,Number],default:null},t},{}),Lh=mo.reduce((t,e)=>{const n="order"+cn(e);return t[n]={type:[String,Number],default:null},t},{}),yc={col:Object.keys(Fh),offset:Object.keys(Vh),order:Object.keys(Lh)};function U1(t,e,n){let s=t;if(!(n==null||n===!1)){if(e){const r=e.replace(t,"");s+=`-${r}`}return t==="col"&&(s="v-"+s),t==="col"&&(n===""||n===!0)||(s+=`-${n}`),s.toLowerCase()}}const j1=["auto","start","end","center","baseline","stretch"],W1=G({cols:{type:[Boolean,String,Number],default:!1},...Fh,offset:{type:[String,Number],default:null},...Vh,order:{type:[String,Number],default:null},...Lh,alignSelf:{type:String,default:null,validator:t=>j1.includes(t)},...we(),...Je()},"VCol"),G1=he()({name:"VCol",props:W1(),setup(t,e){let{slots:n}=e;const s=I(()=>{const r=[];let i;for(i in yc)yc[i].forEach(a=>{const l=t[a],u=U1(i,a,l);u&&r.push(u)});const o=r.some(a=>a.startsWith("v-col-"));return r.push({"v-col":!o||!t.cols,[`v-col-${t.cols}`]:t.cols,[`offset-${t.offset}`]:t.offset,[`order-${t.order}`]:t.order,[`align-self-${t.alignSelf}`]:t.alignSelf}),r});return()=>{var r;return On(t.tag,{class:[s.value,t.class],style:t.style},(r=n.default)==null?void 0:r.call(n))}}}),bl=["start","end","center"],Ph=["space-between","space-around","space-evenly"];function xl(t,e){return mo.reduce((n,s)=>{const r=t+cn(s);return n[r]=e(),n},{})}const q1=[...bl,"baseline","stretch"],Oh=t=>q1.includes(t),Nh=xl("align",()=>({type:String,default:null,validator:Oh})),Y1=[...bl,...Ph],Bh=t=>Y1.includes(t),zh=xl("justify",()=>({type:String,default:null,validator:Bh})),X1=[...bl,...Ph,"stretch"],Dh=t=>X1.includes(t),$h=xl("alignContent",()=>({type:String,default:null,validator:Dh})),bc={align:Object.keys(Nh),justify:Object.keys(zh),alignContent:Object.keys($h)},Z1={align:"align",justify:"justify",alignContent:"align-content"};function K1(t,e,n){let s=Z1[t];if(n!=null){if(e){const r=e.replace(t,"");s+=`-${r}`}return s+=`-${n}`,s.toLowerCase()}}const J1=G({dense:Boolean,noGutters:Boolean,align:{type:String,default:null,validator:Oh},...Nh,justify:{type:String,default:null,validator:Bh},...zh,alignContent:{type:String,default:null,validator:Dh},...$h,...we(),...Je()},"VRow"),Q1=he()({name:"VRow",props:J1(),setup(t,e){let{slots:n}=e;const s=I(()=>{const r=[];let i;for(i in bc)bc[i].forEach(o=>{const a=t[o],l=K1(i,o,a);l&&r.push(l)});return r.push({"v-row--no-gutters":t.noGutters,"v-row--dense":t.dense,[`align-${t.align}`]:t.align,[`justify-${t.justify}`]:t.justify,[`align-content-${t.alignContent}`]:t.alignContent}),r});return()=>{var r;return On(t.tag,{class:["v-row",s.value,t.class],style:t.style},(r=n.default)==null?void 0:r.call(n))}}}),Aa=ao("v-spacer","div","VSpacer"),xr=Symbol.for("vuetify:v-expansion-panel"),eb=["default","accordion","inset","popout"],tb=G({color:String,flat:Boolean,focusable:Boolean,static:Boolean,tile:Boolean,variant:{type:String,default:"default",validator:t=>eb.includes(t)},readonly:Boolean,...we(),...Hf(),...Je(),...dt()},"VExpansionPanels"),nb=he()({name:"VExpansionPanels",props:tb(),emits:{"update:modelValue":t=>!0},setup(t,e){let{slots:n}=e;Wf(t,xr);const{themeClasses:s}=It(t),r=I(()=>t.variant&&`v-expansion-panels--variant-${t.variant}`);return hn({VExpansionPanel:{color:de(t,"color"),readonly:de(t,"readonly")},VExpansionPanelTitle:{focusable:de(t,"focusable"),static:de(t,"static")}}),ye(()=>_(t.tag,{class:["v-expansion-panels",{"v-expansion-panels--flat":t.flat,"v-expansion-panels--tile":t.tile},s.value,r.value,t.class],style:t.style},n)),{}}}),sb=G({...we(),...vl()},"VExpansionPanelText"),Hh=he()({name:"VExpansionPanelText",props:sb(),setup(t,e){let{slots:n}=e;const s=Pe(xr);if(!s)throw new Error("[Vuetify] v-expansion-panel-text needs to be placed inside v-expansion-panel");const{hasContent:r,onAfterLeave:i}=Sh(t,s.isSelected);return ye(()=>_(Pf,{onAfterLeave:i},{default:()=>{var o;return[Gt(_("div",{class:["v-expansion-panel-text",t.class],style:t.style},[n.default&&r.value&&_("div",{class:"v-expansion-panel-text__wrapper"},[(o=n.default)==null?void 0:o.call(n)])]),[[io,s.isSelected.value]])]}})),{}}}),Uh=G({color:String,expandIcon:{type:Ke,default:"$expand"},collapseIcon:{type:Ke,default:"$collapse"},hideActions:Boolean,focusable:Boolean,static:Boolean,ripple:{type:[Boolean,Object],default:!1},readonly:Boolean,...we()},"VExpansionPanelTitle"),jh=he()({name:"VExpansionPanelTitle",directives:{Ripple:Bs},props:Uh(),setup(t,e){let{slots:n}=e;const s=Pe(xr);if(!s)throw new Error("[Vuetify] v-expansion-panel-title needs to be placed inside v-expansion-panel");const{backgroundColorClasses:r,backgroundColorStyles:i}=Tt(t,"color"),o=I(()=>({collapseIcon:t.collapseIcon,disabled:s.disabled.value,expanded:s.isSelected.value,expandIcon:t.expandIcon,readonly:t.readonly}));return ye(()=>{var a;return Gt(_("button",{class:["v-expansion-panel-title",{"v-expansion-panel-title--active":s.isSelected.value,"v-expansion-panel-title--focusable":t.focusable,"v-expansion-panel-title--static":t.static},r.value,t.class],style:[i.value,t.style],type:"button",tabindex:s.disabled.value?-1:void 0,disabled:s.disabled.value,"aria-expanded":s.isSelected.value,onClick:t.readonly?void 0:s.toggle},[_("span",{class:"v-expansion-panel-title__overlay"},null),(a=n.default)==null?void 0:a.call(n,o.value),!t.hideActions&&_("span",{class:"v-expansion-panel-title__icon"},[n.actions?n.actions(o.value):_(an,{icon:s.isSelected.value?t.collapseIcon:t.expandIcon},null)])]),[[as("ripple"),t.ripple]])}),{}}}),rb=G({title:String,text:String,bgColor:String,...we(),...ls(),...Uf(),...vl(),...Kt(),...Je(),...Uh()},"VExpansionPanel"),ib=he()({name:"VExpansionPanel",props:rb(),emits:{"group:selected":t=>!0},setup(t,e){let{slots:n}=e;const s=jf(t,xr),{backgroundColorClasses:r,backgroundColorStyles:i}=Tt(t,"bgColor"),{elevationClasses:o}=us(t),{roundedClasses:a}=Jt(t),l=I(()=>(s==null?void 0:s.disabled.value)||t.disabled),u=I(()=>s.group.items.value.reduce((f,h,g)=>(s.group.selected.value.includes(h.id)&&f.push(g),f),[])),c=I(()=>{const f=s.group.items.value.findIndex(h=>h.id===s.id);return!s.isSelected.value&&u.value.some(h=>h-f===1)}),d=I(()=>{const f=s.group.items.value.findIndex(h=>h.id===s.id);return!s.isSelected.value&&u.value.some(h=>h-f===-1)});return ut(xr,s),hn({VExpansionPanelText:{eager:de(t,"eager")},VExpansionPanelTitle:{readonly:de(t,"readonly")}}),ye(()=>{const f=!!(n.text||t.text),h=!!(n.title||t.title);return _(t.tag,{class:["v-expansion-panel",{"v-expansion-panel--active":s.isSelected.value,"v-expansion-panel--before-active":c.value,"v-expansion-panel--after-active":d.value,"v-expansion-panel--disabled":l.value},a.value,r.value,t.class],style:[i.value,t.style]},{default:()=>{var g;return[_("div",{class:["v-expansion-panel__shadow",...o.value]},null),h&&_(jh,{key:"title",collapseIcon:t.collapseIcon,color:t.color,expandIcon:t.expandIcon,hideActions:t.hideActions,ripple:t.ripple},{default:()=>[n.title?n.title():t.title]}),f&&_(Hh,{key:"text"},{default:()=>[n.text?n.text():t.text]}),(g=n.default)==null?void 0:g.call(n)]}})}),{}}}),ob=G({...we(),...$p()},"VLayout"),ab=he()({name:"VLayout",props:ob(),setup(t,e){let{slots:n}=e;const{layoutClasses:s,layoutStyles:r,getLayoutItem:i,items:o,layoutRef:a}=Gp(t);return ye(()=>{var l;return _("div",{ref:a,class:[s.value,t.class],style:[r.value,t.style]},[(l=n.default)==null?void 0:l.call(n)])}),{getLayoutItem:i,items:o}}}),lb=G({scrollable:Boolean,...we(),...Je({tag:"main"})},"VMain"),ub=he()({name:"VMain",props:lb(),setup(t,e){let{slots:n}=e;const{mainStyles:s}=Up(),{ssrBootStyles:r}=Df();return ye(()=>_(t.tag,{class:["v-main",{"v-main--scrollable":t.scrollable},t.class],style:[s.value,r.value,t.style]},{default:()=>{var i,o;return[t.scrollable?_("div",{class:"v-main__scroller"},[(i=n.default)==null?void 0:i.call(n)]):(o=n.default)==null?void 0:o.call(n)]}})),{}}});function cb(t){let{rootEl:e,isSticky:n,layoutItemStyles:s}=t;const r=be(!1),i=be(0),o=I(()=>{const u=typeof r.value=="boolean"?"top":r.value;return[n.value?{top:"auto",bottom:"auto",height:void 0}:void 0,r.value?{[u]:ie(i.value)}:{top:s.value.top}]});Yt(()=>{le(n,u=>{u?window.addEventListener("scroll",l,{passive:!0}):window.removeEventListener("scroll",l)},{immediate:!0})}),Rt(()=>{window.removeEventListener("scroll",l)});let a=0;function l(){const u=a>window.scrollY?"up":"down",c=e.value.getBoundingClientRect(),d=parseFloat(s.value.top??0),f=window.scrollY-Math.max(0,i.value-d),h=c.height+Math.max(i.value,d)-window.scrollY-window.innerHeight,g=parseFloat(getComputedStyle(e.value).getPropertyValue("--v-body-scroll-y"))||0;c.height<window.innerHeight-d?(r.value="top",i.value=d):u==="up"&&r.value==="bottom"||u==="down"&&r.value==="top"?(i.value=window.scrollY+c.top-g,r.value=!0):u==="down"&&h<=0?(i.value=0,r.value="bottom"):u==="up"&&f<=0&&(g?r.value!=="top"&&(i.value=-f+g+d,r.value="top"):(i.value=c.top+f,r.value="top")),a=window.scrollY}return{isStuck:r,stickyStyles:o}}const db=100,fb=20;function xc(t){return(t<0?-1:1)*Math.sqrt(Math.abs(t))*1.41421356237}function wc(t){if(t.length<2)return 0;if(t.length===2)return t[1].t===t[0].t?0:(t[1].d-t[0].d)/(t[1].t-t[0].t);let e=0;for(let n=t.length-1;n>0;n--){if(t[n].t===t[n-1].t)continue;const s=xc(e),r=(t[n].d-t[n-1].d)/(t[n].t-t[n-1].t);e+=(r-s)*Math.abs(r),n===t.length-1&&(e*=.5)}return xc(e)*1e3}function hb(){const t={};function e(r){Array.from(r.changedTouches).forEach(i=>{(t[i.identifier]??(t[i.identifier]=new sp(fb))).push([r.timeStamp,i])})}function n(r){Array.from(r.changedTouches).forEach(i=>{delete t[i.identifier]})}function s(r){var u;const i=(u=t[r])==null?void 0:u.values().reverse();if(!i)throw new Error(`No samples for touch id ${r}`);const o=i[0],a=[],l=[];for(const c of i){if(o[0]-c[0]>db)break;a.push({t:c[0],d:c[1].clientX}),l.push({t:c[0],d:c[1].clientY})}return{x:wc(a),y:wc(l),get direction(){const{x:c,y:d}=this,[f,h]=[Math.abs(c),Math.abs(d)];return f>h&&c>=0?"right":f>h&&c<=0?"left":h>f&&d>=0?"down":h>f&&d<=0?"up":mb()}}}return{addMovement:e,endTouch:n,getVelocity:s}}function mb(){throw new Error}function gb(t){let{isActive:e,isTemporary:n,width:s,touchless:r,position:i}=t;Yt(()=>{window.addEventListener("touchstart",S,{passive:!0}),window.addEventListener("touchmove",m,{passive:!1}),window.addEventListener("touchend",p,{passive:!0})}),Rt(()=>{window.removeEventListener("touchstart",S),window.removeEventListener("touchmove",m),window.removeEventListener("touchend",p)});const o=I(()=>["left","right"].includes(i.value)),{addMovement:a,endTouch:l,getVelocity:u}=hb();let c=!1;const d=be(!1),f=be(0),h=be(0);let g;function y(b,x){return(i.value==="left"?b:i.value==="right"?document.documentElement.clientWidth-b:i.value==="top"?b:i.value==="bottom"?document.documentElement.clientHeight-b:vs())-(x?s.value:0)}function v(b){let x=arguments.length>1&&arguments[1]!==void 0?arguments[1]:!0;const C=i.value==="left"?(b-h.value)/s.value:i.value==="right"?(document.documentElement.clientWidth-b-h.value)/s.value:i.value==="top"?(b-h.value)/s.value:i.value==="bottom"?(document.documentElement.clientHeight-b-h.value)/s.value:vs();return x?Math.max(0,Math.min(1,C)):C}function S(b){if(r.value)return;const x=b.changedTouches[0].clientX,C=b.changedTouches[0].clientY,M=25,T=i.value==="left"?x<M:i.value==="right"?x>document.documentElement.clientWidth-M:i.value==="top"?C<M:i.value==="bottom"?C>document.documentElement.clientHeight-M:vs(),V=e.value&&(i.value==="left"?x<s.value:i.value==="right"?x>document.documentElement.clientWidth-s.value:i.value==="top"?C<s.value:i.value==="bottom"?C>document.documentElement.clientHeight-s.value:vs());(T||V||e.value&&n.value)&&(c=!0,g=[x,C],h.value=y(o.value?x:C,e.value),f.value=v(o.value?x:C),l(b),a(b))}function m(b){const x=b.changedTouches[0].clientX,C=b.changedTouches[0].clientY;if(c){if(!b.cancelable){c=!1;return}const T=Math.abs(x-g[0]),V=Math.abs(C-g[1]);(o.value?T>V&&T>3:V>T&&V>3)?(d.value=!0,c=!1):(o.value?V:T)>3&&(c=!1)}if(!d.value)return;b.preventDefault(),a(b);const M=v(o.value?x:C,!1);f.value=Math.max(0,Math.min(1,M)),M>1?h.value=y(o.value?x:C,!0):M<0&&(h.value=y(o.value?x:C,!1))}function p(b){if(c=!1,!d.value)return;a(b),d.value=!1;const x=u(b.changedTouches[0].identifier),C=Math.abs(x.x),M=Math.abs(x.y);(o.value?C>M&&C>400:M>C&&M>3)?e.value=x.direction===({left:"right",right:"left",top:"down",bottom:"up"}[i.value]||vs()):e.value=f.value>.5}const w=I(()=>d.value?{transform:i.value==="left"?`translateX(calc(-100% + ${f.value*s.value}px))`:i.value==="right"?`translateX(calc(100% - ${f.value*s.value}px))`:i.value==="top"?`translateY(calc(-100% + ${f.value*s.value}px))`:i.value==="bottom"?`translateY(calc(100% - ${f.value*s.value}px))`:vs(),transition:"none"}:void 0);return{isDragging:d,dragProgress:f,dragStyles:w}}function vs(){throw new Error}const vb=["start","end","left","right","top","bottom"],pb=G({color:String,disableResizeWatcher:Boolean,disableRouteWatcher:Boolean,expandOnHover:Boolean,floating:Boolean,modelValue:{type:Boolean,default:null},permanent:Boolean,rail:{type:Boolean,default:null},railWidth:{type:[Number,String],default:56},scrim:{type:[Boolean,String],default:!0},image:String,temporary:Boolean,touchless:Boolean,width:{type:[Number,String],default:256},location:{type:String,default:"start",validator:t=>vb.includes(t)},sticky:Boolean,...Ir(),...we(),...n0(),...ls(),...Hp(),...Kt(),...Je({tag:"nav"}),...dt()},"VNavigationDrawer"),yb=he()({name:"VNavigationDrawer",props:pb(),emits:{"update:modelValue":t=>!0,"update:rail":t=>!0},setup(t,e){let{attrs:n,emit:s,slots:r}=e;const{isRtl:i}=Zt(),{themeClasses:o}=It(t),{borderClasses:a}=kr(t),{backgroundColorClasses:l,backgroundColorStyles:u}=Tt(de(t,"color")),{elevationClasses:c}=us(t),{displayClasses:d,mobile:f}=bh(t),{roundedClasses:h}=Jt(t),g=nh(),y=ct(t,"modelValue",null,Q=>!!Q),{ssrBootStyles:v}=Df(),{scopeId:S}=go(),m=se(),p=be(!1),w=I(()=>t.rail&&t.expandOnHover&&p.value?Number(t.width):Number(t.rail?t.railWidth:t.width)),b=I(()=>pa(t.location,i.value)),x=I(()=>!t.permanent&&(f.value||t.temporary)),C=I(()=>t.sticky&&!x.value&&b.value!=="bottom");on(()=>t.expandOnHover&&t.rail!=null,()=>{le(p,Q=>s("update:rail",!Q))}),on(()=>!t.disableResizeWatcher,()=>{le(x,Q=>!t.permanent&&xt(()=>y.value=!Q))}),on(()=>!t.disableRouteWatcher&&!!g,()=>{le(g.currentRoute,()=>x.value&&(y.value=!1))}),le(()=>t.permanent,Q=>{Q&&(y.value=!0)}),to(()=>{t.modelValue!=null||x.value||(y.value=t.permanent||!f.value)});const{isDragging:M,dragProgress:T,dragStyles:V}=gb({isActive:y,isTemporary:x,width:w,touchless:de(t,"touchless"),position:b}),L=I(()=>{const Q=x.value?0:t.rail&&t.expandOnHover?Number(t.railWidth):w.value;return M.value?Q*T.value:Q}),{layoutItemStyles:A,layoutItemScrimStyles:k}=jp({id:t.name,order:I(()=>parseInt(t.order,10)),position:b,layoutSize:L,elementSize:w,active:I(()=>y.value||M.value),disableTransitions:I(()=>M.value),absolute:I(()=>t.absolute||C.value&&typeof O.value!="string")}),{isStuck:O,stickyStyles:H}=cb({rootEl:m,isSticky:C,layoutItemStyles:A}),q=Tt(I(()=>typeof t.scrim=="string"?t.scrim:null)),j=I(()=>({...M.value?{opacity:T.value*.2,transition:"none"}:void 0,...k.value}));hn({VList:{bgColor:"transparent"}});function Y(){p.value=!0}function ae(){p.value=!1}return ye(()=>{const Q=r.image||t.image;return _(Re,null,[_(t.tag,Le({ref:m,onMouseenter:Y,onMouseleave:ae,class:["v-navigation-drawer",`v-navigation-drawer--${b.value}`,{"v-navigation-drawer--expand-on-hover":t.expandOnHover,"v-navigation-drawer--floating":t.floating,"v-navigation-drawer--is-hovering":p.value,"v-navigation-drawer--rail":t.rail,"v-navigation-drawer--temporary":x.value,"v-navigation-drawer--active":y.value,"v-navigation-drawer--sticky":C.value},o.value,l.value,a.value,d.value,c.value,h.value,t.class],style:[u.value,A.value,V.value,v.value,H.value,t.style]},S,n),{default:()=>{var te,ge,me,Ce;return[Q&&_("div",{key:"image",class:"v-navigation-drawer__img"},[r.image?(te=r.image)==null?void 0:te.call(r,{image:t.image}):_("img",{src:t.image,alt:""},null)]),r.prepend&&_("div",{class:"v-navigation-drawer__prepend"},[(ge=r.prepend)==null?void 0:ge.call(r)]),_("div",{class:"v-navigation-drawer__content"},[(me=r.default)==null?void 0:me.call(r)]),r.append&&_("div",{class:"v-navigation-drawer__append"},[(Ce=r.append)==null?void 0:Ce.call(r)])]}}),_(qt,{name:"fade-transition"},{default:()=>[x.value&&(M.value||y.value)&&!!t.scrim&&_("div",Le({class:["v-navigation-drawer__scrim",q.backgroundColorClasses.value],style:[j.value,q.backgroundColorStyles.value],onClick:()=>y.value=!1},S),null)]})])}),{isStuck:O}}}),bb=G({...we(),...Fr({variant:"text"})},"VToolbarItems"),_c=he()({name:"VToolbarItems",props:bb(),setup(t,e){let{slots:n}=e;return hn({VBtn:{color:de(t,"color"),height:"inherit",variant:de(t,"variant")}}),ye(()=>{var s;return _("div",{class:["v-toolbar-items",t.class],style:t.style},[(s=n.default)==null?void 0:s.call(n)])}),{}}}),xb=G({id:String,text:String,...rl(pl({closeOnBack:!1,location:"end",locationStrategy:"connected",eager:!0,minWidth:0,offset:10,openOnClick:!1,openOnHover:!0,origin:"auto",scrim:!1,scrollStrategy:"reposition",transition:!1}),["absolute","persistent"])},"VTooltip"),Xe=he()({name:"VTooltip",props:xb(),emits:{"update:modelValue":t=>!0},setup(t,e){let{slots:n}=e;const s=ct(t,"modelValue"),{scopeId:r}=go(),i=mn(),o=I(()=>t.id||`v-tooltip-${i}`),a=se(),l=I(()=>t.location.split(" ").length>1?t.location:t.location+" center"),u=I(()=>t.origin==="auto"||t.origin==="overlap"||t.origin.split(" ").length>1||t.location.split(" ").length>1?t.origin:t.origin+" center"),c=I(()=>t.transition?t.transition:s.value?"scale-transition":"fade-transition"),d=I(()=>Le({"aria-describedby":o.value},t.activatorProps));return ye(()=>{const f=Oi.filterProps(t);return _(Oi,Le({ref:a,class:["v-tooltip",t.class],style:t.style,id:o.value},f,{modelValue:s.value,"onUpdate:modelValue":h=>s.value=h,transition:c.value,absolute:!0,location:l.value,origin:u.value,persistent:!0,role:"tooltip",activatorProps:d.value,_disableGlobalStack:!0},r),{activator:n.activator,default:function(){var v;for(var h=arguments.length,g=new Array(h),y=0;y<h;y++)g[y]=arguments[y];return((v=n.default)==null?void 0:v.call(n,...g))??t.text}})}),Th({},a)}});var wb="M15.41,16.58L10.83,12L15.41,7.41L14,6L8,12L14,18L15.41,16.58Z",_b="M8.59,16.58L13.17,12L8.59,7.41L10,6L16,12L10,18L8.59,16.58Z",Sb="M18 10V8H20V10H18M18 12V10H16V12H18M18 8V6H16V8H18M16 2.84V4H18C17.37 3.54 16.71 3.15 16 2.84M18 4V6H20C19.42 5.25 18.75 4.58 18 4M20 6V8H21.16C20.85 7.29 20.46 6.63 20 6M22 12C22 11.32 21.93 10.65 21.8 10H20V12H22M16 6V4H14V6H16M16 16H18V14H16V16M18 18H20L20 18V16H18V18M16 20H18L18 20V18H16V20M14 21.8C14.7 21.66 15.36 21.44 16 21.16V20H14V21.8M18 14H20V12H18V14M16 8H14V10H16V8M20 16H21.16C21.44 15.36 21.66 14.7 21.8 14H20V16M16 12H14V14H16V12M12 18V16H14V14H12V12H14V10H12V8H14V6H12V4H14V2.2C13.35 2.07 12.69 2 12 2C6.5 2 2 6.5 2 12S6.5 22 12 22V20H14V18H12M14 18H16V16H14V18Z",Ra="M19,6.41L17.59,5L12,10.59L6.41,5L5,6.41L10.59,12L5,17.59L6.41,19L12,13.41L17.59,19L19,17.59L13.41,12L19,6.41Z",Cb="M12,8A4,4 0 0,1 16,12A4,4 0 0,1 12,16A4,4 0 0,1 8,12A4,4 0 0,1 12,8M3.05,13H1V11H3.05C3.5,6.83 6.83,3.5 11,3.05V1H13V3.05C17.17,3.5 20.5,6.83 20.95,11H23V13H20.95C20.5,17.17 17.17,20.5 13,20.95V23H11V20.95C6.83,20.5 3.5,17.17 3.05,13M12,5A7,7 0 0,0 5,12A7,7 0 0,0 12,19A7,7 0 0,0 19,12A7,7 0 0,0 12,5Z",$o="M21,16.5C21,16.88 20.79,17.21 20.47,17.38L12.57,21.82C12.41,21.94 12.21,22 12,22C11.79,22 11.59,21.94 11.43,21.82L3.53,17.38C3.21,17.21 3,16.88 3,16.5V7.5C3,7.12 3.21,6.79 3.53,6.62L11.43,2.18C11.59,2.06 11.79,2 12,2C12.21,2 12.41,2.06 12.57,2.18L20.47,6.62C20.79,6.79 21,7.12 21,7.5V16.5M12,4.15L6.04,7.5L12,10.85L17.96,7.5L12,4.15Z",Rw="M21,16.5C21,16.88 20.79,17.21 20.47,17.38L12.57,21.82C12.41,21.94 12.21,22 12,22C11.79,22 11.59,21.94 11.43,21.82L3.53,17.38C3.21,17.21 3,16.88 3,16.5V7.5C3,7.12 3.21,6.79 3.53,6.62L11.43,2.18C11.59,2.06 11.79,2 12,2C12.21,2 12.41,2.06 12.57,2.18L20.47,6.62C20.79,6.79 21,7.12 21,7.5V16.5M12,4.15L6.04,7.5L12,10.85L17.96,7.5L12,4.15M5,15.91L11,19.29V12.58L5,9.21V15.91M19,15.91V9.21L13,12.58V19.29L19,15.91Z",Iw="M10.76,8.69A0.76,0.76 0 0,0 10,9.45V20.9C10,21.32 10.34,21.66 10.76,21.66C10.95,21.66 11.11,21.6 11.24,21.5L13.15,19.95L14.81,23.57C14.94,23.84 15.21,24 15.5,24C15.61,24 15.72,24 15.83,23.92L18.59,22.64C18.97,22.46 19.15,22 18.95,21.63L17.28,18L19.69,17.55C19.85,17.5 20,17.43 20.12,17.29C20.39,16.97 20.35,16.5 20,16.21L11.26,8.86L11.25,8.87C11.12,8.76 10.95,8.69 10.76,8.69M15,10V8H20V10H15M13.83,4.76L16.66,1.93L18.07,3.34L15.24,6.17L13.83,4.76M10,0H12V5H10V0M3.93,14.66L6.76,11.83L8.17,13.24L5.34,16.07L3.93,14.66M3.93,3.34L5.34,1.93L8.17,4.76L6.76,6.17L3.93,3.34M7,10H2V8H7V10",Eb="M19,4H15.5L14.5,3H9.5L8.5,4H5V6H19M6,19A2,2 0 0,0 8,21H16A2,2 0 0,0 18,19V7H6V19Z",Mb="M5,20H19V18H5M19,9H15V3H9V9H5L12,16L19,9Z",kw="M20,20A2,2 0 0,1 18,22H4A2,2 0 0,1 2,20V6A2,2 0 0,1 4,4H9.5C9.2,4.8 9,5.6 9,6.5C9,10.1 11.9,13 15.5,13C16.3,13 17,12.9 17.6,12.6L20,15V20M19.3,8.9C19.7,8.2 20,7.4 20,6.5C20,4 18,2 15.5,2C13,2 11,4 11,6.5C11,9 13,11 15.5,11C16.4,11 17.2,10.7 17.9,10.3L21,13.4L22.4,12L19.3,8.9M15.5,9C14.1,9 13,7.9 13,6.5C13,5.1 14.1,4 15.5,4C16.9,4 18,5.1 18,6.5C18,7.9 16.9,9 15.5,9Z",Tb="M12,2A10,10 0 0,0 2,12C2,16.42 4.87,20.17 8.84,21.5C9.34,21.58 9.5,21.27 9.5,21C9.5,20.77 9.5,20.14 9.5,19.31C6.73,19.91 6.14,17.97 6.14,17.97C5.68,16.81 5.03,16.5 5.03,16.5C4.12,15.88 5.1,15.9 5.1,15.9C6.1,15.97 6.63,16.93 6.63,16.93C7.5,18.45 8.97,18 9.54,17.76C9.63,17.11 9.89,16.67 10.17,16.42C7.95,16.17 5.62,15.31 5.62,11.5C5.62,10.39 6,9.5 6.65,8.79C6.55,8.54 6.2,7.5 6.75,6.15C6.75,6.15 7.59,5.88 9.5,7.17C10.29,6.95 11.15,6.84 12,6.84C12.85,6.84 13.71,6.95 14.5,7.17C16.41,5.88 17.25,6.15 17.25,6.15C17.8,7.5 17.45,8.54 17.35,8.79C18,9.5 18.38,10.39 18.38,11.5C18.38,15.32 16.04,16.16 13.81,16.41C14.17,16.72 14.5,17.33 14.5,18.26C14.5,19.6 14.5,20.68 14.5,21C14.5,21.27 14.66,21.59 15.17,21.5C19.14,20.16 22,16.42 22,12A10,10 0 0,0 12,2Z",Ab="M9 10A3.04 3.04 0 0 1 12 7A3.04 3.04 0 0 1 15 10A3.04 3.04 0 0 1 12 13A3.04 3.04 0 0 1 9 10M12 19L16 20V16.92A7.54 7.54 0 0 1 12 18A7.54 7.54 0 0 1 8 16.92V20M12 4A5.78 5.78 0 0 0 7.76 5.74A5.78 5.78 0 0 0 6 10A5.78 5.78 0 0 0 7.76 14.23A5.78 5.78 0 0 0 12 16A5.78 5.78 0 0 0 16.24 14.23A5.78 5.78 0 0 0 18 10A5.78 5.78 0 0 0 16.24 5.74A5.78 5.78 0 0 0 12 4M20 10A8.04 8.04 0 0 1 19.43 12.8A7.84 7.84 0 0 1 18 15.28V23L12 21L6 23V15.28A7.9 7.9 0 0 1 4 10A7.68 7.68 0 0 1 6.33 4.36A7.73 7.73 0 0 1 12 2A7.73 7.73 0 0 1 17.67 4.36A7.68 7.68 0 0 1 20 10Z",Rb="M19,13H13V19H11V13H5V11H11V5H13V11H19V13Z",Ib="M16,6C14.87,6 13.77,6.35 12.84,7H4C2.89,7 2,7.89 2,9V15C2,16.11 2.89,17 4,17H5V18A1,1 0 0,0 6,19H8A1,1 0 0,0 9,18V17H15V18A1,1 0 0,0 16,19H18A1,1 0 0,0 19,18V17H20C21.11,17 22,16.11 22,15V9C22,7.89 21.11,7 20,7H19.15C18.23,6.35 17.13,6 16,6M16,7.5A3.5,3.5 0 0,1 19.5,11A3.5,3.5 0 0,1 16,14.5A3.5,3.5 0 0,1 12.5,11A3.5,3.5 0 0,1 16,7.5M4,9H8V10H4V9M16,9A2,2 0 0,0 14,11A2,2 0 0,0 16,13A2,2 0 0,0 18,11A2,2 0 0,0 16,9M4,11H8V12H4V11M4,13H8V14H4V13Z",kb="M4,6V19H20V6H4Z",Fb="M4,6V19H20V6H4M18,17H6V8H18V17Z",Fw="M1.39,18.36L3.16,16.6L4.58,18L5.64,16.95L4.22,15.54L5.64,14.12L8.11,16.6L9.17,15.54L6.7,13.06L8.11,11.65L9.53,13.06L10.59,12L9.17,10.59L10.59,9.17L13.06,11.65L14.12,10.59L11.65,8.11L13.06,6.7L14.47,8.11L15.54,7.05L14.12,5.64L15.54,4.22L18,6.7L19.07,5.64L16.6,3.16L18.36,1.39L22.61,5.64L5.64,22.61L1.39,18.36Z",Ho="M21,9L17,5V8H10V10H17V13M7,11L3,15L7,19V16H14V14H7V11Z",Vb="M2,4H8V6H16V4H22V10H20V14H22V20H16V18H8V20H2V14H4V10H2V4M16,10V8H8V10H6V14H8V16H16V14H18V10H16M4,6V8H6V6H4M18,6V8H20V6H18M4,16V18H6V16H4M18,16V18H20V16H18Z";const wl=(t,e)=>{const n=t.__vccOpts||t;for(const[s,r]of e)n[s]=r;return n},Sc={mdi:{size:24,viewbox:"0 0 24 24"},"simple-icons":{size:24,viewbox:"0 0 24 24"},default:{size:0,viewbox:"0 0 0 0"}},Lb={name:"icon",props:{type:String,path:{type:String,required:!0},size:{type:[String,Number],default:24},viewbox:String,flip:{type:String,validator:t=>["horizontal","vertical","both","none"].includes(t)},rotate:{type:Number,default:0}},computed:{styles(){return{"--sx":["both","horizontal"].includes(this.flip)?"-1":"1","--sy":["both","vertical"].includes(this.flip)?"-1":"1","--r":isNaN(this.rotate)?this.rotate:this.rotate+"deg"}},defaults(){return Sc[this.type]||Sc.default},sizeValue(){return this.size||this.defaults.size},viewboxValue(){return this.viewbox||this.defaults.viewbox}}},Pb=["width","height","viewBox"],Ob=["d"];function Nb(t,e,n,s,r,i){return at(),Mr("svg",{width:i.sizeValue,height:i.sizeValue,viewBox:i.viewboxValue,style:Yi(i.styles)},[jt("path",{d:n.path},null,8,Ob)],12,Pb)}const Be=wl(Lb,[["render",Nb],["__scopeId","data-v-76aa6b74"]]),Bb=dn({__name:"Sidebar",props:{openedInit:{},side:{},width:{}},setup(t){const e=t;let n=se(e.openedInit.valueOf());const s=e.side==="left"?_b:wb;return(r,i)=>(at(),Mr(Re,null,[_(F(it),{icon:"",onClick:i[0]||(i[0]=o=>Te(n)?n.value=!F(n):n=!F(n)),class:Xi(["open-button",r.side])},{default:X(()=>[_(F(Be),{type:"mdi",path:F(s)},null,8,["path"])]),_:1},8,["class"]),_(F(yb),{modelValue:F(n),"onUpdate:modelValue":i[3]||(i[3]=o=>Te(n)?n.value=o:n=o),permanent:"",location:r.side,width:e.width},{default:X(()=>[_(F(zf),{density:"compact"},{default:X(()=>[r.side=="right"?(at(),ln(F(_c),{key:0},{default:X(()=>[Gs(r.$slots,"toolbar-items",{},void 0,!0),_(F(it),{icon:"",onClick:i[1]||(i[1]=o=>Te(n)?n.value=!F(n):n=!F(n))},{default:X(()=>[_(F(Be),{type:"mdi",path:F(Ra)},null,8,["path"])]),_:1})]),_:3})):da("",!0),Gs(r.$slots,"toolbar",{},void 0,!0),r.side=="left"?(at(),ln(F(_c),{key:1},{default:X(()=>[Gs(r.$slots,"toolbar-items",{},void 0,!0),_(F(it),{icon:"",onClick:i[2]||(i[2]=o=>Te(n)?n.value=!F(n):n=!F(n))},{default:X(()=>[_(F(Be),{type:"mdi",path:F(Ra)},null,8,["path"])]),_:1})]),_:3})):da("",!0)]),_:3}),Gs(r.$slots,"default",{},void 0,!0)]),_:3},8,["modelValue","location","width"])],64))}}),Cc=wl(Bb,[["__scopeId","data-v-7849b5fc"]]),Wh=dn({__name:"Loading",setup(t){return(e,n)=>(at(),ln(F(H1),null,{default:X(()=>[_(F(Q1),{justify:"center",style:{height:"100%"}},{default:X(()=>[_(F(G1),{"align-self":"center"},{default:X(()=>[_(F(Kf),{indeterminate:"",style:{display:"block",margin:"0 auto"}}),Gs(e.$slots,"default")]),_:3})]),_:3})]),_:3}))}});function zb(t){return t&&t.__esModule&&Object.prototype.hasOwnProperty.call(t,"default")?t.default:t}class Gh extends HTMLElement{constructor(e,n){super(),this.camera=e,this.options=Object.assign({size:90,padding:8,bubbleSizePrimary:8,bubbleSizeSeconday:6,showSecondary:!0,lineWidth:2,fontSize:"11px",fontFamily:"arial",fontWeight:"bold",fontColor:"#151515",fontYAdjust:0,colors:{x:["#f73c3c","#942424"],y:["#6ccb26","#417a17"],z:["#178cf0","#0e5490"]}},n),this.onAxisSelected=null,this.bubbles=[{axis:"x",direction:new THREE.Vector3(1,0,0),size:this.options.bubbleSizePrimary,color:this.options.colors.x,line:this.options.lineWidth,label:"X"},{axis:"y",direction:new THREE.Vector3(0,1,0),size:this.options.bubbleSizePrimary,color:this.options.colors.y,line:this.options.lineWidth,label:"Y"},{axis:"z",direction:new THREE.Vector3(0,0,1),size:this.options.bubbleSizePrimary,color:this.options.colors.z,line:this.options.lineWidth,label:"Z"},{axis:"-x",direction:new THREE.Vector3(-1,0,0),size:this.options.bubbleSizeSeconday,color:this.options.colors.x},{axis:"-y",direction:new THREE.Vector3(0,-1,0),size:this.options.bubbleSizeSeconday,color:this.options.colors.y},{axis:"-z",direction:new THREE.Vector3(0,0,-1),size:this.options.bubbleSizeSeconday,color:this.options.colors.z}],this.center=new THREE.Vector3(this.options.size/2,this.options.size/2,0),this.selectedAxis=null,this.innerHTML="<canvas width='"+this.options.size+"' height='"+this.options.size+"'></canvas>",this.onMouseMove=this.onMouseMove.bind(this),this.onMouseOut=this.onMouseOut.bind(this),this.onMouseClick=this.onMouseClick.bind(this)}connectedCallback(){this.canvas=this.querySelector("canvas"),this.context=this.canvas.getContext("2d"),this.canvas.addEventListener("mousemove",this.onMouseMove,!1),this.canvas.addEventListener("mouseout",this.onMouseOut,!1),this.canvas.addEventListener("click",this.onMouseClick,!1)}disconnectedCallback(){this.canvas.removeEventListener("mousemove",this.onMouseMove,!1),this.canvas.removeEventListener("mouseout",this.onMouseOut,!1),this.canvas.removeEventListener("click",this.onMouseClick,!1)}onMouseMove(e){const n=this.canvas.getBoundingClientRect();this.mouse=new THREE.Vector3(e.clientX-n.left,e.clientY-n.top,0)}onMouseOut(e){this.mouse=null}onMouseClick(e){this.onAxisSelected&&typeof this.onAxisSelected=="function"&&this.onAxisSelected({axis:this.selectedAxis.axis,direction:this.selectedAxis.direction.clone()})}clear(){this.context.clearRect(0,0,this.canvas.width,this.canvas.height)}drawCircle(e,n=10,s="#FF0000"){this.context.beginPath(),this.context.arc(e.x,e.y,n,0,2*Math.PI,!1),this.context.fillStyle=s,this.context.fill(),this.context.closePath()}drawLine(e,n,s=1,r="#FF0000"){this.context.beginPath(),this.context.moveTo(e.x,e.y),this.context.lineTo(n.x,n.y),this.context.lineWidth=s,this.context.strokeStyle=r,this.context.stroke(),this.context.closePath()}update(){this.clear();let n=new THREE.Matrix4().makeRotationFromEuler(this.camera.rotation).clone().invert();for(var s of this.bubbles)s.position=this.getBubblePosition(s.direction.clone().applyMatrix4(n));const r=[];for(let i in this.bubbles)(this.options.showSecondary===!0||i[0]!=="-")&&r.push(this.bubbles[i]);if(r.sort((i,o)=>i.position.z>o.position.z?1:-1),this.selectedAxis=null,this.mouse){let i=1/0;for(var s of r){const a=this.mouse.distanceTo(s.position);(a<i||a<s.size)&&(i=a,this.selectedAxis=s)}}this.drawLayers(r)}drawLayers(e){for(let n of e){let s=n.color;this.selectedAxis===n?s="#FFFFFF":n.position.z>=-.01?s=n.color[0]:s=n.color[1],this.drawCircle(n.position,n.size,s),n.line&&this.drawLine(this.center,n.position,n.line,s),n.label&&(this.context.font=[this.options.fontWeight,this.options.fontSize,this.options.fontFamily].join(" "),this.context.fillStyle=this.options.fontColor,this.context.textBaseline="middle",this.context.textAlign="center",this.context.fillText(n.label,n.position.x,n.position.y+this.options.fontYAdjust))}}getBubblePosition(e){return new THREE.Vector3(e.x*(this.center.x-this.options.bubbleSizePrimary/2-this.options.padding)+this.center.x,this.center.y-e.y*(this.center.y-this.options.bubbleSizePrimary/2-this.options.padding),e.z)}}window.customElements.define("orientation-gizmo",Gh);var Db=Gh;const $b=zb(Db),qe=["00","01","02","03","04","05","06","07","08","09","0a","0b","0c","0d","0e","0f","10","11","12","13","14","15","16","17","18","19","1a","1b","1c","1d","1e","1f","20","21","22","23","24","25","26","27","28","29","2a","2b","2c","2d","2e","2f","30","31","32","33","34","35","36","37","38","39","3a","3b","3c","3d","3e","3f","40","41","42","43","44","45","46","47","48","49","4a","4b","4c","4d","4e","4f","50","51","52","53","54","55","56","57","58","59","5a","5b","5c","5d","5e","5f","60","61","62","63","64","65","66","67","68","69","6a","6b","6c","6d","6e","6f","70","71","72","73","74","75","76","77","78","79","7a","7b","7c","7d","7e","7f","80","81","82","83","84","85","86","87","88","89","8a","8b","8c","8d","8e","8f","90","91","92","93","94","95","96","97","98","99","9a","9b","9c","9d","9e","9f","a0","a1","a2","a3","a4","a5","a6","a7","a8","a9","aa","ab","ac","ad","ae","af","b0","b1","b2","b3","b4","b5","b6","b7","b8","b9","ba","bb","bc","bd","be","bf","c0","c1","c2","c3","c4","c5","c6","c7","c8","c9","ca","cb","cc","cd","ce","cf","d0","d1","d2","d3","d4","d5","d6","d7","d8","d9","da","db","dc","dd","de","df","e0","e1","e2","e3","e4","e5","e6","e7","e8","e9","ea","eb","ec","ed","ee","ef","f0","f1","f2","f3","f4","f5","f6","f7","f8","f9","fa","fb","fc","fd","fe","ff"];function Hb(){const t=Math.random()*4294967295|0,e=Math.random()*4294967295|0,n=Math.random()*4294967295|0,s=Math.random()*4294967295|0;return(qe[t&255]+qe[t>>8&255]+qe[t>>16&255]+qe[t>>24&255]+"-"+qe[e&255]+qe[e>>8&255]+"-"+qe[e>>16&15|64]+qe[e>>24&255]+"-"+qe[n&63|128]+qe[n>>8&255]+"-"+qe[n>>16&255]+qe[n>>24&255]+qe[s&255]+qe[s>>8&255]+qe[s>>16&255]+qe[s>>24&255]).toLowerCase()}function ot(t,e,n){return Math.max(e,Math.min(n,t))}function Ub(t,e){return(t%e+e)%e}function Uo(t,e,n){return(1-n)*t+n*e}class Vr{constructor(e=0,n=0,s=0,r=1){this.isQuaternion=!0,this._x=e,this._y=n,this._z=s,this._w=r}static slerpFlat(e,n,s,r,i,o,a){let l=s[r+0],u=s[r+1],c=s[r+2],d=s[r+3];const f=i[o+0],h=i[o+1],g=i[o+2],y=i[o+3];if(a===0){e[n+0]=l,e[n+1]=u,e[n+2]=c,e[n+3]=d;return}if(a===1){e[n+0]=f,e[n+1]=h,e[n+2]=g,e[n+3]=y;return}if(d!==y||l!==f||u!==h||c!==g){let v=1-a;const S=l*f+u*h+c*g+d*y,m=S>=0?1:-1,p=1-S*S;if(p>Number.EPSILON){const b=Math.sqrt(p),x=Math.atan2(b,S*m);v=Math.sin(v*x)/b,a=Math.sin(a*x)/b}const w=a*m;if(l=l*v+f*w,u=u*v+h*w,c=c*v+g*w,d=d*v+y*w,v===1-a){const b=1/Math.sqrt(l*l+u*u+c*c+d*d);l*=b,u*=b,c*=b,d*=b}}e[n]=l,e[n+1]=u,e[n+2]=c,e[n+3]=d}static multiplyQuaternionsFlat(e,n,s,r,i,o){const a=s[r],l=s[r+1],u=s[r+2],c=s[r+3],d=i[o],f=i[o+1],h=i[o+2],g=i[o+3];return e[n]=a*g+c*d+l*h-u*f,e[n+1]=l*g+c*f+u*d-a*h,e[n+2]=u*g+c*h+a*f-l*d,e[n+3]=c*g-a*d-l*f-u*h,e}get x(){return this._x}set x(e){this._x=e,this._onChangeCallback()}get y(){return this._y}set y(e){this._y=e,this._onChangeCallback()}get z(){return this._z}set z(e){this._z=e,this._onChangeCallback()}get w(){return this._w}set w(e){this._w=e,this._onChangeCallback()}set(e,n,s,r){return this._x=e,this._y=n,this._z=s,this._w=r,this._onChangeCallback(),this}clone(){return new this.constructor(this._x,this._y,this._z,this._w)}copy(e){return this._x=e.x,this._y=e.y,this._z=e.z,this._w=e.w,this._onChangeCallback(),this}setFromEuler(e,n=!0){const s=e._x,r=e._y,i=e._z,o=e._order,a=Math.cos,l=Math.sin,u=a(s/2),c=a(r/2),d=a(i/2),f=l(s/2),h=l(r/2),g=l(i/2);switch(o){case"XYZ":this._x=f*c*d+u*h*g,this._y=u*h*d-f*c*g,this._z=u*c*g+f*h*d,this._w=u*c*d-f*h*g;break;case"YXZ":this._x=f*c*d+u*h*g,this._y=u*h*d-f*c*g,this._z=u*c*g-f*h*d,this._w=u*c*d+f*h*g;break;case"ZXY":this._x=f*c*d-u*h*g,this._y=u*h*d+f*c*g,this._z=u*c*g+f*h*d,this._w=u*c*d-f*h*g;break;case"ZYX":this._x=f*c*d-u*h*g,this._y=u*h*d+f*c*g,this._z=u*c*g-f*h*d,this._w=u*c*d+f*h*g;break;case"YZX":this._x=f*c*d+u*h*g,this._y=u*h*d+f*c*g,this._z=u*c*g-f*h*d,this._w=u*c*d-f*h*g;break;case"XZY":this._x=f*c*d-u*h*g,this._y=u*h*d-f*c*g,this._z=u*c*g+f*h*d,this._w=u*c*d+f*h*g;break;default:console.warn("THREE.Quaternion: .setFromEuler() encountered an unknown order: "+o)}return n===!0&&this._onChangeCallback(),this}setFromAxisAngle(e,n){const s=n/2,r=Math.sin(s);return this._x=e.x*r,this._y=e.y*r,this._z=e.z*r,this._w=Math.cos(s),this._onChangeCallback(),this}setFromRotationMatrix(e){const n=e.elements,s=n[0],r=n[4],i=n[8],o=n[1],a=n[5],l=n[9],u=n[2],c=n[6],d=n[10],f=s+a+d;if(f>0){const h=.5/Math.sqrt(f+1);this._w=.25/h,this._x=(c-l)*h,this._y=(i-u)*h,this._z=(o-r)*h}else if(s>a&&s>d){const h=2*Math.sqrt(1+s-a-d);this._w=(c-l)/h,this._x=.25*h,this._y=(r+o)/h,this._z=(i+u)/h}else if(a>d){const h=2*Math.sqrt(1+a-s-d);this._w=(i-u)/h,this._x=(r+o)/h,this._y=.25*h,this._z=(l+c)/h}else{const h=2*Math.sqrt(1+d-s-a);this._w=(o-r)/h,this._x=(i+u)/h,this._y=(l+c)/h,this._z=.25*h}return this._onChangeCallback(),this}setFromUnitVectors(e,n){let s=e.dot(n)+1;return s<Number.EPSILON?(s=0,Math.abs(e.x)>Math.abs(e.z)?(this._x=-e.y,this._y=e.x,this._z=0,this._w=s):(this._x=0,this._y=-e.z,this._z=e.y,this._w=s)):(this._x=e.y*n.z-e.z*n.y,this._y=e.z*n.x-e.x*n.z,this._z=e.x*n.y-e.y*n.x,this._w=s),this.normalize()}angleTo(e){return 2*Math.acos(Math.abs(ot(this.dot(e),-1,1)))}rotateTowards(e,n){const s=this.angleTo(e);if(s===0)return this;const r=Math.min(1,n/s);return this.slerp(e,r),this}identity(){return this.set(0,0,0,1)}invert(){return this.conjugate()}conjugate(){return this._x*=-1,this._y*=-1,this._z*=-1,this._onChangeCallback(),this}dot(e){return this._x*e._x+this._y*e._y+this._z*e._z+this._w*e._w}lengthSq(){return this._x*this._x+this._y*this._y+this._z*this._z+this._w*this._w}length(){return Math.sqrt(this._x*this._x+this._y*this._y+this._z*this._z+this._w*this._w)}normalize(){let e=this.length();return e===0?(this._x=0,this._y=0,this._z=0,this._w=1):(e=1/e,this._x=this._x*e,this._y=this._y*e,this._z=this._z*e,this._w=this._w*e),this._onChangeCallback(),this}multiply(e){return this.multiplyQuaternions(this,e)}premultiply(e){return this.multiplyQuaternions(e,this)}multiplyQuaternions(e,n){const s=e._x,r=e._y,i=e._z,o=e._w,a=n._x,l=n._y,u=n._z,c=n._w;return this._x=s*c+o*a+r*u-i*l,this._y=r*c+o*l+i*a-s*u,this._z=i*c+o*u+s*l-r*a,this._w=o*c-s*a-r*l-i*u,this._onChangeCallback(),this}slerp(e,n){if(n===0)return this;if(n===1)return this.copy(e);const s=this._x,r=this._y,i=this._z,o=this._w;let a=o*e._w+s*e._x+r*e._y+i*e._z;if(a<0?(this._w=-e._w,this._x=-e._x,this._y=-e._y,this._z=-e._z,a=-a):this.copy(e),a>=1)return this._w=o,this._x=s,this._y=r,this._z=i,this;const l=1-a*a;if(l<=Number.EPSILON){const h=1-n;return this._w=h*o+n*this._w,this._x=h*s+n*this._x,this._y=h*r+n*this._y,this._z=h*i+n*this._z,this.normalize(),this}const u=Math.sqrt(l),c=Math.atan2(u,a),d=Math.sin((1-n)*c)/u,f=Math.sin(n*c)/u;return this._w=o*d+this._w*f,this._x=s*d+this._x*f,this._y=r*d+this._y*f,this._z=i*d+this._z*f,this._onChangeCallback(),this}slerpQuaternions(e,n,s){return this.copy(e).slerp(n,s)}random(){const e=Math.random(),n=Math.sqrt(1-e),s=Math.sqrt(e),r=2*Math.PI*Math.random(),i=2*Math.PI*Math.random();return this.set(n*Math.cos(r),s*Math.sin(i),s*Math.cos(i),n*Math.sin(r))}equals(e){return e._x===this._x&&e._y===this._y&&e._z===this._z&&e._w===this._w}fromArray(e,n=0){return this._x=e[n],this._y=e[n+1],this._z=e[n+2],this._w=e[n+3],this._onChangeCallback(),this}toArray(e=[],n=0){return e[n]=this._x,e[n+1]=this._y,e[n+2]=this._z,e[n+3]=this._w,e}fromBufferAttribute(e,n){return this._x=e.getX(n),this._y=e.getY(n),this._z=e.getZ(n),this._w=e.getW(n),this._onChangeCallback(),this}toJSON(){return this.toArray()}_onChange(e){return this._onChangeCallback=e,this}_onChangeCallback(){}*[Symbol.iterator](){yield this._x,yield this._y,yield this._z,yield this._w}}class J{constructor(e=0,n=0,s=0){J.prototype.isVector3=!0,this.x=e,this.y=n,this.z=s}set(e,n,s){return s===void 0&&(s=this.z),this.x=e,this.y=n,this.z=s,this}setScalar(e){return this.x=e,this.y=e,this.z=e,this}setX(e){return this.x=e,this}setY(e){return this.y=e,this}setZ(e){return this.z=e,this}setComponent(e,n){switch(e){case 0:this.x=n;break;case 1:this.y=n;break;case 2:this.z=n;break;default:throw new Error("index is out of range: "+e)}return this}getComponent(e){switch(e){case 0:return this.x;case 1:return this.y;case 2:return this.z;default:throw new Error("index is out of range: "+e)}}clone(){return new this.constructor(this.x,this.y,this.z)}copy(e){return this.x=e.x,this.y=e.y,this.z=e.z,this}add(e){return this.x+=e.x,this.y+=e.y,this.z+=e.z,this}addScalar(e){return this.x+=e,this.y+=e,this.z+=e,this}addVectors(e,n){return this.x=e.x+n.x,this.y=e.y+n.y,this.z=e.z+n.z,this}addScaledVector(e,n){return this.x+=e.x*n,this.y+=e.y*n,this.z+=e.z*n,this}sub(e){return this.x-=e.x,this.y-=e.y,this.z-=e.z,this}subScalar(e){return this.x-=e,this.y-=e,this.z-=e,this}subVectors(e,n){return this.x=e.x-n.x,this.y=e.y-n.y,this.z=e.z-n.z,this}multiply(e){return this.x*=e.x,this.y*=e.y,this.z*=e.z,this}multiplyScalar(e){return this.x*=e,this.y*=e,this.z*=e,this}multiplyVectors(e,n){return this.x=e.x*n.x,this.y=e.y*n.y,this.z=e.z*n.z,this}applyEuler(e){return this.applyQuaternion(Ec.setFromEuler(e))}applyAxisAngle(e,n){return this.applyQuaternion(Ec.setFromAxisAngle(e,n))}applyMatrix3(e){const n=this.x,s=this.y,r=this.z,i=e.elements;return this.x=i[0]*n+i[3]*s+i[6]*r,this.y=i[1]*n+i[4]*s+i[7]*r,this.z=i[2]*n+i[5]*s+i[8]*r,this}applyNormalMatrix(e){return this.applyMatrix3(e).normalize()}applyMatrix4(e){const n=this.x,s=this.y,r=this.z,i=e.elements,o=1/(i[3]*n+i[7]*s+i[11]*r+i[15]);return this.x=(i[0]*n+i[4]*s+i[8]*r+i[12])*o,this.y=(i[1]*n+i[5]*s+i[9]*r+i[13])*o,this.z=(i[2]*n+i[6]*s+i[10]*r+i[14])*o,this}applyQuaternion(e){const n=this.x,s=this.y,r=this.z,i=e.x,o=e.y,a=e.z,l=e.w,u=2*(o*r-a*s),c=2*(a*n-i*r),d=2*(i*s-o*n);return this.x=n+l*u+o*d-a*c,this.y=s+l*c+a*u-i*d,this.z=r+l*d+i*c-o*u,this}project(e){return this.applyMatrix4(e.matrixWorldInverse).applyMatrix4(e.projectionMatrix)}unproject(e){return this.applyMatrix4(e.projectionMatrixInverse).applyMatrix4(e.matrixWorld)}transformDirection(e){const n=this.x,s=this.y,r=this.z,i=e.elements;return this.x=i[0]*n+i[4]*s+i[8]*r,this.y=i[1]*n+i[5]*s+i[9]*r,this.z=i[2]*n+i[6]*s+i[10]*r,this.normalize()}divide(e){return this.x/=e.x,this.y/=e.y,this.z/=e.z,this}divideScalar(e){return this.multiplyScalar(1/e)}min(e){return this.x=Math.min(this.x,e.x),this.y=Math.min(this.y,e.y),this.z=Math.min(this.z,e.z),this}max(e){return this.x=Math.max(this.x,e.x),this.y=Math.max(this.y,e.y),this.z=Math.max(this.z,e.z),this}clamp(e,n){return this.x=Math.max(e.x,Math.min(n.x,this.x)),this.y=Math.max(e.y,Math.min(n.y,this.y)),this.z=Math.max(e.z,Math.min(n.z,this.z)),this}clampScalar(e,n){return this.x=Math.max(e,Math.min(n,this.x)),this.y=Math.max(e,Math.min(n,this.y)),this.z=Math.max(e,Math.min(n,this.z)),this}clampLength(e,n){const s=this.length();return this.divideScalar(s||1).multiplyScalar(Math.max(e,Math.min(n,s)))}floor(){return this.x=Math.floor(this.x),this.y=Math.floor(this.y),this.z=Math.floor(this.z),this}ceil(){return this.x=Math.ceil(this.x),this.y=Math.ceil(this.y),this.z=Math.ceil(this.z),this}round(){return this.x=Math.round(this.x),this.y=Math.round(this.y),this.z=Math.round(this.z),this}roundToZero(){return this.x=Math.trunc(this.x),this.y=Math.trunc(this.y),this.z=Math.trunc(this.z),this}negate(){return this.x=-this.x,this.y=-this.y,this.z=-this.z,this}dot(e){return this.x*e.x+this.y*e.y+this.z*e.z}lengthSq(){return this.x*this.x+this.y*this.y+this.z*this.z}length(){return Math.sqrt(this.x*this.x+this.y*this.y+this.z*this.z)}manhattanLength(){return Math.abs(this.x)+Math.abs(this.y)+Math.abs(this.z)}normalize(){return this.divideScalar(this.length()||1)}setLength(e){return this.normalize().multiplyScalar(e)}lerp(e,n){return this.x+=(e.x-this.x)*n,this.y+=(e.y-this.y)*n,this.z+=(e.z-this.z)*n,this}lerpVectors(e,n,s){return this.x=e.x+(n.x-e.x)*s,this.y=e.y+(n.y-e.y)*s,this.z=e.z+(n.z-e.z)*s,this}cross(e){return this.crossVectors(this,e)}crossVectors(e,n){const s=e.x,r=e.y,i=e.z,o=n.x,a=n.y,l=n.z;return this.x=r*l-i*a,this.y=i*o-s*l,this.z=s*a-r*o,this}projectOnVector(e){const n=e.lengthSq();if(n===0)return this.set(0,0,0);const s=e.dot(this)/n;return this.copy(e).multiplyScalar(s)}projectOnPlane(e){return jo.copy(this).projectOnVector(e),this.sub(jo)}reflect(e){return this.sub(jo.copy(e).multiplyScalar(2*this.dot(e)))}angleTo(e){const n=Math.sqrt(this.lengthSq()*e.lengthSq());if(n===0)return Math.PI/2;const s=this.dot(e)/n;return Math.acos(ot(s,-1,1))}distanceTo(e){return Math.sqrt(this.distanceToSquared(e))}distanceToSquared(e){const n=this.x-e.x,s=this.y-e.y,r=this.z-e.z;return n*n+s*s+r*r}manhattanDistanceTo(e){return Math.abs(this.x-e.x)+Math.abs(this.y-e.y)+Math.abs(this.z-e.z)}setFromSpherical(e){return this.setFromSphericalCoords(e.radius,e.phi,e.theta)}setFromSphericalCoords(e,n,s){const r=Math.sin(n)*e;return this.x=r*Math.sin(s),this.y=Math.cos(n)*e,this.z=r*Math.cos(s),this}setFromCylindrical(e){return this.setFromCylindricalCoords(e.radius,e.theta,e.y)}setFromCylindricalCoords(e,n,s){return this.x=e*Math.sin(n),this.y=s,this.z=e*Math.cos(n),this}setFromMatrixPosition(e){const n=e.elements;return this.x=n[12],this.y=n[13],this.z=n[14],this}setFromMatrixScale(e){const n=this.setFromMatrixColumn(e,0).length(),s=this.setFromMatrixColumn(e,1).length(),r=this.setFromMatrixColumn(e,2).length();return this.x=n,this.y=s,this.z=r,this}setFromMatrixColumn(e,n){return this.fromArray(e.elements,n*4)}setFromMatrix3Column(e,n){return this.fromArray(e.elements,n*3)}setFromEuler(e){return this.x=e._x,this.y=e._y,this.z=e._z,this}setFromColor(e){return this.x=e.r,this.y=e.g,this.z=e.b,this}equals(e){return e.x===this.x&&e.y===this.y&&e.z===this.z}fromArray(e,n=0){return this.x=e[n],this.y=e[n+1],this.z=e[n+2],this}toArray(e=[],n=0){return e[n]=this.x,e[n+1]=this.y,e[n+2]=this.z,e}fromBufferAttribute(e,n){return this.x=e.getX(n),this.y=e.getY(n),this.z=e.getZ(n),this}random(){return this.x=Math.random(),this.y=Math.random(),this.z=Math.random(),this}randomDirection(){const e=(Math.random()-.5)*2,n=Math.random()*Math.PI*2,s=Math.sqrt(1-e**2);return this.x=s*Math.cos(n),this.y=s*Math.sin(n),this.z=e,this}*[Symbol.iterator](){yield this.x,yield this.y,yield this.z}}const jo=new J,Ec=new Vr,jb=0,Wb=1,Gb="",Sn="srgb",_l="srgb-linear",qb="display-p3",qh="display-p3-linear",Ia="linear",Mc="srgb",Tc="rec709",Ac="p3",Xs=2e3,Rc=2001;class He{constructor(e,n,s,r,i,o,a,l,u,c,d,f,h,g,y,v){He.prototype.isMatrix4=!0,this.elements=[1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1],e!==void 0&&this.set(e,n,s,r,i,o,a,l,u,c,d,f,h,g,y,v)}set(e,n,s,r,i,o,a,l,u,c,d,f,h,g,y,v){const S=this.elements;return S[0]=e,S[4]=n,S[8]=s,S[12]=r,S[1]=i,S[5]=o,S[9]=a,S[13]=l,S[2]=u,S[6]=c,S[10]=d,S[14]=f,S[3]=h,S[7]=g,S[11]=y,S[15]=v,this}identity(){return this.set(1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1),this}clone(){return new He().fromArray(this.elements)}copy(e){const n=this.elements,s=e.elements;return n[0]=s[0],n[1]=s[1],n[2]=s[2],n[3]=s[3],n[4]=s[4],n[5]=s[5],n[6]=s[6],n[7]=s[7],n[8]=s[8],n[9]=s[9],n[10]=s[10],n[11]=s[11],n[12]=s[12],n[13]=s[13],n[14]=s[14],n[15]=s[15],this}copyPosition(e){const n=this.elements,s=e.elements;return n[12]=s[12],n[13]=s[13],n[14]=s[14],this}setFromMatrix3(e){const n=e.elements;return this.set(n[0],n[3],n[6],0,n[1],n[4],n[7],0,n[2],n[5],n[8],0,0,0,0,1),this}extractBasis(e,n,s){return e.setFromMatrixColumn(this,0),n.setFromMatrixColumn(this,1),s.setFromMatrixColumn(this,2),this}makeBasis(e,n,s){return this.set(e.x,n.x,s.x,0,e.y,n.y,s.y,0,e.z,n.z,s.z,0,0,0,0,1),this}extractRotation(e){const n=this.elements,s=e.elements,r=1/ps.setFromMatrixColumn(e,0).length(),i=1/ps.setFromMatrixColumn(e,1).length(),o=1/ps.setFromMatrixColumn(e,2).length();return n[0]=s[0]*r,n[1]=s[1]*r,n[2]=s[2]*r,n[3]=0,n[4]=s[4]*i,n[5]=s[5]*i,n[6]=s[6]*i,n[7]=0,n[8]=s[8]*o,n[9]=s[9]*o,n[10]=s[10]*o,n[11]=0,n[12]=0,n[13]=0,n[14]=0,n[15]=1,this}makeRotationFromEuler(e){const n=this.elements,s=e.x,r=e.y,i=e.z,o=Math.cos(s),a=Math.sin(s),l=Math.cos(r),u=Math.sin(r),c=Math.cos(i),d=Math.sin(i);if(e.order==="XYZ"){const f=o*c,h=o*d,g=a*c,y=a*d;n[0]=l*c,n[4]=-l*d,n[8]=u,n[1]=h+g*u,n[5]=f-y*u,n[9]=-a*l,n[2]=y-f*u,n[6]=g+h*u,n[10]=o*l}else if(e.order==="YXZ"){const f=l*c,h=l*d,g=u*c,y=u*d;n[0]=f+y*a,n[4]=g*a-h,n[8]=o*u,n[1]=o*d,n[5]=o*c,n[9]=-a,n[2]=h*a-g,n[6]=y+f*a,n[10]=o*l}else if(e.order==="ZXY"){const f=l*c,h=l*d,g=u*c,y=u*d;n[0]=f-y*a,n[4]=-o*d,n[8]=g+h*a,n[1]=h+g*a,n[5]=o*c,n[9]=y-f*a,n[2]=-o*u,n[6]=a,n[10]=o*l}else if(e.order==="ZYX"){const f=o*c,h=o*d,g=a*c,y=a*d;n[0]=l*c,n[4]=g*u-h,n[8]=f*u+y,n[1]=l*d,n[5]=y*u+f,n[9]=h*u-g,n[2]=-u,n[6]=a*l,n[10]=o*l}else if(e.order==="YZX"){const f=o*l,h=o*u,g=a*l,y=a*u;n[0]=l*c,n[4]=y-f*d,n[8]=g*d+h,n[1]=d,n[5]=o*c,n[9]=-a*c,n[2]=-u*c,n[6]=h*d+g,n[10]=f-y*d}else if(e.order==="XZY"){const f=o*l,h=o*u,g=a*l,y=a*u;n[0]=l*c,n[4]=-d,n[8]=u*c,n[1]=f*d+y,n[5]=o*c,n[9]=h*d-g,n[2]=g*d-h,n[6]=a*c,n[10]=y*d+f}return n[3]=0,n[7]=0,n[11]=0,n[12]=0,n[13]=0,n[14]=0,n[15]=1,this}makeRotationFromQuaternion(e){return this.compose(Yb,e,Xb)}lookAt(e,n,s){const r=this.elements;return ht.subVectors(e,n),ht.lengthSq()===0&&(ht.z=1),ht.normalize(),yn.crossVectors(s,ht),yn.lengthSq()===0&&(Math.abs(s.z)===1?ht.x+=1e-4:ht.z+=1e-4,ht.normalize(),yn.crossVectors(s,ht)),yn.normalize(),Yr.crossVectors(ht,yn),r[0]=yn.x,r[4]=Yr.x,r[8]=ht.x,r[1]=yn.y,r[5]=Yr.y,r[9]=ht.y,r[2]=yn.z,r[6]=Yr.z,r[10]=ht.z,this}multiply(e){return this.multiplyMatrices(this,e)}premultiply(e){return this.multiplyMatrices(e,this)}multiplyMatrices(e,n){const s=e.elements,r=n.elements,i=this.elements,o=s[0],a=s[4],l=s[8],u=s[12],c=s[1],d=s[5],f=s[9],h=s[13],g=s[2],y=s[6],v=s[10],S=s[14],m=s[3],p=s[7],w=s[11],b=s[15],x=r[0],C=r[4],M=r[8],T=r[12],V=r[1],L=r[5],A=r[9],k=r[13],O=r[2],H=r[6],q=r[10],j=r[14],Y=r[3],ae=r[7],Q=r[11],te=r[15];return i[0]=o*x+a*V+l*O+u*Y,i[4]=o*C+a*L+l*H+u*ae,i[8]=o*M+a*A+l*q+u*Q,i[12]=o*T+a*k+l*j+u*te,i[1]=c*x+d*V+f*O+h*Y,i[5]=c*C+d*L+f*H+h*ae,i[9]=c*M+d*A+f*q+h*Q,i[13]=c*T+d*k+f*j+h*te,i[2]=g*x+y*V+v*O+S*Y,i[6]=g*C+y*L+v*H+S*ae,i[10]=g*M+y*A+v*q+S*Q,i[14]=g*T+y*k+v*j+S*te,i[3]=m*x+p*V+w*O+b*Y,i[7]=m*C+p*L+w*H+b*ae,i[11]=m*M+p*A+w*q+b*Q,i[15]=m*T+p*k+w*j+b*te,this}multiplyScalar(e){const n=this.elements;return n[0]*=e,n[4]*=e,n[8]*=e,n[12]*=e,n[1]*=e,n[5]*=e,n[9]*=e,n[13]*=e,n[2]*=e,n[6]*=e,n[10]*=e,n[14]*=e,n[3]*=e,n[7]*=e,n[11]*=e,n[15]*=e,this}determinant(){const e=this.elements,n=e[0],s=e[4],r=e[8],i=e[12],o=e[1],a=e[5],l=e[9],u=e[13],c=e[2],d=e[6],f=e[10],h=e[14],g=e[3],y=e[7],v=e[11],S=e[15];return g*(+i*l*d-r*u*d-i*a*f+s*u*f+r*a*h-s*l*h)+y*(+n*l*h-n*u*f+i*o*f-r*o*h+r*u*c-i*l*c)+v*(+n*u*d-n*a*h-i*o*d+s*o*h+i*a*c-s*u*c)+S*(-r*a*c-n*l*d+n*a*f+r*o*d-s*o*f+s*l*c)}transpose(){const e=this.elements;let n;return n=e[1],e[1]=e[4],e[4]=n,n=e[2],e[2]=e[8],e[8]=n,n=e[6],e[6]=e[9],e[9]=n,n=e[3],e[3]=e[12],e[12]=n,n=e[7],e[7]=e[13],e[13]=n,n=e[11],e[11]=e[14],e[14]=n,this}setPosition(e,n,s){const r=this.elements;return e.isVector3?(r[12]=e.x,r[13]=e.y,r[14]=e.z):(r[12]=e,r[13]=n,r[14]=s),this}invert(){const e=this.elements,n=e[0],s=e[1],r=e[2],i=e[3],o=e[4],a=e[5],l=e[6],u=e[7],c=e[8],d=e[9],f=e[10],h=e[11],g=e[12],y=e[13],v=e[14],S=e[15],m=d*v*u-y*f*u+y*l*h-a*v*h-d*l*S+a*f*S,p=g*f*u-c*v*u-g*l*h+o*v*h+c*l*S-o*f*S,w=c*y*u-g*d*u+g*a*h-o*y*h-c*a*S+o*d*S,b=g*d*l-c*y*l-g*a*f+o*y*f+c*a*v-o*d*v,x=n*m+s*p+r*w+i*b;if(x===0)return this.set(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);const C=1/x;return e[0]=m*C,e[1]=(y*f*i-d*v*i-y*r*h+s*v*h+d*r*S-s*f*S)*C,e[2]=(a*v*i-y*l*i+y*r*u-s*v*u-a*r*S+s*l*S)*C,e[3]=(d*l*i-a*f*i-d*r*u+s*f*u+a*r*h-s*l*h)*C,e[4]=p*C,e[5]=(c*v*i-g*f*i+g*r*h-n*v*h-c*r*S+n*f*S)*C,e[6]=(g*l*i-o*v*i-g*r*u+n*v*u+o*r*S-n*l*S)*C,e[7]=(o*f*i-c*l*i+c*r*u-n*f*u-o*r*h+n*l*h)*C,e[8]=w*C,e[9]=(g*d*i-c*y*i-g*s*h+n*y*h+c*s*S-n*d*S)*C,e[10]=(o*y*i-g*a*i+g*s*u-n*y*u-o*s*S+n*a*S)*C,e[11]=(c*a*i-o*d*i-c*s*u+n*d*u+o*s*h-n*a*h)*C,e[12]=b*C,e[13]=(c*y*r-g*d*r+g*s*f-n*y*f-c*s*v+n*d*v)*C,e[14]=(g*a*r-o*y*r-g*s*l+n*y*l+o*s*v-n*a*v)*C,e[15]=(o*d*r-c*a*r+c*s*l-n*d*l-o*s*f+n*a*f)*C,this}scale(e){const n=this.elements,s=e.x,r=e.y,i=e.z;return n[0]*=s,n[4]*=r,n[8]*=i,n[1]*=s,n[5]*=r,n[9]*=i,n[2]*=s,n[6]*=r,n[10]*=i,n[3]*=s,n[7]*=r,n[11]*=i,this}getMaxScaleOnAxis(){const e=this.elements,n=e[0]*e[0]+e[1]*e[1]+e[2]*e[2],s=e[4]*e[4]+e[5]*e[5]+e[6]*e[6],r=e[8]*e[8]+e[9]*e[9]+e[10]*e[10];return Math.sqrt(Math.max(n,s,r))}makeTranslation(e,n,s){return e.isVector3?this.set(1,0,0,e.x,0,1,0,e.y,0,0,1,e.z,0,0,0,1):this.set(1,0,0,e,0,1,0,n,0,0,1,s,0,0,0,1),this}makeRotationX(e){const n=Math.cos(e),s=Math.sin(e);return this.set(1,0,0,0,0,n,-s,0,0,s,n,0,0,0,0,1),this}makeRotationY(e){const n=Math.cos(e),s=Math.sin(e);return this.set(n,0,s,0,0,1,0,0,-s,0,n,0,0,0,0,1),this}makeRotationZ(e){const n=Math.cos(e),s=Math.sin(e);return this.set(n,-s,0,0,s,n,0,0,0,0,1,0,0,0,0,1),this}makeRotationAxis(e,n){const s=Math.cos(n),r=Math.sin(n),i=1-s,o=e.x,a=e.y,l=e.z,u=i*o,c=i*a;return this.set(u*o+s,u*a-r*l,u*l+r*a,0,u*a+r*l,c*a+s,c*l-r*o,0,u*l-r*a,c*l+r*o,i*l*l+s,0,0,0,0,1),this}makeScale(e,n,s){return this.set(e,0,0,0,0,n,0,0,0,0,s,0,0,0,0,1),this}makeShear(e,n,s,r,i,o){return this.set(1,s,i,0,e,1,o,0,n,r,1,0,0,0,0,1),this}compose(e,n,s){const r=this.elements,i=n._x,o=n._y,a=n._z,l=n._w,u=i+i,c=o+o,d=a+a,f=i*u,h=i*c,g=i*d,y=o*c,v=o*d,S=a*d,m=l*u,p=l*c,w=l*d,b=s.x,x=s.y,C=s.z;return r[0]=(1-(y+S))*b,r[1]=(h+w)*b,r[2]=(g-p)*b,r[3]=0,r[4]=(h-w)*x,r[5]=(1-(f+S))*x,r[6]=(v+m)*x,r[7]=0,r[8]=(g+p)*C,r[9]=(v-m)*C,r[10]=(1-(f+y))*C,r[11]=0,r[12]=e.x,r[13]=e.y,r[14]=e.z,r[15]=1,this}decompose(e,n,s){const r=this.elements;let i=ps.set(r[0],r[1],r[2]).length();const o=ps.set(r[4],r[5],r[6]).length(),a=ps.set(r[8],r[9],r[10]).length();this.determinant()<0&&(i=-i),e.x=r[12],e.y=r[13],e.z=r[14],Ft.copy(this);const u=1/i,c=1/o,d=1/a;return Ft.elements[0]*=u,Ft.elements[1]*=u,Ft.elements[2]*=u,Ft.elements[4]*=c,Ft.elements[5]*=c,Ft.elements[6]*=c,Ft.elements[8]*=d,Ft.elements[9]*=d,Ft.elements[10]*=d,n.setFromRotationMatrix(Ft),s.x=i,s.y=o,s.z=a,this}makePerspective(e,n,s,r,i,o,a=Xs){const l=this.elements,u=2*i/(n-e),c=2*i/(s-r),d=(n+e)/(n-e),f=(s+r)/(s-r);let h,g;if(a===Xs)h=-(o+i)/(o-i),g=-2*o*i/(o-i);else if(a===Rc)h=-o/(o-i),g=-o*i/(o-i);else throw new Error("THREE.Matrix4.makePerspective(): Invalid coordinate system: "+a);return l[0]=u,l[4]=0,l[8]=d,l[12]=0,l[1]=0,l[5]=c,l[9]=f,l[13]=0,l[2]=0,l[6]=0,l[10]=h,l[14]=g,l[3]=0,l[7]=0,l[11]=-1,l[15]=0,this}makeOrthographic(e,n,s,r,i,o,a=Xs){const l=this.elements,u=1/(n-e),c=1/(s-r),d=1/(o-i),f=(n+e)*u,h=(s+r)*c;let g,y;if(a===Xs)g=(o+i)*d,y=-2*d;else if(a===Rc)g=i*d,y=-1*d;else throw new Error("THREE.Matrix4.makeOrthographic(): Invalid coordinate system: "+a);return l[0]=2*u,l[4]=0,l[8]=0,l[12]=-f,l[1]=0,l[5]=2*c,l[9]=0,l[13]=-h,l[2]=0,l[6]=0,l[10]=y,l[14]=-g,l[3]=0,l[7]=0,l[11]=0,l[15]=1,this}equals(e){const n=this.elements,s=e.elements;for(let r=0;r<16;r++)if(n[r]!==s[r])return!1;return!0}fromArray(e,n=0){for(let s=0;s<16;s++)this.elements[s]=e[s+n];return this}toArray(e=[],n=0){const s=this.elements;return e[n]=s[0],e[n+1]=s[1],e[n+2]=s[2],e[n+3]=s[3],e[n+4]=s[4],e[n+5]=s[5],e[n+6]=s[6],e[n+7]=s[7],e[n+8]=s[8],e[n+9]=s[9],e[n+10]=s[10],e[n+11]=s[11],e[n+12]=s[12],e[n+13]=s[13],e[n+14]=s[14],e[n+15]=s[15],e}}const ps=new J,Ft=new He,Yb=new J(0,0,0),Xb=new J(1,1,1),yn=new J,Yr=new J,ht=new J,Zb=dn({__name:"OrientationGizmo",props:{elem:{},scene:{}},setup(t){globalThis.THREE={Vector3:J,Matrix4:He};const e=t;function n(a,l){let u=new $b(l.camera,{size:a.clientWidth,bubbleSizePrimary:a.clientWidth/12,bubbleSizeSeconday:a.clientWidth/14,fontSize:a.clientWidth/10+"px"});for(let c of[["y","-z"],["z","-y"],["z","-z"]]){let d=u.bubbles.findIndex(y=>y.axis==c[0]),f=u.bubbles.findIndex(y=>y.axis==c[1]),h=u.bubbles[d].direction.clone(),g=u.bubbles[f].direction.clone();u.bubbles[d].direction.copy(g),u.bubbles[f].direction.copy(h)}return u.onAxisSelected=c=>{let d=l.getCamera().position.clone(),f=l.getTarget().clone().add(l.target.position),h=d.clone().sub(f).length(),g=new J(c.direction.x,c.direction.y,c.direction.z),y=f.clone().add(g.clone().multiplyScalar(h));l.getCamera().position.copy(y),l.getCamera().lookAt(f),l.__perspectiveCamera&&(l.__perspectiveCamera.position.copy(y),l.__perspectiveCamera.lookAt(f)),l.queueRender(),requestIdleCallback(()=>{var v;return(v=e.elem)==null?void 0:v.dispatchEvent(new CustomEvent("camera-change",{detail:{source:"none"}}))},{timeout:100})},u}let s=se(null),r;function i(){r.isConnected&&(r.update(),requestIdleCallback(i,{timeout:250}))}let o=()=>{s.value&&(r&&s.value.removeChild(r),r=n(s.value,e.scene),s.value.appendChild(r),requestIdleCallback(i,{timeout:250}))};return Yt(o),Ka(o),(a,l)=>(at(),Mr("div",{ref_key:"container",ref:s,class:"orientation-gizmo"},null,512))}});let Kb=class{addEventListener(e,n){this._listeners===void 0&&(this._listeners={});const s=this._listeners;s[e]===void 0&&(s[e]=[]),s[e].indexOf(n)===-1&&s[e].push(n)}hasEventListener(e,n){if(this._listeners===void 0)return!1;const s=this._listeners;return s[e]!==void 0&&s[e].indexOf(n)!==-1}removeEventListener(e,n){if(this._listeners===void 0)return;const r=this._listeners[e];if(r!==void 0){const i=r.indexOf(n);i!==-1&&r.splice(i,1)}}dispatchEvent(e){if(this._listeners===void 0)return;const s=this._listeners[e.type];if(s!==void 0){e.target=this;const r=s.slice(0);for(let i=0,o=r.length;i<o;i++)r[i].call(this,e);e.target=null}}};const Ic=new He,kc=new Vr;class vo{constructor(e=0,n=0,s=0,r=vo.DEFAULT_ORDER){this.isEuler=!0,this._x=e,this._y=n,this._z=s,this._order=r}get x(){return this._x}set x(e){this._x=e,this._onChangeCallback()}get y(){return this._y}set y(e){this._y=e,this._onChangeCallback()}get z(){return this._z}set z(e){this._z=e,this._onChangeCallback()}get order(){return this._order}set order(e){this._order=e,this._onChangeCallback()}set(e,n,s,r=this._order){return this._x=e,this._y=n,this._z=s,this._order=r,this._onChangeCallback(),this}clone(){return new this.constructor(this._x,this._y,this._z,this._order)}copy(e){return this._x=e._x,this._y=e._y,this._z=e._z,this._order=e._order,this._onChangeCallback(),this}setFromRotationMatrix(e,n=this._order,s=!0){const r=e.elements,i=r[0],o=r[4],a=r[8],l=r[1],u=r[5],c=r[9],d=r[2],f=r[6],h=r[10];switch(n){case"XYZ":this._y=Math.asin(ot(a,-1,1)),Math.abs(a)<.9999999?(this._x=Math.atan2(-c,h),this._z=Math.atan2(-o,i)):(this._x=Math.atan2(f,u),this._z=0);break;case"YXZ":this._x=Math.asin(-ot(c,-1,1)),Math.abs(c)<.9999999?(this._y=Math.atan2(a,h),this._z=Math.atan2(l,u)):(this._y=Math.atan2(-d,i),this._z=0);break;case"ZXY":this._x=Math.asin(ot(f,-1,1)),Math.abs(f)<.9999999?(this._y=Math.atan2(-d,h),this._z=Math.atan2(-o,u)):(this._y=0,this._z=Math.atan2(l,i));break;case"ZYX":this._y=Math.asin(-ot(d,-1,1)),Math.abs(d)<.9999999?(this._x=Math.atan2(f,h),this._z=Math.atan2(l,i)):(this._x=0,this._z=Math.atan2(-o,u));break;case"YZX":this._z=Math.asin(ot(l,-1,1)),Math.abs(l)<.9999999?(this._x=Math.atan2(-c,u),this._y=Math.atan2(-d,i)):(this._x=0,this._y=Math.atan2(a,h));break;case"XZY":this._z=Math.asin(-ot(o,-1,1)),Math.abs(o)<.9999999?(this._x=Math.atan2(f,u),this._y=Math.atan2(a,i)):(this._x=Math.atan2(-c,h),this._y=0);break;default:console.warn("THREE.Euler: .setFromRotationMatrix() encountered an unknown order: "+n)}return this._order=n,s===!0&&this._onChangeCallback(),this}setFromQuaternion(e,n,s){return Ic.makeRotationFromQuaternion(e),this.setFromRotationMatrix(Ic,n,s)}setFromVector3(e,n=this._order){return this.set(e.x,e.y,e.z,n)}reorder(e){return kc.setFromEuler(this),this.setFromQuaternion(kc,e)}equals(e){return e._x===this._x&&e._y===this._y&&e._z===this._z&&e._order===this._order}fromArray(e){return this._x=e[0],this._y=e[1],this._z=e[2],e[3]!==void 0&&(this._order=e[3]),this._onChangeCallback(),this}toArray(e=[],n=0){return e[n]=this._x,e[n+1]=this._y,e[n+2]=this._z,e[n+3]=this._order,e}_onChange(e){return this._onChangeCallback=e,this}_onChangeCallback(){}*[Symbol.iterator](){yield this._x,yield this._y,yield this._z,yield this._order}}vo.DEFAULT_ORDER="XYZ";class Jb{constructor(){this.mask=1}set(e){this.mask=(1<<e|0)>>>0}enable(e){this.mask|=1<<e|0}enableAll(){this.mask=-1}toggle(e){this.mask^=1<<e|0}disable(e){this.mask&=~(1<<e|0)}disableAll(){this.mask=0}test(e){return(this.mask&e.mask)!==0}isEnabled(e){return(this.mask&(1<<e|0))!==0}}class ds{constructor(e,n,s,r,i,o,a,l,u){ds.prototype.isMatrix3=!0,this.elements=[1,0,0,0,1,0,0,0,1],e!==void 0&&this.set(e,n,s,r,i,o,a,l,u)}set(e,n,s,r,i,o,a,l,u){const c=this.elements;return c[0]=e,c[1]=r,c[2]=a,c[3]=n,c[4]=i,c[5]=l,c[6]=s,c[7]=o,c[8]=u,this}identity(){return this.set(1,0,0,0,1,0,0,0,1),this}copy(e){const n=this.elements,s=e.elements;return n[0]=s[0],n[1]=s[1],n[2]=s[2],n[3]=s[3],n[4]=s[4],n[5]=s[5],n[6]=s[6],n[7]=s[7],n[8]=s[8],this}extractBasis(e,n,s){return e.setFromMatrix3Column(this,0),n.setFromMatrix3Column(this,1),s.setFromMatrix3Column(this,2),this}setFromMatrix4(e){const n=e.elements;return this.set(n[0],n[4],n[8],n[1],n[5],n[9],n[2],n[6],n[10]),this}multiply(e){return this.multiplyMatrices(this,e)}premultiply(e){return this.multiplyMatrices(e,this)}multiplyMatrices(e,n){const s=e.elements,r=n.elements,i=this.elements,o=s[0],a=s[3],l=s[6],u=s[1],c=s[4],d=s[7],f=s[2],h=s[5],g=s[8],y=r[0],v=r[3],S=r[6],m=r[1],p=r[4],w=r[7],b=r[2],x=r[5],C=r[8];return i[0]=o*y+a*m+l*b,i[3]=o*v+a*p+l*x,i[6]=o*S+a*w+l*C,i[1]=u*y+c*m+d*b,i[4]=u*v+c*p+d*x,i[7]=u*S+c*w+d*C,i[2]=f*y+h*m+g*b,i[5]=f*v+h*p+g*x,i[8]=f*S+h*w+g*C,this}multiplyScalar(e){const n=this.elements;return n[0]*=e,n[3]*=e,n[6]*=e,n[1]*=e,n[4]*=e,n[7]*=e,n[2]*=e,n[5]*=e,n[8]*=e,this}determinant(){const e=this.elements,n=e[0],s=e[1],r=e[2],i=e[3],o=e[4],a=e[5],l=e[6],u=e[7],c=e[8];return n*o*c-n*a*u-s*i*c+s*a*l+r*i*u-r*o*l}invert(){const e=this.elements,n=e[0],s=e[1],r=e[2],i=e[3],o=e[4],a=e[5],l=e[6],u=e[7],c=e[8],d=c*o-a*u,f=a*l-c*i,h=u*i-o*l,g=n*d+s*f+r*h;if(g===0)return this.set(0,0,0,0,0,0,0,0,0);const y=1/g;return e[0]=d*y,e[1]=(r*u-c*s)*y,e[2]=(a*s-r*o)*y,e[3]=f*y,e[4]=(c*n-r*l)*y,e[5]=(r*i-a*n)*y,e[6]=h*y,e[7]=(s*l-u*n)*y,e[8]=(o*n-s*i)*y,this}transpose(){let e;const n=this.elements;return e=n[1],n[1]=n[3],n[3]=e,e=n[2],n[2]=n[6],n[6]=e,e=n[5],n[5]=n[7],n[7]=e,this}getNormalMatrix(e){return this.setFromMatrix4(e).invert().transpose()}transposeIntoArray(e){const n=this.elements;return e[0]=n[0],e[1]=n[3],e[2]=n[6],e[3]=n[1],e[4]=n[4],e[5]=n[7],e[6]=n[2],e[7]=n[5],e[8]=n[8],this}setUvTransform(e,n,s,r,i,o,a){const l=Math.cos(i),u=Math.sin(i);return this.set(s*l,s*u,-s*(l*o+u*a)+o+e,-r*u,r*l,-r*(-u*o+l*a)+a+n,0,0,1),this}scale(e,n){return this.premultiply(Wo.makeScale(e,n)),this}rotate(e){return this.premultiply(Wo.makeRotation(-e)),this}translate(e,n){return this.premultiply(Wo.makeTranslation(e,n)),this}makeTranslation(e,n){return e.isVector2?this.set(1,0,e.x,0,1,e.y,0,0,1):this.set(1,0,e,0,1,n,0,0,1),this}makeRotation(e){const n=Math.cos(e),s=Math.sin(e);return this.set(n,-s,0,s,n,0,0,0,1),this}makeScale(e,n){return this.set(e,0,0,0,n,0,0,0,1),this}equals(e){const n=this.elements,s=e.elements;for(let r=0;r<9;r++)if(n[r]!==s[r])return!1;return!0}fromArray(e,n=0){for(let s=0;s<9;s++)this.elements[s]=e[s+n];return this}toArray(e=[],n=0){const s=this.elements;return e[n]=s[0],e[n+1]=s[1],e[n+2]=s[2],e[n+3]=s[3],e[n+4]=s[4],e[n+5]=s[5],e[n+6]=s[6],e[n+7]=s[7],e[n+8]=s[8],e}clone(){return new this.constructor().fromArray(this.elements)}}const Wo=new ds;let Qb=0;const Fc=new J,ys=new Vr,en=new He,Xr=new J,Us=new J,ex=new J,tx=new Vr,Vc=new J(1,0,0),Lc=new J(0,1,0),Pc=new J(0,0,1),nx={type:"added"},sx={type:"removed"};class kn extends Kb{constructor(){super(),this.isObject3D=!0,Object.defineProperty(this,"id",{value:Qb++}),this.uuid=Hb(),this.name="",this.type="Object3D",this.parent=null,this.children=[],this.up=kn.DEFAULT_UP.clone();const e=new J,n=new vo,s=new Vr,r=new J(1,1,1);function i(){s.setFromEuler(n,!1)}function o(){n.setFromQuaternion(s,void 0,!1)}n._onChange(i),s._onChange(o),Object.defineProperties(this,{position:{configurable:!0,enumerable:!0,value:e},rotation:{configurable:!0,enumerable:!0,value:n},quaternion:{configurable:!0,enumerable:!0,value:s},scale:{configurable:!0,enumerable:!0,value:r},modelViewMatrix:{value:new He},normalMatrix:{value:new ds}}),this.matrix=new He,this.matrixWorld=new He,this.matrixAutoUpdate=kn.DEFAULT_MATRIX_AUTO_UPDATE,this.matrixWorldAutoUpdate=kn.DEFAULT_MATRIX_WORLD_AUTO_UPDATE,this.matrixWorldNeedsUpdate=!1,this.layers=new Jb,this.visible=!0,this.castShadow=!1,this.receiveShadow=!1,this.frustumCulled=!0,this.renderOrder=0,this.animations=[],this.userData={}}onBeforeShadow(){}onAfterShadow(){}onBeforeRender(){}onAfterRender(){}applyMatrix4(e){this.matrixAutoUpdate&&this.updateMatrix(),this.matrix.premultiply(e),this.matrix.decompose(this.position,this.quaternion,this.scale)}applyQuaternion(e){return this.quaternion.premultiply(e),this}setRotationFromAxisAngle(e,n){this.quaternion.setFromAxisAngle(e,n)}setRotationFromEuler(e){this.quaternion.setFromEuler(e,!0)}setRotationFromMatrix(e){this.quaternion.setFromRotationMatrix(e)}setRotationFromQuaternion(e){this.quaternion.copy(e)}rotateOnAxis(e,n){return ys.setFromAxisAngle(e,n),this.quaternion.multiply(ys),this}rotateOnWorldAxis(e,n){return ys.setFromAxisAngle(e,n),this.quaternion.premultiply(ys),this}rotateX(e){return this.rotateOnAxis(Vc,e)}rotateY(e){return this.rotateOnAxis(Lc,e)}rotateZ(e){return this.rotateOnAxis(Pc,e)}translateOnAxis(e,n){return Fc.copy(e).applyQuaternion(this.quaternion),this.position.add(Fc.multiplyScalar(n)),this}translateX(e){return this.translateOnAxis(Vc,e)}translateY(e){return this.translateOnAxis(Lc,e)}translateZ(e){return this.translateOnAxis(Pc,e)}localToWorld(e){return this.updateWorldMatrix(!0,!1),e.applyMatrix4(this.matrixWorld)}worldToLocal(e){return this.updateWorldMatrix(!0,!1),e.applyMatrix4(en.copy(this.matrixWorld).invert())}lookAt(e,n,s){e.isVector3?Xr.copy(e):Xr.set(e,n,s);const r=this.parent;this.updateWorldMatrix(!0,!1),Us.setFromMatrixPosition(this.matrixWorld),this.isCamera||this.isLight?en.lookAt(Us,Xr,this.up):en.lookAt(Xr,Us,this.up),this.quaternion.setFromRotationMatrix(en),r&&(en.extractRotation(r.matrixWorld),ys.setFromRotationMatrix(en),this.quaternion.premultiply(ys.invert()))}add(e){if(arguments.length>1){for(let n=0;n<arguments.length;n++)this.add(arguments[n]);return this}return e===this?(console.error("THREE.Object3D.add: object can't be added as a child of itself.",e),this):(e&&e.isObject3D?(e.parent!==null&&e.parent.remove(e),e.parent=this,this.children.push(e),e.dispatchEvent(nx)):console.error("THREE.Object3D.add: object not an instance of THREE.Object3D.",e),this)}remove(e){if(arguments.length>1){for(let s=0;s<arguments.length;s++)this.remove(arguments[s]);return this}const n=this.children.indexOf(e);return n!==-1&&(e.parent=null,this.children.splice(n,1),e.dispatchEvent(sx)),this}removeFromParent(){const e=this.parent;return e!==null&&e.remove(this),this}clear(){return this.remove(...this.children)}attach(e){return this.updateWorldMatrix(!0,!1),en.copy(this.matrixWorld).invert(),e.parent!==null&&(e.parent.updateWorldMatrix(!0,!1),en.multiply(e.parent.matrixWorld)),e.applyMatrix4(en),this.add(e),e.updateWorldMatrix(!1,!0),this}getObjectById(e){return this.getObjectByProperty("id",e)}getObjectByName(e){return this.getObjectByProperty("name",e)}getObjectByProperty(e,n){if(this[e]===n)return this;for(let s=0,r=this.children.length;s<r;s++){const o=this.children[s].getObjectByProperty(e,n);if(o!==void 0)return o}}getObjectsByProperty(e,n,s=[]){this[e]===n&&s.push(this);const r=this.children;for(let i=0,o=r.length;i<o;i++)r[i].getObjectsByProperty(e,n,s);return s}getWorldPosition(e){return this.updateWorldMatrix(!0,!1),e.setFromMatrixPosition(this.matrixWorld)}getWorldQuaternion(e){return this.updateWorldMatrix(!0,!1),this.matrixWorld.decompose(Us,e,ex),e}getWorldScale(e){return this.updateWorldMatrix(!0,!1),this.matrixWorld.decompose(Us,tx,e),e}getWorldDirection(e){this.updateWorldMatrix(!0,!1);const n=this.matrixWorld.elements;return e.set(n[8],n[9],n[10]).normalize()}raycast(){}traverse(e){e(this);const n=this.children;for(let s=0,r=n.length;s<r;s++)n[s].traverse(e)}traverseVisible(e){if(this.visible===!1)return;e(this);const n=this.children;for(let s=0,r=n.length;s<r;s++)n[s].traverseVisible(e)}traverseAncestors(e){const n=this.parent;n!==null&&(e(n),n.traverseAncestors(e))}updateMatrix(){this.matrix.compose(this.position,this.quaternion,this.scale),this.matrixWorldNeedsUpdate=!0}updateMatrixWorld(e){this.matrixAutoUpdate&&this.updateMatrix(),(this.matrixWorldNeedsUpdate||e)&&(this.parent===null?this.matrixWorld.copy(this.matrix):this.matrixWorld.multiplyMatrices(this.parent.matrixWorld,this.matrix),this.matrixWorldNeedsUpdate=!1,e=!0);const n=this.children;for(let s=0,r=n.length;s<r;s++){const i=n[s];(i.matrixWorldAutoUpdate===!0||e===!0)&&i.updateMatrixWorld(e)}}updateWorldMatrix(e,n){const s=this.parent;if(e===!0&&s!==null&&s.matrixWorldAutoUpdate===!0&&s.updateWorldMatrix(!0,!1),this.matrixAutoUpdate&&this.updateMatrix(),this.parent===null?this.matrixWorld.copy(this.matrix):this.matrixWorld.multiplyMatrices(this.parent.matrixWorld,this.matrix),n===!0){const r=this.children;for(let i=0,o=r.length;i<o;i++){const a=r[i];a.matrixWorldAutoUpdate===!0&&a.updateWorldMatrix(!1,!0)}}}toJSON(e){const n=e===void 0||typeof e=="string",s={};n&&(e={geometries:{},materials:{},textures:{},images:{},shapes:{},skeletons:{},animations:{},nodes:{}},s.metadata={version:4.6,type:"Object",generator:"Object3D.toJSON"});const r={};r.uuid=this.uuid,r.type=this.type,this.name!==""&&(r.name=this.name),this.castShadow===!0&&(r.castShadow=!0),this.receiveShadow===!0&&(r.receiveShadow=!0),this.visible===!1&&(r.visible=!1),this.frustumCulled===!1&&(r.frustumCulled=!1),this.renderOrder!==0&&(r.renderOrder=this.renderOrder),Object.keys(this.userData).length>0&&(r.userData=this.userData),r.layers=this.layers.mask,r.matrix=this.matrix.toArray(),r.up=this.up.toArray(),this.matrixAutoUpdate===!1&&(r.matrixAutoUpdate=!1),this.isInstancedMesh&&(r.type="InstancedMesh",r.count=this.count,r.instanceMatrix=this.instanceMatrix.toJSON(),this.instanceColor!==null&&(r.instanceColor=this.instanceColor.toJSON())),this.isBatchedMesh&&(r.type="BatchedMesh",r.perObjectFrustumCulled=this.perObjectFrustumCulled,r.sortObjects=this.sortObjects,r.drawRanges=this._drawRanges,r.reservedRanges=this._reservedRanges,r.visibility=this._visibility,r.active=this._active,r.bounds=this._bounds.map(a=>({boxInitialized:a.boxInitialized,boxMin:a.box.min.toArray(),boxMax:a.box.max.toArray(),sphereInitialized:a.sphereInitialized,sphereRadius:a.sphere.radius,sphereCenter:a.sphere.center.toArray()})),r.maxGeometryCount=this._maxGeometryCount,r.maxVertexCount=this._maxVertexCount,r.maxIndexCount=this._maxIndexCount,r.geometryInitialized=this._geometryInitialized,r.geometryCount=this._geometryCount,r.matricesTexture=this._matricesTexture.toJSON(e),this.boundingSphere!==null&&(r.boundingSphere={center:r.boundingSphere.center.toArray(),radius:r.boundingSphere.radius}),this.boundingBox!==null&&(r.boundingBox={min:r.boundingBox.min.toArray(),max:r.boundingBox.max.toArray()}));function i(a,l){return a[l.uuid]===void 0&&(a[l.uuid]=l.toJSON(e)),l.uuid}if(this.isScene)this.background&&(this.background.isColor?r.background=this.background.toJSON():this.background.isTexture&&(r.background=this.background.toJSON(e).uuid)),this.environment&&this.environment.isTexture&&this.environment.isRenderTargetTexture!==!0&&(r.environment=this.environment.toJSON(e).uuid);else if(this.isMesh||this.isLine||this.isPoints){r.geometry=i(e.geometries,this.geometry);const a=this.geometry.parameters;if(a!==void 0&&a.shapes!==void 0){const l=a.shapes;if(Array.isArray(l))for(let u=0,c=l.length;u<c;u++){const d=l[u];i(e.shapes,d)}else i(e.shapes,l)}}if(this.isSkinnedMesh&&(r.bindMode=this.bindMode,r.bindMatrix=this.bindMatrix.toArray(),this.skeleton!==void 0&&(i(e.skeletons,this.skeleton),r.skeleton=this.skeleton.uuid)),this.material!==void 0)if(Array.isArray(this.material)){const a=[];for(let l=0,u=this.material.length;l<u;l++)a.push(i(e.materials,this.material[l]));r.material=a}else r.material=i(e.materials,this.material);if(this.children.length>0){r.children=[];for(let a=0;a<this.children.length;a++)r.children.push(this.children[a].toJSON(e).object)}if(this.animations.length>0){r.animations=[];for(let a=0;a<this.animations.length;a++){const l=this.animations[a];r.animations.push(i(e.animations,l))}}if(n){const a=o(e.geometries),l=o(e.materials),u=o(e.textures),c=o(e.images),d=o(e.shapes),f=o(e.skeletons),h=o(e.animations),g=o(e.nodes);a.length>0&&(s.geometries=a),l.length>0&&(s.materials=l),u.length>0&&(s.textures=u),c.length>0&&(s.images=c),d.length>0&&(s.shapes=d),f.length>0&&(s.skeletons=f),h.length>0&&(s.animations=h),g.length>0&&(s.nodes=g)}return s.object=r,s;function o(a){const l=[];for(const u in a){const c=a[u];delete c.metadata,l.push(c)}return l}}clone(e){return new this.constructor().copy(this,e)}copy(e,n=!0){if(this.name=e.name,this.up.copy(e.up),this.position.copy(e.position),this.rotation.order=e.rotation.order,this.quaternion.copy(e.quaternion),this.scale.copy(e.scale),this.matrix.copy(e.matrix),this.matrixWorld.copy(e.matrixWorld),this.matrixAutoUpdate=e.matrixAutoUpdate,this.matrixWorldAutoUpdate=e.matrixWorldAutoUpdate,this.matrixWorldNeedsUpdate=e.matrixWorldNeedsUpdate,this.layers.mask=e.layers.mask,this.visible=e.visible,this.castShadow=e.castShadow,this.receiveShadow=e.receiveShadow,this.frustumCulled=e.frustumCulled,this.renderOrder=e.renderOrder,this.animations=e.animations.slice(),this.userData=JSON.parse(JSON.stringify(e.userData)),n===!0)for(let s=0;s<e.children.length;s++){const r=e.children[s];this.add(r.clone())}return this}}kn.DEFAULT_UP=new J(0,1,0);kn.DEFAULT_MATRIX_AUTO_UPDATE=!0;kn.DEFAULT_MATRIX_WORLD_AUTO_UPDATE=!0;class rx extends kn{constructor(){super(),this.isCamera=!0,this.type="Camera",this.matrixWorldInverse=new He,this.projectionMatrix=new He,this.projectionMatrixInverse=new He,this.coordinateSystem=Xs}copy(e,n){return super.copy(e,n),this.matrixWorldInverse.copy(e.matrixWorldInverse),this.projectionMatrix.copy(e.projectionMatrix),this.projectionMatrixInverse.copy(e.projectionMatrixInverse),this.coordinateSystem=e.coordinateSystem,this}getWorldDirection(e){return super.getWorldDirection(e).negate()}updateMatrixWorld(e){super.updateMatrixWorld(e),this.matrixWorldInverse.copy(this.matrixWorld).invert()}updateWorldMatrix(e,n){super.updateWorldMatrix(e,n),this.matrixWorldInverse.copy(this.matrixWorld).invert()}clone(){return new this.constructor().copy(this)}}class ix extends rx{constructor(e=-1,n=1,s=1,r=-1,i=.1,o=2e3){super(),this.isOrthographicCamera=!0,this.type="OrthographicCamera",this.zoom=1,this.view=null,this.left=e,this.right=n,this.top=s,this.bottom=r,this.near=i,this.far=o,this.updateProjectionMatrix()}copy(e,n){return super.copy(e,n),this.left=e.left,this.right=e.right,this.top=e.top,this.bottom=e.bottom,this.near=e.near,this.far=e.far,this.zoom=e.zoom,this.view=e.view===null?null:Object.assign({},e.view),this}setViewOffset(e,n,s,r,i,o){this.view===null&&(this.view={enabled:!0,fullWidth:1,fullHeight:1,offsetX:0,offsetY:0,width:1,height:1}),this.view.enabled=!0,this.view.fullWidth=e,this.view.fullHeight=n,this.view.offsetX=s,this.view.offsetY=r,this.view.width=i,this.view.height=o,this.updateProjectionMatrix()}clearViewOffset(){this.view!==null&&(this.view.enabled=!1),this.updateProjectionMatrix()}updateProjectionMatrix(){const e=(this.right-this.left)/(2*this.zoom),n=(this.top-this.bottom)/(2*this.zoom),s=(this.right+this.left)/2,r=(this.top+this.bottom)/2;let i=s-e,o=s+e,a=r+n,l=r-n;if(this.view!==null&&this.view.enabled){const u=(this.right-this.left)/this.view.fullWidth/this.zoom,c=(this.top-this.bottom)/this.view.fullHeight/this.zoom;i+=u*this.view.offsetX,o=i+u*this.view.width,a-=c*this.view.offsetY,l=a-c*this.view.height}this.projectionMatrix.makeOrthographic(i,o,a,l,this.near,this.far,this.coordinateSystem),this.projectionMatrixInverse.copy(this.projectionMatrix).invert()}toJSON(e){const n=super.toJSON(e);return n.object.zoom=this.zoom,n.object.left=this.left,n.object.right=this.right,n.object.top=this.top,n.object.bottom=this.bottom,n.object.near=this.near,n.object.far=this.far,this.view!==null&&(n.object.view=Object.assign({},this.view)),n}}const ox=jt("h5",null,"Camera",-1),ax={class:"icon-detail"},lx=jt("br",null,null,-1),ux=jt("h5",null,"Extras",-1),cx={ref:"statsHolder"},dx=dn({__name:"Tools",props:{viewer:{}},emits:["findModel"],setup(t,{emit:e}){const n=ra({loader:()=>ha(()=>import("./Selection-AWIjaBb7.js"),__vite__mapDeps([0,1]),import.meta.url),loadingComponent:()=>"Loading...",delay:0});let s=se(null);const r=ra({loader:()=>ha(()=>import("./LicensesDialogContent-DtmHXTA6.js"),__vite__mapDeps([]),import.meta.url),loadingComponent:Wh,delay:0});let i=t;const o=e;let a=se([]),l=()=>a.value.filter(S=>S.object.type=="Mesh"||S.object.type=="SkinnedMesh").length,u=()=>a.value.filter(S=>S.object.type=="Line").length,c=()=>a.value.filter(S=>S.object.type=="Points").length;function d(S){var w;let m=(w=i.viewer)==null?void 0:w.scene;if(!m)return;let p=m.__perspectiveCamera;if(S||p&&m.camera!=p){let b=m.getTarget().clone().add(m.target.position),x=2*Math.tan(p.fov*Math.PI/180/2)*p.position.distanceTo(b),C=x,M=x/m.aspect;m.camera=new ix(-C,C,M,-M,p.near,p.far),m.camera.position.copy(p.position),m.camera.lookAt(b),S&&m.queueRender(),requestAnimationFrame(()=>d(!1))}}let f=se("PERSP");function h(){var w;let S=(w=i.viewer)==null?void 0:w.scene;if(!S)return;let m=S.camera,p=m.isPerspectiveCamera;p?(S.__perspectiveCamera=m,d(!0)):(S.camera=S.__perspectiveCamera,S.queueRender()),f.value=p?"ORTHO":"PERSP",requestIdleCallback(()=>{var b,x;return(x=(b=i.viewer)==null?void 0:b.elem)==null?void 0:x.dispatchEvent(new CustomEvent("camera-change",{detail:{source:"none"}}))},{timeout:100})}async function g(){var m;let S=(m=i.viewer)==null?void 0:m.elem;S&&(await S.updateFraming(),S.zoom(3))}async function y(){var b;let S=(b=i.viewer)==null?void 0:b.elem;if(!S)return;const m=await S.exportScene({onlyVisible:!0,binary:!0}),p=new File([m],"export.glb"),w=document.createElement("a");w.download=p.name,w.href=URL.createObjectURL(p),w.click()}async function v(){window.open("https://github.com/yeicor-3d/yet-another-cad-viewer","_blank")}return window.addEventListener("keydown",S=>{S.key==="p"?h():S.key==="c"?g():S.key==="d"?y():S.key==="g"&&v()}),(S,m)=>{var p;return at(),Mr(Re,null,[(p=F(i).viewer)!=null&&p.scene?(at(),ln(Zb,{key:0,scene:F(i).viewer.scene,elem:F(i).viewer.elem},null,8,["scene","elem"])):da("",!0),_(F(hi)),ox,_(F(it),{icon:"",onClick:h},{default:X(()=>[jt("span",ax,sn(F(f)),1),_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Toggle (P)rojection"),lx,Ve("(currently "+sn(F(f)==="PERSP"?"perspective":"orthographic")+") ",1)]),_:1}),_(F(Be),{type:"mdi",path:F(Ib)},null,8,["path"])]),_:1}),_(F(it),{icon:"",onClick:g},{default:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Re(c)enter Camera")]),_:1}),_(F(Be),{type:"mdi",path:F(Cb)},null,8,["path"])]),_:1}),_(F(hi)),jt("h5",null,"Selection ("+sn(F(l)())+"F "+sn(F(u)())+"E "+sn(F(c)())+"V)",1),_(F(n),{ref:F(s),viewer:F(i).viewer,modelValue:F(a),"onUpdate:modelValue":m[0]||(m[0]=w=>Te(a)?a.value=w:a=w),onFindModel:m[1]||(m[1]=w=>o("findModel",w))},null,8,["viewer","modelValue"]),_(F(hi)),_(F(Aa)),ux,_(F(it),{icon:"",onClick:y},{default:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("(D)ownload Scene")]),_:1}),_(F(Be),{type:"mdi",path:F(Mb)},null,8,["path"])]),_:1}),_(F(R0),{id:"licenses-dialog",fullscreen:""},{activator:X(({props:w})=>[_(F(it),Le({icon:""},w),{default:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Show Licenses")]),_:1}),_(F(Be),{type:"mdi",path:F(Ab)},null,8,["path"])]),_:2},1040)]),default:X(({isActive:w})=>[_(F(O0),null,{default:X(()=>[_(F(zf),null,{default:X(()=>[_(F(Vi),null,{default:X(()=>[Ve("Licenses")]),_:1}),_(F(Aa)),_(F(it),{icon:"",onClick:b=>w.value=!1},{default:X(()=>[_(F(Be),{type:"mdi",path:F(Ra)},null,8,["path"])]),_:2},1032,["onClick"])]),_:2},1024),_(F(Ah),null,{default:X(()=>[_(F(r))]),_:1})]),_:2},1024)]),_:1}),_(F(it),{icon:"",onClick:v},{default:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Open (G)itHub")]),_:1}),_(F(Be),{type:"mdi",path:F(Tb)},null,8,["path"])]),_:1}),jt("div",cx,null,512)],64)}}});class Sl{constructor(){this._listeners={}}addEventListener(e,n){const s=this._listeners;return s[e]===void 0&&(s[e]=[]),s[e].indexOf(n)===-1&&s[e].push(n),this}removeEventListener(e,n){if(this._listeners===void 0)return this;const r=this._listeners[e];if(r!==void 0){const i=r.indexOf(n);i!==-1&&r.splice(i,1)}return this}dispatchEvent(e){if(this._listeners===void 0)return this;const s=this._listeners[e.type];if(s!==void 0){const r=s.slice(0);for(let i=0,o=r.length;i<o;i++)r[i].call(this,e)}return this}dispose(){for(const e in this._listeners)delete this._listeners[e]}}class Ls extends Sl{constructor(e,n,s,r={}){if(super(),this._name=void 0,this._parent=void 0,this._child=void 0,this._attributes=void 0,this._disposed=!1,this._name=e,this._parent=n,this._child=s,this._attributes=r,!n.isOnGraph(s))throw new Error("Cannot connect disconnected graphs.")}getName(){return this._name}getParent(){return this._parent}getChild(){return this._child}setChild(e){return this._child=e,this}getAttributes(){return this._attributes}dispose(){this._disposed||(this._disposed=!0,this.dispatchEvent({type:"dispose",target:this}),super.dispose())}isDisposed(){return this._disposed}}class fx extends Sl{constructor(...e){super(...e),this._emptySet=new Set,this._edges=new Set,this._parentEdges=new Map,this._childEdges=new Map}listEdges(){return Array.from(this._edges)}listParentEdges(e){return Array.from(this._childEdges.get(e)||this._emptySet)}listParents(e){return this.listParentEdges(e).map(n=>n.getParent())}listChildEdges(e){return Array.from(this._parentEdges.get(e)||this._emptySet)}listChildren(e){return this.listChildEdges(e).map(n=>n.getChild())}disconnectParents(e,n){let s=this.listParentEdges(e);return n&&(s=s.filter(r=>n(r.getParent()))),s.forEach(r=>r.dispose()),this}createEdge(e,n,s,r){return this._registerEdge(new Ls(e,n,s,r))}_registerEdge(e){this._edges.add(e);const n=e.getParent();this._parentEdges.has(n)||this._parentEdges.set(n,new Set),this._parentEdges.get(n).add(e);const s=e.getChild();return this._childEdges.has(s)||this._childEdges.set(s,new Set),this._childEdges.get(s).add(e),e.addEventListener("dispose",()=>this._removeEdge(e)),e}_removeEdge(e){return this._edges.delete(e),this._parentEdges.get(e.getParent()).delete(e),this._childEdges.get(e.getChild()).delete(e),this}}function or(){return or=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var n=arguments[e];for(var s in n)Object.prototype.hasOwnProperty.call(n,s)&&(t[s]=n[s])}return t},or.apply(this,arguments)}function ka(t){return t instanceof Ls}function Zs(t){return Array.isArray(t)&&t[0]instanceof Ls}function Ks(t){return!!(mx(t)&&hx(t)instanceof Ls)}function hx(t){for(const e in t)return t[e]}function mx(t){return!!t&&Object.getPrototypeOf(t)===Object.prototype}const Ee=Symbol("attributes"),Es=Symbol("immutableKeys");class Cl extends Sl{constructor(e){super(),this._disposed=!1,this.graph=void 0,this[Ee]=void 0,this[Es]=void 0,this.graph=e,this[Es]=new Set,this[Ee]=this._createAttributes()}getDefaults(){return{}}_createAttributes(){const e=this.getDefaults(),n={};for(const s in e){const r=e[s];if(r instanceof Cl){const i=this.graph.createEdge(s,this,r);i.addEventListener("dispose",()=>r.dispose()),this[Es].add(s),n[s]=i}else n[s]=r}return n}isOnGraph(e){return this.graph===e.graph}isDisposed(){return this._disposed}dispose(){this._disposed||(this.graph.listChildEdges(this).forEach(e=>e.dispose()),this.graph.disconnectParents(this),this._disposed=!0,this.dispatchEvent({type:"dispose"}))}detach(){return this.graph.disconnectParents(this),this}swap(e,n){for(const s in this[Ee]){const r=this[Ee][s];if(ka(r)){const i=r;i.getChild()===e&&this.setRef(s,n,i.getAttributes())}else if(Zs(r)){const o=r.find(a=>a.getChild()===e);if(o){const a=o.getAttributes();this.removeRef(s,e).addRef(s,n,a)}}else if(Ks(r)){const i=r;for(const o in i){const a=i[o];a.getChild()===e&&this.setRefMap(s,o,n,a.getAttributes())}}}return this}get(e){return this[Ee][e]}set(e,n){return this[Ee][e]=n,this.dispatchEvent({type:"change",attribute:e})}getRef(e){const n=this[Ee][e];return n?n.getChild():null}setRef(e,n,s){if(this[Es].has(e))throw new Error(`Cannot overwrite immutable attribute, "${e}".`);const r=this[Ee][e];if(r&&r.dispose(),!n)return this;const i=this.graph.createEdge(e,this,n,s);return i.addEventListener("dispose",()=>{delete this[Ee][e],this.dispatchEvent({type:"change",attribute:e})}),this[Ee][e]=i,this.dispatchEvent({type:"change",attribute:e})}listRefs(e){return this[Ee][e].map(s=>s.getChild())}addRef(e,n,s){const r=this.graph.createEdge(e,this,n,s),i=this[Ee][e];return i.push(r),r.addEventListener("dispose",()=>{let o;for(;(o=i.indexOf(r))!==-1;)i.splice(o,1);this.dispatchEvent({type:"change",attribute:e})}),this.dispatchEvent({type:"change",attribute:e})}removeRef(e,n){return this[Ee][e].filter(i=>i.getChild()===n).forEach(i=>i.dispose()),this}listRefMapKeys(e){return Object.keys(this[Ee][e])}listRefMapValues(e){return Object.values(this[Ee][e]).map(n=>n.getChild())}getRefMap(e,n){const s=this[Ee][e];return s[n]?s[n].getChild():null}setRefMap(e,n,s,r){const i=this[Ee][e],o=i[n];if(o&&o.dispose(),!s)return this;r=Object.assign(r||{},{key:n});const a=this.graph.createEdge(e,this,s,or({},r,{key:n}));return a.addEventListener("dispose",()=>{delete i[n],this.dispatchEvent({type:"change",attribute:e,key:n})}),i[n]=a,this.dispatchEvent({type:"change",attribute:e,key:n})}dispatchEvent(e){return super.dispatchEvent(or({},e,{target:this})),this.graph.dispatchEvent(or({},e,{target:this,type:`node:${e.type}`})),this}}const Yh="v3.10.0",Ni="@glb.bin";var ne,Bi,yt,Fa,Qn;(function(t){t.ACCESSOR="Accessor",t.ANIMATION="Animation",t.ANIMATION_CHANNEL="AnimationChannel",t.ANIMATION_SAMPLER="AnimationSampler",t.BUFFER="Buffer",t.CAMERA="Camera",t.MATERIAL="Material",t.MESH="Mesh",t.PRIMITIVE="Primitive",t.PRIMITIVE_TARGET="PrimitiveTarget",t.NODE="Node",t.ROOT="Root",t.SCENE="Scene",t.SKIN="Skin",t.TEXTURE="Texture",t.TEXTURE_INFO="TextureInfo"})(ne||(ne={})),function(t){t.INTERLEAVED="interleaved",t.SEPARATE="separate"}(Bi||(Bi={})),function(t){t.ARRAY_BUFFER="ARRAY_BUFFER",t.ELEMENT_ARRAY_BUFFER="ELEMENT_ARRAY_BUFFER",t.INVERSE_BIND_MATRICES="INVERSE_BIND_MATRICES",t.OTHER="OTHER",t.SPARSE="SPARSE"}(yt||(yt={})),function(t){t[t.R=4096]="R",t[t.G=256]="G",t[t.B=16]="B",t[t.A=1]="A"}(Fa||(Fa={})),function(t){t.GLTF="GLTF",t.GLB="GLB"}(Qn||(Qn={}));const zi={5120:Int8Array,5121:Uint8Array,5122:Int16Array,5123:Uint16Array,5125:Uint32Array,5126:Float32Array};var Zr,Va=typeof Float32Array<"u"?Float32Array:Array;function Go(t){return Math.hypot(t[0],t[1],t[2])}Math.hypot||(Math.hypot=function(){for(var t=0,e=arguments.length;e--;)t+=arguments[e]*arguments[e];return Math.sqrt(t)}),Zr=new Va(3),Va!=Float32Array&&(Zr[0]=0,Zr[1]=0,Zr[2]=0);class Se{static createBufferFromDataURI(e){if(typeof Buffer>"u"){const n=atob(e.split(",")[1]),s=new Uint8Array(n.length);for(let r=0;r<n.length;r++)s[r]=n.charCodeAt(r);return s}{const n=e.split(",")[1],s=e.indexOf("base64")>=0;return Buffer.from(n,s?"base64":"utf8")}}static encodeText(e){return typeof TextEncoder<"u"?new TextEncoder().encode(e):Buffer.from(e)}static decodeText(e){return typeof TextDecoder<"u"?new TextDecoder().decode(e):Buffer.from(e).toString("utf8")}static concat(e){let n=0;for(const i of e)n+=i.byteLength;const s=new Uint8Array(n);let r=0;for(const i of e)s.set(i,r),r+=i.byteLength;return s}static pad(e,n=0){const s=this.padNumber(e.byteLength);if(s===e.byteLength)return e;const r=new Uint8Array(s);if(r.set(e),n!==0)for(let i=e.byteLength;i<s;i++)r[i]=n;return r}static padNumber(e){return 4*Math.ceil(e/4)}static equals(e,n){if(e===n)return!0;if(e.byteLength!==n.byteLength)return!1;let s=e.byteLength;for(;s--;)if(e[s]!==n[s])return!1;return!0}static toView(e,n=0,s=1/0){return new Uint8Array(e.buffer,e.byteOffset+n,Math.min(e.byteLength,s))}static assertView(e){if(e&&!ArrayBuffer.isView(e))throw new Error(`Method requires Uint8Array parameter; received "${typeof e}".`);return e}}class Kr{static hexToFactor(e,n){e=Math.floor(e);const s=n;return s[0]=(e>>16&255)/255,s[1]=(e>>8&255)/255,s[2]=(255&e)/255,this.convertSRGBToLinear(n,n)}static factorToHex(e){const n=[...e],[s,r,i]=this.convertLinearToSRGB(e,n);return 255*s<<16^255*r<<8^255*i<<0}static convertSRGBToLinear(e,n){const s=e,r=n;for(let i=0;i<3;i++)r[i]=s[i]<.04045?.0773993808*s[i]:Math.pow(.9478672986*s[i]+.0521327014,2.4);return n}static convertLinearToSRGB(e,n){const s=e,r=n;for(let i=0;i<3;i++)r[i]=s[i]<.0031308?12.92*s[i]:1.055*Math.pow(s[i],.41666)-.055;return n}}class po{match(e){return e.length>=8&&e[0]===137&&e[1]===80&&e[2]===78&&e[3]===71&&e[4]===13&&e[5]===10&&e[6]===26&&e[7]===10}getSize(e){const n=new DataView(e.buffer,e.byteOffset);return Se.decodeText(e.slice(12,16))===po.PNG_FRIED_CHUNK_NAME?[n.getUint32(32,!1),n.getUint32(36,!1)]:[n.getUint32(16,!1),n.getUint32(20,!1)]}getChannels(e){return 4}}po.PNG_FRIED_CHUNK_NAME="CgBI";class es{static registerFormat(e,n){this.impls[e]=n}static getMimeType(e){for(const n in this.impls)if(this.impls[n].match(e))return n;return null}static getSize(e,n){return this.impls[n]?this.impls[n].getSize(e):null}static getChannels(e,n){return this.impls[n]?this.impls[n].getChannels(e):null}static getVRAMByteLength(e,n){if(!this.impls[n])return null;if(this.impls[n].getVRAMByteLength)return this.impls[n].getVRAMByteLength(e);let s=0;const r=this.getSize(e,n);if(!r)return null;for(;r[0]>1||r[1]>1;)s+=r[0]*r[1]*4,r[0]=Math.max(Math.floor(r[0]/2),1),r[1]=Math.max(Math.floor(r[1]/2),1);return s+=4,s}static mimeTypeToExtension(e){return e==="image/jpeg"?"jpg":e.split("/").pop()}static extensionToMimeType(e){return e==="jpg"?"image/jpeg":e?`image/${e}`:""}}function gx(t,e){if(e>t.byteLength)throw new TypeError("Corrupt JPG, exceeded buffer limits");if(t.getUint8(e)!==255)throw new TypeError("Invalid JPG, marker table corrupted");return t}es.impls={"image/jpeg":new class{match(t){return t.length>=3&&t[0]===255&&t[1]===216&&t[2]===255}getSize(t){let e,n,s=new DataView(t.buffer,t.byteOffset+4);for(;s.byteLength;){if(e=s.getUint16(0,!1),gx(s,e),n=s.getUint8(e+1),n===192||n===193||n===194)return[s.getUint16(e+7,!1),s.getUint16(e+5,!1)];s=new DataView(t.buffer,s.byteOffset+e+2)}throw new TypeError("Invalid JPG, no size found")}getChannels(t){return 3}},"image/png":new po};class Ps{static basename(e){const n=e.split(/[\\/]/).pop();return n.substring(0,n.lastIndexOf("."))}static extension(e){if(e.startsWith("data:image/")){const n=e.match(/data:(image\/\w+)/)[1];return es.mimeTypeToExtension(n)}return e.startsWith("data:model/gltf+json")?"gltf":e.startsWith("data:model/gltf-binary")?"glb":e.startsWith("data:application/")?"bin":e.split(/[\\/]/).pop().split(/[.]/).pop()}}function Oc(t){return Object.prototype.toString.call(t)==="[object Object]"}function Yn(t){if(Oc(t)===!1)return!1;const e=t.constructor;if(e===void 0)return!0;const n=e.prototype;return Oc(n)!==!1&&Object.prototype.hasOwnProperty.call(n,"isPrototypeOf")!==!1}var qo,La;(function(t){t[t.SILENT=4]="SILENT",t[t.ERROR=3]="ERROR",t[t.WARN=2]="WARN",t[t.INFO=1]="INFO",t[t.DEBUG=0]="DEBUG"})(La||(La={}));class Et{constructor(e){this.verbosity=void 0,this.verbosity=e}debug(e){this.verbosity<=Et.Verbosity.DEBUG&&console.debug(e)}info(e){this.verbosity<=Et.Verbosity.INFO&&console.info(e)}warn(e){this.verbosity<=Et.Verbosity.WARN&&console.warn(e)}error(e){this.verbosity<=Et.Verbosity.ERROR&&console.error(e)}}function vx(t,e,n){var s=e[0],r=e[1],i=e[2],o=e[3],a=e[4],l=e[5],u=e[6],c=e[7],d=e[8],f=e[9],h=e[10],g=e[11],y=e[12],v=e[13],S=e[14],m=e[15],p=n[0],w=n[1],b=n[2],x=n[3];return t[0]=p*s+w*a+b*d+x*y,t[1]=p*r+w*l+b*f+x*v,t[2]=p*i+w*u+b*h+x*S,t[3]=p*o+w*c+b*g+x*m,t[4]=(p=n[4])*s+(w=n[5])*a+(b=n[6])*d+(x=n[7])*y,t[5]=p*r+w*l+b*f+x*v,t[6]=p*i+w*u+b*h+x*S,t[7]=p*o+w*c+b*g+x*m,t[8]=(p=n[8])*s+(w=n[9])*a+(b=n[10])*d+(x=n[11])*y,t[9]=p*r+w*l+b*f+x*v,t[10]=p*i+w*u+b*h+x*S,t[11]=p*o+w*c+b*g+x*m,t[12]=(p=n[12])*s+(w=n[13])*a+(b=n[14])*d+(x=n[15])*y,t[13]=p*r+w*l+b*f+x*v,t[14]=p*i+w*u+b*h+x*S,t[15]=p*o+w*c+b*g+x*m,t}qo=Et,Et.Verbosity=La,Et.DEFAULT_INSTANCE=new qo(qo.Verbosity.INFO);class ze{static identity(e){return e}static eq(e,n,s=1e-5){if(e.length!==n.length)return!1;for(let r=0;r<e.length;r++)if(Math.abs(e[r]-n[r])>s)return!1;return!0}static decodeNormalizedInt(e,n){switch(n){case 5126:return e;case 5123:return e/65535;case 5121:return e/255;case 5122:return Math.max(e/32767,-1);case 5120:return Math.max(e/127,-1);default:throw new Error("Invalid component type.")}}static denormalize(e,n){return ze.decodeNormalizedInt(e,n)}static encodeNormalizedInt(e,n){switch(n){case 5126:return e;case 5123:return Math.round(65535*e);case 5121:return Math.round(255*e);case 5122:return Math.round(32767*e);case 5120:return Math.round(127*e);default:throw new Error("Invalid component type.")}}static normalize(e,n){return ze.encodeNormalizedInt(e,n)}static decompose(e,n,s,r){let i=Go([e[0],e[1],e[2]]);const o=Go([e[4],e[5],e[6]]),a=Go([e[8],e[9],e[10]]);var l,u,c,d,f,h,g,y,v,S,m,p,w,b,x,C,M;((u=(l=e)[0])*(g=l[5])-(c=l[1])*(h=l[4]))*((p=l[10])*(M=l[15])-(w=l[11])*(C=l[14]))-(u*(y=l[6])-(d=l[2])*h)*((m=l[9])*M-w*(x=l[13]))+(u*(v=l[7])-(f=l[3])*h)*(m*C-p*x)+(c*y-d*g)*((S=l[8])*M-w*(b=l[12]))-(c*v-f*g)*(S*C-p*b)+(d*v-f*y)*(S*x-m*b)<0&&(i=-i),n[0]=e[12],n[1]=e[13],n[2]=e[14];const T=e.slice(),V=1/i,L=1/o,A=1/a;T[0]*=V,T[1]*=V,T[2]*=V,T[4]*=L,T[5]*=L,T[6]*=L,T[8]*=A,T[9]*=A,T[10]*=A,function(k,O){var H=new Va(3);(function(_t,Qe){var Nn=Qe[4],E=Qe[5],R=Qe[6],P=Qe[8],N=Qe[9],B=Qe[10];_t[0]=Math.hypot(Qe[0],Qe[1],Qe[2]),_t[1]=Math.hypot(Nn,E,R),_t[2]=Math.hypot(P,N,B)})(H,O);var q=1/H[0],j=1/H[1],Y=1/H[2],ae=O[0]*q,Q=O[1]*j,te=O[2]*Y,ge=O[4]*q,me=O[5]*j,Ce=O[6]*Y,We=O[8]*q,Z=O[9]*j,xe=O[10]*Y,gn=ae+me+xe,ke=0;gn>0?(ke=2*Math.sqrt(gn+1),k[3]=.25*ke,k[0]=(Ce-Z)/ke,k[1]=(We-te)/ke,k[2]=(Q-ge)/ke):ae>me&&ae>xe?(ke=2*Math.sqrt(1+ae-me-xe),k[3]=(Ce-Z)/ke,k[0]=.25*ke,k[1]=(Q+ge)/ke,k[2]=(We+te)/ke):me>xe?(ke=2*Math.sqrt(1+me-ae-xe),k[3]=(We-te)/ke,k[0]=(Q+ge)/ke,k[1]=.25*ke,k[2]=(Ce+Z)/ke):(ke=2*Math.sqrt(1+xe-ae-me),k[3]=(Q-ge)/ke,k[0]=(We+te)/ke,k[1]=(Ce+Z)/ke,k[2]=.25*ke)}(s,T),r[0]=i,r[1]=o,r[2]=a}static compose(e,n,s,r){const i=r,o=n[0],a=n[1],l=n[2],u=n[3],c=o+o,d=a+a,f=l+l,h=o*c,g=o*d,y=o*f,v=a*d,S=a*f,m=l*f,p=u*c,w=u*d,b=u*f,x=s[0],C=s[1],M=s[2];return i[0]=(1-(v+m))*x,i[1]=(g+b)*x,i[2]=(y-w)*x,i[3]=0,i[4]=(g-b)*C,i[5]=(1-(h+m))*C,i[6]=(S+p)*C,i[7]=0,i[8]=(y+w)*M,i[9]=(S-p)*M,i[10]=(1-(h+v))*M,i[11]=0,i[12]=e[0],i[13]=e[1],i[14]=e[2],i[15]=1,i}}function px(t,e){if(!!t!=!!e)return!1;const n=t.getChild(),s=e.getChild();return n===s||n.equals(s)}function yx(t,e){if(!!t!=!!e||t.length!==e.length)return!1;for(let n=0;n<t.length;n++){const s=t[n],r=e[n];if(s.getChild()!==r.getChild()&&!s.getChild().equals(r.getChild()))return!1}return!0}function bx(t,e){if(!!t!=!!e)return!1;const n=Object.keys(t),s=Object.keys(e);if(n.length!==s.length)return!1;for(const r in t){const i=t[r],o=e[r];if(!!i!=!!o)return!1;const a=i.getChild(),l=o.getChild();if(a!==l&&!a.equals(l))return!1}return!0}function Xh(t,e){if(t===e)return!0;if(!!t!=!!e||!t||!e||t.length!==e.length)return!1;for(let n=0;n<t.length;n++)if(t[n]!==e[n])return!1;return!0}function Zh(t,e){if(t===e)return!0;if(!!t!=!!e)return!1;if(!Yn(t)||!Yn(e))return t===e;const n=t,s=e;let r,i=0,o=0;for(r in n)i++;for(r in s)o++;if(i!==o)return!1;for(r in n){const a=n[r],l=s[r];if(Di(a)&&Di(l)){if(!Xh(a,l))return!1}else if(Yn(a)&&Yn(l)){if(!Zh(a,l))return!1}else if(a!==l)return!1}return!0}function Di(t){return Array.isArray(t)||ArrayBuffer.isView(t)}const Nc=new Set,xx=function(){let t="";for(let e=0;e<6;e++)t+="23456789abdegjkmnpqrvwxyzABDEGJKMNPQRVWXYZ".charAt(Math.floor(42*Math.random()));return t},wx=function(){for(let t=0;t<999;t++){const e=xx();if(!Nc.has(e))return Nc.add(e),e}return""},Bc="https://null.example";let ar=class{static dirname(e){const n=e.lastIndexOf("/");return n===-1?"./":e.substring(0,n+1)}static basename(e){return Ps.basename(new URL(e,Bc).pathname)}static extension(e){return Ps.extension(new URL(e,Bc).pathname)}static resolve(e,n){if(!this.isRelativePath(n))return n;const s=e.split("/"),r=n.split("/");s.pop();for(let i=0;i<r.length;i++)r[i]!=="."&&(r[i]===".."?s.pop():s.push(r[i]));return s.join("/")}static isAbsoluteURL(e){return this.PROTOCOL_REGEXP.test(e)}static isRelativePath(e){return!/^(?:[a-zA-Z]+:)?\//.test(e)}};ar.DEFAULT_INIT={},ar.PROTOCOL_REGEXP=/^[a-zA-Z]+:\/\//;const un=t=>t,_x=new Set;class El extends Cl{constructor(e,n=""){super(e),this[Ee].name=n,this.init(),this.dispatchEvent({type:"create"})}getGraph(){return this.graph}getDefaults(){return Object.assign(super.getDefaults(),{name:"",extras:{}})}set(e,n){return Array.isArray(n)&&(n=n.slice()),super.set(e,n)}getName(){return this.get("name")}setName(e){return this.set("name",e)}getExtras(){return this.get("extras")}setExtras(e){return this.set("extras",e)}clone(){return new this.constructor(this.graph).copy(this,un)}copy(e,n=un){for(const s in this[Ee]){const r=this[Ee][s];if(r instanceof Ls)this[Es].has(s)||r.dispose();else if(Zs(r))for(const i of r)i.dispose();else if(Ks(r))for(const i in r)r[i].dispose()}for(const s in e[Ee]){const r=this[Ee][s],i=e[Ee][s];if(i instanceof Ls)this[Es].has(s)?r.getChild().copy(n(i.getChild()),n):this.setRef(s,n(i.getChild()),i.getAttributes());else if(Zs(i))for(const o of i)this.addRef(s,n(o.getChild()),o.getAttributes());else if(Ks(i))for(const o in i){const a=i[o];this.setRefMap(s,o,n(a.getChild()),a.getAttributes())}else this[Ee][s]=Yn(i)?JSON.parse(JSON.stringify(i)):Array.isArray(i)||i instanceof ArrayBuffer||ArrayBuffer.isView(i)?i.slice():i}return this}equals(e,n=_x){if(this===e)return!0;if(this.propertyType!==e.propertyType)return!1;for(const s in this[Ee]){if(n.has(s))continue;const r=this[Ee][s],i=e[Ee][s];if(ka(r)||ka(i)){if(!px(r,i))return!1}else if(Zs(r)||Zs(i)){if(!yx(r,i))return!1}else if(Ks(r)||Ks(i)){if(!bx(r,i))return!1}else if(Yn(r)||Yn(i)){if(!Zh(r,i))return!1}else if(Di(r)||Di(i)){if(!Xh(r,i))return!1}else if(r!==i)return!1}return!0}detach(){return this.graph.disconnectParents(this,e=>e.propertyType!=="Root"),this}listParents(){return this.graph.listParents(this)}}class st extends El{getDefaults(){return Object.assign(super.getDefaults(),{extensions:{}})}getExtension(e){return this.getRefMap("extensions",e)}setExtension(e,n){return n&&n.t(this),this.setRefMap("extensions",e,n)}listExtensions(){return this.listRefMapValues("extensions")}}class ue extends st{constructor(...e){super(...e),this.i=ze.identity,this.o=ze.identity}init(){this.propertyType=ne.ACCESSOR}getDefaults(){return Object.assign(super.getDefaults(),{array:null,type:ue.Type.SCALAR,componentType:ue.ComponentType.FLOAT,normalized:!1,sparse:!1,buffer:null})}copy(e,n=un){return super.copy(e,n),this.i=e.i,this.o=e.o,this}static getElementSize(e){switch(e){case ue.Type.SCALAR:return 1;case ue.Type.VEC2:return 2;case ue.Type.VEC3:return 3;case ue.Type.VEC4:case ue.Type.MAT2:return 4;case ue.Type.MAT3:return 9;case ue.Type.MAT4:return 16;default:throw new Error("Unexpected type: "+e)}}static getComponentSize(e){switch(e){case ue.ComponentType.BYTE:case ue.ComponentType.UNSIGNED_BYTE:return 1;case ue.ComponentType.SHORT:case ue.ComponentType.UNSIGNED_SHORT:return 2;case ue.ComponentType.UNSIGNED_INT:case ue.ComponentType.FLOAT:return 4;default:throw new Error("Unexpected component type: "+e)}}getMinNormalized(e){const n=this.getElementSize();this.getMin(e);for(let s=0;s<n;s++)e[s]=this.o(e[s]);return e}getMin(e){const n=this.get("array"),s=this.getCount(),r=this.getElementSize();for(let i=0;i<r;i++)e[i]=1/0;for(let i=0;i<s*r;i+=r)for(let o=0;o<r;o++){const a=n[i+o];Number.isFinite(a)&&(e[o]=Math.min(e[o],a))}return e}getMaxNormalized(e){const n=this.getElementSize();this.getMax(e);for(let s=0;s<n;s++)e[s]=this.o(e[s]);return e}getMax(e){const n=this.get("array"),s=this.getCount(),r=this.getElementSize();for(let i=0;i<r;i++)e[i]=-1/0;for(let i=0;i<s*r;i+=r)for(let o=0;o<r;o++){const a=n[i+o];Number.isFinite(a)&&(e[o]=Math.max(e[o],a))}return e}getCount(){const e=this.get("array");return e?e.length/this.getElementSize():0}getType(){return this.get("type")}setType(e){return this.set("type",e)}getElementSize(){return ue.getElementSize(this.get("type"))}getComponentSize(){return this.get("array").BYTES_PER_ELEMENT}getComponentType(){return this.get("componentType")}getNormalized(){return this.get("normalized")}setNormalized(e){return this.set("normalized",e),e?(this.o=n=>ze.decodeNormalizedInt(n,this.get("componentType")),this.i=n=>ze.encodeNormalizedInt(n,this.get("componentType"))):(this.o=ze.identity,this.i=ze.identity),this}getScalar(e){const n=this.getElementSize();return this.o(this.get("array")[e*n])}setScalar(e,n){return this.get("array")[e*this.getElementSize()]=this.i(n),this}getElement(e,n){const s=this.getElementSize(),r=this.get("array");for(let i=0;i<s;i++)n[i]=this.o(r[e*s+i]);return n}setElement(e,n){const s=this.getElementSize(),r=this.get("array");for(let i=0;i<s;i++)r[e*s+i]=this.i(n[i]);return this}getSparse(){return this.get("sparse")}setSparse(e){return this.set("sparse",e)}getBuffer(){return this.getRef("buffer")}setBuffer(e){return this.setRef("buffer",e)}getArray(){return this.get("array")}setArray(e){return this.set("componentType",e?function(n){switch(n.constructor){case Float32Array:return ue.ComponentType.FLOAT;case Uint32Array:return ue.ComponentType.UNSIGNED_INT;case Uint16Array:return ue.ComponentType.UNSIGNED_SHORT;case Uint8Array:return ue.ComponentType.UNSIGNED_BYTE;case Int16Array:return ue.ComponentType.SHORT;case Int8Array:return ue.ComponentType.BYTE;default:throw new Error("Unknown accessor componentType.")}}(e):ue.ComponentType.FLOAT),this.set("array",e),this}getByteLength(){const e=this.get("array");return e?e.byteLength:0}}ue.Type={SCALAR:"SCALAR",VEC2:"VEC2",VEC3:"VEC3",VEC4:"VEC4",MAT2:"MAT2",MAT3:"MAT3",MAT4:"MAT4"},ue.ComponentType={BYTE:5120,UNSIGNED_BYTE:5121,SHORT:5122,UNSIGNED_SHORT:5123,UNSIGNED_INT:5125,FLOAT:5126};class Kh extends st{init(){this.propertyType=ne.ANIMATION}getDefaults(){return Object.assign(super.getDefaults(),{channels:[],samplers:[]})}addChannel(e){return this.addRef("channels",e)}removeChannel(e){return this.removeRef("channels",e)}listChannels(){return this.listRefs("channels")}addSampler(e){return this.addRef("samplers",e)}removeSampler(e){return this.removeRef("samplers",e)}listSamplers(){return this.listRefs("samplers")}}class Ml extends st{init(){this.propertyType=ne.ANIMATION_CHANNEL}getDefaults(){return Object.assign(super.getDefaults(),{targetPath:null,targetNode:null,sampler:null})}getTargetPath(){return this.get("targetPath")}setTargetPath(e){return this.set("targetPath",e)}getTargetNode(){return this.getRef("targetNode")}setTargetNode(e){return this.setRef("targetNode",e)}getSampler(){return this.getRef("sampler")}setSampler(e){return this.setRef("sampler",e)}}Ml.TargetPath={TRANSLATION:"translation",ROTATION:"rotation",SCALE:"scale",WEIGHTS:"weights"};class Lr extends st{init(){this.propertyType=ne.ANIMATION_SAMPLER}getDefaultAttributes(){return Object.assign(super.getDefaults(),{interpolation:Lr.Interpolation.LINEAR,input:null,output:null})}getInterpolation(){return this.get("interpolation")}setInterpolation(e){return this.set("interpolation",e)}getInput(){return this.getRef("input")}setInput(e){return this.setRef("input",e,{usage:yt.OTHER})}getOutput(){return this.getRef("output")}setOutput(e){return this.setRef("output",e,{usage:yt.OTHER})}}Lr.Interpolation={LINEAR:"LINEAR",STEP:"STEP",CUBICSPLINE:"CUBICSPLINE"};class Tl extends st{init(){this.propertyType=ne.BUFFER}getDefaults(){return Object.assign(super.getDefaults(),{uri:""})}getURI(){return this.get("uri")}setURI(e){return this.set("uri",e)}}class fs extends st{init(){this.propertyType=ne.CAMERA}getDefaults(){return Object.assign(super.getDefaults(),{type:fs.Type.PERSPECTIVE,znear:.1,zfar:100,aspectRatio:null,yfov:2*Math.PI*50/360,xmag:1,ymag:1})}getType(){return this.get("type")}setType(e){return this.set("type",e)}getZNear(){return this.get("znear")}setZNear(e){return this.set("znear",e)}getZFar(){return this.get("zfar")}setZFar(e){return this.set("zfar",e)}getAspectRatio(){return this.get("aspectRatio")}setAspectRatio(e){return this.set("aspectRatio",e)}getYFov(){return this.get("yfov")}setYFov(e){return this.set("yfov",e)}getXMag(){return this.get("xmag")}setXMag(e){return this.set("xmag",e)}getYMag(){return this.get("ymag")}setYMag(e){return this.set("ymag",e)}}fs.Type={PERSPECTIVE:"perspective",ORTHOGRAPHIC:"orthographic"};class Sx extends El{t(e){if(!this.parentTypes.includes(e.propertyType))throw new Error(`Parent "${e.propertyType}" invalid for child "${this.propertyType}".`)}}Sx.EXTENSION_NAME=void 0;class Pt extends st{init(){this.propertyType=ne.TEXTURE_INFO}getDefaults(){return Object.assign(super.getDefaults(),{texCoord:0,magFilter:null,minFilter:null,wrapS:Pt.WrapMode.REPEAT,wrapT:Pt.WrapMode.REPEAT})}getTexCoord(){return this.get("texCoord")}setTexCoord(e){return this.set("texCoord",e)}getMagFilter(){return this.get("magFilter")}setMagFilter(e){return this.set("magFilter",e)}getMinFilter(){return this.get("minFilter")}setMinFilter(e){return this.set("minFilter",e)}getWrapS(){return this.get("wrapS")}setWrapS(e){return this.set("wrapS",e)}getWrapT(){return this.get("wrapT")}setWrapT(e){return this.set("wrapT",e)}}Pt.WrapMode={CLAMP_TO_EDGE:33071,MIRRORED_REPEAT:33648,REPEAT:10497},Pt.MagFilter={NEAREST:9728,LINEAR:9729},Pt.MinFilter={NEAREST:9728,LINEAR:9729,NEAREST_MIPMAP_NEAREST:9984,LINEAR_MIPMAP_NEAREST:9985,NEAREST_MIPMAP_LINEAR:9986,LINEAR_MIPMAP_LINEAR:9987};const{R:Jr,G:Qr,B:ei,A:Cx}=Fa;class ss extends st{init(){this.propertyType=ne.MATERIAL}getDefaults(){return Object.assign(super.getDefaults(),{alphaMode:ss.AlphaMode.OPAQUE,alphaCutoff:.5,doubleSided:!1,baseColorFactor:[1,1,1,1],baseColorTexture:null,baseColorTextureInfo:new Pt(this.graph,"baseColorTextureInfo"),emissiveFactor:[0,0,0],emissiveTexture:null,emissiveTextureInfo:new Pt(this.graph,"emissiveTextureInfo"),normalScale:1,normalTexture:null,normalTextureInfo:new Pt(this.graph,"normalTextureInfo"),occlusionStrength:1,occlusionTexture:null,occlusionTextureInfo:new Pt(this.graph,"occlusionTextureInfo"),roughnessFactor:1,metallicFactor:1,metallicRoughnessTexture:null,metallicRoughnessTextureInfo:new Pt(this.graph,"metallicRoughnessTextureInfo")})}getDoubleSided(){return this.get("doubleSided")}setDoubleSided(e){return this.set("doubleSided",e)}getAlpha(){return this.get("baseColorFactor")[3]}setAlpha(e){const n=this.get("baseColorFactor").slice();return n[3]=e,this.set("baseColorFactor",n)}getAlphaMode(){return this.get("alphaMode")}setAlphaMode(e){return this.set("alphaMode",e)}getAlphaCutoff(){return this.get("alphaCutoff")}setAlphaCutoff(e){return this.set("alphaCutoff",e)}getBaseColorFactor(){return this.get("baseColorFactor")}setBaseColorFactor(e){return this.set("baseColorFactor",e)}getBaseColorHex(){return Kr.factorToHex(this.get("baseColorFactor"))}setBaseColorHex(e){const n=this.get("baseColorFactor").slice();return this.set("baseColorFactor",Kr.hexToFactor(e,n))}getBaseColorTexture(){return this.getRef("baseColorTexture")}getBaseColorTextureInfo(){return this.getRef("baseColorTexture")?this.getRef("baseColorTextureInfo"):null}setBaseColorTexture(e){return this.setRef("baseColorTexture",e,{channels:Jr|Qr|ei|Cx,isColor:!0})}getEmissiveFactor(){return this.get("emissiveFactor")}setEmissiveFactor(e){return this.set("emissiveFactor",e)}getEmissiveHex(){return Kr.factorToHex(this.get("emissiveFactor"))}setEmissiveHex(e){const n=this.get("emissiveFactor").slice();return this.set("emissiveFactor",Kr.hexToFactor(e,n))}getEmissiveTexture(){return this.getRef("emissiveTexture")}getEmissiveTextureInfo(){return this.getRef("emissiveTexture")?this.getRef("emissiveTextureInfo"):null}setEmissiveTexture(e){return this.setRef("emissiveTexture",e,{channels:Jr|Qr|ei,isColor:!0})}getNormalScale(){return this.get("normalScale")}setNormalScale(e){return this.set("normalScale",e)}getNormalTexture(){return this.getRef("normalTexture")}getNormalTextureInfo(){return this.getRef("normalTexture")?this.getRef("normalTextureInfo"):null}setNormalTexture(e){return this.setRef("normalTexture",e,{channels:Jr|Qr|ei})}getOcclusionStrength(){return this.get("occlusionStrength")}setOcclusionStrength(e){return this.set("occlusionStrength",e)}getOcclusionTexture(){return this.getRef("occlusionTexture")}getOcclusionTextureInfo(){return this.getRef("occlusionTexture")?this.getRef("occlusionTextureInfo"):null}setOcclusionTexture(e){return this.setRef("occlusionTexture",e,{channels:Jr})}getRoughnessFactor(){return this.get("roughnessFactor")}setRoughnessFactor(e){return this.set("roughnessFactor",e)}getMetallicFactor(){return this.get("metallicFactor")}setMetallicFactor(e){return this.set("metallicFactor",e)}getMetallicRoughnessTexture(){return this.getRef("metallicRoughnessTexture")}getMetallicRoughnessTextureInfo(){return this.getRef("metallicRoughnessTexture")?this.getRef("metallicRoughnessTextureInfo"):null}setMetallicRoughnessTexture(e){return this.setRef("metallicRoughnessTexture",e,{channels:Qr|ei})}}ss.AlphaMode={OPAQUE:"OPAQUE",MASK:"MASK",BLEND:"BLEND"};class Jh extends st{init(){this.propertyType=ne.MESH}getDefaults(){return Object.assign(super.getDefaults(),{weights:[],primitives:[]})}addPrimitive(e){return this.addRef("primitives",e)}removePrimitive(e){return this.removeRef("primitives",e)}listPrimitives(){return this.listRefs("primitives")}getWeights(){return this.get("weights")}setWeights(e){return this.set("weights",e)}}class Qh extends st{constructor(...e){super(...e),this.u=null,this.h=new Set}init(){this.propertyType=ne.NODE}getDefaults(){return Object.assign(super.getDefaults(),{translation:[0,0,0],rotation:[0,0,0,1],scale:[1,1,1],weights:[],camera:null,mesh:null,skin:null,children:[]})}copy(e,n=un){if(n===un)throw new Error("Node cannot be copied.");return super.copy(e,n)}getTranslation(){return this.get("translation")}getRotation(){return this.get("rotation")}getScale(){return this.get("scale")}setTranslation(e){return this.set("translation",e)}setRotation(e){return this.set("rotation",e)}setScale(e){return this.set("scale",e)}getMatrix(){return ze.compose(this.get("translation"),this.get("rotation"),this.get("scale"),[])}setMatrix(e){const n=this.get("translation").slice(),s=this.get("rotation").slice(),r=this.get("scale").slice();return ze.decompose(e,n,s,r),this.set("translation",n).set("rotation",s).set("scale",r)}getWorldTranslation(){const e=[0,0,0];return ze.decompose(this.getWorldMatrix(),e,[0,0,0,1],[1,1,1]),e}getWorldRotation(){const e=[0,0,0,1];return ze.decompose(this.getWorldMatrix(),[0,0,0],e,[1,1,1]),e}getWorldScale(){const e=[1,1,1];return ze.decompose(this.getWorldMatrix(),[0,0,0],[0,0,0,1],e),e}getWorldMatrix(){const e=[];for(let r=this;r!=null;r=r.u)e.push(r);let n;const s=e.pop().getMatrix();for(;n=e.pop();)vx(s,s,n.getMatrix());return s}addChild(e){if(e.u&&e.u.removeChild(e),e.h.size)for(const s of e.h)s.removeChild(e);this.addRef("children",e),e.u=this;const n=this[Ee].children;return n[n.length-1].addEventListener("dispose",()=>e.u=null),this}removeChild(e){return this.removeRef("children",e)}listChildren(){return this.listRefs("children")}getParent(){return this.u?this.u:this.listParents().find(e=>e.propertyType===ne.SCENE)||null}getParentNode(){return this.u}getMesh(){return this.getRef("mesh")}setMesh(e){return this.setRef("mesh",e)}getCamera(){return this.getRef("camera")}setCamera(e){return this.setRef("camera",e)}getSkin(){return this.getRef("skin")}setSkin(e){return this.setRef("skin",e)}getWeights(){return this.get("weights")}setWeights(e){return this.set("weights",e)}traverse(e){e(this);for(const n of this.listChildren())n.traverse(e);return this}}class yo extends st{init(){this.propertyType=ne.PRIMITIVE}getDefaults(){return Object.assign(super.getDefaults(),{mode:yo.Mode.TRIANGLES,material:null,indices:null,attributes:{},targets:[]})}getIndices(){return this.getRef("indices")}setIndices(e){return this.setRef("indices",e,{usage:yt.ELEMENT_ARRAY_BUFFER})}getAttribute(e){return this.getRefMap("attributes",e)}setAttribute(e,n){return this.setRefMap("attributes",e,n,{usage:yt.ARRAY_BUFFER})}listAttributes(){return this.listRefMapValues("attributes")}listSemantics(){return this.listRefMapKeys("attributes")}getMaterial(){return this.getRef("material")}setMaterial(e){return this.setRef("material",e)}getMode(){return this.get("mode")}setMode(e){return this.set("mode",e)}listTargets(){return this.listRefs("targets")}addTarget(e){return this.addRef("targets",e)}removeTarget(e){return this.removeRef("targets",e)}}yo.Mode={POINTS:0,LINES:1,LINE_LOOP:2,LINE_STRIP:3,TRIANGLES:4,TRIANGLE_STRIP:5,TRIANGLE_FAN:6};class Ex extends El{init(){this.propertyType=ne.PRIMITIVE_TARGET}getDefaults(){return Object.assign(super.getDefaults(),{attributes:{}})}getAttribute(e){return this.getRefMap("attributes",e)}setAttribute(e,n){return this.setRefMap("attributes",e,n,{usage:yt.ARRAY_BUFFER})}listAttributes(){return this.listRefMapValues("attributes")}listSemantics(){return this.listRefMapKeys("attributes")}}function nt(){return nt=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var n=arguments[e];for(var s in n)Object.prototype.hasOwnProperty.call(n,s)&&(t[s]=n[s])}return t},nt.apply(this,arguments)}class Al extends st{init(){this.propertyType=ne.SCENE}getDefaults(){return Object.assign(super.getDefaults(),{children:[]})}copy(e,n=un){if(n===un)throw new Error("Scene cannot be copied.");return super.copy(e,n)}addChild(e){e.u&&e.u.removeChild(e),this.addRef("children",e),e.h.add(this);const n=this[Ee].children;return n[n.length-1].addEventListener("dispose",()=>e.h.delete(this)),this}removeChild(e){return this.removeRef("children",e)}listChildren(){return this.listRefs("children")}traverse(e){for(const n of this.listChildren())n.traverse(e);return this}}class em extends st{init(){this.propertyType=ne.SKIN}getDefaults(){return Object.assign(super.getDefaults(),{skeleton:null,inverseBindMatrices:null,joints:[]})}getSkeleton(){return this.getRef("skeleton")}setSkeleton(e){return this.setRef("skeleton",e)}getInverseBindMatrices(){return this.getRef("inverseBindMatrices")}setInverseBindMatrices(e){return this.setRef("inverseBindMatrices",e,{usage:yt.INVERSE_BIND_MATRICES})}addJoint(e){return this.addRef("joints",e)}removeJoint(e){return this.removeRef("joints",e)}listJoints(){return this.listRefs("joints")}}class tm extends st{init(){this.propertyType=ne.TEXTURE}getDefaults(){return Object.assign(super.getDefaults(),{image:null,mimeType:"",uri:""})}getMimeType(){return this.get("mimeType")||es.extensionToMimeType(Ps.extension(this.get("uri")))}setMimeType(e){return this.set("mimeType",e)}getURI(){return this.get("uri")}setURI(e){this.set("uri",e);const n=es.extensionToMimeType(Ps.extension(e));return n&&this.set("mimeType",n),this}getImage(){return this.get("image")}setImage(e){return this.set("image",Se.assertView(e))}getSize(){const e=this.get("image");return e?es.getSize(e,this.getMimeType()):null}}class Mx extends st{init(){this.propertyType=ne.ROOT}getDefaults(){return Object.assign(super.getDefaults(),{asset:{generator:`glTF-Transform ${Yh}`,version:"2.0"},defaultScene:null,accessors:[],animations:[],buffers:[],cameras:[],materials:[],meshes:[],nodes:[],scenes:[],skins:[],textures:[]})}constructor(e){super(e),this.l=new Set,e.addEventListener("node:create",n=>{this.g(n.target)})}clone(){throw new Error("Root cannot be cloned.")}copy(e,n=un){if(n===un)throw new Error("Root cannot be copied.");this.set("asset",nt({},e.get("asset"))),this.setName(e.getName()),this.setExtras(nt({},e.getExtras())),this.setDefaultScene(e.getDefaultScene()?n(e.getDefaultScene()):null);for(const s of e.listRefMapKeys("extensions")){const r=e.getExtension(s);this.setExtension(s,n(r))}return this}g(e){return e instanceof Al?this.addRef("scenes",e):e instanceof Qh?this.addRef("nodes",e):e instanceof fs?this.addRef("cameras",e):e instanceof em?this.addRef("skins",e):e instanceof Jh?this.addRef("meshes",e):e instanceof ss?this.addRef("materials",e):e instanceof tm?this.addRef("textures",e):e instanceof Kh?this.addRef("animations",e):e instanceof ue?this.addRef("accessors",e):e instanceof Tl&&this.addRef("buffers",e),this}getAsset(){return this.get("asset")}listExtensionsUsed(){return Array.from(this.l)}listExtensionsRequired(){return this.listExtensionsUsed().filter(e=>e.isRequired())}p(e){return this.l.add(e),this}m(e){return this.l.delete(e),this}listScenes(){return this.listRefs("scenes")}setDefaultScene(e){return this.setRef("defaultScene",e)}getDefaultScene(){return this.getRef("defaultScene")}listNodes(){return this.listRefs("nodes")}listCameras(){return this.listRefs("cameras")}listSkins(){return this.listRefs("skins")}listMeshes(){return this.listRefs("meshes")}listMaterials(){return this.listRefs("materials")}listTextures(){return this.listRefs("textures")}listAnimations(){return this.listRefs("animations")}listAccessors(){return this.listRefs("accessors")}listBuffers(){return this.listRefs("buffers")}}class Fn{static fromGraph(e){return Fn.v.get(e)||null}constructor(){this.T=new fx,this.S=new Mx(this.T),this.M=Et.DEFAULT_INSTANCE,Fn.v.set(this.T,this)}getRoot(){return this.S}getGraph(){return this.T}getLogger(){return this.M}setLogger(e){return this.M=e,this}clone(){return new Fn().setLogger(this.M).merge(this)}merge(e){for(const i of e.getRoot().listExtensionsUsed()){const o=this.createExtension(i.constructor);i.isRequired()&&o.setRequired(!0)}const n=new Set,s=new Map;n.add(e.S),s.set(e.S,this.S);for(const i of e.T.listEdges())for(const o of[i.getParent(),i.getChild()]){if(n.has(o))continue;let a;a=o.propertyType===ne.TEXTURE_INFO?o:new o.constructor(this.T),s.set(o,a),n.add(o)}const r=i=>{const o=s.get(i);if(!o)throw new Error("Could resolve property.");return o};for(const i of n){const o=s.get(i);if(!o)throw new Error("Could resolve property.");o.propertyType!==ne.TEXTURE_INFO&&o.copy(i,r)}return this}async transform(...e){const n=e.map(s=>s.name);for(const s of e)await s(this,{stack:n});return this}createExtension(e){const n=e.EXTENSION_NAME;return this.getRoot().listExtensionsUsed().find(s=>s.extensionName===n)||new e(this)}createScene(e=""){return new Al(this.T,e)}createNode(e=""){return new Qh(this.T,e)}createCamera(e=""){return new fs(this.T,e)}createSkin(e=""){return new em(this.T,e)}createMesh(e=""){return new Jh(this.T,e)}createPrimitive(){return new yo(this.T)}createPrimitiveTarget(e=""){return new Ex(this.T,e)}createMaterial(e=""){return new ss(this.T,e)}createTexture(e=""){return new tm(this.T,e)}createAnimation(e=""){return new Kh(this.T,e)}createAnimationChannel(e=""){return new Ml(this.T,e)}createAnimationSampler(e=""){return new Lr(this.T,e)}createAccessor(e="",n=null){return n||(n=this.getRoot().listBuffers()[0]),new ue(this.T,e).setBuffer(n)}createBuffer(e=""){return new Tl(this.T,e)}}Fn.v=new WeakMap;class Tx{constructor(e){this.jsonDoc=void 0,this.buffers=[],this.bufferViews=[],this.bufferViewBuffers=[],this.accessors=[],this.textures=[],this.textureInfos=new Map,this.materials=[],this.meshes=[],this.cameras=[],this.nodes=[],this.skins=[],this.animations=[],this.scenes=[],this.jsonDoc=e}setTextureInfo(e,n){this.textureInfos.set(e,n),n.texCoord!==void 0&&e.setTexCoord(n.texCoord),n.extras!==void 0&&e.setExtras(n.extras);const s=this.jsonDoc.json.textures[n.index];if(s.sampler===void 0)return;const r=this.jsonDoc.json.samplers[s.sampler];r.magFilter!==void 0&&e.setMagFilter(r.magFilter),r.minFilter!==void 0&&e.setMinFilter(r.minFilter),r.wrapS!==void 0&&e.setWrapS(r.wrapS),r.wrapT!==void 0&&e.setWrapT(r.wrapT)}}const zc={logger:Et.DEFAULT_INSTANCE,extensions:[],dependencies:{}};class Ax{static read(e,n=zc){const s=nt({},zc,n),{json:r}=e,i=new Fn().setLogger(s.logger);this.validate(e,s);const o=new Tx(e),a=r.asset,l=i.getRoot().getAsset();a.copyright&&(l.copyright=a.copyright),a.extras&&(l.extras=a.extras),r.extras!==void 0&&i.getRoot().setExtras(nt({},r.extras));const u=r.extensionsUsed||[],c=r.extensionsRequired||[];for(const m of s.extensions)if(u.includes(m.EXTENSION_NAME)){const p=i.createExtension(m).setRequired(c.includes(m.EXTENSION_NAME));for(const w of p.readDependencies)p.install(w,s.dependencies[w])}const d=r.buffers||[];i.getRoot().listExtensionsUsed().filter(m=>m.prereadTypes.includes(ne.BUFFER)).forEach(m=>m.preread(o,ne.BUFFER)),o.buffers=d.map(m=>{const p=i.createBuffer(m.name);return m.extras&&p.setExtras(m.extras),m.uri&&m.uri.indexOf("__")!==0&&p.setURI(m.uri),p}),o.bufferViewBuffers=(r.bufferViews||[]).map((m,p)=>{if(!o.bufferViews[p]){const w=e.json.buffers[m.buffer];o.bufferViews[p]=Se.toView(w.uri?e.resources[w.uri]:e.resources[Ni],m.byteOffset||0,m.byteLength)}return o.buffers[m.buffer]});const f=r.accessors||[];o.accessors=f.map(m=>{const p=i.createAccessor(m.name,o.bufferViewBuffers[m.bufferView]).setType(m.type);return m.extras&&p.setExtras(m.extras),m.normalized!==void 0&&p.setNormalized(m.normalized),m.bufferView===void 0||p.setArray(ti(m,o)),p});const h=r.images||[],g=r.textures||[];i.getRoot().listExtensionsUsed().filter(m=>m.prereadTypes.includes(ne.TEXTURE)).forEach(m=>m.preread(o,ne.TEXTURE)),o.textures=h.map(m=>{const p=i.createTexture(m.name);if(m.extras&&p.setExtras(m.extras),m.bufferView!==void 0){const w=r.bufferViews[m.bufferView],b=e.json.buffers[w.buffer],x=w.byteOffset||0,C=(b.uri?e.resources[b.uri]:e.resources[Ni]).slice(x,x+w.byteLength);p.setImage(C)}else m.uri!==void 0&&(p.setImage(e.resources[m.uri]),m.uri.indexOf("__")!==0&&p.setURI(m.uri));if(m.mimeType!==void 0)p.setMimeType(m.mimeType);else if(m.uri){const w=Ps.extension(m.uri);p.setMimeType(es.extensionToMimeType(w))}return p}),o.materials=(r.materials||[]).map(m=>{const p=i.createMaterial(m.name);m.extras&&p.setExtras(m.extras),m.alphaMode!==void 0&&p.setAlphaMode(m.alphaMode),m.alphaCutoff!==void 0&&p.setAlphaCutoff(m.alphaCutoff),m.doubleSided!==void 0&&p.setDoubleSided(m.doubleSided);const w=m.pbrMetallicRoughness||{};if(w.baseColorFactor!==void 0&&p.setBaseColorFactor(w.baseColorFactor),m.emissiveFactor!==void 0&&p.setEmissiveFactor(m.emissiveFactor),w.metallicFactor!==void 0&&p.setMetallicFactor(w.metallicFactor),w.roughnessFactor!==void 0&&p.setRoughnessFactor(w.roughnessFactor),w.baseColorTexture!==void 0){const b=w.baseColorTexture;p.setBaseColorTexture(o.textures[g[b.index].source]),o.setTextureInfo(p.getBaseColorTextureInfo(),b)}if(m.emissiveTexture!==void 0){const b=m.emissiveTexture;p.setEmissiveTexture(o.textures[g[b.index].source]),o.setTextureInfo(p.getEmissiveTextureInfo(),b)}if(m.normalTexture!==void 0){const b=m.normalTexture;p.setNormalTexture(o.textures[g[b.index].source]),o.setTextureInfo(p.getNormalTextureInfo(),b),m.normalTexture.scale!==void 0&&p.setNormalScale(m.normalTexture.scale)}if(m.occlusionTexture!==void 0){const b=m.occlusionTexture;p.setOcclusionTexture(o.textures[g[b.index].source]),o.setTextureInfo(p.getOcclusionTextureInfo(),b),m.occlusionTexture.strength!==void 0&&p.setOcclusionStrength(m.occlusionTexture.strength)}if(w.metallicRoughnessTexture!==void 0){const b=w.metallicRoughnessTexture;p.setMetallicRoughnessTexture(o.textures[g[b.index].source]),o.setTextureInfo(p.getMetallicRoughnessTextureInfo(),b)}return p});const y=r.meshes||[];i.getRoot().listExtensionsUsed().filter(m=>m.prereadTypes.includes(ne.PRIMITIVE)).forEach(m=>m.preread(o,ne.PRIMITIVE)),o.meshes=y.map(m=>{const p=i.createMesh(m.name);return m.extras&&p.setExtras(m.extras),m.weights!==void 0&&p.setWeights(m.weights),(m.primitives||[]).forEach(w=>{const b=i.createPrimitive();w.extras&&b.setExtras(w.extras),w.material!==void 0&&b.setMaterial(o.materials[w.material]),w.mode!==void 0&&b.setMode(w.mode);for(const[C,M]of Object.entries(w.attributes||{}))b.setAttribute(C,o.accessors[M]);w.indices!==void 0&&b.setIndices(o.accessors[w.indices]);const x=m.extras&&m.extras.targetNames||[];(w.targets||[]).forEach((C,M)=>{const T=x[M]||M.toString(),V=i.createPrimitiveTarget(T);for(const[L,A]of Object.entries(C))V.setAttribute(L,o.accessors[A]);b.addTarget(V)}),p.addPrimitive(b)}),p}),o.cameras=(r.cameras||[]).map(m=>{const p=i.createCamera(m.name).setType(m.type);if(m.extras&&p.setExtras(m.extras),m.type===fs.Type.PERSPECTIVE){const w=m.perspective;p.setYFov(w.yfov),p.setZNear(w.znear),w.zfar!==void 0&&p.setZFar(w.zfar),w.aspectRatio!==void 0&&p.setAspectRatio(w.aspectRatio)}else{const w=m.orthographic;p.setZNear(w.znear).setZFar(w.zfar).setXMag(w.xmag).setYMag(w.ymag)}return p});const v=r.nodes||[];i.getRoot().listExtensionsUsed().filter(m=>m.prereadTypes.includes(ne.NODE)).forEach(m=>m.preread(o,ne.NODE)),o.nodes=v.map(m=>{const p=i.createNode(m.name);if(m.extras&&p.setExtras(m.extras),m.translation!==void 0&&p.setTranslation(m.translation),m.rotation!==void 0&&p.setRotation(m.rotation),m.scale!==void 0&&p.setScale(m.scale),m.matrix!==void 0){const w=[0,0,0],b=[0,0,0,1],x=[1,1,1];ze.decompose(m.matrix,w,b,x),p.setTranslation(w),p.setRotation(b),p.setScale(x)}return m.weights!==void 0&&p.setWeights(m.weights),p}),o.skins=(r.skins||[]).map(m=>{const p=i.createSkin(m.name);m.extras&&p.setExtras(m.extras),m.inverseBindMatrices!==void 0&&p.setInverseBindMatrices(o.accessors[m.inverseBindMatrices]),m.skeleton!==void 0&&p.setSkeleton(o.nodes[m.skeleton]);for(const w of m.joints)p.addJoint(o.nodes[w]);return p}),v.map((m,p)=>{const w=o.nodes[p];(m.children||[]).forEach(b=>w.addChild(o.nodes[b])),m.mesh!==void 0&&w.setMesh(o.meshes[m.mesh]),m.camera!==void 0&&w.setCamera(o.cameras[m.camera]),m.skin!==void 0&&w.setSkin(o.skins[m.skin])}),o.animations=(r.animations||[]).map(m=>{const p=i.createAnimation(m.name);m.extras&&p.setExtras(m.extras);const w=(m.samplers||[]).map(b=>{const x=i.createAnimationSampler().setInput(o.accessors[b.input]).setOutput(o.accessors[b.output]).setInterpolation(b.interpolation||Lr.Interpolation.LINEAR);return b.extras&&x.setExtras(b.extras),p.addSampler(x),x});return(m.channels||[]).forEach(b=>{const x=i.createAnimationChannel().setSampler(w[b.sampler]).setTargetPath(b.target.path);b.target.node!==void 0&&x.setTargetNode(o.nodes[b.target.node]),b.extras&&x.setExtras(b.extras),p.addChannel(x)}),p});const S=r.scenes||[];return i.getRoot().listExtensionsUsed().filter(m=>m.prereadTypes.includes(ne.SCENE)).forEach(m=>m.preread(o,ne.SCENE)),o.scenes=S.map(m=>{const p=i.createScene(m.name);return m.extras&&p.setExtras(m.extras),(m.nodes||[]).map(w=>o.nodes[w]).forEach(w=>p.addChild(w)),p}),r.scene!==void 0&&i.getRoot().setDefaultScene(o.scenes[r.scene]),i.getRoot().listExtensionsUsed().forEach(m=>m.read(o)),f.forEach((m,p)=>{const w=o.accessors[p],b=!!m.sparse,x=!m.bufferView&&!w.getArray();(b||x)&&w.setSparse(!0).setArray(function(C,M){const T=zi[C.componentType],V=ue.getElementSize(C.type);let L;L=C.bufferView!==void 0?ti(C,M):new T(C.count*V);const A=C.sparse;if(!A)return L;const k=A.count,O=nt({},C,A.indices,{count:k,type:"SCALAR"}),H=nt({},C,A.values,{count:k}),q=ti(O,M),j=ti(H,M);for(let Y=0;Y<O.count;Y++)for(let ae=0;ae<V;ae++)L[q[Y]*V+ae]=j[Y*V+ae];return L}(m,o))}),i}static validate(e,n){const s=e.json;if(s.asset.version!=="2.0")throw new Error(`Unsupported glTF version, "${s.asset.version}".`);if(s.extensionsRequired){for(const r of s.extensionsRequired)if(!n.extensions.find(i=>i.EXTENSION_NAME===r))throw new Error(`Missing required extension, "${r}".`)}if(s.extensionsUsed)for(const r of s.extensionsUsed)n.extensions.find(i=>i.EXTENSION_NAME===r)||n.logger.warn(`Missing optional extension, "${r}".`)}}function ti(t,e){const n=e.bufferViews[t.bufferView],s=e.jsonDoc.json.bufferViews[t.bufferView],r=zi[t.componentType],i=ue.getElementSize(t.type),o=r.BYTES_PER_ELEMENT;if(s.byteStride!==void 0&&s.byteStride!==i*o)return function(l,u){const c=u.bufferViews[l.bufferView],d=u.jsonDoc.json.bufferViews[l.bufferView],f=zi[l.componentType],h=ue.getElementSize(l.type),g=f.BYTES_PER_ELEMENT,y=l.byteOffset||0,v=new f(l.count*h),S=new DataView(c.buffer,c.byteOffset,c.byteLength),m=d.byteStride;for(let p=0;p<l.count;p++)for(let w=0;w<h;w++){const b=y+p*m+w*g;let x;switch(l.componentType){case ue.ComponentType.FLOAT:x=S.getFloat32(b,!0);break;case ue.ComponentType.UNSIGNED_INT:x=S.getUint32(b,!0);break;case ue.ComponentType.UNSIGNED_SHORT:x=S.getUint16(b,!0);break;case ue.ComponentType.UNSIGNED_BYTE:x=S.getUint8(b);break;case ue.ComponentType.SHORT:x=S.getInt16(b,!0);break;case ue.ComponentType.BYTE:x=S.getInt8(b);break;default:throw new Error(`Unexpected componentType "${l.componentType}".`)}v[p*h+w]=x}return v}(t,e);const a=n.byteOffset+(t.byteOffset||0);return new r(n.buffer.slice(a,a+t.count*i*o))}var lr;(function(t){t[t.ARRAY_BUFFER=34962]="ARRAY_BUFFER",t[t.ELEMENT_ARRAY_BUFFER=34963]="ELEMENT_ARRAY_BUFFER"})(lr||(lr={}));class ts{constructor(e,n,s){this.C=void 0,this.jsonDoc=void 0,this.options=void 0,this.accessorIndexMap=new Map,this.animationIndexMap=new Map,this.bufferIndexMap=new Map,this.cameraIndexMap=new Map,this.skinIndexMap=new Map,this.materialIndexMap=new Map,this.meshIndexMap=new Map,this.nodeIndexMap=new Map,this.imageIndexMap=new Map,this.textureDefIndexMap=new Map,this.textureInfoDefMap=new Map,this.samplerDefIndexMap=new Map,this.sceneIndexMap=new Map,this.imageBufferViews=[],this.otherBufferViews=new Map,this.otherBufferViewsIndexMap=new Map,this.extensionData={},this.bufferURIGenerator=void 0,this.imageURIGenerator=void 0,this.logger=void 0,this.F=new Map,this.accessorUsageGroupedByParent=new Set(["ARRAY_BUFFER"]),this.accessorParents=new Map,this.C=e,this.jsonDoc=n,this.options=s;const r=e.getRoot(),i=r.listBuffers().length,o=r.listTextures().length;this.bufferURIGenerator=new Dc(i>1,()=>s.basename||"buffer"),this.imageURIGenerator=new Dc(o>1,a=>function(l,u){const c=l.getGraph().listParentEdges(u).find(d=>d.getParent()!==l.getRoot());return c?c.getName().replace(/texture$/i,""):""}(e,a)||s.basename||"texture"),this.logger=e.getLogger()}createTextureInfoDef(e,n){const s={magFilter:n.getMagFilter()||void 0,minFilter:n.getMinFilter()||void 0,wrapS:n.getWrapS(),wrapT:n.getWrapT()},r=JSON.stringify(s);this.samplerDefIndexMap.has(r)||(this.samplerDefIndexMap.set(r,this.jsonDoc.json.samplers.length),this.jsonDoc.json.samplers.push(s));const i={source:this.imageIndexMap.get(e),sampler:this.samplerDefIndexMap.get(r)},o=JSON.stringify(i);this.textureDefIndexMap.has(o)||(this.textureDefIndexMap.set(o,this.jsonDoc.json.textures.length),this.jsonDoc.json.textures.push(i));const a={index:this.textureDefIndexMap.get(o)};return n.getTexCoord()!==0&&(a.texCoord=n.getTexCoord()),Object.keys(n.getExtras()).length>0&&(a.extras=n.getExtras()),this.textureInfoDefMap.set(n,a),a}createPropertyDef(e){const n={};return e.getName()&&(n.name=e.getName()),Object.keys(e.getExtras()).length>0&&(n.extras=e.getExtras()),n}createAccessorDef(e){const n=this.createPropertyDef(e);return n.type=e.getType(),n.componentType=e.getComponentType(),n.count=e.getCount(),this.C.getGraph().listParentEdges(e).some(s=>s.getName()==="attributes"&&s.getAttributes().key==="POSITION"||s.getName()==="input")&&(n.max=e.getMax([]).map(Math.fround),n.min=e.getMin([]).map(Math.fround)),e.getNormalized()&&(n.normalized=e.getNormalized()),n}createImageData(e,n,s){if(this.options.format===Qn.GLB)this.imageBufferViews.push(n),e.bufferView=this.jsonDoc.json.bufferViews.length,this.jsonDoc.json.bufferViews.push({buffer:0,byteOffset:-1,byteLength:n.byteLength});else{const r=es.mimeTypeToExtension(s.getMimeType());e.uri=this.imageURIGenerator.createURI(s,r),this.jsonDoc.resources[e.uri]=n}}getAccessorUsage(e){const n=this.F.get(e);if(n)return n;if(e.getSparse())return yt.SPARSE;for(const s of this.C.getGraph().listParentEdges(e)){const{usage:r}=s.getAttributes();if(r)return r;s.getParent().propertyType!==ne.ROOT&&this.logger.warn(`Missing attribute ".usage" on edge, "${s.getName()}".`)}return yt.OTHER}addAccessorToUsageGroup(e,n){const s=this.F.get(e);if(s&&s!==n)throw new Error(`Accessor with usage "${s}" cannot be reused as "${n}".`);return this.F.set(e,n),this}listAccessorUsageGroups(){const e={};for(const[n,s]of Array.from(this.F.entries()))e[s]=e[s]||[],e[s].push(n);return e}}ts.BufferViewTarget=lr,ts.BufferViewUsage=yt,ts.USAGE_TO_TARGET={[yt.ARRAY_BUFFER]:lr.ARRAY_BUFFER,[yt.ELEMENT_ARRAY_BUFFER]:lr.ELEMENT_ARRAY_BUFFER};class Dc{constructor(e,n){this.multiple=void 0,this.basename=void 0,this.counter={},this.multiple=e,this.basename=n}createURI(e,n){if(e.getURI())return e.getURI();if(this.multiple){const s=this.basename(e);return this.counter[s]=this.counter[s]||1,`${s}_${this.counter[s]++}.${n}`}return`${this.basename(e)}.${n}`}}const{BufferViewUsage:ni}=ts,{UNSIGNED_INT:Rx,UNSIGNED_SHORT:Ix,UNSIGNED_BYTE:kx}=ue.ComponentType;class Fx{static write(e,n){const s=e.getRoot(),r={asset:nt({generator:`glTF-Transform ${Yh}`},s.getAsset()),extras:nt({},s.getExtras())},i={json:r,resources:{}},o=new ts(e,i,n),a=n.logger||Et.DEFAULT_INSTANCE,l=new Set(n.extensions.map(v=>v.EXTENSION_NAME)),u=e.getRoot().listExtensionsUsed().filter(v=>l.has(v.extensionName)),c=e.getRoot().listExtensionsRequired().filter(v=>l.has(v.extensionName));u.length<e.getRoot().listExtensionsUsed().length&&a.warn("Some extensions were not registered for I/O, and will not be written.");for(const v of u)for(const S of v.writeDependencies)v.install(S,n.dependencies[S]);function d(v,S,m,p){const w=[];let b=0;for(const C of v){const M=o.createAccessorDef(C);M.bufferView=r.bufferViews.length;const T=C.getArray(),V=Se.pad(Se.toView(T));M.byteOffset=b,b+=V.byteLength,w.push(V),o.accessorIndexMap.set(C,r.accessors.length),r.accessors.push(M)}const x={buffer:S,byteOffset:m,byteLength:Se.concat(w).byteLength};return p&&(x.target=p),r.bufferViews.push(x),{buffers:w,byteLength:b}}function f(v,S,m){const p=v[0].getCount();let w=0;for(const M of v){const T=o.createAccessorDef(M);T.bufferView=r.bufferViews.length,T.byteOffset=w;const V=M.getElementSize(),L=M.getComponentSize();w+=Se.padNumber(V*L),o.accessorIndexMap.set(M,r.accessors.length),r.accessors.push(T)}const b=p*w,x=new ArrayBuffer(b),C=new DataView(x);for(let M=0;M<p;M++){let T=0;for(const V of v){const L=V.getElementSize(),A=V.getComponentSize(),k=V.getComponentType(),O=V.getArray();for(let H=0;H<L;H++){const q=M*w+T+H*A,j=O[M*L+H];switch(k){case ue.ComponentType.FLOAT:C.setFloat32(q,j,!0);break;case ue.ComponentType.BYTE:C.setInt8(q,j);break;case ue.ComponentType.SHORT:C.setInt16(q,j,!0);break;case ue.ComponentType.UNSIGNED_BYTE:C.setUint8(q,j);break;case ue.ComponentType.UNSIGNED_SHORT:C.setUint16(q,j,!0);break;case ue.ComponentType.UNSIGNED_INT:C.setUint32(q,j,!0);break;default:throw new Error("Unexpected component type: "+k)}}T+=Se.padNumber(L*A)}}return r.bufferViews.push({buffer:S,byteOffset:m,byteLength:b,byteStride:w,target:ts.BufferViewTarget.ARRAY_BUFFER}),{byteLength:b,buffers:[new Uint8Array(x)]}}function h(v,S,m){const p=[];let w=0;const b=new Map;let x=-1/0;for(const k of v){const O=o.createAccessorDef(k);r.accessors.push(O),o.accessorIndexMap.set(k,r.accessors.length-1);const H=[],q=[],j=[],Y=new Array(k.getElementSize()).fill(0);for(let ge=0,me=k.getCount();ge<me;ge++)if(k.getElement(ge,j),!ze.eq(j,Y,0)){x=Math.max(ge,x),H.push(ge);for(let Ce=0;Ce<j.length;Ce++)q.push(j[Ce])}const ae=H.length,Q={accessorDef:O,count:ae};if(b.set(k,Q),ae===0)continue;if(ae>k.getCount()/3){const ge=(100*H.length/k.getCount()).toFixed(1);a.warn(`Sparse accessor with many non-zero elements (${ge}%) may increase file size.`)}const te=zi[k.getComponentType()];Q.indices=H,Q.values=new te(q)}if(!Number.isFinite(x))return{buffers:p,byteLength:w};const C=x<255?Uint8Array:x<65535?Uint16Array:Uint32Array,M=x<255?kx:x<65535?Ix:Rx,T={buffer:S,byteOffset:m+w,byteLength:0};for(const k of v){const O=b.get(k);if(O.count===0)continue;O.indicesByteOffset=T.byteLength;const H=Se.pad(Se.toView(new C(O.indices)));p.push(H),w+=H.byteLength,T.byteLength+=H.byteLength}r.bufferViews.push(T);const V=r.bufferViews.length-1,L={buffer:S,byteOffset:m+w,byteLength:0};for(const k of v){const O=b.get(k);if(O.count===0)continue;O.valuesByteOffset=L.byteLength;const H=Se.pad(Se.toView(O.values));p.push(H),w+=H.byteLength,L.byteLength+=H.byteLength}r.bufferViews.push(L);const A=r.bufferViews.length-1;for(const k of v){const O=b.get(k);O.count!==0&&(O.accessorDef.sparse={count:O.count,indices:{bufferView:V,byteOffset:O.indicesByteOffset,componentType:M},values:{bufferView:A,byteOffset:O.valuesByteOffset}})}return{buffers:p,byteLength:w}}const g=new Map;for(const v of e.getGraph().listEdges()){if(v.getParent()===s)continue;const S=v.getChild();if(S instanceof ue){const m=g.get(S)||[];m.push(v),g.set(S,m)}}if(r.accessors=[],r.bufferViews=[],r.samplers=[],r.textures=[],r.images=s.listTextures().map((v,S)=>{const m=o.createPropertyDef(v);v.getMimeType()&&(m.mimeType=v.getMimeType());const p=v.getImage();return p&&o.createImageData(m,p,v),o.imageIndexMap.set(v,S),m}),u.filter(v=>v.prewriteTypes.includes(ne.ACCESSOR)).forEach(v=>v.prewrite(o,ne.ACCESSOR)),s.listAccessors().forEach(v=>{const S=o.accessorUsageGroupedByParent,m=o.accessorParents;if(o.accessorIndexMap.has(v))return;const p=g.get(v)||[],w=o.getAccessorUsage(v);if(o.addAccessorToUsageGroup(v,w),S.has(w)){const b=p[0].getParent(),x=m.get(b)||new Set;x.add(v),m.set(b,x)}}),u.filter(v=>v.prewriteTypes.includes(ne.BUFFER)).forEach(v=>v.prewrite(o,ne.BUFFER)),(s.listAccessors().length>0||s.listTextures().length>0||o.otherBufferViews.size>0)&&s.listBuffers().length===0)throw new Error("Buffer required for Document resources, but none was found.");r.buffers=[],s.listBuffers().forEach((v,S)=>{const m=o.createPropertyDef(v),p=o.accessorUsageGroupedByParent,w=o.accessorParents,b=v.listParents().filter(L=>L instanceof ue),x=new Set(b),C=[],M=r.buffers.length;let T=0;const V=o.listAccessorUsageGroups();for(const L in V)if(p.has(L))for(const A of Array.from(w.values())){const k=Array.from(A).filter(O=>x.has(O)).filter(O=>o.getAccessorUsage(O)===L);if(k.length)if(L!==ni.ARRAY_BUFFER||n.vertexLayout===Bi.INTERLEAVED){const O=L===ni.ARRAY_BUFFER?f(k,M,T):d(k,M,T);T+=O.byteLength,C.push(...O.buffers)}else for(const O of k){const H=f([O],M,T);T+=H.byteLength,C.push(...H.buffers)}}else{const A=V[L].filter(H=>x.has(H));if(!A.length)continue;const k=L===ni.ELEMENT_ARRAY_BUFFER?ts.BufferViewTarget.ELEMENT_ARRAY_BUFFER:void 0,O=L===ni.SPARSE?h(A,M,T):d(A,M,T,k);T+=O.byteLength,C.push(...O.buffers)}if(o.imageBufferViews.length&&S===0){for(let L=0;L<o.imageBufferViews.length;L++)if(r.bufferViews[r.images[L].bufferView].byteOffset=T,T+=o.imageBufferViews[L].byteLength,C.push(o.imageBufferViews[L]),T%8){const A=8-T%8;T+=A,C.push(new Uint8Array(A))}}if(o.otherBufferViews.has(v))for(const L of o.otherBufferViews.get(v))r.bufferViews.push({buffer:M,byteOffset:T,byteLength:L.byteLength}),o.otherBufferViewsIndexMap.set(L,r.bufferViews.length-1),T+=L.byteLength,C.push(L);if(T){let L;n.format===Qn.GLB?L=Ni:(L=o.bufferURIGenerator.createURI(v,"bin"),m.uri=L),m.byteLength=T,i.resources[L]=Se.concat(C)}r.buffers.push(m),o.bufferIndexMap.set(v,S)}),s.listAccessors().find(v=>!v.getBuffer())&&a.warn("Skipped writing one or more Accessors: no Buffer assigned."),r.materials=s.listMaterials().map((v,S)=>{const m=o.createPropertyDef(v);if(v.getAlphaMode()!==ss.AlphaMode.OPAQUE&&(m.alphaMode=v.getAlphaMode()),v.getAlphaMode()===ss.AlphaMode.MASK&&(m.alphaCutoff=v.getAlphaCutoff()),v.getDoubleSided()&&(m.doubleSided=!0),m.pbrMetallicRoughness={},ze.eq(v.getBaseColorFactor(),[1,1,1,1])||(m.pbrMetallicRoughness.baseColorFactor=v.getBaseColorFactor()),ze.eq(v.getEmissiveFactor(),[0,0,0])||(m.emissiveFactor=v.getEmissiveFactor()),v.getRoughnessFactor()!==1&&(m.pbrMetallicRoughness.roughnessFactor=v.getRoughnessFactor()),v.getMetallicFactor()!==1&&(m.pbrMetallicRoughness.metallicFactor=v.getMetallicFactor()),v.getBaseColorTexture()){const p=v.getBaseColorTexture(),w=v.getBaseColorTextureInfo();m.pbrMetallicRoughness.baseColorTexture=o.createTextureInfoDef(p,w)}if(v.getEmissiveTexture()){const p=v.getEmissiveTexture(),w=v.getEmissiveTextureInfo();m.emissiveTexture=o.createTextureInfoDef(p,w)}if(v.getNormalTexture()){const p=v.getNormalTexture(),w=v.getNormalTextureInfo(),b=o.createTextureInfoDef(p,w);v.getNormalScale()!==1&&(b.scale=v.getNormalScale()),m.normalTexture=b}if(v.getOcclusionTexture()){const p=v.getOcclusionTexture(),w=v.getOcclusionTextureInfo(),b=o.createTextureInfoDef(p,w);v.getOcclusionStrength()!==1&&(b.strength=v.getOcclusionStrength()),m.occlusionTexture=b}if(v.getMetallicRoughnessTexture()){const p=v.getMetallicRoughnessTexture(),w=v.getMetallicRoughnessTextureInfo();m.pbrMetallicRoughness.metallicRoughnessTexture=o.createTextureInfoDef(p,w)}return o.materialIndexMap.set(v,S),m}),r.meshes=s.listMeshes().map((v,S)=>{const m=o.createPropertyDef(v);let p=null;return m.primitives=v.listPrimitives().map(w=>{const b={attributes:{}};b.mode=w.getMode();const x=w.getMaterial();x&&(b.material=o.materialIndexMap.get(x)),Object.keys(w.getExtras()).length&&(b.extras=w.getExtras());const C=w.getIndices();C&&(b.indices=o.accessorIndexMap.get(C));for(const M of w.listSemantics())b.attributes[M]=o.accessorIndexMap.get(w.getAttribute(M));for(const M of w.listTargets()){const T={};for(const V of M.listSemantics())T[V]=o.accessorIndexMap.get(M.getAttribute(V));b.targets=b.targets||[],b.targets.push(T)}return w.listTargets().length&&!p&&(p=w.listTargets().map(M=>M.getName())),b}),v.getWeights().length&&(m.weights=v.getWeights()),p&&(m.extras=m.extras||{},m.extras.targetNames=p),o.meshIndexMap.set(v,S),m}),r.cameras=s.listCameras().map((v,S)=>{const m=o.createPropertyDef(v);if(m.type=v.getType(),m.type===fs.Type.PERSPECTIVE){m.perspective={znear:v.getZNear(),zfar:v.getZFar(),yfov:v.getYFov()};const p=v.getAspectRatio();p!==null&&(m.perspective.aspectRatio=p)}else m.orthographic={znear:v.getZNear(),zfar:v.getZFar(),xmag:v.getXMag(),ymag:v.getYMag()};return o.cameraIndexMap.set(v,S),m}),r.nodes=s.listNodes().map((v,S)=>{const m=o.createPropertyDef(v);return ze.eq(v.getTranslation(),[0,0,0])||(m.translation=v.getTranslation()),ze.eq(v.getRotation(),[0,0,0,1])||(m.rotation=v.getRotation()),ze.eq(v.getScale(),[1,1,1])||(m.scale=v.getScale()),v.getWeights().length&&(m.weights=v.getWeights()),o.nodeIndexMap.set(v,S),m}),r.skins=s.listSkins().map((v,S)=>{const m=o.createPropertyDef(v),p=v.getInverseBindMatrices();p&&(m.inverseBindMatrices=o.accessorIndexMap.get(p));const w=v.getSkeleton();return w&&(m.skeleton=o.nodeIndexMap.get(w)),m.joints=v.listJoints().map(b=>o.nodeIndexMap.get(b)),o.skinIndexMap.set(v,S),m}),s.listNodes().forEach((v,S)=>{const m=r.nodes[S],p=v.getMesh();p&&(m.mesh=o.meshIndexMap.get(p));const w=v.getCamera();w&&(m.camera=o.cameraIndexMap.get(w));const b=v.getSkin();b&&(m.skin=o.skinIndexMap.get(b)),v.listChildren().length>0&&(m.children=v.listChildren().map(x=>o.nodeIndexMap.get(x)))}),r.animations=s.listAnimations().map((v,S)=>{const m=o.createPropertyDef(v),p=new Map;return m.samplers=v.listSamplers().map((w,b)=>{const x=o.createPropertyDef(w);return x.input=o.accessorIndexMap.get(w.getInput()),x.output=o.accessorIndexMap.get(w.getOutput()),x.interpolation=w.getInterpolation(),p.set(w,b),x}),m.channels=v.listChannels().map(w=>{const b=o.createPropertyDef(w);return b.sampler=p.get(w.getSampler()),b.target={node:o.nodeIndexMap.get(w.getTargetNode()),path:w.getTargetPath()},b}),o.animationIndexMap.set(v,S),m}),r.scenes=s.listScenes().map((v,S)=>{const m=o.createPropertyDef(v);return m.nodes=v.listChildren().map(p=>o.nodeIndexMap.get(p)),o.sceneIndexMap.set(v,S),m});const y=s.getDefaultScene();return y&&(r.scene=s.listScenes().indexOf(y)),r.extensionsUsed=u.map(v=>v.extensionName),r.extensionsRequired=c.map(v=>v.extensionName),u.forEach(v=>v.write(o)),function(v){const S=[];for(const m in v){const p=v[m];(Array.isArray(p)&&p.length===0||p===null||p===""||p&&typeof p=="object"&&Object.keys(p).length===0)&&S.push(m)}for(const m of S)delete v[m]}(r),i}}var $i;(function(t){t[t.JSON=1313821514]="JSON",t[t.BIN=5130562]="BIN"})($i||($i={}));class Vx{constructor(){this.M=Et.DEFAULT_INSTANCE,this.l=new Set,this.U={},this.P=Bi.INTERLEAVED,this.lastReadBytes=0,this.lastWriteBytes=0}setLogger(e){return this.M=e,this}registerExtensions(e){for(const n of e)this.l.add(n),n.register();return this}registerDependencies(e){return Object.assign(this.U,e),this}setVertexLayout(e){return this.P=e,this}async read(e){return await this.readJSON(await this.readAsJSON(e))}async readAsJSON(e){const n=await this.readURI(e,"view");this.lastReadBytes=n.byteLength;const s=$c(n)?this.L(n):{json:JSON.parse(Se.decodeText(n)),resources:{}};return await this.j(s,this.dirname(e)),this.D(s),s}async readJSON(e){return e=this._(e),this.D(e),Ax.read(e,{extensions:Array.from(this.l),dependencies:this.U,logger:this.M})}async binaryToJSON(e){const n=this.L(Se.assertView(e));this.D(n);const s=n.json;if(s.buffers&&s.buffers.some(r=>function(i,o){return o.uri!==void 0&&!(o.uri in i.resources)}(n,r)))throw new Error("Cannot resolve external buffers with binaryToJSON().");if(s.images&&s.images.some(r=>function(i,o){return o.uri!==void 0&&!(o.uri in i.resources)&&o.bufferView===void 0}(n,r)))throw new Error("Cannot resolve external images with binaryToJSON().");return n}async readBinary(e){return this.readJSON(await this.binaryToJSON(Se.assertView(e)))}async writeJSON(e,n={}){if(n.format===Qn.GLB&&e.getRoot().listBuffers().length>1)throw new Error("GLB must have 0–1 buffers.");return Fx.write(e,{format:n.format||Qn.GLTF,basename:n.basename||"",logger:this.M,vertexLayout:this.P,dependencies:nt({},this.U),extensions:Array.from(this.l)})}async writeBinary(e){const{json:n,resources:s}=await this.writeJSON(e,{format:Qn.GLB}),r=new Uint32Array([1179937895,2,12]),i=JSON.stringify(n),o=Se.pad(Se.encodeText(i),32),a=Se.toView(new Uint32Array([o.byteLength,1313821514])),l=Se.concat([a,o]);r[r.length-1]+=l.byteLength;const u=Object.values(s)[0];if(!u||!u.byteLength)return Se.concat([Se.toView(r),l]);const c=Se.pad(u,0),d=Se.toView(new Uint32Array([c.byteLength,5130562])),f=Se.concat([d,c]);return r[r.length-1]+=f.byteLength,Se.concat([Se.toView(r),l,f])}async j(e,n){var s=this;const r=[...e.json.images||[],...e.json.buffers||[]].map(async function(i){const o=i.uri;if(!o||o.match(/data:/))return Promise.resolve();e.resources[o]=await s.readURI(s.resolve(n,o),"view"),s.lastReadBytes+=e.resources[o].byteLength});await Promise.all(r)}D(e){function n(s){if(s.uri){if(s.uri in e.resources)Se.assertView(e.resources[s.uri]);else if(s.uri.match(/data:/)){const r=`__${wx()}.${Ps.extension(s.uri)}`;e.resources[r]=Se.createBufferFromDataURI(s.uri),s.uri=r}}}(e.json.images||[]).forEach(s=>{if(s.bufferView===void 0&&s.uri===void 0)throw new Error("Missing resource URI or buffer view.");n(s)}),(e.json.buffers||[]).forEach(n)}_(e){const{images:n,buffers:s}=e.json;return e={json:nt({},e.json),resources:nt({},e.resources)},n&&(e.json.images=n.map(r=>nt({},r))),s&&(e.json.buffers=s.map(r=>nt({},r))),e}L(e){if(!$c(e))throw new Error("Invalid glTF 2.0 binary.");const n=new Uint32Array(e.buffer,e.byteOffset+12,2);if(n[1]!==$i.JSON)throw new Error("Missing required GLB JSON chunk.");const s=n[0],r=Se.decodeText(Se.toView(e,20,s)),i=JSON.parse(r),o=20+s;if(e.byteLength<=o)return{json:i,resources:{}};const a=new Uint32Array(e.buffer,e.byteOffset+o,2);if(a[1]!==$i.BIN)throw new Error("Expected GLB BIN in second chunk.");const l=Se.toView(e,o+8,a[0]);return{json:i,resources:{[Ni]:l}}}}function $c(t){if(t.byteLength<3*Uint32Array.BYTES_PER_ELEMENT)return!1;const e=new Uint32Array(t.buffer,t.byteOffset,3);return e[0]===1179937895&&e[1]===2}class Lx extends Vx{constructor(e=ar.DEFAULT_INIT){super(),this.J=void 0,this.J=e}async readURI(e,n){const s=await fetch(e,this.J);switch(n){case"view":return new Uint8Array(await s.arrayBuffer());case"text":return s.text()}}resolve(e,n){return ar.resolve(e,n)}dirname(e){return ar.dirname(e)}}function Hi(){return Hi=Object.assign?Object.assign.bind():function(t){for(var e=1;e<arguments.length;e++){var n=arguments[e];for(var s in n)Object.prototype.hasOwnProperty.call(n,s)&&(t[s]=n[s])}return t},Hi.apply(this,arguments)}function Px(t,e){return Object.defineProperty(e,"name",{value:t}),e}var Ui=typeof Float32Array<"u"?Float32Array:Array;function Ox(){var t=new Ui(3);return Ui!=Float32Array&&(t[0]=0,t[1]=0,t[2]=0),t}Math.hypot||(Math.hypot=function(){for(var t=0,e=arguments.length;e--;)t+=arguments[e]*arguments[e];return Math.sqrt(t)});function Nx(){var t=new Ui(4);return Ui!=Float32Array&&(t[0]=0,t[1]=0,t[2]=0,t[3]=0),t}Ox();Nx();ne.ACCESSOR,ne.MESH,ne.TEXTURE,ne.MATERIAL,ne.SKIN;ne.NODE,ne.SKIN,ne.MESH,ne.CAMERA,ne.PRIMITIVE,ne.PRIMITIVE_TARGET,ne.ANIMATION,ne.MATERIAL,ne.TEXTURE,ne.ACCESSOR,ne.BUFFER;Ml.TargetPath;const Bx={pattern:/.*/,quantizationVolume:"mesh",quantizePosition:14,quantizeNormal:10,quantizeTexcoord:12,quantizeColor:8,quantizeWeight:8,quantizeGeneric:12,normalizeWeights:!0};Hi({level:"high"},Bx);var Hc;function js(t,e,n){for(let s=0,r=n.length;s<r;s++)n[s]=t[e*r+s];return n}function Uc(t,e,n){for(let s=0,r=n.length;s<r;s++)t[e*r+s]=n[s]}function si(t,e,n=0){if(t.length!==e.length)return!1;for(let s=0;s<t.length;s++)if(Math.abs(t[s]-e[s])>n)return!1;return!0}function zx(t,e,n){return t*(1-n)+e*n}function Dx(t,e,n,s){for(let r=0;r<e.length;r++)t[r]=zx(e[r],n[r],s);return t}function $x(t,e,n,s){let r,i,o,a,l,u=e[0],c=e[1],d=e[2],f=e[3],h=n[0],g=n[1],y=n[2],v=n[3];return i=u*h+c*g+d*y+f*v,i<0&&(i=-i,h=-h,g=-g,y=-y,v=-v),1-i>1e-6?(r=Math.acos(i),o=Math.sin(r),a=Math.sin((1-s)*r)/o,l=Math.sin(s*r)/o):(a=1-s,l=s),t[0]=a*u+l*h,t[1]=a*c+l*g,t[2]=a*d+l*y,t[3]=a*f+l*v,t}function jc(t,e){const n=function(s,r){return s[0]*r[0]+s[1]*r[1]+s[2]*r[2]+s[3]*r[3]}(t,e);return Math.acos(2*n*n-1)}(function(t){t[t.STEP=0]="STEP",t[t.LERP=1]="LERP",t[t.SLERP=2]="SLERP"})(Hc||(Hc={}));Promise.resolve();var ji;(function(t){t.LANCZOS3="lanczos3",t.LANCZOS2="lanczos2"})(ji||(ji={}));ji.LANCZOS3;ji.LANCZOS3;const Wc="unpartition",Gc={};function Hx(t=Gc){return Hi({},Gc,t),Px(Wc,async e=>{const n=e.getLogger(),s=e.getRoot().listBuffers()[0];e.getRoot().listAccessors().forEach(r=>r.setBuffer(s)),e.getRoot().listBuffers().forEach((r,i)=>i>0?r.dispose():null),n.debug(`${Wc}: Complete.`)})}let nm=new Lx,Ht="__yacv_name",gi="__helpers";async function Ux(t,e,n){let s=await nm.read(t);return await n.transform(sm(e)),await s.transform(Gx(e)),n.merge(s)}async function jx(t){return await t.transform(qx(),Hx())}async function qc(t){return nm.writeBinary(t)}async function Wx(t,e){return await e.transform(sm(t))}function Gx(t){return e=>{for(let n of e.getGraph().listEdges().map(s=>s.getChild()))n.getExtras()||n.setExtras({}),n.getExtras()[Ht]=t;return e}}function sm(t){return e=>{var n;for(let s of e.getGraph().listEdges().map(r=>r.getChild()))s.getExtras()==null||s instanceof Al||s instanceof Tl||(((n=s.getExtras()[Ht])==null?void 0:n.toString())??"")==t&&s.dispose();return e}}function qx(){return t=>{let e=t.getRoot(),n=e.getDefaultScene()??e.listScenes()[0];for(let s of e.listScenes())if(s!==n){for(let r of s.listChildren())n.addChild(r);s.dispose()}}}class Rl{constructor(e=0,n=0){Rl.prototype.isVector2=!0,this.x=e,this.y=n}get width(){return this.x}set width(e){this.x=e}get height(){return this.y}set height(e){this.y=e}set(e,n){return this.x=e,this.y=n,this}setScalar(e){return this.x=e,this.y=e,this}setX(e){return this.x=e,this}setY(e){return this.y=e,this}setComponent(e,n){switch(e){case 0:this.x=n;break;case 1:this.y=n;break;default:throw new Error("index is out of range: "+e)}return this}getComponent(e){switch(e){case 0:return this.x;case 1:return this.y;default:throw new Error("index is out of range: "+e)}}clone(){return new this.constructor(this.x,this.y)}copy(e){return this.x=e.x,this.y=e.y,this}add(e){return this.x+=e.x,this.y+=e.y,this}addScalar(e){return this.x+=e,this.y+=e,this}addVectors(e,n){return this.x=e.x+n.x,this.y=e.y+n.y,this}addScaledVector(e,n){return this.x+=e.x*n,this.y+=e.y*n,this}sub(e){return this.x-=e.x,this.y-=e.y,this}subScalar(e){return this.x-=e,this.y-=e,this}subVectors(e,n){return this.x=e.x-n.x,this.y=e.y-n.y,this}multiply(e){return this.x*=e.x,this.y*=e.y,this}multiplyScalar(e){return this.x*=e,this.y*=e,this}divide(e){return this.x/=e.x,this.y/=e.y,this}divideScalar(e){return this.multiplyScalar(1/e)}applyMatrix3(e){const n=this.x,s=this.y,r=e.elements;return this.x=r[0]*n+r[3]*s+r[6],this.y=r[1]*n+r[4]*s+r[7],this}min(e){return this.x=Math.min(this.x,e.x),this.y=Math.min(this.y,e.y),this}max(e){return this.x=Math.max(this.x,e.x),this.y=Math.max(this.y,e.y),this}clamp(e,n){return this.x=Math.max(e.x,Math.min(n.x,this.x)),this.y=Math.max(e.y,Math.min(n.y,this.y)),this}clampScalar(e,n){return this.x=Math.max(e,Math.min(n,this.x)),this.y=Math.max(e,Math.min(n,this.y)),this}clampLength(e,n){const s=this.length();return this.divideScalar(s||1).multiplyScalar(Math.max(e,Math.min(n,s)))}floor(){return this.x=Math.floor(this.x),this.y=Math.floor(this.y),this}ceil(){return this.x=Math.ceil(this.x),this.y=Math.ceil(this.y),this}round(){return this.x=Math.round(this.x),this.y=Math.round(this.y),this}roundToZero(){return this.x=Math.trunc(this.x),this.y=Math.trunc(this.y),this}negate(){return this.x=-this.x,this.y=-this.y,this}dot(e){return this.x*e.x+this.y*e.y}cross(e){return this.x*e.y-this.y*e.x}lengthSq(){return this.x*this.x+this.y*this.y}length(){return Math.sqrt(this.x*this.x+this.y*this.y)}manhattanLength(){return Math.abs(this.x)+Math.abs(this.y)}normalize(){return this.divideScalar(this.length()||1)}angle(){return Math.atan2(-this.y,-this.x)+Math.PI}angleTo(e){const n=Math.sqrt(this.lengthSq()*e.lengthSq());if(n===0)return Math.PI/2;const s=this.dot(e)/n;return Math.acos(ot(s,-1,1))}distanceTo(e){return Math.sqrt(this.distanceToSquared(e))}distanceToSquared(e){const n=this.x-e.x,s=this.y-e.y;return n*n+s*s}manhattanDistanceTo(e){return Math.abs(this.x-e.x)+Math.abs(this.y-e.y)}setLength(e){return this.normalize().multiplyScalar(e)}lerp(e,n){return this.x+=(e.x-this.x)*n,this.y+=(e.y-this.y)*n,this}lerpVectors(e,n,s){return this.x=e.x+(n.x-e.x)*s,this.y=e.y+(n.y-e.y)*s,this}equals(e){return e.x===this.x&&e.y===this.y}fromArray(e,n=0){return this.x=e[n],this.y=e[n+1],this}toArray(e=[],n=0){return e[n]=this.x,e[n+1]=this.y,e}fromBufferAttribute(e,n){return this.x=e.getX(n),this.y=e.getY(n),this}rotateAround(e,n){const s=Math.cos(n),r=Math.sin(n),i=this.x-e.x,o=this.y-e.y;return this.x=i*s-o*r+e.x,this.y=i*r+o*s+e.y,this}random(){return this.x=Math.random(),this.y=Math.random(),this}*[Symbol.iterator](){yield this.x,yield this.y}}const bs={x:[[247,60,60],[148,36,36]],z:[[108,203,38],[65,122,23]],y:[[23,140,240],[14,84,144]]};function Js(t,e,n,s,r,i="__helper",o=WebGL2RenderingContext.LINES){const a=t.getRoot().listBuffers()[0]??t.createBuffer(i+"Buffer"),l=t.getRoot().getDefaultScene()??t.getRoot().listScenes()[0]??t.createScene(i+"Scene"),u=t.createAccessor(i+"Position").setArray(new Float32Array(e)).setType("VEC3").setBuffer(a),c=t.createAccessor(i+"Indices").setArray(new Uint32Array(n)).setType("SCALAR").setBuffer(a);let d=null;s&&(d=t.createAccessor(i+"Color").setArray(new Float32Array(s)).setType("VEC3").setBuffer(a));const f=t.createMaterial(i+"Material").setAlphaMode("OPAQUE"),h=t.createPrimitive().setIndices(c).setAttribute("POSITION",u).setMode(o).setMaterial(f);s&&h.setAttribute("COLOR_0",d);const g=t.createMesh(i+"Mesh").addPrimitive(h),y=t.createNode(i+"Node").setMesh(g).setMatrix(r.elements);l.addChild(y)}function Yx(t,e,n){let s=[[0,0,0,e.x,0,0],[0,0,0,0,e.y,0],[0,0,0,0,0,-e.z]],r=[0,1],i=[[...bs.x[0],...bs.x[1]],[...bs.y[0],...bs.y[1]],[...bs.z[0],...bs.z[1]]].map(o=>o.map(a=>a/255));Js(t,s[0],r,i[0],n,"__helper_axes"),Js(t,s[1],r,i[1],n,"__helper_axes"),Js(t,s[2],r,i[2],n,"__helper_axes"),Js(t,[0,0,0],[0],null,n,"__helper_axes",WebGL2RenderingContext.POINTS)}function Xx(t,e,n=new He,s=10){for(let r of[new J(1,0,0),new J(0,1,0),new J(0,0,-1)])for(let i of[1,-1]){let o=r.clone().multiply(e.clone().multiplyScalar(.5*i)),a=new He().makeTranslation(o.x,o.y,o.z),l=new He().lookAt(new J,o,new J(0,1,0)),u=new Rl;r.x&&u.set(e.z,e.y),r.y&&u.set(e.x,e.z),r.z&&u.set(e.x,e.y);let c=n.clone().multiply(a).multiply(l);Zx(t,u,c,s)}}function Zx(t,e,n=new He,s=10,r=.002){const i=[],o=[];for(let a=0;a<=s;a++){const l=-e.x/2+e.x*a/s,u=-e.y/2+e.y*a/s;i.push(l-r*e.x/2,-e.y/2,0),i.push(l+r*e.x/2,-e.y/2,0),i.push(l+r*e.x/2,e.y/2,0),i.push(l-r*e.x/2,e.y/2,0);const c=a*4;o.push(c,c+1,c+2),o.push(c,c+2,c+3),i.push(-e.x/2,u-r*e.y/2,0),i.push(e.x/2,u-r*e.y/2,0),i.push(e.x/2,u+r*e.y/2,0),i.push(-e.x/2,u+r*e.y/2,0);const d=(s+1+a)*4;o.push(d,d+1,d+2),o.push(d,d+2,d+3)}Js(t,i,o,null,n,"__helper_grid",WebGL2RenderingContext.TRIANGLES)}class rm{constructor(e=new J(1/0,1/0,1/0),n=new J(-1/0,-1/0,-1/0)){this.isBox3=!0,this.min=e,this.max=n}set(e,n){return this.min.copy(e),this.max.copy(n),this}setFromArray(e){this.makeEmpty();for(let n=0,s=e.length;n<s;n+=3)this.expandByPoint(Vt.fromArray(e,n));return this}setFromBufferAttribute(e){this.makeEmpty();for(let n=0,s=e.count;n<s;n++)this.expandByPoint(Vt.fromBufferAttribute(e,n));return this}setFromPoints(e){this.makeEmpty();for(let n=0,s=e.length;n<s;n++)this.expandByPoint(e[n]);return this}setFromCenterAndSize(e,n){const s=Vt.copy(n).multiplyScalar(.5);return this.min.copy(e).sub(s),this.max.copy(e).add(s),this}setFromObject(e,n=!1){return this.makeEmpty(),this.expandByObject(e,n)}clone(){return new this.constructor().copy(this)}copy(e){return this.min.copy(e.min),this.max.copy(e.max),this}makeEmpty(){return this.min.x=this.min.y=this.min.z=1/0,this.max.x=this.max.y=this.max.z=-1/0,this}isEmpty(){return this.max.x<this.min.x||this.max.y<this.min.y||this.max.z<this.min.z}getCenter(e){return this.isEmpty()?e.set(0,0,0):e.addVectors(this.min,this.max).multiplyScalar(.5)}getSize(e){return this.isEmpty()?e.set(0,0,0):e.subVectors(this.max,this.min)}expandByPoint(e){return this.min.min(e),this.max.max(e),this}expandByVector(e){return this.min.sub(e),this.max.add(e),this}expandByScalar(e){return this.min.addScalar(-e),this.max.addScalar(e),this}expandByObject(e,n=!1){e.updateWorldMatrix(!1,!1);const s=e.geometry;if(s!==void 0){const i=s.getAttribute("position");if(n===!0&&i!==void 0&&e.isInstancedMesh!==!0)for(let o=0,a=i.count;o<a;o++)e.isMesh===!0?e.getVertexPosition(o,Vt):Vt.fromBufferAttribute(i,o),Vt.applyMatrix4(e.matrixWorld),this.expandByPoint(Vt);else e.boundingBox!==void 0?(e.boundingBox===null&&e.computeBoundingBox(),ri.copy(e.boundingBox)):(s.boundingBox===null&&s.computeBoundingBox(),ri.copy(s.boundingBox)),ri.applyMatrix4(e.matrixWorld),this.union(ri)}const r=e.children;for(let i=0,o=r.length;i<o;i++)this.expandByObject(r[i],n);return this}containsPoint(e){return!(e.x<this.min.x||e.x>this.max.x||e.y<this.min.y||e.y>this.max.y||e.z<this.min.z||e.z>this.max.z)}containsBox(e){return this.min.x<=e.min.x&&e.max.x<=this.max.x&&this.min.y<=e.min.y&&e.max.y<=this.max.y&&this.min.z<=e.min.z&&e.max.z<=this.max.z}getParameter(e,n){return n.set((e.x-this.min.x)/(this.max.x-this.min.x),(e.y-this.min.y)/(this.max.y-this.min.y),(e.z-this.min.z)/(this.max.z-this.min.z))}intersectsBox(e){return!(e.max.x<this.min.x||e.min.x>this.max.x||e.max.y<this.min.y||e.min.y>this.max.y||e.max.z<this.min.z||e.min.z>this.max.z)}intersectsSphere(e){return this.clampPoint(e.center,Vt),Vt.distanceToSquared(e.center)<=e.radius*e.radius}intersectsPlane(e){let n,s;return e.normal.x>0?(n=e.normal.x*this.min.x,s=e.normal.x*this.max.x):(n=e.normal.x*this.max.x,s=e.normal.x*this.min.x),e.normal.y>0?(n+=e.normal.y*this.min.y,s+=e.normal.y*this.max.y):(n+=e.normal.y*this.max.y,s+=e.normal.y*this.min.y),e.normal.z>0?(n+=e.normal.z*this.min.z,s+=e.normal.z*this.max.z):(n+=e.normal.z*this.max.z,s+=e.normal.z*this.min.z),n<=-e.constant&&s>=-e.constant}intersectsTriangle(e){if(this.isEmpty())return!1;this.getCenter(Ws),ii.subVectors(this.max,Ws),xs.subVectors(e.a,Ws),ws.subVectors(e.b,Ws),_s.subVectors(e.c,Ws),bn.subVectors(ws,xs),xn.subVectors(_s,ws),Un.subVectors(xs,_s);let n=[0,-bn.z,bn.y,0,-xn.z,xn.y,0,-Un.z,Un.y,bn.z,0,-bn.x,xn.z,0,-xn.x,Un.z,0,-Un.x,-bn.y,bn.x,0,-xn.y,xn.x,0,-Un.y,Un.x,0];return!Yo(n,xs,ws,_s,ii)||(n=[1,0,0,0,1,0,0,0,1],!Yo(n,xs,ws,_s,ii))?!1:(oi.crossVectors(bn,xn),n=[oi.x,oi.y,oi.z],Yo(n,xs,ws,_s,ii))}clampPoint(e,n){return n.copy(e).clamp(this.min,this.max)}distanceToPoint(e){return this.clampPoint(e,Vt).distanceTo(e)}getBoundingSphere(e){return this.isEmpty()?e.makeEmpty():(this.getCenter(e.center),e.radius=this.getSize(Vt).length()*.5),e}intersect(e){return this.min.max(e.min),this.max.min(e.max),this.isEmpty()&&this.makeEmpty(),this}union(e){return this.min.min(e.min),this.max.max(e.max),this}applyMatrix4(e){return this.isEmpty()?this:(tn[0].set(this.min.x,this.min.y,this.min.z).applyMatrix4(e),tn[1].set(this.min.x,this.min.y,this.max.z).applyMatrix4(e),tn[2].set(this.min.x,this.max.y,this.min.z).applyMatrix4(e),tn[3].set(this.min.x,this.max.y,this.max.z).applyMatrix4(e),tn[4].set(this.max.x,this.min.y,this.min.z).applyMatrix4(e),tn[5].set(this.max.x,this.min.y,this.max.z).applyMatrix4(e),tn[6].set(this.max.x,this.max.y,this.min.z).applyMatrix4(e),tn[7].set(this.max.x,this.max.y,this.max.z).applyMatrix4(e),this.setFromPoints(tn),this)}translate(e){return this.min.add(e),this.max.add(e),this}equals(e){return e.min.equals(this.min)&&e.max.equals(this.max)}}const tn=[new J,new J,new J,new J,new J,new J,new J,new J],Vt=new J,ri=new rm,xs=new J,ws=new J,_s=new J,bn=new J,xn=new J,Un=new J,Ws=new J,ii=new J,oi=new J,jn=new J;function Yo(t,e,n,s,r){for(let i=0,o=t.length-3;i<=o;i+=3){jn.fromArray(t,i);const a=r.x*Math.abs(jn.x)+r.y*Math.abs(jn.y)+r.z*Math.abs(jn.z),l=e.dot(jn),u=n.dot(jn),c=s.dot(jn);if(Math.max(-Math.max(l,u,c),Math.min(l,u,c))>a)return!1}return!0}class ns{static async loadModel(e,n,s,r,i=!0,o=!0){let a=performance.now();if(n=await Ux(r,s,n),console.log("Model",s,"loaded in",performance.now()-a,"ms"),i&&(await this.reloadHelpers(e,n,o),o=!1),o){let l=performance.now();n=await this.showCurrentDoc(e,n),console.log("Scene displayed in",performance.now()-l,"ms")}return n}static async reloadHelpers(e,n,s){let r=ns.getBoundingBox(n);if(!r)return n;let i=new Fn,o=new He().makeTranslation(r.getCenter(new J));Yx(i,r.getSize(new J).multiplyScalar(.5),o),Xx(i,r.getSize(new J),o);let a=URL.createObjectURL(new Blob([await qc(i)]));return await ns.loadModel(e,n,gi,a,!1,s)}static getBoundingBox(e){if(e.getRoot().listNodes().length===0)return null;let n=[1e6,1e6,1e6],s=[-1e6,-1e6,-1e6];e.getRoot().listNodes().forEach(o=>{var l;if((o.getExtras()[Ht]??gi)===gi)return;let a=new He(...o.getWorldMatrix());for(let u of((l=o.getMesh())==null?void 0:l.listPrimitives())??[]){let c=u.getAttribute("POSITION");if(!c)continue;let d=new J(...c.getMin([0,0,0])).applyMatrix4(a),f=new J(...c.getMax([0,0,0])).applyMatrix4(a);n=n.map((h,g)=>Math.min(h,d.getComponent(g))),s=s.map((h,g)=>Math.max(h,f.getComponent(g)))}});let r=new J().fromArray(s).sub(new J().fromArray(n)),i=new J().fromArray(n).add(r.clone().multiplyScalar(.5));return new rm().setFromCenterAndSize(i,r)}static async removeModel(e,n,s,r=!0,i=!0){let o=performance.now();return n=await Wx(s,n),console.log("Model",s,"removed in",performance.now()-o,"ms"),r&&(n=await this.reloadHelpers(e,n,i)),n}static async showCurrentDoc(e,n){n=await jx(n);let s=await qc(n),r=new Blob([s],{type:"model/gltf-binary"});return console.debug("Showing current doc",n,"as",Array.from(s)),e.value=URL.createObjectURL(r),n}}const Yc=new ds().set(.8224621,.177538,0,.0331941,.9668058,0,.0170827,.0723974,.9105199),Xc=new ds().set(1.2249401,-.2249404,0,-.0420569,1.0420571,0,-.0196376,-.0786361,1.0982735),ai={[_l]:{transfer:Ia,primaries:Tc,toReference:t=>t,fromReference:t=>t},[Sn]:{transfer:Mc,primaries:Tc,toReference:t=>t.convertSRGBToLinear(),fromReference:t=>t.convertLinearToSRGB()},[qh]:{transfer:Ia,primaries:Ac,toReference:t=>t.applyMatrix3(Xc),fromReference:t=>t.applyMatrix3(Yc)},[qb]:{transfer:Mc,primaries:Ac,toReference:t=>t.convertSRGBToLinear().applyMatrix3(Xc),fromReference:t=>t.applyMatrix3(Yc).convertLinearToSRGB()}},Kx=new Set([_l,qh]),Lt={enabled:!0,_workingColorSpace:_l,get workingColorSpace(){return this._workingColorSpace},set workingColorSpace(t){if(!Kx.has(t))throw new Error(`Unsupported working color space, "${t}".`);this._workingColorSpace=t},convert:function(t,e,n){if(this.enabled===!1||e===n||!e||!n)return t;const s=ai[e].toReference,r=ai[n].fromReference;return r(s(t))},fromWorkingColorSpace:function(t,e){return this.convert(t,this._workingColorSpace,e)},toWorkingColorSpace:function(t,e){return this.convert(t,e,this._workingColorSpace)},getPrimaries:function(t){return ai[t].primaries},getTransfer:function(t){return t===Gb?Ia:ai[t].transfer}};function Xo(t){return t<.04045?t*.0773993808:Math.pow(t*.9478672986+.0521327014,2.4)}function Zo(t){return t<.0031308?t*12.92:1.055*Math.pow(t,.41666)-.055}const im={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074},wn={h:0,s:0,l:0},li={h:0,s:0,l:0};function Ko(t,e,n){return n<0&&(n+=1),n>1&&(n-=1),n<1/6?t+(e-t)*6*n:n<1/2?e:n<2/3?t+(e-t)*6*(2/3-n):t}class Il{constructor(e,n,s){return this.isColor=!0,this.r=1,this.g=1,this.b=1,this.set(e,n,s)}set(e,n,s){if(n===void 0&&s===void 0){const r=e;r&&r.isColor?this.copy(r):typeof r=="number"?this.setHex(r):typeof r=="string"&&this.setStyle(r)}else this.setRGB(e,n,s);return this}setScalar(e){return this.r=e,this.g=e,this.b=e,this}setHex(e,n=Sn){return e=Math.floor(e),this.r=(e>>16&255)/255,this.g=(e>>8&255)/255,this.b=(e&255)/255,Lt.toWorkingColorSpace(this,n),this}setRGB(e,n,s,r=Lt.workingColorSpace){return this.r=e,this.g=n,this.b=s,Lt.toWorkingColorSpace(this,r),this}setHSL(e,n,s,r=Lt.workingColorSpace){if(e=Ub(e,1),n=ot(n,0,1),s=ot(s,0,1),n===0)this.r=this.g=this.b=s;else{const i=s<=.5?s*(1+n):s+n-s*n,o=2*s-i;this.r=Ko(o,i,e+1/3),this.g=Ko(o,i,e),this.b=Ko(o,i,e-1/3)}return Lt.toWorkingColorSpace(this,r),this}setStyle(e,n=Sn){function s(i){i!==void 0&&parseFloat(i)<1&&console.warn("THREE.Color: Alpha component of "+e+" will be ignored.")}let r;if(r=/^(\w+)\(([^\)]*)\)/.exec(e)){let i;const o=r[1],a=r[2];switch(o){case"rgb":case"rgba":if(i=/^\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*(?:,\s*(\d*\.?\d+)\s*)?$/.exec(a))return s(i[4]),this.setRGB(Math.min(255,parseInt(i[1],10))/255,Math.min(255,parseInt(i[2],10))/255,Math.min(255,parseInt(i[3],10))/255,n);if(i=/^\s*(\d+)\%\s*,\s*(\d+)\%\s*,\s*(\d+)\%\s*(?:,\s*(\d*\.?\d+)\s*)?$/.exec(a))return s(i[4]),this.setRGB(Math.min(100,parseInt(i[1],10))/100,Math.min(100,parseInt(i[2],10))/100,Math.min(100,parseInt(i[3],10))/100,n);break;case"hsl":case"hsla":if(i=/^\s*(\d*\.?\d+)\s*,\s*(\d*\.?\d+)\%\s*,\s*(\d*\.?\d+)\%\s*(?:,\s*(\d*\.?\d+)\s*)?$/.exec(a))return s(i[4]),this.setHSL(parseFloat(i[1])/360,parseFloat(i[2])/100,parseFloat(i[3])/100,n);break;default:console.warn("THREE.Color: Unknown color model "+e)}}else if(r=/^\#([A-Fa-f\d]+)$/.exec(e)){const i=r[1],o=i.length;if(o===3)return this.setRGB(parseInt(i.charAt(0),16)/15,parseInt(i.charAt(1),16)/15,parseInt(i.charAt(2),16)/15,n);if(o===6)return this.setHex(parseInt(i,16),n);console.warn("THREE.Color: Invalid hex color "+e)}else if(e&&e.length>0)return this.setColorName(e,n);return this}setColorName(e,n=Sn){const s=im[e.toLowerCase()];return s!==void 0?this.setHex(s,n):console.warn("THREE.Color: Unknown color "+e),this}clone(){return new this.constructor(this.r,this.g,this.b)}copy(e){return this.r=e.r,this.g=e.g,this.b=e.b,this}copySRGBToLinear(e){return this.r=Xo(e.r),this.g=Xo(e.g),this.b=Xo(e.b),this}copyLinearToSRGB(e){return this.r=Zo(e.r),this.g=Zo(e.g),this.b=Zo(e.b),this}convertSRGBToLinear(){return this.copySRGBToLinear(this),this}convertLinearToSRGB(){return this.copyLinearToSRGB(this),this}getHex(e=Sn){return Lt.fromWorkingColorSpace(Ye.copy(this),e),Math.round(ot(Ye.r*255,0,255))*65536+Math.round(ot(Ye.g*255,0,255))*256+Math.round(ot(Ye.b*255,0,255))}getHexString(e=Sn){return("000000"+this.getHex(e).toString(16)).slice(-6)}getHSL(e,n=Lt.workingColorSpace){Lt.fromWorkingColorSpace(Ye.copy(this),n);const s=Ye.r,r=Ye.g,i=Ye.b,o=Math.max(s,r,i),a=Math.min(s,r,i);let l,u;const c=(a+o)/2;if(a===o)l=0,u=0;else{const d=o-a;switch(u=c<=.5?d/(o+a):d/(2-o-a),o){case s:l=(r-i)/d+(r<i?6:0);break;case r:l=(i-s)/d+2;break;case i:l=(s-r)/d+4;break}l/=6}return e.h=l,e.s=u,e.l=c,e}getRGB(e,n=Lt.workingColorSpace){return Lt.fromWorkingColorSpace(Ye.copy(this),n),e.r=Ye.r,e.g=Ye.g,e.b=Ye.b,e}getStyle(e=Sn){Lt.fromWorkingColorSpace(Ye.copy(this),e);const n=Ye.r,s=Ye.g,r=Ye.b;return e!==Sn?`color(${e} ${n.toFixed(3)} ${s.toFixed(3)} ${r.toFixed(3)})`:`rgb(${Math.round(n*255)},${Math.round(s*255)},${Math.round(r*255)})`}offsetHSL(e,n,s){return this.getHSL(wn),this.setHSL(wn.h+e,wn.s+n,wn.l+s)}add(e){return this.r+=e.r,this.g+=e.g,this.b+=e.b,this}addColors(e,n){return this.r=e.r+n.r,this.g=e.g+n.g,this.b=e.b+n.b,this}addScalar(e){return this.r+=e,this.g+=e,this.b+=e,this}sub(e){return this.r=Math.max(0,this.r-e.r),this.g=Math.max(0,this.g-e.g),this.b=Math.max(0,this.b-e.b),this}multiply(e){return this.r*=e.r,this.g*=e.g,this.b*=e.b,this}multiplyScalar(e){return this.r*=e,this.g*=e,this.b*=e,this}lerp(e,n){return this.r+=(e.r-this.r)*n,this.g+=(e.g-this.g)*n,this.b+=(e.b-this.b)*n,this}lerpColors(e,n,s){return this.r=e.r+(n.r-e.r)*s,this.g=e.g+(n.g-e.g)*s,this.b=e.b+(n.b-e.b)*s,this}lerpHSL(e,n){this.getHSL(wn),e.getHSL(li);const s=Uo(wn.h,li.h,n),r=Uo(wn.s,li.s,n),i=Uo(wn.l,li.l,n);return this.setHSL(s,r,i),this}setFromVector3(e){return this.r=e.x,this.g=e.y,this.b=e.z,this}applyMatrix3(e){const n=this.r,s=this.g,r=this.b,i=e.elements;return this.r=i[0]*n+i[3]*s+i[6]*r,this.g=i[1]*n+i[4]*s+i[7]*r,this.b=i[2]*n+i[5]*s+i[8]*r,this}equals(e){return e.r===this.r&&e.g===this.g&&e.b===this.b}fromArray(e,n=0){return this.r=e[n],this.g=e[n+1],this.b=e[n+2],this}toArray(e=[],n=0){return e[n]=this.r,e[n+1]=this.g,e[n+2]=this.b,e}fromBufferAttribute(e,n){return this.r=e.getX(n),this.g=e.getY(n),this.b=e.getZ(n),this}toJSON(){return this.getHex()}*[Symbol.iterator](){yield this.r,yield this.g,yield this.b}}const Ye=new Il;Il.NAMES=im;const Jo=new J,Jx=new J,Qx=new ds;class Qo{constructor(e=new J(1,0,0),n=0){this.isPlane=!0,this.normal=e,this.constant=n}set(e,n){return this.normal.copy(e),this.constant=n,this}setComponents(e,n,s,r){return this.normal.set(e,n,s),this.constant=r,this}setFromNormalAndCoplanarPoint(e,n){return this.normal.copy(e),this.constant=-n.dot(this.normal),this}setFromCoplanarPoints(e,n,s){const r=Jo.subVectors(s,n).cross(Jx.subVectors(e,n)).normalize();return this.setFromNormalAndCoplanarPoint(r,e),this}copy(e){return this.normal.copy(e.normal),this.constant=e.constant,this}normalize(){const e=1/this.normal.length();return this.normal.multiplyScalar(e),this.constant*=e,this}negate(){return this.constant*=-1,this.normal.negate(),this}distanceToPoint(e){return this.normal.dot(e)+this.constant}distanceToSphere(e){return this.distanceToPoint(e.center)-e.radius}projectPoint(e,n){return n.copy(e).addScaledVector(this.normal,-this.distanceToPoint(e))}intersectLine(e,n){const s=e.delta(Jo),r=this.normal.dot(s);if(r===0)return this.distanceToPoint(e.start)===0?n.copy(e.start):null;const i=-(e.start.dot(this.normal)+this.constant)/r;return i<0||i>1?null:n.copy(e.start).addScaledVector(s,i)}intersectsLine(e){const n=this.distanceToPoint(e.start),s=this.distanceToPoint(e.end);return n<0&&s>0||s<0&&n>0}intersectsBox(e){return e.intersectsPlane(this)}intersectsSphere(e){return e.intersectsPlane(this)}coplanarPoint(e){return e.copy(this.normal).multiplyScalar(-this.constant)}applyMatrix4(e,n){const s=n||Qx.getNormalMatrix(e),r=this.coplanarPoint(Jo).applyMatrix4(e),i=this.normal.applyMatrix3(s).normalize();return this.constant=-r.dot(i),this}translate(e){return this.constant-=e.dot(this.normal),this}equals(e){return e.normal.equals(this.normal)&&e.constant===this.constant}clone(){return new this.constructor().copy(this)}}const ew={class:"model-name"},tw=dn({__name:"Model",props:ql({meshes:{},viewer:{}},{enabledFeatures:{default:[0,1,2]},enabledFeaturesModifiers:{},opacity:{default:1},opacityModifiers:{}}),emits:ql(["remove"],["update:enabledFeatures","update:opacity"]),setup(t,{emit:e}){const n=t,s=e;let r=n.meshes[0].getExtras()[Ht];const i=au(t,"enabledFeatures"),o=au(t,"opacity"),a=se(1),l=se(!1),u=se(1),c=se(!1),d=se(1),f=se(!1);let h=n.meshes.map(b=>b.listPrimitives().filter(x=>x.getMode()===WebGL2RenderingContext.TRIANGLES).length).reduce((b,x)=>b+x,0),g=n.meshes.map(b=>b.listPrimitives().filter(x=>x.getMode()in[WebGL2RenderingContext.LINE_STRIP,WebGL2RenderingContext.LINES]).length).reduce((b,x)=>b+x,0),y=n.meshes.map(b=>b.listPrimitives().filter(x=>x.getMode()===WebGL2RenderingContext.POINTS).length).reduce((b,x)=>b+x,0);h===0&&(i.value=i.value.filter(b=>b!==0)),g===0&&(i.value=i.value.filter(b=>b!==1)),y===0&&(i.value=i.value.filter(b=>b!==2));function v(b){var M;let x=(M=n.viewer)==null?void 0:M.scene,C=x==null?void 0:x._model;!x||!C||(C.traverse(T=>{if(T.userData[Ht]===r){let V=T.type=="Mesh"||T.type=="SkinnedMesh",L=T.type=="Line"||T.type=="LineSegments",A=T.type=="Points";if(V||L||A){let k=b.includes(V?0:L?1:A?2:-1);T.visible!==k&&(T.visible=k,T.userData.backChild&&(T.userData.backChild.visible=k))}}}),x.queueRender())}le(i,v);function S(b){var M;let x=(M=n.viewer)==null?void 0:M.scene,C=x==null?void 0:x._model;!x||!C||(C.traverse(T=>{T.userData[Ht]===r&&T.material&&T.material.opacity!==b&&(T.material.transparent=b<1,T.material.opacity=b,T.material.needsUpdate=!0)}),x.queueRender())}le(o,S);let{sceneDocument:m}=Pe("sceneDocument");function p(){var L,A;let b=(L=n.viewer)==null?void 0:L.scene,x=b==null?void 0:b._model;if(!b||!x)return;let C=a.value<1&&!l.value||a.value>0&&l.value,M=u.value<1&&!c.value||u.value>0&&c.value,T=d.value<1&&!f.value||d.value>0&&f.value,V;if((A=n.viewer)!=null&&A.renderer&&(C||M||T)){n.viewer.renderer.threeRenderer.localClippingEnabled=!0;let k=ns.getBoundingBox(m.value);if(!k)return;V=k.translate(b.getTarget())}x.traverse(k=>{if(k.userData[Ht]===r&&k.material)if(V){let O=V.min.x+a.value*(V.max.x-V.min.x),H=V.min.z+u.value*(V.max.z-V.min.z),q=V.min.y+d.value*(V.max.y-V.min.y),j=[new Qo(new J(-1,0,0),O),new Qo(new J(0,0,1),H),new Qo(new J(0,-1,0),q)];l.value&&j[0].negate(),c.value&&j[1].negate(),f.value&&j[2].negate(),T||j.pop(),M||j.splice(1,1),C||j.shift(),k.material.clippingPlanes=j,k.userData.backChild&&k.userData.backChild.material&&(k.userData.backChild.material.clippingPlanes=j)}else k.material.clippingPlanes=[],k.userData.backChild&&k.userData.backChild.material&&(k.userData.backChild.material.clippingPlanes=[])}),b.queueRender()}le(a,p),le(u,p),le(d,p),le(l,p),le(c,p),le(f,p),n.viewer.onElemReady(b=>b.addEventListener("camera-change",p));function w(){var M;let b=(M=n.viewer)==null?void 0:M.scene,x=b==null?void 0:b._model;if(!b||!x)return;let C=[];x.traverse(T=>{if(T.userData[Ht]===r){if((T.type=="Mesh"||T.type=="SkinnedMesh")&&(T.material.side=jb,r!==gi)){let V=T.clone();V.material=T.material.clone(),V.material.side=Wb,V.material.color=new Il(.25,.25,.25),T.userData.backChild=V,C.push(V)}T.type=="Points"&&(T.material.size=5,T.material.needsUpdate=!0)}}),C.forEach(T=>x.add(T)),b.queueRender(),v(i.value),S(o.value),p()}return n.viewer.onElemReady(b=>b.addEventListener("load",w)),(b,x)=>(at(),ln(F(ib),{value:F(r)},{default:X(()=>[_(F(jh),{"expand-icon":"hide-this-icon","collapse-icon":"hide-this-icon"},{default:X(()=>[_(F(by),{modelValue:i.value,"onUpdate:modelValue":x[0]||(x[0]=C=>i.value=C),multiple:"",onClick:x[1]||(x[1]=wu(()=>{},["stop"])),color:"surface-light"},{default:X(()=>[_(F(it),{icon:""},{default:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Toggle Faces ("+sn(F(h))+")",1)]),_:1}),_(F(Be),{type:"mdi",path:F(kb),rotate:90},null,8,["path"])]),_:1}),_(F(it),{icon:""},{default:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Toggle Edges ("+sn(F(g))+")",1)]),_:1}),_(F(Be),{type:"mdi",path:F(Fb),rotate:90},null,8,["path"])]),_:1}),_(F(it),{icon:""},{default:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Toggle Vertices ("+sn(F(y))+")",1)]),_:1}),_(F(Be),{type:"mdi",path:F(Vb),rotate:90},null,8,["path"])]),_:1})]),_:1},8,["modelValue"]),jt("div",ew,sn(F(r)),1),_(F(Aa)),_(F(it),{icon:"",onClick:x[2]||(x[2]=wu(C=>s("remove"),["stop"]))},{default:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Remove")]),_:1}),_(F(Be),{type:"mdi",path:F(Eb)},null,8,["path"])]),_:1})]),_:1}),_(F(Hh),null,{default:X(()=>[_(F(qr),{modelValue:o.value,"onUpdate:modelValue":x[3]||(x[3]=C=>o.value=C),"hide-details":"",min:"0",max:"1",step:.1},{prepend:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Change opacity")]),_:1}),_(F(Be),{type:"mdi",path:F(Sb)},null,8,["path"])]),_:1},8,["modelValue"]),_(F(hi)),_(F(qr),{modelValue:a.value,"onUpdate:modelValue":x[5]||(x[5]=C=>a.value=C),"hide-details":"",min:"0",max:"1"},{prepend:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Clip plane X")]),_:1}),_(F(Be),{type:"mdi",path:F($o),rotate:120},null,8,["path"]),Ve(" X ")]),append:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Swap clip plane X")]),_:1}),_(F(Oo),{trueIcon:"mdi-checkbox-marked-outline",falseIcon:"mdi-checkbox-blank-outline",modelValue:l.value,"onUpdate:modelValue":x[4]||(x[4]=C=>l.value=C)},{label:X(()=>[_(F(Be),{type:"mdi",path:F(Ho)},null,8,["path"])]),_:1},8,["modelValue"])]),_:1},8,["modelValue"]),_(F(qr),{modelValue:u.value,"onUpdate:modelValue":x[7]||(x[7]=C=>u.value=C),"hide-details":"",min:"0",max:"1"},{prepend:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Clip plane Y")]),_:1}),_(F(Be),{type:"mdi",path:F($o),rotate:-120},null,8,["path"]),Ve(" Y ")]),append:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Swap clip plane Y")]),_:1}),_(F(Oo),{trueIcon:"mdi-checkbox-marked-outline",falseIcon:"mdi-checkbox-blank-outline",modelValue:c.value,"onUpdate:modelValue":x[6]||(x[6]=C=>c.value=C)},{label:X(()=>[_(F(Be),{type:"mdi",path:F(Ho)},null,8,["path"])]),_:1},8,["modelValue"])]),_:1},8,["modelValue"]),_(F(qr),{modelValue:d.value,"onUpdate:modelValue":x[9]||(x[9]=C=>d.value=C),"hide-details":"",min:"0",max:"1"},{prepend:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Clip plane Z")]),_:1}),_(F(Be),{type:"mdi",path:F($o)},null,8,["path"]),Ve(" Z ")]),append:X(()=>[_(F(Xe),{activator:"parent"},{default:X(()=>[Ve("Swap clip plane Z")]),_:1}),_(F(Oo),{trueIcon:"mdi-checkbox-marked-outline",falseIcon:"mdi-checkbox-blank-outline",modelValue:f.value,"onUpdate:modelValue":x[8]||(x[8]=C=>f.value=C)},{label:X(()=>[_(F(Be),{type:"mdi",path:F(Ho)},null,8,["path"])]),_:1},8,["modelValue"])]),_:1},8,["modelValue"])]),_:1})]),_:1},8,["value"]))}}),nw=wl(tw,[["__scopeId","data-v-5c681122"]]),sw=dn({__name:"Models",props:{viewer:{}},emits:["remove"],setup(t,{expose:e,emit:n}){const s=t,r=n;let{sceneDocument:i}=Pe("sceneDocument"),o=se([]);function a(d){return d.getRoot().listMeshes().reduce((f,h)=>{var v;let g=((v=h.getExtras()[Ht])==null?void 0:v.toString())??"Unnamed",y=f.find(S=>l(S[0])===g);return y?y.push(h):f.push([h]),f},[])}function l(d){var f;return((f=d.getExtras()[Ht])==null?void 0:f.toString())??"Unnamed"}function u(d){r("remove",l(d))}function c(d){console.log("Find model",d),o.value.includes(d)||o.value.push(d),console.log("Expanded",o.value)}return e({findModel:c}),(d,f)=>(at(!0),Mr(Re,null,Rg(a(F(i)),h=>(at(),ln(F(nb),{key:l(h[0]),modelValue:F(o),"onUpdate:modelValue":f[0]||(f[0]=g=>Te(o)?o.value=g:o=g),multiple:""},{default:X(()=>[_(nw,{meshes:h,viewer:s.viewer,onRemove:g=>u(h[0])},null,8,["meshes","viewer","onRemove"])]),_:2},1032,["modelValue"]))),128))}}),wr={preload:["dev+http://127.0.0.1:32323/"],displayLoadingEveryMs:1e3,monitorEveryMs:100,monitorOpenTimeoutMs:1e3,autoplay:!0,arModes:"webxr scene-viewer quick-look",exposure:1,shadowIntensity:0,background:""},Zc=[];function om(t,e){let n=t.endsWith(".0");n&&(t=t.slice(0,-2));let s=wr[t];if(s===void 0)throw new Error(`Unknown setting: ${t}`);if(Array.isArray(s))if(n)s=s[0];else{let r=[];return Zc.includes(t)?r=s:Zc.push(t),r.push(om(t+".0",e)),r}switch(typeof s){case"boolean":return e==="true";case"number":return Number(e);case"string":return e;default:throw new Error(`Unknown setting type: ${typeof s} -- ${s}`)}}const rw=new URL(window.location.href);rw.searchParams.forEach((t,e)=>{e in wr&&(wr[e]=om(e,t))});const iw=250;class ow{constructor(e,n,s,r){Qt(this,"name");Qt(this,"url");Qt(this,"hash");Qt(this,"isRemove");this.name=e,this.url=n,this.hash=s,this.isRemove=r}}class aw extends Event{constructor(n){super("update");Qt(this,"models");this.models=n}}class lw extends EventTarget{constructor(){super(...arguments);Qt(this,"knownObjectHashes",{});Qt(this,"bufferedUpdates",[]);Qt(this,"batchTimeout",null)}async load(n){if(n.startsWith("dev+")){let s=new URL(n.slice(4));s.searchParams.set("api_updates","true"),await this.monitorDevServer(s)}else{let s=n.split("/").pop();s=(s==null?void 0:s.split(".")[0])||`unknown-${Math.random()}`;let i=(await fetch(n,{method:"HEAD"})).headers.get("etag");this.foundModel(s,i,n,!1)}}async monitorDevServer(n){try{let s=await fetch(n.toString());if(s.status===200){let r=uw(s.body.getReader());for await(let i of r){if(!i||!i.startsWith("data:"))continue;let o=JSON.parse(i.slice(5)),a=new URL(n);a.searchParams.delete("api_updates"),a.searchParams.set("api_object",o.name),this.foundModel(o.name,o.hash,a.toString(),o.is_remove)}}}catch{}setTimeout(()=>this.monitorDevServer(n),wr.monitorEveryMs)}foundModel(n,s,r,i){let o=this.knownObjectHashes[n];if(!s||s!==o||i){i?delete this.knownObjectHashes[n]:this.knownObjectHashes[n]=s;let a=new ow(n,r,s,i);this.bufferedUpdates.push(a),this.batchTimeout!==null&&clearTimeout(this.batchTimeout),this.batchTimeout=setTimeout(()=>{this.dispatchEvent(new aw(this.bufferedUpdates)),this.bufferedUpdates=[]},iw)}}}async function*uw(t){let e=new TextDecoder,n=new Uint8Array;for(;;){let{value:s,done:r}=await t.read();if(r||!s)break;n=new Uint8Array([...n,...s]);let o=e.decode(n).split(`
`);for(let a=0;a<o.length-1;a++)yield o[a];n=new Uint8Array([...n.slice(-1)])}}const cw=dn({__name:"App",setup(t){const e=ra({loader:()=>ha(()=>import("./ModelViewerWrapper-CThr6USm.js"),__vite__mapDeps([2,1]),import.meta.url),loadingComponent:Wh,delay:0});let n=se(window.innerWidth>1200);const s=se(""),r=se(null),i=be(new Fn);ut("sceneDocument",{sceneDocument:i});const o=se(null),a=se(!1);ut("disableTap",{disableTap:a,setDisableTap:h=>a.value=h});async function u(h){console.log("Received model update request",h.models);let g=i.value;for(let y in h.models){let v=parseInt(y)===h.models.length-1,S=h.models[y];S.isRemove?g=await ns.removeModel(s,g,S.name,v):g=await ns.loadModel(s,g,S.name,S.url,v,v)}i.value=g,Hl(i)}async function c(h){i.value=await ns.removeModel(s,i.value,h),Hl(i)}let d=new lw;d.addEventListener("update",h=>u(h));for(let h of wr.preload)d.load(h);async function f(){const h=prompt(`For an improved experience in viewing CAD/GLTF models with automatic updates, it's recommended to use the official yacv_server Python package. This ensures seamless serving of models and automatic updates.

Otherwise, enter the URL of the model to load:`);h&&await d.load(h)}return(h,g)=>(at(),ln(F(ab),{"full-height":""},{default:X(()=>[_(F(ub),{id:"main"},{default:X(()=>[_(F(e),{ref_key:"viewer",ref:r,src:s.value},null,8,["src"])]),_:1}),_(Cc,{"opened-init":F(n),side:"left",width:300},{toolbar:X(()=>[_(F(Vi),null,{default:X(()=>[Ve("Models")]),_:1})]),"toolbar-items":X(()=>[_(F(it),{icon:"",onClick:f},{default:X(()=>[_(F(Be),{type:"mdi",path:F(Rb)},null,8,["path"])]),_:1})]),default:X(()=>[_(sw,{ref_key:"models",ref:o,viewer:r.value,onRemove:c},null,8,["viewer"])]),_:1},8,["opened-init"]),_(Cc,{"opened-init":F(n),side:"right",width:48*3+1},{toolbar:X(()=>[_(F(Vi),null,{default:X(()=>[Ve("Tools")]),_:1})]),default:X(()=>[_(dx,{viewer:r.value,onFindModel:g[0]||(g[0]=y=>{var v;return(v=o.value)==null?void 0:v.findModel(y)})},null,8,["viewer"])]),_:1},8,["opened-init"])]),_:1}))}});function am(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};const{blueprint:e,...n}=t,s=vt(e,n),{aliases:r={},components:i={},directives:o={}}=s,a=Fp(s.defaults),l=t0(s.display,s.ssr),u=ty(s.theme),c=Sy(s.icons),d=Zp(s.locale),f=O1(s.date,d),h=D1(s.goTo,d);return{install:y=>{for(const v in o)y.directive(v,o[v]);for(const v in i)y.component(v,i[v]);for(const v in r)y.component(v,Rr({...r[v],name:v,aliasName:r[v].name}));if(u.install(y),y.provide(Vs,a),y.provide(Sa,l),y.provide(Fi,u),y.provide(xa,c),y.provide(ki,d),y.provide(P1,f.options),y.provide(pc,f.instance),y.provide(B1,h),Oe&&s.ssr)if(y.$nuxt)y.$nuxt.hook("app:suspense:resolve",()=>{l.update()});else{const{mount:v}=y;y.mount=function(){const S=v(...arguments);return xt(()=>l.update()),y.mount=v,S}}mn.reset(),y.mixin({computed:{$vuetify(){return tt({defaults:Ss.call(this,Vs),display:Ss.call(this,Sa),theme:Ss.call(this,Fi),icons:Ss.call(this,xa),locale:Ss.call(this,ki),date:Ss.call(this,pc)})}}})},defaults:a,display:l,theme:u,icons:c,locale:d,date:f,goTo:h}}const dw="3.5.8";am.version=dw;function Ss(t){var s,r;const e=this.$,n=((s=e.parent)==null?void 0:s.provides)??((r=e.vnode.appContext)==null?void 0:r.provides);if(n&&t in n)return n[t]}function fw(t,e){const n=e.modifiers||{},s=e.value,{once:r,immediate:i,...o}=n,a=!Object.keys(o).length,{handler:l,options:u}=typeof s=="object"?s:{handler:s,options:{attributes:(o==null?void 0:o.attr)??a,characterData:(o==null?void 0:o.char)??a,childList:(o==null?void 0:o.child)??a,subtree:(o==null?void 0:o.sub)??a}},c=new MutationObserver(function(){let d=arguments.length>0&&arguments[0]!==void 0?arguments[0]:[],f=arguments.length>1?arguments[1]:void 0;l==null||l(d,f),r&&lm(t,e)});i&&(l==null||l([],c)),t._mutate=Object(t._mutate),t._mutate[e.instance.$.uid]={observer:c},c.observe(t,u)}function lm(t,e){var n;(n=t._mutate)!=null&&n[e.instance.$.uid]&&(t._mutate[e.instance.$.uid].observer.disconnect(),delete t._mutate[e.instance.$.uid])}const hw={mounted:fw,unmounted:lm};function mw(t,e){var r,i;const n=e.value,s={passive:!((r=e.modifiers)!=null&&r.active)};window.addEventListener("resize",n,s),t._onResize=Object(t._onResize),t._onResize[e.instance.$.uid]={handler:n,options:s},(i=e.modifiers)!=null&&i.quiet||n()}function gw(t,e){var r;if(!((r=t._onResize)!=null&&r[e.instance.$.uid]))return;const{handler:n,options:s}=t._onResize[e.instance.$.uid];window.removeEventListener("resize",n,s),delete t._onResize[e.instance.$.uid]}const vw={mounted:mw,unmounted:gw};function um(t,e){const{self:n=!1}=e.modifiers??{},s=e.value,r=typeof s=="object"&&s.options||{passive:!0},i=typeof s=="function"||"handleEvent"in s?s:s.handler,o=n?t:e.arg?document.querySelector(e.arg):window;o&&(o.addEventListener("scroll",i,r),t._onScroll=Object(t._onScroll),t._onScroll[e.instance.$.uid]={handler:i,options:r,target:n?void 0:o})}function cm(t,e){var i;if(!((i=t._onScroll)!=null&&i[e.instance.$.uid]))return;const{handler:n,options:s,target:r=t}=t._onScroll[e.instance.$.uid];r.removeEventListener("scroll",n,s),delete t._onScroll[e.instance.$.uid]}function pw(t,e){e.value!==e.oldValue&&(cm(t,e),um(t,e))}const yw={mounted:um,unmounted:cm,updated:pw},bw=Object.freeze(Object.defineProperty({__proto__:null,ClickOutside:Mh,Intersect:Bf,Mutate:hw,Resize:vw,Ripple:Bs,Scroll:yw,Touch:j0},Symbol.toStringTag,{value:"Module"}));console.log("Starting yet-another-cad-viewer v0.6.15 (2370fd72ed656b1137bb7cfdd807791d50b275d0)...");const xw=am({directives:bw,theme:{defaultTheme:window.matchMedia("(prefers-color-scheme: dark)").matches?"dark":"light"}}),dm=Hv(cw);dm.use(xw);dm.mount("body");export{Uf as $,Hf as A,he as B,Zt as C,bh as D,Wf as E,Re as F,be as G,I as H,Ke as I,Af as J,Oe as K,ye as L,an as M,vf as N,vr as O,oo as P,dt as Q,Fr as R,It as S,hn as T,de as U,Tw as V,Ti as W,Ir as X,cs as Y,ls as Z,ha as _,at as a,uo as a$,Kt as a0,rh as a1,fo as a2,Bs as a3,Kp as a4,kr as a5,hl as a6,Ns as a7,us as a8,Jt as a9,Cs as aA,hh as aB,Jf as aC,ph as aD,Qf as aE,Xy as aF,Gy as aG,Tt as aH,Ln as aI,xf as aJ,ie as aK,Ys as aL,ba as aM,Iy as aN,hf as aO,Qv as aP,yh as aQ,Bf as aR,tp as aS,ic as aT,Vn as aU,Cw as aV,Pn as aW,At as aX,Sw as aY,lo as aZ,Ge as a_,ho as aa,ct as ab,jf as ac,sh as ad,Gt as ae,as as af,fl as ag,Aw as ah,Nt as ai,io as aj,sc as ak,rl as al,pl as am,iy as an,go as ao,mn as ap,v0 as aq,ut as ar,Ew as as,Oi as at,Th as au,xt as av,rp as aw,ip as ax,cl as ay,ay as az,jt as b,on as b0,Np as b1,In as b2,Ky as b3,Oo as b4,op as b5,pf as b6,ql as b7,au as b8,Xe as b9,Be as ba,Iw as bb,it as bc,Te as bd,kw as be,Rw as bf,Fw as bg,ns as bh,bs as bi,Ht as bj,pe as bk,Rt as bl,Rr as bm,Df as bn,Pf as bo,ao as bp,Mw as bq,hi as br,_w as bs,Mr as c,dn as d,Gs as e,Ve as f,Rg as g,ln as h,Pe as i,X as j,_ as k,Uo as l,Le as m,da as n,Yt as o,Xi as p,wl as q,se as r,wr as s,sn as t,F as u,G as v,le as w,we as x,n0 as y,Je as z};
